Package: autoharp
Title: Semi-Automatic Grading of R and Rmd Scripts
Version: 0.3.2
Authors@R: c(
    person("Vik", "Gopal", , "vik.gopal@nus.edu.sg", role = "aut"),
    person("Agrawal", "Naman", , "naman.a@nus.edu.sg", role = c("aut", "cre")),
    person("Yuting", "Huang", , "yhuang@nus.edu.sg", role = "ctb"),
    person("Samuel", "Seah", , "samuelseah@u.nus.edu", role = "aut"),
    person("Viknesh", "Jeya Kumar", , "viknesh@u.nus.edu", role = "aut"),
    person("Gabriel", "Ang", , "gabrielang@u.nus.edu", role = "aut"),
    person("Ruofan", "Liu", , "kelseyliu1998@gmail.com", role = "ctb"),
    person("National University of Singapore", role = "cph")
  )
Description: A customisable set of tools for assessing and grading 
    R or R-markdown scripts from students. It allows for checking correctness 
    of code output, runtime statistics and static code analysis. The latter 
    feature is made possible by representing R expressions using a tree
    structure.
License: MIT + file LICENSE
Encoding: UTF-8
RoxygenNote: 7.3.3.9000
Imports: magrittr, dplyr, stringr, rlang, tibble, knitr, rmarkdown,
        shiny, lintr, methods, igraph, xfun
Collate: 'treeharp.R' 'th_getter-length.R' 'as.matrix.R' 'autoharp.R'
        'check_correctness.R' 'check_rmd.R' 'check_runtime.R'
        'count_lints.R' 'env_size.R' 'extracton_section_text.R'
        'forestharp.R' 'forestharp_helpers.R' 'generate_thumbnails.R'
        'join_treeharp.R' 'lang_2_tree.R' 'lang_2_tree_helpers.R'
        'log_summary.R' 'nlp_related.R' 'populate_soln_env.R'
        'pre_checks.R' 'render_one.R' 'reset_path.R' 'to_BFS.R'
        'tree_kernel.R' 'tree_routines.R' 'utils-pipe.R' 'utils.R'
        'write_html.R' 'zzz.R'
Suggests: testthat
SystemRequirements: pandoc (>= 1.14) - http://pandoc.org
URL: https://namanlab.github.io/autoharp/
VignetteBuilder: knitr
NeedsCompilation: no
Packaged: 2026-04-11 08:03:48 UTC; namanagrawal
Author: Vik Gopal [aut],
  Agrawal Naman [aut, cre],
  Yuting Huang [ctb],
  Samuel Seah [aut],
  Viknesh Jeya Kumar [aut],
  Gabriel Ang [aut],
  Ruofan Liu [ctb],
  National University of Singapore [cph]
Maintainer: Agrawal Naman <naman.a@nus.edu.sg>
Repository: CRAN
Date/Publication: 2026-04-11 12:00:12 UTC
Built: R 4.6.0; ; 2026-04-28 23:27:19 UTC; unix

# Generated by roxygen2: do not edit by hand

export("%>%")
export(K2)
export(TreeHarp)
export(apply_can_improve_this)
export(carve_mst)
export(carve_subtree)
export(check_correctness)
export(check_rmd)
export(check_runtime)
export(clean_dir)
export(copy_e2e)
export(count_fn_call)
export(count_fn_in_fn)
export(count_lam_fn)
export(count_lints_all)
export(count_lints_one)
export(count_num_lines_for_loop)
export(count_self_fn)
export(detect_fn_arg)
export(detect_fn_call_in_for)
export(detect_for_in_fn_def)
export(detect_growing)
export(detect_nested_for)
export(env_size)
export(extract_actual_args)
export(extract_assigned_objects)
export(extract_chunks)
export(extract_fn_call)
export(extract_formal_args)
export(extract_non_chunks)
export(extract_section_text)
export(extract_self_fn)
export(fapply)
export(find_branch_num)
export(generate_all_subtrees)
export(generate_thumbnails)
export(get_adj_list)
export(get_child_ids)
export(get_libraries)
export(get_next_subtree)
export(get_node_types)
export(get_parent_call_id)
export(get_parent_id)
export(get_recursive_index)
export(is_subtree_rooted_at)
export(jaccard_treeharp)
export(join_treeharps)
export(keep_branches)
export(lang_2_tree)
export(log_summary)
export(only_actual_args_differ)
export(path_to_root)
export(pipe_can_improve_this)
export(populate_soln_env)
export(prune_depth)
export(render_one)
export(render_prechecks)
export(replace_sp_chars_filename)
export(reset_path)
export(rmd_to_forestharp)
export(rmd_to_token_count)
export(subtree_at)
exportMethods(TreeHarp)
exportMethods(get_adj_list)
exportMethods(get_child_ids)
exportMethods(get_node_types)
exportMethods(get_parent_id)
exportMethods(length)
exportMethods(names)
exportMethods(plot)
exportMethods(show)
importFrom(magrittr,"%>%")
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,setClass)
importFrom(methods,setValidity)
importFrom(methods,show)
importFrom(methods,slot)
importFrom(parallel,clusterCall)
importFrom(parallel,clusterEvalQ)
importFrom(parallel,clusterExport)
importFrom(parallel,makePSOCKcluster)
importFrom(parallel,stopCluster)
importFrom(rlang,.data)
importFrom(rlang,call_args)
importFrom(rlang,call_args_names)
importFrom(rlang,call_name)
importFrom(rlang,enexpr)
importFrom(rlang,is_formula)
importFrom(rlang,is_na)
importFrom(rlang,is_pairlist)
importFrom(rlang,is_symbol)
importFrom(rlang,is_syntactic_literal)
importFrom(rmarkdown,render)
importFrom(stats,na.omit)
importFrom(utils,tail)

# autoharp <img src="man/figures/logo.png" align="right" height="80" alt="autoharp logo" />

<!-- badges: start -->
<!--
[![R-CMD-check](https://img.shields.io/github/actions/workflow/status/namanlab/test_autoharp_2/R-CMD-check.yaml?logo=r&logoColor=white&label=R-CMD-check&style=flat-square)](https://github.com/namanlab/test_autoharp_2/actions/workflows/R-CMD-check.yaml)
[![pkgdown](https://img.shields.io/github/actions/workflow/status/namanlab/test_autoharp_2/pkgdown.yaml?logo=github&logoColor=white&label=pkgdown&style=flat-square)](https://namanlab.github.io/test_autoharp_2/)
[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg?style=flat-square&logo=opensourceinitiative&logoColor=white)](https://opensource.org/licenses/MIT)
-->
<!-- badges: end -->

> **Semi-automatic grading of R and Rmd scripts**: powerful, flexible, and
> designed for real-world teaching at scale.

`autoharp` is an R package that
provides a customisable set of tools for **assessing and grading R or
R Markdown scripts** submitted by students. It goes beyond simple correctness
checking to offer a complete, four-layer grading framework used in production
across undergraduate statistics courses.

---

## Why autoharp?

Grading programming assignments at scale is hard. Students submit diverse
code that may be *correct* but structured differently, or *incorrect* in
subtle ways that output-comparison alone cannot catch. `autoharp` solves
this with four complementary layers:

<div class="feature-grid">

| Feature | What it Does |
|---------|-------------|
| ✅ **Output correctness** | Compare student objects against a reference solution using typed, tolerance-aware checks |
| ⏱️ **Runtime & memory profiling** | Measure how long student code takes and how much memory it uses |
| 🌳 **Static code analysis (AST)** | Inspect Abstract Syntax Trees to check *how* students solved a problem: e.g., "did they use `apply` instead of a `for` loop?" |
| 🔍 **Similarity detection** | Identify suspiciously similar submissions using cosine, Jaccard, or edit-distance metrics |

</div>

---

## The autoharp Workflow

<p align="center">
  <img src="man/figures/workflow.png" width="85%" alt="autoharp four-phase grading workflow: Prepare → Distribute → Grade → Review" />
</p>

The workflow has **four phases**:

1. 📝 **Prepare**: Write a question sheet and a *solution template* (an Rmd
   with special `autoharp.objs` / `autoharp.scalars` chunk options)
2. 📤 **Distribute**: Share the question sheet and a blank student template
3. ⚙️ **Grade**: Run `populate_soln_env()` then `render_one()` per student;
   each submission runs in a sandboxed R process
4. 📊 **Review**: Inspect results via logs, thumbnail galleries, and the
   interactive Grading App

---

## Installation


You can install **autoharp** from CRAN with:

```r
install.packages("autoharp")
```

You can also install the development version of autoharp from
[GitHub](https://github.com/namanlab/autoharp) with:

```r
# install.packages("devtools")
devtools::install_github("namanlab/autoharp")
```

---

## Quick Start

```r
library(autoharp)

# 1. Render the solution template to build a reference environment
# Returns a list with $env (solution environment) and $test_fname (test script path) 
soln <- populate_soln_env("solution_template.Rmd")

# 2. Grade a student submission (runs in a sandboxed process)
result <- render_one(
  rmd_name      = "student_submission.Rmd",
  out_dir       = "grading_output/",
  knit_root_dir = getwd(),
  soln_stuff    = soln
)

# 3. Inspect the grading result
print(result)
```

---

## The Three Pillars of autoharp

### 1. 📄 Complete Documents, Not Snippets

`autoharp` grades **complete, reproducible documents**: entire R scripts or
R Markdown files, rather than isolated code snippets. This encourages good
scientific computing practices and tests the full submission pipeline.

### 2. 🌳 Static Analysis via AST Trees

The `TreeHarp` S4 class represents any R expression as a **rooted tree**,
enabling structural code checks that go beyond output comparison:

```r
# Convert an expression to a tree
tree <- TreeHarp(quote(mean(x + rnorm(10))), TRUE)
plot(tree)

# Check for absence of for-loops across an entire student script
forest <- rmd_to_forestharp("student_script.Rmd")
has_for_loop <- any(sapply(forest$forest, function(t) {
  any(t@nodeTypes$name == "for")
}))
```

### 3. 🖥️ Integrated Shiny Applications

`autoharp` ships with **three ready-to-use Shiny applications**:

| App | Audience | Purpose |
|-----|----------|---------|
| **Grading App** | Instructor | Multi-tab GUI for batch grading, session management, and manual plot review |
| **Similarity App** | Instructor | Detect code similarities using cosine, Jaccard, or edit-distance metrics |
| **Solution Checker** | Student | Self-service portal to check submissions before the deadline |

---

## The Grading App

The Grading App provides a complete browser-based interface for the grading
workflow - no command-line required. It features session management,
auto-generated object tests, plot comparison, and batch grading with a live
progress bar.

<p align="center">
  <img src="man/figures/grading_app_autoharp.png" width="80%"
       alt="Grading App: Autoharp Grading tab" />
  <br><em>Batch grading: run render_one() for all student submissions with live progress tracking</em>
</p>

---

## Similarity Detection

The Similarity App helps identify suspiciously similar submissions using
three metrics: cosine similarity, Jaccard index, and edit distance.

<p align="center">
  <img src="man/figures/similarity_app_heatmap.png" width="80%"
       alt="Similarity App: pairwise heatmap of student submissions" />
  <br><em>Pairwise similarity heatmap: darker cells indicate higher similarity between submissions</em>
</p>

---

## The Solution Checker (Student Portal)

The Solution Checker is a **student-facing portal** where students can
self-check their submission before the deadline.

<p align="center">
  <img src="man/figures/soln_checker_ui_02.png" width="80%"
       alt="Solution Checker: student interface" />
  <br><em>Students upload their script and instantly see correctness results, lint feedback, and runtime stats</em>
</p>

In a NUS course with **122 students** across two assignments, the Solution
Checker saw peak usage of **19 logins/day** in the final 3 days before
submission, evidence that students actively use it to iterate on their work.

---

## Real-World Validation

`autoharp` was developed and validated at the **National University of
Singapore** across multiple cohorts. Key highlights from production use:

- ✅ Graded hundreds of R/Rmd submissions across statistics courses
- ⏱️ Each submission grades with a few minutes minutes in a sandboxed process
- 🔍 Identified multiple cases of academic dishonesty via the Similarity App
- 📝 Students using the Solution Checker showed higher submission quality

---

## Learning More

- 📖 [Get Started](https://namanlab.github.io/autoharp/articles/getting-started.html): step-by-step introduction
- 🌳 [TreeHarp Tutorial](https://namanlab.github.io/autoharp/articles/treeharp_class.html): deep dive into AST-based analysis
- 🔄 [Grading Workflow](https://namanlab.github.io/autoharp/articles/demo_workflow.html): complete end-to-end example
- 📚 [Function Reference](https://namanlab.github.io/autoharp/reference/index.html): full API documentation

---

## Citation

If you use `autoharp` in your teaching or research, please cite:

```bibtex
@article{gopal2026autoharp,
  title   = {autoharp: An R Package for Autograding R/Rmd/qmd Scripts},
  author  = {Gopal, Vikneswaran and Naman, Agrawal and Huang, Yuting},
  year    = {2026}
}
```

---

## License

MIT © [National University of Singapore](https://www.nus.edu.sg/)

# Script Similarity Analysis App

A instructor-facing Shiny application for analyzing and visualizing code
similarity between R, Rmd, and Qmd scripts. This app helps instructors identify
potential plagiarism, similar approaches, and understand patterns in student
submissions.

## Features

### Similarity Analysis Methods

The app supports three different similarity calculation methods:

1. **Cosine Similarity (Token-based)**: Compares scripts based on token
   frequency vectors using cosine distance
2. **Jaccard Similarity (Set-based)**: Compares unique tokens using Jaccard
   index (intersection/union)
3. **Edit Distance**: Compares raw code using Levenshtein distance, normalized
   by length

### Visualization Methods

Choose from two dimensionality reduction techniques to visualize script similarities:

1. **t-SNE (t-Distributed Stochastic Neighbor Embedding)**:
   - Better for discovering local structure and clusters
   - Adjustable perplexity parameter (1-50)
   - Recommended for larger datasets (>10 scripts)

2. **MDS (Multi-Dimensional Scaling)**:
   - Better for preserving global distances
   - More stable and deterministic
   - Works well with smaller datasets

### Interactive Features

- **2D Scatter Plot**: Interactive visualization where closer points indicate
  more similar scripts
- **Click to View**: Click on any point to view the script content (can open
  multiple files)
- **Similarity Matrix Heatmap**: Hierarchically clustered heatmap showing all
  pairwise similarities
- **Cluster Detection**: Automatic identification of similarity clusters above a
  threshold
- **Top Similar Pairs**: Table showing the most similar script pairs
- **Summary Statistics**: Mean, median, standard deviation of similarities
- **Export Results**: Download all pairwise similarities as CSV

## How to Run the Application:

```r
library(shiny)
runApp(system.file(file.path('shiny', 'similarity_app', 'app.R'), 
                   package='autoharp'))
```

## Usage Instructions

### Basic Workflow

1. **Specify Scripts Folder**:
   - Enter the path to a folder containing R, Rmd, or Qmd files
   - Use the "Browse..." button for a graphical folder selector (if shinyFiles
     is installed)
   - Can use absolute paths (e.g., `/home/user/scripts`) or relative paths
     (e.g., `sol`)
   - The app will automatically find all `.R`, `.Rmd`, and `.qmd` files

2. **Configure Analysis**:
   - **Similarity Method**: Choose how to compare scripts (default: cosine)
   - **Visualization Method**: Choose how to display results (default: t-SNE)
   - **t-SNE Perplexity**: Adjust if using t-SNE (lower for smaller datasets)
   - **Similarity Threshold**: Set threshold for cluster detection (0.5-0.95)

3. **Run Analysis**:
   - Click "Run Similarity Analysis" to process all scripts
   - Progress indicator shows calculation status
   - Results appear in multiple sections below

4. **Explore Results**:
   - **Summary Statistics**: See overall similarity distribution
   - **Scatter Plot**: Explore the 2D projection, click points to view scripts
   - **Heatmap**: See all pairwise similarities in hierarchically clustered heatmap
   - **Similar Pairs**: Review the top 20 most similar script pairs
   - **Clusters**: See detected groups of similar scripts
   - **Full Matrix**: Browse the complete similarity matrix

5. **Export Results**:
   - Click "Download Results" to save all pairwise similarities as CSV
   - File includes Script1, Script2, Similarity score, and Method used

### Understanding the Results

#### Similarity Scores

- **1.0**: Identical scripts (or extremely similar)
- **0.7-0.99**: High similarity - may indicate collaboration or plagiarism
- **0.4-0.69**: Medium similarity - similar approaches or shared patterns
- **0.0-0.39**: Low similarity - different approaches

#### Scatter Plot Interpretation

- **Proximity**: Scripts closer together in the 2D plot are more similar
- **Colors**: Different colors indicate detected clusters (gray = singleton)
- **Interactivity**: Click on points to view script content in a modal
- **Multiple Views**: Can open multiple scripts to compare side-by-side

#### Clusters

- Clusters are groups of scripts with similarity above the threshold
- Useful for identifying potential collaboration or common approaches
- Singleton scripts (not in any cluster) are shown in gray

## Method Selection Guide

### When to Use Each Similarity Method

**Cosine (Token-based)** 
- Default, recommended for most cases:
- Fast and efficient
- Robust to code formatting differences
- Good balance of precision and recall
- Works well for both small and large codebases

**Jaccard (Set-based)** - When order doesn't matter:
- Ignores token frequency (only presence/absence)
- More lenient for varied implementations
- Good for detecting use of similar vocabulary
- Much slower to run on large batch of scripts.

**Edit Distance** - For exact matching:
- Sensitive to whitespace and formatting
- Best for detecting near-exact copies
- More strict than token-based methods
- Slower for large files

### When to Use Each Visualization Method

**t-SNE** - Recommended for most cases:
- Better at revealing local clusters
- More visually intuitive groupings
- Adjust perplexity based on dataset size:
  - Small (5-10 scripts): perplexity = 2-5
  - Medium (10-30 scripts): perplexity = 5-15
  - Large (>30 scripts): perplexity = 15-30

**MDS** - For distance preservation:
- More deterministic (reproducible results)
- Better preserves global structure
- Faster computation
- Works better with small datasets

## Configuration Options

### Similarity Threshold

Controls cluster detection sensitivity:
- **0.5**: Very lenient - large clusters, may include false positives
- **0.7**: Balanced - good for general use (default)
- **0.8**: Strict - only very similar scripts grouped
- **0.95**: Very strict - near-identical scripts only

### t-SNE Perplexity

Balances local vs. global structure:
- **Low (1-5)**: Emphasizes local structure, tighter clusters
- **Medium (5-15)**: Balanced view
- **High (15-30)**: Emphasizes global structure, looser clusters
- **Note**: Automatically adjusted for small datasets

## File Handling

### Supported File Types

- `.R` - R scripts
- `.r` - R scripts (lowercase)
- `.Rmd` - R Markdown files (code chunks extracted)
- `.rmd` - R Markdown files
- `.qmd` - Quarto documents (code chunks extracted)
- `.Qmd` - Quarto documents (code chunks extracted)

### Code Extraction

For Rmd and Qmd files:
- Only code chunks are analyzed (prose is ignored)
- Uses knitr patterns for robust chunk detection
- Handles multi-line chunk options
- Empty chunks are handled gracefully

## Technical Notes

### Performance

- Pairwise similarity: O(n²) where n is number of scripts
- Large datasets (>100 scripts) may take several minutes
- Progress indicator shows calculation status

### Dependencies

Required packages:
- `shiny`: Application framework
- `DT`: Interactive data tables
- `Rtsne`: t-SNE dimensionality reduction
- `stringr`: String manipulation
- `plotly`: Interactive plots
- `shinyFiles`: Folder selection UI
- `knitr`: Required for Rmd/Qmd chunk extraction

# Introduction

The `solution_checker` shiny application is a website for students to double-check
their assignment *before* actual grading. From experience, especially in the
initial portion of a course, many student mistakes are due to incorrect path
settings, or not coding with the intent of making it easier for someone else to
run their code.

The solution checker allows students to upload code to a server. The instructor 
may want to prepare a stripped-down solution template to make basic checks, e.g.
whether the script can be knitted and rendered, existence of particular functions
for the particular assignment, etc. 

## File Structure Overview

* `data/` folder can be used to store datasets that the script will use.
* `R/` contains scripts that will be executed by R before the shiny application
  begins. At present, there is only one script: `globals.R`. It can be used to 
  customise options for the shiny application.
* `soln_templates/` contains all the solution templates that the server will use.
* `app.R` contains the UI and server logic.
* `db_creation.R` is a helper script to guide the instructor to setting up an
  SQLite database.
  * Remember to set the value for AUTOHARP_TUNER_DB_KEY in .Renviron,
    which should be stored in ~/.Renviron
  * To generate one, you can use: `openssl rand -hex 16`
  
## Global Variables

The following variables are initialised in `globals.R`:

1.  `soln_templates_dir`: Full path to solution templates that this application will use.
2.  `knit_wd`: The working directory for knitting.
3.  `tabs`: The tabs to show. By default, the checker does three things:
    1. Lint the file.
    2. Render it.
    3. Display a summary of checks conducted.
4.  `app_title`: The title to display on the page.
5.  `permission_to_install`: Whether the application should have permission to install
    new packages. You are recommended to set this as `FALSE`.
6.  `max_time`: The maximum time in seconds before the applications gives up 
    running a script. The purpose of this is to avoid hogging the CPU and resulting 
    in a DoS to others.
7.  `summary_header`: The summary header to look for in the solution template. The 
    text in this chunk will be displayed on the correctness checks tab. 
8.  `corr_cols_to_drop`: The correctness output contains several columns that are 
    probably not important to the student. Examples are running time, memory used, 
    and so on. These columns can be dropped for purpose of these basic, student-facing 
    checks.
    
# Deploying on Shiny Server

## Securing the UI

Before deployment, use `db_creation.R` to create an SQLite database with
usernames and passwords for students, and an admin.

The key for the database can be stored as an environment variable in `.Renviron`
(see `R/globals.R` for how it is read in and used).

If you need to generate a secure hex string, you can use this command:

```
openssl rand -hex 16
```

## Shiny Server

It is also recommended to have a web proxy in front of your shiny server, and to
run applications as a non-sudo user.

# Autoharp Instructor Grading App

A Shiny application designed to help instructors grade student R/Rmd scripts
more easily when using the autoharp package.

## Features

### Tab 0: Session Management

This tab allows you to save and resume grading sessions:

1. **Load Previous Session**: Upload a previously saved session file (.rds) to resume grading
2. **Current Session Info**: View information about your active grading session
3. **Manual Save**: Save your session manually at any time
4. **Auto-Save**: Sessions are automatically saved after:
   - Completing automated grading (Tab 1)
   - Saving plot grades (Tab 2)
   - Completing autoharp grading (Tab 4)

**Session Data Includes**:
- Solution template and objects
- Student files list
- All grading results
- Plot grades and comments
- Current grading position
- Output directory configuration

### Tab 1: Automated Object Testing

This tab allows instructors to:

1. **Load a Solution Template**: Upload a plain RMD file containing the correct solutions (no autoharp annotations needed)
2. **Specify Student Scripts Folder**: Point to a folder containing student R/Rmd submissions (use the "Browse..." button for a graphical folder selector, or type the path manually)
3. **Specify Output Directory**: Configure where all outputs (plots, results, session files) are saved (default: "grading_output")
4. **Specify Working Directory**: **NEW!** Set the working directory from which student files are executed (default: student file location)
5. **Select Objects to Test**: Choose which objects from the solution should be tested against student submissions
6. **Run Automated Grading**: Execute comprehensive tests on all student submissions

**Note**: When you select a folder in Tab 1, the same folder path is automatically populated in Tab 4 (Autoharp Grading) for convenience. You can change it in Tab 4 if needed.

### Tab 2: Plot Grading

This tab provides an interface for manually grading student-generated plots:

1. **Navigate Through Students**: Use Previous/Next buttons or dropdown to select students
2. **View Student Plots**: See all plots generated by each student (plots are saved to a `student_plots` folder during grading)
3. **Assign Grades**: Enter numeric grades (0-100) for each student
4. **Add Comments**: Provide optional feedback comments
5. **Track Progress**: Visual progress bar shows grading completion status

**Note**: Plots are automatically rendered and saved to a `student_plots` subfolder within the configured output directory during the automated grading process. Sessions are auto-saved after each grade entry.

### Tab 3: Template Download

This tab allows you to generate an autoharp-compatible solution template with test chunks:

1. **Upload Solution**: Upload a solution RMD file in the 'Object Testing' tab
2. **Select Objects**: Choose which objects to include in tests
3. **Generate Template**: Create the autoharp version with test chunks
4. **Download Template**: Download the generated template for use with `render_one()`

The generated template includes:
- Proper R Markdown chunks with `test` prefix names
- `autoharp.objs` option containing solution objects for comparison
- `autoharp.scalars` option containing test result scalars
- Comprehensive type-based tests for all selected objects

### Tab 4: Autoharp Grading

This tab allows you to use the autoharp `render_one` function for comprehensive grading:

1. **Upload Template File**: Upload an autoharp-compatible template file with test chunks (generated from Tab 3)
2. **Specify Student Scripts Folder**: Point to a folder containing student scripts (use the "Browse..." button or type the path). This field is automatically populated with the folder selected in Tab 1, but you can change it if needed.
3. **Specify Output Directory**: Configure where render_one outputs are saved (default: "student_out")
4. **Specify Working Directory**: **NEW!** Set the working directory from which student files are executed (default: student file location)
5. **Specify Test Pattern**: Enter the pattern to identify test chunks (default: "test")
6. **Allow Package Installation**: Choose whether to allow automatic package installation during grading
7. **Run Autoharp Grading**: Execute `render_one` on all student scripts using `lapply`
8. **View Results**: See results in a comprehensive table including:
   - Run status (SUCCESS/FAIL)
   - Run time and memory usage
   - All correctness check results
   - Plot grades from Tab 2 (if available)
9. **Download Excel**: Export all results as an Excel file with two sheets:
   - Sheet 1: Grading Results (autoharp correctness checks)
   - Sheet 2: Plot Grades (manual plot grades and comments from Tab 2)

**Configuration**:
- Template file: Upload via file input (supports .Rmd and .qmd files)
- Student scripts: Specified via folder path (automatically synced from Tab 1)
- Output directory: Configurable (default: "student_out")
- Working directory: Configurable (default: student file location)

**Workflow**:
1. First, use Tab 1 to upload a solution and select the student scripts folder
2. Configure output directory in Tab 1 or Tab 4 (where all outputs will be saved)
3. Use Tab 3 to generate/download an autoharp template
4. In Tab 4, upload the generated autoharp template
5. The student scripts folder is automatically populated from Tab 1, but can be changed if needed
6. Optionally, use Tab 2 to manually grade plots
7. Use Tab 4 to run autoharp grading with the specified test pattern
8. Download comprehensive results as an Excel file with two sheets
9. Your session is automatically saved throughout - return anytime via Tab 0!

---

## Output Directory Management

You can configure where all grading outputs are stored:

- **Default Locations**: 
  - Tab 1: `grading_output/` (for object testing results and plots)
  - Tab 4: `student_out/` (for autoharp render_one outputs)
- **Custom Directories**: Specify any path for organized storage
- **Structure**:
  ```
  <output_directory>/
  ├── student_plots/        # Student-generated plots
  │   ├── student_01/
  │   └── student_02/
  ├── sessions/            # Auto-saved session files
  │   ├── session_20260123_120000.rds
  │   └── session_20260123_130000.rds
  └── [render_one outputs]  # HTML, logs, etc.
  ```

---

## Session Management

 Sessions are automatically saved and can be resumed:

### Auto-Save Behavior
- Sessions are saved automatically after:
  - Running automated grading (Tab 1)
  - Saving each plot grade (Tab 2)
  - Completing autoharp grading (Tab 4)
- Session files are stored in `<output_directory>/sessions/`
- Each session has a unique ID with timestamp

### Resuming a Session
1. Go to the **Session** tab
2. Click "Upload Session File" and select a `.rds` file
3. Click "Load Session" to restore:
   - All grading results
   - Plot grades and comments
   - Student files
   - Configuration settings
   - Your position in the grading workflow

### Manual Save
- Use the "Save Session Now" button in the Session tab anytime
- Useful before closing the app or switching tasks

---

## Comprehensive Default Tests

The app includes comprehensive built-in tests based on object types:

### Numeric Vectors

| Test | Description | Details |
|------|-------------|---------|
| `length` | Compares length of vectors | `length(obj) == length(.obj)` |
| `mean` | Compares mean values | Tolerance: 1e-6, handles NA |
| `sd` | Compares standard deviation | Tolerance: 1e-6, handles NA |
| `min` | Compares minimum values | Tolerance: 1e-6, handles NA |
| `max` | Compares maximum values | Tolerance: 1e-6, handles NA |
| `sum` | Compares sum of values | Tolerance: 1e-6, handles NA |
| `na_count` | Compares count of NA values | `sum(is.na(obj)) == sum(is.na(.obj))` |
| `identical` | Checks if all values match | Uses `all.equal()` with tolerance |

### Character Vectors

| Test | Description | Details |
|------|-------------|---------|
| `length` | Compares length of vectors | `length(obj) == length(.obj)` |
| `identical` | Checks exact match | Uses `identical()` |
| `unique_count` | Compares unique value count | `length(unique(obj)) == length(unique(.obj))` |
| `nchar_total` | Compares total character count | `sum(nchar(obj)) == sum(nchar(.obj))` |

### Logical Vectors

| Test | Description | Details |
|------|-------------|---------|
| `length` | Compares length of vectors | `length(obj) == length(.obj)` |
| `identical` | Checks exact match | Uses `identical()` |
| `sum_true` | Compares sum of TRUE values | `sum(obj, na.rm = TRUE) == sum(.obj, na.rm = TRUE)` |
| `na_count` | Compares count of NA values | `sum(is.na(obj)) == sum(is.na(.obj))` |

### Data Frames

| Test | Description | Details |
|------|-------------|---------|
| `nrow` | Compares number of rows | `nrow(obj) == nrow(.obj)` |
| `ncol` | Compares number of columns | `ncol(obj) == ncol(.obj)` |
| `colnames` | Checks if column names match | Uses `setequal()` |
| `coltypes` | Checks if column types match | Compares `sapply(obj, class)` |
| `column_mean_*` | Compares means of numeric columns | For each numeric column (up to 3) |
| `column_sd_*` | Compares SDs of numeric columns | For each numeric column |
| `factor_levels_*` | Compares factor levels | For each factor column |
| `correlation_matrix` | Compares correlation structure | Max correlation difference < 0.01 |

### Matrices

| Test | Description | Details |
|------|-------------|---------|
| `nrow` | Compares number of rows | `nrow(obj) == nrow(.obj)` |
| `ncol` | Compares number of columns | `ncol(obj) == ncol(.obj)` |
| `dim` | Compares dimensions | `identical(dim(obj), dim(.obj))` |
| `rownames` | Checks if row names match | Uses `identical()` |
| `colnames` | Checks if column names match | Uses `identical()` |
| `mean` | Compares overall mean | Tolerance: 1e-6 |
| `max_diff` | Maximum element difference | `max(abs(obj - .obj)) < 1e-6` |

### Lists

| Test | Description | Details |
|------|-------------|---------|
| `length` | Compares number of elements | `length(obj) == length(.obj)` |
| `names` | Checks if element names match | Uses `setequal()` |
| `element_*` | Tests each named element | Uses `all.equal()` for each (up to 3) |
| `element_class_*` | Checks element classes | For each named element |
| `element_length_*` | Checks element lengths | For atomic and list elements |
| `element_mean_*` | Checks element means | For numeric elements |

### Functions

| Test | Description | Details |
|------|-------------|---------|
| `num_args` | Compares number of arguments | `length(formals(obj)) == length(formals(.obj))` |
| `arg_names` | Checks if argument names match | Uses `setequal()` |
| `is_function` | Verifies it's a function | `is.function(obj)` |
| `body_length` | Compares body complexity | Info only - line count comparison |

### Factors

| Test | Description | Details |
|------|-------------|---------|
| `length` | Compares length | `length(obj) == length(.obj)` |
| `nlevels` | Compares number of levels | `nlevels(obj) == nlevels(.obj)` |
| `levels` | Checks if levels match | Uses `setequal()` |
| `level_order` | Checks if level order matches | Uses `identical()` |
| `frequencies` | Compares level frequencies | Compares `table()` output |

---

## How to Run

### From the app directory:

```r
library(shiny)
runApp("app")
```

### Or from any R session:

```r
library(shiny)
runApp("/path/to/autoharp/app")
```

## Usage Instructions

1. **Start the App**: Run the app using one of the methods above

2. **Resume or Start New Session** (Optional):
   - Go to the **Session** tab to load a previous session
   - Or start fresh by going directly to Tab 1

3. **Configure Output Directory**:
   - In Tab 1 or Tab 4, specify your desired output directory
   - All grading outputs, plots, and session files will be saved here
   - Defaults: `grading_output` (Tab 1) or `student_out` (Tab 4)

4. **Load Solution Template**: 
   - Upload an RMD file containing the correct solutions
   - This can be a plain RMD file without any autoharp-specific annotations
   - The app will execute the RMD and extract all objects created

5. **Specify Student Folder**:
   - Enter the full path to the folder containing student R/Rmd files
   - The app will find all `.R`, `.r`, `.Rmd`, and `.rmd` files

6. **Select Objects**:
   - After loading the solution, checkboxes will appear for each object
   - Select which objects should be tested against student submissions

7. **Run Grading**:
   - Click "Run Automated Grading" to process all student files
   - View results in the summary table and detailed test output
   - Session is automatically saved after completion

8. **Grade Plots (Tab 2)**:
   - Navigate through students using the selector or buttons
   - View each student's plots
   - Assign grades and comments
   - Click "Save Grade" to record the grade (auto-saves session)

9. **Generate Template (Tab 3)**:
   - Click "Generate Template" to create an autoharp-compatible solution
   - Preview the generated template with test chunks
   - Download the template for use with `render_one()`

10. **Download Results**:
    - Click "Download Results as CSV" to export all data
    - The CSV includes both automated test scores and manual plot grades

11. **Pause and Resume**:
    - Your session is automatically saved throughout the grading process
    - Close the app at any time
    - Return later and load your session from the Session tab to continue

## Generated Template Format

The generated template adds test chunks in the following format:


```{r test_01_X, autoharp.objs=c("X"), autoharp.scalars=c("X_length_check", "X_mean_check", "X_identical_check")}
# Tests for X (numeric)
# Length check
X_length_check <- length(X) == length(.X)

# Mean check (with tolerance)
X_mean_check <- isTRUE(all.equal(mean(X, na.rm = TRUE), mean(.X, na.rm = TRUE), tolerance = 1e-6))

# Identical values check
X_identical_check <- isTRUE(all.equal(X, .X))
```

This format is compatible with autoharp's `render_one()` function for automated grading.

## Technical Notes

- Student scripts are executed in isolated environments
- Errors during execution are caught and recorded
- Tests use tolerance of 1e-6 for numeric comparisons
- Plot collection works by saving figures to a dedicated `student_plots` directory during knitting
- Tab 4 uses the autoharp `populate_soln_env` and `render_one` functions for comprehensive grading
- **Sessions are saved as RDS files** containing all grading state, allowing you to pause and resume work
- **Output directories are configurable** for flexible organization of grading materials

## File Structure

```
app/
├── app.R              # Main Shiny application
├── www/helpers.R      # Test functions and utilities
└── README.md          # This file
```

**Generated directories during grading (configurable):**
```
<output_directory>/          # User-specified output directory
├── student_plots/           # Contains rendered plots from each student
│   ├── student_01/
│   │   ├── plot1.png
│   │   └── plot2.png
│   └── student_02/
│       └── plot1.png
├── sessions/                # Auto-saved session files
│   ├── session_20260123_120000.rds
│   └── session_20260123_130000.rds
├── [student_name].html      # Rendered output from render_one
├── [student_name].knit.md   # Knitted markdown files
└── render_one.log           # Log file from render_one
```

## Dependencies

The app uses base R functions and requires:
- shiny (for the UI)
- DT (for data tables)
- knitr (for knitting RMD files)
- rmarkdown (for rendering Rmd files)
- base64enc (for plot encoding)
- autoharp (for render_one and populate_soln_env functions)
- shinyFiles (optional but recommended, for enhanced folder selection UI - will gracefully degrade if not available)
- openxlsx (optional, for Excel export in Tab 4 - will prompt to install if not available)

These are already dependencies of the autoharp package, except for `shinyFiles` and `openxlsx` which are optional.

**Note:** This version no longer requires `shinydashboard` package. If you're upgrading from a previous version, you can remove the `shinydashboard` dependency from your environment.

# Student Script 2 - Very Similar Linear Regression (Slight Variation)

# Load dataset
data(mtcars)

# Build linear regression model
linear_model <- lm(mpg ~ wt + hp, data = mtcars)

# Display model summary
summary(linear_model)

# Generate predictions
pred_values <- predict(linear_model, mtcars)

# Compute residual values
resid_values <- residuals(linear_model)

# Visualize the relationship
plot(mtcars$wt, mtcars$mpg, main = "Miles per Gallon vs Weight",
     xlab = "Weight (1000 lbs)", ylab = "MPG", pch = 19, col = "darkblue")
abline(linear_model, col = "red", lwd = 2)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----similarity-app-run, eval=FALSE, echo=TRUE--------------------------------
# library(shiny)
# runApp(system.file(file.path('shiny', 'similarity_app', 'app.R'),
#                    package = 'autoharp'))

## ----similarity-app-files, echo=TRUE------------------------------------------
list.files(system.file("shiny", "similarity_app", package = "autoharp"))

# Student Script 4 - Plotting with ggplot2

library(ggplot2)

# Load data
data(diamonds)

# Create scatter plot
p1 <- ggplot(diamonds, aes(x = carat, y = price, color = cut)) +
  geom_point(alpha = 0.5) +
  theme_minimal() +
  labs(title = "Diamond Price vs Carat",
       x = "Carat",
       y = "Price ($)")

print(p1)

# Create boxplot
p2 <- ggplot(diamonds, aes(x = cut, y = price, fill = cut)) +
  geom_boxplot() +
  theme_classic() +
  labs(title = "Price Distribution by Cut Quality",
       x = "Cut",
       y = "Price ($)")

print(p2)

# Create histogram
p3 <- ggplot(diamonds, aes(x = price)) +
  geom_histogram(bins = 50, fill = "steelblue", color = "black") +
  theme_bw() +
  labs(title = "Distribution of Diamond Prices",
       x = "Price ($)",
       y = "Frequency")

print(p3)

# helper functions for App

# constants for test generation
DEFAULT_NUMERIC_TOLERANCE <- 1e-6
MAX_ELEMENTS_TO_TEST <- 3

#' Sanitize object name for use as R variable prefix
#' 
#' @param obj_name Object name to sanitize
#' @return Sanitized name safe for use in R variable names
sanitize_name <- function(obj_name) {
  gsub("[^a-zA-Z0-9_]", "_", obj_name)
}

#' Create a test result data frame
#' Helper function to create consistent test result format
#' 
#' @param object Object name
#' @param test Test name
#' @param result Test result (PASS, FAIL, SKIP, INFO)
#' @param details Details about the test
#' @return A data.frame with one row
create_test_result <- function(object, test, result, details) {
  data.frame(
    object = object,
    test = test,
    result = result,
    details = details,
    stringsAsFactors = FALSE
  )
}

#' Bind multiple data frames by row
#' Helper function that works like dplyr::bind_rows but with base R
#' 
#' @param df_list List of data frames to bind
#' @return A single data frame with all rows
bind_test_results <- function(df_list) {
  if (length(df_list) == 0) {
    return(data.frame(
      object = character(0),
      test = character(0),
      result = character(0),
      details = character(0),
      stringsAsFactors = FALSE
    ))
  }
  do.call(rbind, df_list)
}

#' Source an R file with each expression wrapped in try-catch
#' 
#' This function reads an R file, parses it into expressions, and evaluates
#' each expression independently wrapped in tryCatch. This allows the script
#' to continue even if some expressions have errors.
#' 
#' @param file Path to the R file to source
#' @param envir Environment in which to evaluate the expressions
#' @return Invisible NULL. Expressions are evaluated for their side effects.
source_with_trycatch <- function(file, envir = parent.frame()) {
  # Read and parse the file
  exprs <- tryCatch({
    parse(file = file)
  }, error = function(e) {
    warning(paste("Failed to parse file:", file, "-", e$message))
    return(NULL)
  })
  
  if (is.null(exprs)) return(invisible(NULL))
  
  # Evaluate each expression wrapped in tryCatch
  for (i in seq_along(exprs)) {
    tryCatch({
      eval(exprs[[i]], envir = envir)
    }, error = function(e) {
      # Continue execution - the error is caught and the script continues
      # The test framework will detect which objects were not created
      # Warning is issued to help with debugging if needed
      warning(paste("Error in expression", i, "of", file, ":", e$message))
    })
  }
  
  invisible(NULL)
}

#' Run comprehensive tests on objects
#' 
#' @param student_env Environment containing student objects
#' @param solution_env Environment containing solution objects
#' @param objects_to_test Character vector of object names to test
#' @return A data.frame with test results
run_object_tests <- function(student_env, solution_env, objects_to_test) {
  
  results <- list()
  
  for (obj_name in objects_to_test) {
    # Check if object exists in solution
    if (!exists(obj_name, envir = solution_env)) {
      results[[length(results) + 1]] <- create_test_result(
        object = obj_name,
        test = "existence_in_solution",
        result = "SKIP",
        details = "Object not found in solution template"
      )
      next
    }
    
    # Check if object exists in student environment
    if (!exists(obj_name, envir = student_env)) {
      results[[length(results) + 1]] <- create_test_result(
        object = obj_name,
        test = "existence",
        result = "FAIL",
        details = "Object not found in student submission"
      )
      next
    }
    
    # Get both objects
    solution_obj <- get(obj_name, envir = solution_env)
    student_obj <- get(obj_name, envir = student_env)
    
    # Determine object type and run appropriate tests
    obj_tests <- run_type_specific_tests(student_obj, solution_obj, obj_name)
    results <- c(results, obj_tests)
  }
  
  bind_test_results(results)
}

#' Run tests specific to object type
#' 
#' @param student_obj Student's object
#' @param solution_obj Solution object
#' @param obj_name Name of the object
#' @return List of test result data.frame objects
run_type_specific_tests <- function(student_obj, solution_obj, obj_name) {
  
  results <- list()
  
  # Class check
  results[[length(results) + 1]] <- test_class(student_obj, solution_obj, obj_name)
  
  # Type-specific tests
  if (is.vector(solution_obj) && !is.list(solution_obj)) {
    results <- c(results, test_vector(student_obj, solution_obj, obj_name))
  } else if (is.data.frame(solution_obj)) {
    results <- c(results, test_dataframe(student_obj, solution_obj, obj_name))
  } else if (is.matrix(solution_obj)) {
    results <- c(results, test_matrix(student_obj, solution_obj, obj_name))
  } else if (is.list(solution_obj)) {
    results <- c(results, test_list(student_obj, solution_obj, obj_name))
  } else if (is.function(solution_obj)) {
    results <- c(results, test_function(student_obj, solution_obj, obj_name))
  } else if (is.factor(solution_obj)) {
    results <- c(results, test_factor(student_obj, solution_obj, obj_name))
  }
  
  results
}

#' Test class of object
test_class <- function(student_obj, solution_obj, obj_name) {
  student_class <- class(student_obj)[1]
  solution_class <- class(solution_obj)[1]
  
  create_test_result(
    object = obj_name,
    test = "class",
    result = ifelse(student_class == solution_class, "PASS", "FAIL"),
    details = ifelse(
      student_class == solution_class,
      paste("Class:", student_class),
      paste("Expected:", solution_class, "| Got:", student_class)
    )
  )
}

#' Test vector objects
#' 
#' Tests: length, mean, sd, min, max, sum, identical values
test_vector <- function(student_obj, solution_obj, obj_name) {
  results <- list()
  
  # Length test
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "length",
    result = ifelse(length(student_obj) == length(solution_obj), "PASS", "FAIL"),
    details = paste("Expected:", length(solution_obj), "| Got:", length(student_obj))
  )
  
  # For numeric vectors, run statistical tests
  if (is.numeric(solution_obj) && is.numeric(student_obj)) {
    # Mean test
    solution_mean <- mean(solution_obj, na.rm = TRUE)
    student_mean <- mean(student_obj, na.rm = TRUE)
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "mean",
      result = ifelse(
        (is.na(solution_mean) && is.na(student_mean)) || 
          (!is.na(solution_mean) && !is.na(student_mean) && 
             abs(solution_mean - student_mean) < 1e-6),
        "PASS", "FAIL"
      ),
      details = paste("Expected:", round(solution_mean, 6), "| Got:", round(student_mean, 6))
    )
    
    # SD test
    solution_sd <- sd(solution_obj, na.rm = TRUE)
    student_sd <- sd(student_obj, na.rm = TRUE)
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "sd",
      result = ifelse(
        (is.na(solution_sd) && is.na(student_sd)) || 
          (!is.na(solution_sd) && !is.na(student_sd) && 
             abs(solution_sd - student_sd) < 1e-6),
        "PASS", "FAIL"
      ),
      details = paste("Expected:", round(solution_sd, 6), "| Got:", round(student_sd, 6))
    )
    
    # Min test
    solution_min <- min(solution_obj, na.rm = TRUE)
    student_min <- min(student_obj, na.rm = TRUE)
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "min",
      result = ifelse(
        abs(solution_min - student_min) < 1e-6,
        "PASS", "FAIL"
      ),
      details = paste("Expected:", round(solution_min, 6), "| Got:", round(student_min, 6))
    )
    
    # Max test
    solution_max <- max(solution_obj, na.rm = TRUE)
    student_max <- max(student_obj, na.rm = TRUE)
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "max",
      result = ifelse(
        abs(solution_max - student_max) < 1e-6,
        "PASS", "FAIL"
      ),
      details = paste("Expected:", round(solution_max, 6), "| Got:", round(student_max, 6))
    )
    
    # Sum test
    solution_sum <- sum(solution_obj, na.rm = TRUE)
    student_sum <- sum(student_obj, na.rm = TRUE)
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "sum",
      result = ifelse(
        abs(solution_sum - student_sum) < 1e-6,
        "PASS", "FAIL"
      ),
      details = paste("Expected:", round(solution_sum, 6), "| Got:", round(student_sum, 6))
    )
    
    # NA count test
    solution_na <- sum(is.na(solution_obj))
    student_na <- sum(is.na(student_obj))
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "na_count",
      result = ifelse(solution_na == student_na, "PASS", "FAIL"),
      details = paste("Expected NA count:", solution_na, "| Got:", student_na)
    )
  }
  
  # Identical values test (if lengths match)
  if (length(student_obj) == length(solution_obj)) {
    is_identical <- all.equal(student_obj, solution_obj)
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "identical_values",
      result = ifelse(isTRUE(is_identical), "PASS", "FAIL"),
      details = ifelse(
        isTRUE(is_identical),
        "Values are identical",
        paste("Differences found:", 
              ifelse(is.character(is_identical), 
                     substr(paste(is_identical, collapse = "; "), 1, 100), 
                     "Values differ"))
      )
    )
  }
  
  results
}

#' Test data frame objects
#' 
#' Tests: nrow, ncol, column names, column types, means/sds for numeric cols,
#' unique values for factors, correlation matrix
test_dataframe <- function(student_obj, solution_obj, obj_name) {
  results <- list()
  
  # Check if student object is a data frame
  if (!is.data.frame(student_obj)) {
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "is_dataframe",
      result = "FAIL",
      details = paste("Expected data.frame, got:", class(student_obj)[1])
    )
    return(results)
  }
  
  # nrow test
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "nrow",
    result = ifelse(nrow(student_obj) == nrow(solution_obj), "PASS", "FAIL"),
    details = paste("Expected:", nrow(solution_obj), "| Got:", nrow(student_obj))
  )
  
  # ncol test
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "ncol",
    result = ifelse(ncol(student_obj) == ncol(solution_obj), "PASS", "FAIL"),
    details = paste("Expected:", ncol(solution_obj), "| Got:", ncol(student_obj))
  )
  
  # Column names test
  solution_names <- names(solution_obj)
  student_names <- names(student_obj)
  names_match <- setequal(solution_names, student_names)
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "column_names",
    result = ifelse(names_match, "PASS", "FAIL"),
    details = ifelse(
      names_match,
      paste("Columns:", paste(solution_names, collapse = ", ")),
      paste("Missing:", paste(setdiff(solution_names, student_names), collapse = ", "),
            "| Extra:", paste(setdiff(student_names, solution_names), collapse = ", "))
    )
  )
  
  # Column types test
  if (names_match) {
    for (col_name in solution_names) {
      solution_type <- class(solution_obj[[col_name]])[1]
      student_type <- class(student_obj[[col_name]])[1]
      
      results[[length(results) + 1]] <- create_test_result(
        object = obj_name,
        test = paste0("column_type_", col_name),
        result = ifelse(solution_type == student_type, "PASS", "FAIL"),
        details = paste("Column", col_name, "- Expected:", solution_type, "| Got:", student_type)
      )
    }
    
    # For numeric columns, test mean and sd
    numeric_cols <- names(solution_obj)[sapply(solution_obj, is.numeric)]
    for (col_name in numeric_cols) {
      if (col_name %in% names(student_obj) && is.numeric(student_obj[[col_name]])) {
        solution_mean <- mean(solution_obj[[col_name]], na.rm = TRUE)
        student_mean <- mean(student_obj[[col_name]], na.rm = TRUE)
        
        results[[length(results) + 1]] <- create_test_result(
          object = obj_name,
          test = paste0("column_mean_", col_name),
          result = ifelse(
            abs(solution_mean - student_mean) < 1e-6 ||
              (is.na(solution_mean) && is.na(student_mean)),
            "PASS", "FAIL"
          ),
          details = paste("Column", col_name, "mean - Expected:", 
                          round(solution_mean, 4), "| Got:", round(student_mean, 4))
        )
        
        solution_sd <- sd(solution_obj[[col_name]], na.rm = TRUE)
        student_sd <- sd(student_obj[[col_name]], na.rm = TRUE)
        
        results[[length(results) + 1]] <- create_test_result(
          object = obj_name,
          test = paste0("column_sd_", col_name),
          result = ifelse(
            abs(solution_sd - student_sd) < 1e-6 ||
              (is.na(solution_sd) && is.na(student_sd)),
            "PASS", "FAIL"
          ),
          details = paste("Column", col_name, "sd - Expected:", 
                          round(solution_sd, 4), "| Got:", round(student_sd, 4))
        )
      }
    }
    
    # For factor columns, test unique values
    factor_cols <- names(solution_obj)[sapply(solution_obj, is.factor)]
    for (col_name in factor_cols) {
      if (col_name %in% names(student_obj) && is.factor(student_obj[[col_name]])) {
        solution_levels <- levels(solution_obj[[col_name]])
        student_levels <- levels(student_obj[[col_name]])
        levels_match <- setequal(solution_levels, student_levels)
        
        results[[length(results) + 1]] <- create_test_result(
          object = obj_name,
          test = paste0("factor_levels_", col_name),
          result = ifelse(levels_match, "PASS", "FAIL"),
          details = paste("Column", col_name, "levels match:", levels_match)
        )
      }
    }
    
    # Correlation matrix test for numeric columns (if at least 2 numeric cols)
    if (length(numeric_cols) >= 2) {
      tryCatch({
        solution_cor <- cor(solution_obj[, numeric_cols, drop = FALSE], use = "pairwise.complete.obs")
        student_cor <- cor(student_obj[, numeric_cols, drop = FALSE], use = "pairwise.complete.obs")
        
        cor_diff <- max(abs(solution_cor - student_cor), na.rm = TRUE)
        
        results[[length(results) + 1]] <- create_test_result(
          object = obj_name,
          test = "correlation_matrix",
          result = ifelse(cor_diff < 0.01, "PASS", "FAIL"),
          details = paste("Max correlation difference:", round(cor_diff, 6))
        )
      }, error = function(e) {
        results[[length(results) + 1]] <- create_test_result(
          object = obj_name,
          test = "correlation_matrix",
          result = "SKIP",
          details = paste("Could not compute correlation:", e$message)
        )
      })
    }
  }
  
  results
}

#' Test matrix objects
#' 
#' Tests: dimensions, row/col names, element-wise comparison, matrix properties
test_matrix <- function(student_obj, solution_obj, obj_name) {
  results <- list()
  
  # Check if student object is a matrix
  if (!is.matrix(student_obj)) {
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "is_matrix",
      result = "FAIL",
      details = paste("Expected matrix, got:", class(student_obj)[1])
    )
    return(results)
  }
  
  # Dimension tests
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "nrow",
    result = ifelse(nrow(student_obj) == nrow(solution_obj), "PASS", "FAIL"),
    details = paste("Expected:", nrow(solution_obj), "| Got:", nrow(student_obj))
  )
  
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "ncol",
    result = ifelse(ncol(student_obj) == ncol(solution_obj), "PASS", "FAIL"),
    details = paste("Expected:", ncol(solution_obj), "| Got:", ncol(student_obj))
  )
  
  # Row names test
  if (!is.null(rownames(solution_obj))) {
    rownames_match <- identical(rownames(solution_obj), rownames(student_obj))
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "rownames",
      result = ifelse(rownames_match, "PASS", "FAIL"),
      details = ifelse(rownames_match, "Row names match", "Row names differ")
    )
  }
  
  # Column names test
  if (!is.null(colnames(solution_obj))) {
    colnames_match <- identical(colnames(solution_obj), colnames(student_obj))
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "colnames",
      result = ifelse(colnames_match, "PASS", "FAIL"),
      details = ifelse(colnames_match, "Column names match", "Column names differ")
    )
  }
  
  # Element-wise comparison (if dimensions match)
  if (all(dim(student_obj) == dim(solution_obj))) {
    if (is.numeric(solution_obj) && is.numeric(student_obj)) {
      max_diff <- max(abs(solution_obj - student_obj), na.rm = TRUE)
      
      results[[length(results) + 1]] <- create_test_result(
        object = obj_name,
        test = "max_element_difference",
        result = ifelse(max_diff < 1e-6, "PASS", "FAIL"),
        details = paste("Max element difference:", max_diff)
      )
      
      # Mean of all elements
      solution_mean <- mean(solution_obj, na.rm = TRUE)
      student_mean <- mean(student_obj, na.rm = TRUE)
      
      results[[length(results) + 1]] <- create_test_result(
        object = obj_name,
        test = "overall_mean",
        result = ifelse(abs(solution_mean - student_mean) < 1e-6, "PASS", "FAIL"),
        details = paste("Expected:", round(solution_mean, 6), "| Got:", round(student_mean, 6))
      )
    }
  }
  
  results
}

#' Test list objects
#' 
#' Tests: length, element names, recursive element tests
test_list <- function(student_obj, solution_obj, obj_name) {
  results <- list()
  
  # Check if student object is a list
  if (!is.list(student_obj)) {
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "is_list",
      result = "FAIL",
      details = paste("Expected list, got:", class(student_obj)[1])
    )
    return(results)
  }
  
  # Length test
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "length",
    result = ifelse(length(student_obj) == length(solution_obj), "PASS", "FAIL"),
    details = paste("Expected:", length(solution_obj), "| Got:", length(student_obj))
  )
  
  # Names test
  solution_names <- names(solution_obj)
  student_names <- names(student_obj)
  
  if (!is.null(solution_names)) {
    names_match <- setequal(solution_names, student_names)
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "element_names",
      result = ifelse(names_match, "PASS", "FAIL"),
      details = ifelse(
        names_match,
        paste("Names:", paste(solution_names, collapse = ", ")),
        paste("Missing:", paste(setdiff(solution_names, student_names), collapse = ", "))
      )
    )
    
    # Test each named element
    for (elem_name in solution_names) {
      if (elem_name %in% student_names) {
        solution_elem <- solution_obj[[elem_name]]
        student_elem <- student_obj[[elem_name]]
        
        # Class test for element
        results[[length(results) + 1]] <- create_test_result(
          object = obj_name,
          test = paste0("element_class_", elem_name),
          result = ifelse(class(solution_elem)[1] == class(student_elem)[1], "PASS", "FAIL"),
          details = paste("Element", elem_name, "- Expected class:", 
                          class(solution_elem)[1], "| Got:", class(student_elem)[1])
        )
        
        # Length test for element
        if (is.atomic(solution_elem) || is.list(solution_elem)) {
          results[[length(results) + 1]] <- create_test_result(
            object = obj_name,
            test = paste0("element_length_", elem_name),
            result = ifelse(length(solution_elem) == length(student_elem), "PASS", "FAIL"),
            details = paste("Element", elem_name, "- Expected length:", 
                            length(solution_elem), "| Got:", length(student_elem))
          )
        }
        
        # For numeric elements, test mean
        if (is.numeric(solution_elem) && is.numeric(student_elem)) {
          solution_mean <- mean(solution_elem, na.rm = TRUE)
          student_mean <- mean(student_elem, na.rm = TRUE)
          
          results[[length(results) + 1]] <- create_test_result(
            object = obj_name,
            test = paste0("element_mean_", elem_name),
            result = ifelse(
              abs(solution_mean - student_mean) < 1e-6 ||
                (is.na(solution_mean) && is.na(student_mean)),
              "PASS", "FAIL"
            ),
            details = paste("Element", elem_name, "mean - Expected:", 
                            round(solution_mean, 4), "| Got:", round(student_mean, 4))
          )
        }
      }
    }
  }
  
  results
}

#' Test function objects
#' 
#' Tests: number of arguments, argument names, body structure
test_function <- function(student_obj, solution_obj, obj_name) {
  results <- list()
  
  # Check if student object is a function
  if (!is.function(student_obj)) {
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "is_function",
      result = "FAIL",
      details = paste("Expected function, got:", class(student_obj)[1])
    )
    return(results)
  }
  
  # Number of arguments test
  solution_args <- names(formals(solution_obj))
  student_args <- names(formals(student_obj))
  
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "num_arguments",
    result = ifelse(length(solution_args) == length(student_args), "PASS", "FAIL"),
    details = paste("Expected:", length(solution_args), "args | Got:", length(student_args), "args")
  )
  
  # Argument names test
  args_match <- setequal(solution_args, student_args)
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "argument_names",
    result = ifelse(args_match, "PASS", "FAIL"),
    details = ifelse(
      args_match,
      paste("Args:", paste(solution_args, collapse = ", ")),
      paste("Expected:", paste(solution_args, collapse = ", "),
            "| Got:", paste(student_args, collapse = ", "))
    )
  )
  
  # Body length test (rough complexity measure)
  solution_body_length <- length(deparse(body(solution_obj)))
  student_body_length <- length(deparse(body(student_obj)))
  
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "body_length",
    result = "INFO",
    details = paste("Solution body:", solution_body_length, "lines | Student body:", 
                    student_body_length, "lines")
  )
  
  results
}

#' Test factor objects
#' 
#' Tests: levels, number of levels, level order
test_factor <- function(student_obj, solution_obj, obj_name) {
  results <- list()
  
  # Check if student object is a factor
  if (!is.factor(student_obj)) {
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "is_factor",
      result = "FAIL",
      details = paste("Expected factor, got:", class(student_obj)[1])
    )
    return(results)
  }
  
  # Number of levels test
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "num_levels",
    result = ifelse(nlevels(solution_obj) == nlevels(student_obj), "PASS", "FAIL"),
    details = paste("Expected:", nlevels(solution_obj), "levels | Got:", nlevels(student_obj), "levels")
  )
  
  # Levels match test
  levels_match <- setequal(levels(solution_obj), levels(student_obj))
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "levels_match",
    result = ifelse(levels_match, "PASS", "FAIL"),
    details = ifelse(
      levels_match,
      paste("Levels:", paste(levels(solution_obj), collapse = ", ")),
      paste("Missing:", paste(setdiff(levels(solution_obj), levels(student_obj)), collapse = ", "))
    )
  )
  
  # Level order test
  order_match <- identical(levels(solution_obj), levels(student_obj))
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "levels_order",
    result = ifelse(order_match, "PASS", "FAIL"),
    details = ifelse(order_match, "Level order matches", "Level order differs")
  )
  
  # Length test
  results[[length(results) + 1]] <- create_test_result(
    object = obj_name,
    test = "length",
    result = ifelse(length(solution_obj) == length(student_obj), "PASS", "FAIL"),
    details = paste("Expected:", length(solution_obj), "| Got:", length(student_obj))
  )
  
  # Value frequency test
  solution_table <- table(solution_obj)
  student_table <- table(student_obj)
  
  if (levels_match) {
    # Safely compare frequency tables handling potential missing names
    common_names <- intersect(names(solution_table), names(student_table))
    all_names_present <- length(common_names) == length(names(solution_table))
    
    if (all_names_present) {
      freq_match <- all(solution_table[common_names] == student_table[common_names])
    } else {
      freq_match <- FALSE
    }
    
    results[[length(results) + 1]] <- create_test_result(
      object = obj_name,
      test = "level_frequencies",
      result = ifelse(freq_match, "PASS", "FAIL"),
      details = ifelse(freq_match, "Level frequencies match", "Level frequencies differ")
    )
  }
  
  results
}

#' Collect plots generated by a student script
#' 
#' @param student_file Path to student script
#' @param plots_output_dir Optional path to dedicated plots output directory
#' @return Character vector of plot file paths (as data URIs or file paths)
collect_student_plots <- function(student_file, plots_output_dir = NULL) {
  # This function collects plots generated during knitting
  
  # Helper function to find image files in a directory
  find_plot_files <- function(dir_path) {
    if (dir.exists(dir_path)) {
      list.files(dir_path, 
                 pattern = "\\.(png|jpg|jpeg|gif|svg)$",
                 full.names = TRUE,
                 recursive = TRUE)
    } else {
      character(0)
    }
  }
  
  if (grepl("\\.(Rmd|rmd|qmd|Qmd)$", student_file)) {
    student_name <- basename(student_file)
    student_base <- tools::file_path_sans_ext(student_name)
    
    # List of directories to check, in order of priority
    dirs_to_check <- c()
    
    # First, check the dedicated plots output directory if provided
    if (!is.null(plots_output_dir)) {
      # Object Testing layout: {plots_output_dir}/{student_base}/
      dirs_to_check <- c(dirs_to_check, file.path(plots_output_dir, student_base))
      # Autoharp render_one layout: {out_dir}/{student_base}_files/figure-html/
      dirs_to_check <- c(dirs_to_check,
                         file.path(plots_output_dir, paste0(student_base, "_files"), "figure-html"),
                         file.path(plots_output_dir, paste0(student_base, "_files")))
    }
    
    # Fallback: look in the standard _files directory created by rmarkdown/quarto
    base_name <- sub("\\.(Rmd|rmd|qmd|Qmd)$", "", student_file)
    dirs_to_check <- c(dirs_to_check, paste0(base_name, "_files"))
    
    # Also check for figure-html subdirectory (common pattern)
    dirs_to_check <- c(dirs_to_check, 
                       file.path(dirname(student_file), paste0(student_base, "_files"), "figure-html"))
    
    # Check each directory in order, return first non-empty result
    for (dir_path in dirs_to_check) {
      plot_files <- find_plot_files(dir_path)
      if (length(plot_files) > 0) {
        return(plot_files)
      }
    }
  }
  
  # Return empty vector if no plots found
  character(0)
}

#' Generate autoharp-compatible template with test chunks
#' 
#' @param original_content Character vector of original RMD file lines
#' @param solution_env Environment containing solution objects
#' @param objects_to_test Character vector of object names to test
#' @param test_prefix Prefix for test chunk names
#' @param include_all_tests Whether to include all comprehensive tests
#' @return Character vector of new template lines
generate_autoharp_template <- function(original_content, solution_env, 
                                       objects_to_test, test_prefix = "test",
                                       include_all_tests = TRUE) {
  
  # Start with original content
  template_lines <- original_content
  
  # Generate test chunks for each object type group
  test_chunks <- generate_test_chunks(
    solution_env = solution_env,
    objects_to_test = objects_to_test,
    test_prefix = test_prefix,
    include_all_tests = include_all_tests
  )
  
  # Add test chunks at the end of the document
  template_lines <- c(
    template_lines,
    "",
    "## Autoharp Test Chunks",
    "",
    "The following chunks contain automated tests for correctness checking.",
    "",
    test_chunks
  )
  
  template_lines
}

#' Generate test chunks for objects
#' 
#' @param solution_env Environment containing solution objects
#' @param objects_to_test Character vector of object names to test
#' @param test_prefix Prefix for test chunk names
#' @param include_all_tests Whether to include all comprehensive tests
#' @return Character vector of test chunk lines
generate_test_chunks <- function(solution_env, objects_to_test, 
                                 test_prefix = "test",
                                 include_all_tests = TRUE) {
  
  all_chunks <- character(0)
  chunk_counter <- 1
  
  for (obj_name in objects_to_test) {
    if (!exists(obj_name, envir = solution_env)) next
    
    obj <- get(obj_name, envir = solution_env)
    
    # Generate tests based on object type
    test_code <- generate_type_tests(obj, obj_name, include_all_tests)
    
    if (length(test_code$scalars) > 0) {
      # Create chunk header
      chunk_name <- paste0(test_prefix, "_", sprintf("%02d", chunk_counter), "_", obj_name)
      objs_str <- paste0('"', obj_name, '"')
      scalars_str <- paste0('"', test_code$scalars, '"', collapse = ", ")
      
      chunk_header <- paste0(
        "```{r ", chunk_name, 
        ", autoharp.objs=c(", objs_str, ")",
        ", autoharp.scalars=c(", scalars_str, ")}"
      )
      
      # Create full chunk
      chunk <- c(
        chunk_header,
        paste0("# Tests for ", obj_name, " (", class(obj)[1], ")"),
        test_code$code,
        "```",
        ""
      )
      
      all_chunks <- c(all_chunks, chunk)
      chunk_counter <- chunk_counter + 1
    }
  }
  
  all_chunks
}

#' Generate type-specific test code
#' 
#' @param obj The object to test
#' @param obj_name Name of the object
#' @param include_all_tests Whether to include all tests
#' @return List with code (character vector) and scalars (character vector)
generate_type_tests <- function(obj, obj_name, include_all_tests = TRUE) {
  
  if (is.vector(obj) && !is.list(obj)) {
    if (is.numeric(obj)) {
      return(generate_numeric_vector_tests(obj_name, include_all_tests))
    } else if (is.character(obj)) {
      return(generate_character_vector_tests(obj_name, include_all_tests))
    } else if (is.logical(obj)) {
      return(generate_logical_vector_tests(obj_name, include_all_tests))
    } else {
      return(generate_generic_vector_tests(obj_name, include_all_tests))
    }
  } else if (is.data.frame(obj)) {
    return(generate_dataframe_tests(obj, obj_name, include_all_tests))
  } else if (is.matrix(obj)) {
    return(generate_matrix_tests(obj, obj_name, include_all_tests))
  } else if (is.list(obj)) {
    return(generate_list_tests(obj, obj_name, include_all_tests))
  } else if (is.function(obj)) {
    return(generate_function_tests(obj, obj_name, include_all_tests))
  } else if (is.factor(obj)) {
    return(generate_factor_tests(obj_name, include_all_tests))
  } else {
    return(generate_generic_tests(obj_name, include_all_tests))
  }
}

#' Generate tests for numeric vectors
generate_numeric_vector_tests <- function(obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_length_check"),
    paste0(prefix, "_mean_check"),
    paste0(prefix, "_identical_check")
  )
  
  code <- c(
    paste0("# Length check"),
    paste0(prefix, "_length_check <- length(", obj_name, ") == length(", dot_name, ")"),
    "",
    paste0("# Mean check (with tolerance)"),
    paste0(prefix, "_mean_check <- isTRUE(all.equal("),
    paste0("  mean(", obj_name, ", na.rm = TRUE),"),
    paste0("  mean(", dot_name, ", na.rm = TRUE),"),
    paste0("  tolerance = ", DEFAULT_NUMERIC_TOLERANCE, "))"),
    "",
    paste0("# Identical values check"),
    paste0(prefix, "_identical_check <- isTRUE(all.equal(", obj_name, ", ", dot_name, "))")
  )
  
  if (include_all_tests) {
    scalars <- c(scalars,
                 paste0(prefix, "_sd_check"),
                 paste0(prefix, "_min_check"),
                 paste0(prefix, "_max_check"),
                 paste0(prefix, "_sum_check"),
                 paste0(prefix, "_na_count_check")
    )
    
    code <- c(code,
              "",
              paste0("# Standard deviation check"),
              paste0(prefix, "_sd_check <- isTRUE(all.equal("),
              paste0("  sd(", obj_name, ", na.rm = TRUE),"),
              paste0("  sd(", dot_name, ", na.rm = TRUE),"),
              paste0("  tolerance = ", DEFAULT_NUMERIC_TOLERANCE, "))"),
              "",
              paste0("# Min check"),
              paste0(prefix, "_min_check <- isTRUE(all.equal("),
              paste0("  min(", obj_name, ", na.rm = TRUE),"),
              paste0("  min(", dot_name, ", na.rm = TRUE),"),
              paste0("  tolerance = ", DEFAULT_NUMERIC_TOLERANCE, "))"),
              "",
              paste0("# Max check"),
              paste0(prefix, "_max_check <- isTRUE(all.equal("),
              paste0("  max(", obj_name, ", na.rm = TRUE),"),
              paste0("  max(", dot_name, ", na.rm = TRUE),"),
              paste0("  tolerance = ", DEFAULT_NUMERIC_TOLERANCE, "))"),
              "",
              paste0("# Sum check"),
              paste0(prefix, "_sum_check <- isTRUE(all.equal("),
              paste0("  sum(", obj_name, ", na.rm = TRUE),"),
              paste0("  sum(", dot_name, ", na.rm = TRUE),"),
              paste0("  tolerance = ", DEFAULT_NUMERIC_TOLERANCE, "))"),
              "",
              paste0("# NA count check"),
              paste0(prefix, "_na_count_check <- sum(is.na(", obj_name, ")) == sum(is.na(", dot_name, "))")
    )
  }
  
  list(code = code, scalars = scalars)
}

#' Generate tests for character vectors
generate_character_vector_tests <- function(obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_length_check"),
    paste0(prefix, "_identical_check")
  )
  
  code <- c(
    paste0("# Length check"),
    paste0(prefix, "_length_check <- length(", obj_name, ") == length(", dot_name, ")"),
    "",
    paste0("# Identical values check"),
    paste0(prefix, "_identical_check <- identical(", obj_name, ", ", dot_name, ")")
  )
  
  if (include_all_tests) {
    scalars <- c(scalars,
                 paste0(prefix, "_unique_count_check"),
                 paste0(prefix, "_nchar_total_check")
    )
    
    code <- c(code,
              "",
              paste0("# Unique count check"),
              paste0(prefix, "_unique_count_check <- length(unique(", obj_name, ")) == length(unique(", dot_name, "))"),
              "",
              paste0("# Total character count check"),
              paste0(prefix, "_nchar_total_check <- sum(nchar(", obj_name, ")) == sum(nchar(", dot_name, "))")
    )
  }
  
  list(code = code, scalars = scalars)
}

#' Generate tests for logical vectors
generate_logical_vector_tests <- function(obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_length_check"),
    paste0(prefix, "_identical_check"),
    paste0(prefix, "_sum_true_check")
  )
  
  code <- c(
    paste0("# Length check"),
    paste0(prefix, "_length_check <- length(", obj_name, ") == length(", dot_name, ")"),
    "",
    paste0("# Identical values check"),
    paste0(prefix, "_identical_check <- identical(", obj_name, ", ", dot_name, ")"),
    "",
    paste0("# Sum of TRUE values check"),
    paste0(prefix, "_sum_true_check <- sum(", obj_name, ", na.rm = TRUE) == sum(", dot_name, ", na.rm = TRUE)")
  )
  
  if (include_all_tests) {
    scalars <- c(scalars,
                 paste0(prefix, "_na_count_check")
    )
    
    code <- c(code,
              "",
              paste0("# NA count check"),
              paste0(prefix, "_na_count_check <- sum(is.na(", obj_name, ")) == sum(is.na(", dot_name, "))")
    )
  }
  
  list(code = code, scalars = scalars)
}

#' Generate tests for generic vectors
generate_generic_vector_tests <- function(obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_length_check"),
    paste0(prefix, "_class_check"),
    paste0(prefix, "_identical_check")
  )
  
  code <- c(
    paste0("# Length check"),
    paste0(prefix, "_length_check <- length(", obj_name, ") == length(", dot_name, ")"),
    "",
    paste0("# Class check"),
    paste0(prefix, "_class_check <- identical(class(", obj_name, "), class(", dot_name, "))"),
    "",
    paste0("# Identical values check"),
    paste0(prefix, "_identical_check <- isTRUE(all.equal(", obj_name, ", ", dot_name, "))")
  )
  
  list(code = code, scalars = scalars)
}

#' Generate tests for data frames
generate_dataframe_tests <- function(obj, obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_nrow_check"),
    paste0(prefix, "_ncol_check"),
    paste0(prefix, "_colnames_check")
  )
  
  code <- c(
    paste0("# Number of rows check"),
    paste0(prefix, "_nrow_check <- nrow(", obj_name, ") == nrow(", dot_name, ")"),
    "",
    paste0("# Number of columns check"),
    paste0(prefix, "_ncol_check <- ncol(", obj_name, ") == ncol(", dot_name, ")"),
    "",
    paste0("# Column names check"),
    paste0(prefix, "_colnames_check <- setequal(names(", obj_name, "), names(", dot_name, "))")
  )
  
  if (include_all_tests) {
    # Add column type checks
    col_names <- names(obj)
    numeric_cols <- col_names[sapply(obj, is.numeric)]
    
    scalars <- c(scalars, paste0(prefix, "_coltypes_check"))
    code <- c(code,
              "",
              paste0("# Column types check"),
              paste0(prefix, "_coltypes_check <- identical("),
              paste0("  sapply(", obj_name, ", class),"),
              paste0("  sapply(", dot_name, ", class))")
    )
    
    # Add numeric column mean checks (limit to MAX_ELEMENTS_TO_TEST)
    if (length(numeric_cols) > 0) {
      for (col in numeric_cols[1:min(MAX_ELEMENTS_TO_TEST, length(numeric_cols))]) {
        safe_col <- sanitize_name(col)
        scalar_name <- paste0(prefix, "_col_", safe_col, "_mean_check")
        scalars <- c(scalars, scalar_name)
        code <- c(code,
                  "",
                  paste0("# Mean check for column '", col, "'"),
                  paste0(scalar_name, " <- isTRUE(all.equal("),
                  paste0("  mean(", obj_name, "[['", col, "']], na.rm = TRUE),"),
                  paste0("  mean(", dot_name, "[['", col, "']], na.rm = TRUE),"),
                  paste0("  tolerance = ", DEFAULT_NUMERIC_TOLERANCE, "))")
        )
      }
    }
  }
  
  list(code = code, scalars = scalars)
}

#' Generate tests for matrices
generate_matrix_tests <- function(obj, obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_nrow_check"),
    paste0(prefix, "_ncol_check"),
    paste0(prefix, "_dim_check")
  )
  
  code <- c(
    paste0("# Number of rows check"),
    paste0(prefix, "_nrow_check <- nrow(", obj_name, ") == nrow(", dot_name, ")"),
    "",
    paste0("# Number of columns check"),
    paste0(prefix, "_ncol_check <- ncol(", obj_name, ") == ncol(", dot_name, ")"),
    "",
    paste0("# Dimension check"),
    paste0(prefix, "_dim_check <- identical(dim(", obj_name, "), dim(", dot_name, "))")
  )
  
  if (include_all_tests) {
    scalars <- c(scalars,
                 paste0(prefix, "_rownames_check"),
                 paste0(prefix, "_colnames_check"),
                 paste0(prefix, "_mean_check"),
                 paste0(prefix, "_max_diff_check")
    )
    
    code <- c(code,
              "",
              paste0("# Row names check"),
              paste0(prefix, "_rownames_check <- identical("),
              paste0("  rownames(", obj_name, "),"),
              paste0("  rownames(", dot_name, "))"),
              "",
              paste0("# Column names check"),
              paste0(prefix, "_colnames_check <- identical("),
              paste0("  colnames(", obj_name, "),"),
              paste0("  colnames(", dot_name, "))"),
              "",
              paste0("# Overall mean check"),
              paste0(prefix, "_mean_check <- isTRUE(all.equal("),
              paste0("  mean(", obj_name, ", na.rm = TRUE),"),
              paste0("  mean(", dot_name, ", na.rm = TRUE),"),
              paste0("  tolerance = ", DEFAULT_NUMERIC_TOLERANCE, "))"),
              "",
              paste0("# Max element difference check"),
              paste0(prefix, "_max_diff_check <- tryCatch({"),
              paste0("  max(abs(", obj_name, " - ", dot_name, "), na.rm = TRUE) < ", DEFAULT_NUMERIC_TOLERANCE),
              paste0("}, error = function(e) FALSE)")
    )
  }
  
  list(code = code, scalars = scalars)
}

#' Generate tests for lists
generate_list_tests <- function(obj, obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_length_check"),
    paste0(prefix, "_names_check")
  )
  
  code <- c(
    paste0("# Length check"),
    paste0(prefix, "_length_check <- length(", obj_name, ") == length(", dot_name, ")"),
    "",
    paste0("# Names check"),
    paste0(prefix, "_names_check <- setequal(names(", obj_name, "), names(", dot_name, "))")
  )
  
  if (include_all_tests && !is.null(names(obj))) {
    # Add element checks for first few named elements
    elem_names <- names(obj)[1:min(MAX_ELEMENTS_TO_TEST, length(names(obj)))]
    elem_names <- elem_names[!is.na(elem_names) & elem_names != ""]
    
    for (elem in elem_names) {
      safe_elem <- sanitize_name(elem)
      scalar_name <- paste0(prefix, "_elem_", safe_elem, "_check")
      scalars <- c(scalars, scalar_name)
      code <- c(code,
                "",
                paste0("# Element '", elem, "' check"),
                paste0(scalar_name, " <- isTRUE(all.equal("),
                paste0("  ", obj_name, "[['", elem, "']],"),
                paste0("  ", dot_name, "[['", elem, "']]))")
      )
    }
  }
  
  list(code = code, scalars = scalars)
}

#' Generate tests for functions
generate_function_tests <- function(obj, obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_num_args_check"),
    paste0(prefix, "_arg_names_check")
  )
  
  code <- c(
    paste0("# Number of arguments check"),
    paste0(prefix, "_num_args_check <- length(formals(", obj_name, ")) == length(formals(", dot_name, "))"),
    "",
    paste0("# Argument names check"),
    paste0(prefix, "_arg_names_check <- setequal("),
    paste0("  names(formals(", obj_name, ")),"),
    paste0("  names(formals(", dot_name, ")))")
  )
  
  if (include_all_tests) {
    scalars <- c(scalars, paste0(prefix, "_is_function_check"))
    code <- c(code,
              "",
              paste0("# Is function check"),
              paste0(prefix, "_is_function_check <- is.function(", obj_name, ")")
    )
  }
  
  list(code = code, scalars = scalars)
}

#' Generate tests for factors
generate_factor_tests <- function(obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_length_check"),
    paste0(prefix, "_nlevels_check"),
    paste0(prefix, "_levels_check")
  )
  
  code <- c(
    paste0("# Length check"),
    paste0(prefix, "_length_check <- length(", obj_name, ") == length(", dot_name, ")"),
    "",
    paste0("# Number of levels check"),
    paste0(prefix, "_nlevels_check <- nlevels(", obj_name, ") == nlevels(", dot_name, ")"),
    "",
    paste0("# Levels match check"),
    paste0(prefix, "_levels_check <- setequal(levels(", obj_name, "), levels(", dot_name, "))")
  )
  
  if (include_all_tests) {
    scalars <- c(scalars,
                 paste0(prefix, "_level_order_check"),
                 paste0(prefix, "_frequencies_check")
    )
    
    code <- c(code,
              "",
              paste0("# Level order check"),
              paste0(prefix, "_level_order_check <- identical("),
              paste0("  levels(", obj_name, "),"),
              paste0("  levels(", dot_name, "))"),
              "",
              paste0("# Level frequencies check"),
              paste0(prefix, "_frequencies_check <- identical("),
              paste0("  as.vector(table(", obj_name, ")),"),
              paste0("  as.vector(table(", dot_name, ")))")
    )
  }
  
  list(code = code, scalars = scalars)
}

#' Generate tests for generic objects
generate_generic_tests <- function(obj_name, include_all_tests = TRUE) {
  dot_name <- paste0(".", obj_name)
  prefix <- sanitize_name(obj_name)
  
  scalars <- c(
    paste0(prefix, "_class_check"),
    paste0(prefix, "_identical_check")
  )
  
  code <- c(
    paste0("# Class check"),
    paste0(prefix, "_class_check <- identical(class(", obj_name, "), class(", dot_name, "))"),
    "",
    paste0("# Identical check"),
    paste0(prefix, "_identical_check <- isTRUE(all.equal(", obj_name, ", ", dot_name, "))")
  )
  
  list(code = code, scalars = scalars)
}

# Student Script 7 - Similar to Script 1 (Plagiarism Case)

# Import mtcars dataset
data(mtcars)

# Create a linear regression model with weight and horsepower as predictors
regression_model <- lm(mpg ~ wt + hp, data = mtcars)

# Show the model summary statistics
summary(regression_model)

# Create prediction values from the model
predicted_mpg <- predict(regression_model, mtcars)

# Extract the residual values
residual_values <- residuals(regression_model)

# Generate a scatter plot with regression line
plot(mtcars$wt, mtcars$mpg, 
     main = "Fuel Efficiency vs Vehicle Weight",
     xlab = "Weight", ylab = "Miles per Gallon", 
     pch = 19, col = "blue")
abline(regression_model, col = "red", lwd = 2)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----solution-checker-files, echo=TRUE----------------------------------------
list.files(system.file("shiny", "solution_checker", package="autoharp"))

## ----solution-checker-db-creation, eval=FALSE---------------------------------
# create_db(credentials_data = credentials,
#           sqlite_path = "st2137.sqlite",
#           passphrase = Sys.getenv("AUTOHARP_TUNER_DB_KEY")
#           )

## ----eval=FALSE---------------------------------------------------------------
# soln_templates_dir <- "/home/viknesh/NUS/coursesTaught/autoharp/mytesting/secure_tuner/soln_templates"
# knit_wd <- "/home/viknesh/NUS/coursesTaught/autoharp/mytesting/secure_tuner/"
# permission_to_install <- FALSE
# max_time <- 120
# 
# summary_header <- "# Summary Output"
# tabs <- c("lint", "html", "correctness")
# app_title <- "R Solution Checker"
# corr_cols_to_drop = c(1,2,4,5)
# db_key <- Sys.getenv("AUTOHARP_TUNER_DB_KEY")

## LOAD LIBRARIES NEEDED HERE
library(autoharp)

## UTILITY FUNCTIONS
get_summary_output <- function (rmd_file, summary_header = "# Summary Output", 
                                dir = tempdir()){
  all_non_chunks <- extract_non_chunks(rmd_file)
  ind <- which(all_non_chunks == summary_header)
  if(length(ind) != 0) {
    display_chunks <- all_non_chunks[ind:length(all_non_chunks)]
    fp <- file.path(dir, 'summary_chunk_info.Rmd') 
    writeLines(display_chunks, con = fp)
    summary_df <- shiny::withMathJax(shiny::includeMarkdown(fp))
    unlink(fp)
  } else {
    summary_df <- h5(br(), "No summary found")
  }
  return(summary_df)
}

## INITIALISATION OF VARIABLES
soln_templates_dir <- "/home/viknesh/NUS/coursesTaught/autoharp/mytesting/secure_tuner/soln_templates"
knit_wd <- "/home/viknesh/NUS/coursesTaught/autoharp/mytesting/secure_tuner/"
tabs <- c("lint", "html", "correctness")
app_title <- "R Solution Checker"
permission_to_install <- FALSE
max_time <- 120
summary_header <- "# Summary Output"
corr_cols_to_drop = c(1,2,4,5)
db_key <- Sys.getenv("AUTOHARP_TUNER_DB_KEY")

soln_fnames <- list.files(soln_templates_dir, full.names = TRUE)
soln_choices <- sapply(soln_fnames, function(x) rmarkdown::yaml_front_matter(x)$title, 
                       USE.NAMES = FALSE)
if(anyDuplicated(soln_choices)){
  stop("Duplicate titles found in solution templates. Please resolve.")
}

### This is populating the soln env
soln_out_all <- lapply(soln_fnames, populate_soln_env, knit_root_dir = knit_wd)
names(soln_out_all) <- soln_choices
  
### This is populating the summary outputs 
chunk_out_all <- lapply(soln_fnames, get_summary_output)
names(chunk_out_all) <- soln_choices
  
  
### default list of lints
lint_list <- list(lintr::T_and_F_symbol_linter(),
                  lintr::assignment_linter(),
                  lintr::brace_linter(), 
                  lintr::commas_linter(), 
                  lintr::equals_na_linter(), 
                  lintr::function_left_parentheses_linter(), 
                  lintr::infix_spaces_linter(), 
                  lintr::line_length_linter(), 
                  lintr::whitespace_linter(),
                  lintr::absolute_path_linter(), 
                  lintr::pipe_continuation_linter(), 
                  lintr::spaces_inside_linter(), 
                  lintr::trailing_blank_lines_linter(), 
                  lintr::trailing_whitespace_linter(), 
                  lintr::unnecessary_concatenation_linter())

# Helper functions for Script Similarity Analysis App

# maximum number of TreeHarp objects to compare per file to prevent exponential computation
# tree kernel computation has O(n*m) complexity where n and m are the number of trees
# setting a reasonable limit prevents hanging on files with many expressions
MAX_AST_TREES <- 50

# timeout in seconds for each individual K2 tree kernel call
# prevents hanging on deeply nested or complex tree structures
K2_TIMEOUT_SECS <- 5

#' Extract code chunks from RMD or QMD files
#' 
#' @param file_path Path to RMD or QMD file
#' @return Character string containing all extracted code
extract_code_chunks <- function(file_path) {
  tryCatch({
    # Use autoharp's extract_chunks function if available
    if (requireNamespace("autoharp", quietly = TRUE)) {
      chunks <- autoharp::extract_chunks(file_path)
      
      if (is.null(chunks) || length(chunks) == 0) {
        warning(paste("No code chunks found in", file_path))
        return("")
      }
      
      # Extract code from each chunk (excluding chunk headers and footers)
      code_lines <- character(0)
      for (chunk in chunks) {
        # Chunks include header and footer, extract only the code
        if (length(chunk) > 2) {
          chunk_code <- chunk[2:(length(chunk) - 1)]
          code_lines <- c(code_lines, chunk_code)
        }
      }
      
      # Combine all code lines
      if (length(code_lines) > 0) {
        return(paste(code_lines, collapse = "\n"))
      } else {
        return("")
      }
    } else {
      # Fallback to knitr-based extraction
      lines <- readLines(file_path, warn = FALSE)
      
      if (!requireNamespace("knitr", quietly = TRUE)) {
        stop("knitr package is required for chunk extraction")
      }
      
      # Use base R grepl instead of stringr::str_detect
      begin_ids <- which(grepl(knitr::all_patterns$md$chunk.begin, lines))
      end_ids <- which(grepl(knitr::all_patterns$md$chunk.end, lines))
      
      if (length(begin_ids) == 0 || length(end_ids) == 0) {
        warning(paste("No code chunks found in", file_path))
        return("")
      }
      
      if (length(begin_ids) != length(end_ids)) {
        warning(paste("Malformed or mismatched code chunks in", file_path))
        return("")
      }
      
      code_lines <- character(0)
      for (i in seq_along(begin_ids)) {
        if (end_ids[i] > begin_ids[i] + 1) {
          chunk_code <- lines[(begin_ids[i] + 1):(end_ids[i] - 1)]
          code_lines <- c(code_lines, chunk_code)
        }
      }
      
      if (length(code_lines) > 0) {
        return(paste(code_lines, collapse = "\n"))
      } else {
        return("")
      }
    }
  }, error = function(e) {
    warning(paste("Error extracting code chunks from", file_path, ":", e$message))
    return("")
  })
}

#' Parse R script to tokens
#' 
#' @param file_path Path to R script
#' @param include_actuals Whether to include actual arguments/literals (default TRUE)
#' @param ngram_size Size of n-grams to generate (default 1 for unigrams)
#' @param exclude_library_calls Whether to exclude library function calls (default TRUE)
#' @return Character vector of tokens
parse_script_tokens <- function(file_path, include_actuals = TRUE, ngram_size = 1, exclude_library_calls = TRUE) {
  # Check if autoharp is available for more sophisticated token parsing
  if (requireNamespace("autoharp", quietly = TRUE)) {
    # Use autoharp's rmd_to_token_count for better token extraction
    token_df <- tryCatch({
      autoharp::rmd_to_token_count(file_path, include_actuals = include_actuals)
    }, error = function(e) {
      NULL
    })
    
    if (!is.null(token_df) && nrow(token_df) > 0) {
      # Extract tokens from the data frame
      tokens <- rep(token_df$token, token_df$n)
      
      # Exclude library function calls if requested
      if (exclude_library_calls && requireNamespace("autoharp", quietly = TRUE)) {
        # Get libraries used in the file
        libraries <- tryCatch({
          autoharp::get_libraries(file_path)
        }, error = function(e) {
          character(0)
        })
        
        if (length(libraries) > 0) {
          # Get function names from each library and exclude them from tokens
          library_fns <- character(0)
          for (lib in libraries) {
            tryCatch({
              lib_fns <- ls(envir=getNamespace(lib))
              library_fns <- c(library_fns, lib_fns)
            }, error = function(e) {
              NULL
            })
          }
          
          # remove library function names from tokens
          if (length(library_fns) > 0) {
            tokens <- tokens[!tokens %in% library_fns]
          }
        }
      }
      
      # Generate n-grams if requested
      if (ngram_size > 1 && length(tokens) >= ngram_size) {
        tokens <- generate_ngrams(tokens, ngram_size)
      }
      return(tokens)
    }
  }
}

#' Generate n-grams from a token vector
#' 
#' @param tokens Character vector of tokens
#' @param ngram_size Size of n-grams
#' @return Character vector of n-grams (concatenated with underscore)
generate_ngrams <- function(tokens, ngram_size = 2) {
  if (length(tokens) < ngram_size) {
    return(character(0))
  }
  
  ngrams <- character(length(tokens) - ngram_size + 1)
  for (i in 1:(length(tokens) - ngram_size + 1)) {
    ngrams[i] <- paste(tokens[i:(i + ngram_size - 1)], collapse = "_")
  }
  
  return(ngrams)
}

#' Calculate cosine similarity between two token vectors using TF-IDF
#' 
#' @param tokens1 First token vector
#' @param tokens2 Second token vector
#' @param all_tokens Optional list of all token vectors for IDF calculation
#' @return Cosine similarity score (0-1)
cosine_similarity <- function(tokens1, tokens2, all_tokens = NULL) {
  # Handle empty token vectors
  if (length(tokens1) == 0 || length(tokens2) == 0) {
    return(0)
  }
  
  # If all_tokens is provided, use TF-IDF weighting
  if (!is.null(all_tokens) && length(all_tokens) > 1) {
    # Create vocabulary from all documents
    vocab <- unique(unlist(all_tokens))
    n_docs <- length(all_tokens)
    
    # Calculate document frequencies (how many documents contain each term)
    doc_freq <- sapply(vocab, function(term) {
      sum(sapply(all_tokens, function(doc) term %in% doc))
    })
    
    # Calculate IDF: log(N / df)
    # Use pmax to ensure we never divide by zero (though this shouldn't happen with current logic)
    idf <- log(n_docs / pmax(doc_freq, 1))
    
    # Calculate TF for both documents
    tf1 <- sapply(vocab, function(t) sum(tokens1 == t))
    tf2 <- sapply(vocab, function(t) sum(tokens2 == t))
    
    # Apply TF-IDF weighting
    vec1 <- tf1 * idf
    vec2 <- tf2 * idf
  } else {
    # Fallback to simple term frequency if IDF data not available
    vocab <- unique(c(tokens1, tokens2))
    vec1 <- sapply(vocab, function(t) sum(tokens1 == t))
    vec2 <- sapply(vocab, function(t) sum(tokens2 == t))
  }
  
  # Calculate cosine similarity
  dot_product <- sum(vec1 * vec2)
  norm1 <- sqrt(sum(vec1^2))
  norm2 <- sqrt(sum(vec2^2))
  
  if (norm1 == 0 || norm2 == 0) return(0)
  
  similarity <- dot_product / (norm1 * norm2)
  return(similarity)
}

#' Compute cosine similarity matrix for all documents at once (vectorized)
#' 
#' This function computes the entire similarity matrix in one go using vectorized
#' operations, which is much faster than computing pairwise similarities one by one.
#' 
#' @param all_tokens List of token vectors for all documents
#' @param use_tfidf Whether to use TF-IDF weighting (default TRUE)
#' @return Similarity matrix (n x n)
cosine_similarity_matrix <- function(all_tokens, use_tfidf = TRUE) {
  n_docs <- length(all_tokens)
  
  # Create vocabulary from all documents
  vocab <- unique(unlist(all_tokens))
  n_terms <- length(vocab)
  
  if (n_terms == 0) {
    # Return matrix of zeros if no tokens
    sim_matrix <- matrix(0, nrow = n_docs, ncol = n_docs)
    diag(sim_matrix) <- 1
    return(sim_matrix)
  }
  
  # Build document-term matrix (DTM)
  # Rows = documents, Columns = terms
  dtm <- matrix(0, nrow = n_docs, ncol = n_terms)
  
  for (i in 1:n_docs) {
    if (length(all_tokens[[i]]) > 0) {
      # Count term frequencies for this document
      token_counts <- table(factor(all_tokens[[i]], levels = vocab))
      dtm[i, ] <- as.numeric(token_counts)
    }
  }
  
  # Apply TF-IDF weighting if requested
  # Note: TF-IDF requires at least 2 documents for meaningful IDF calculation
  # With n_docs=1, log(1/1)=0, making all TF-IDF values 0
  if (use_tfidf && n_docs > 1) {
    # Calculate document frequency (number of documents containing each term)
    doc_freq <- colSums(dtm > 0)
    
    # Calculate IDF: log(N / df)
    # Use pmax to ensure we never divide by zero
    idf <- log(n_docs / pmax(doc_freq, 1))
    
    # Apply IDF weighting to each row (document)
    # Use sweep to multiply each column by its IDF value
    dtm <- sweep(dtm, 2, idf, "*")
  }
  
  # compute cosine similarity matrix using vectorized operations
  # cosine(i, j) = dot(i, j) / (norm(i) * norm(j))
  
  # Step 1: Compute norms for each document (row)
  norms <- sqrt(rowSums(dtm^2))
  
  # Step 2: Handle zero-norm documents (empty documents)
  # documents with zero norm will have 0 similarity with all others (except self)
  zero_norm_docs <- norms == 0
  
  # Step 3: Compute all pairwise dot products at once
  # dtm %*% t(dtm) gives us all dot products in one operation
  dot_products <- tcrossprod(dtm)  # Equivalent to dtm %*% t(dtm) but faster
  
  # Step 4: Compute outer product of norms to get normalization factors
  norm_products <- outer(norms, norms)
  
  # Step 5: Compute cosine similarities
  # Avoid division by zero by setting zero denominators to 1
  # This will give 0 similarity (since numerator is also 0)
  norm_products[norm_products == 0] <- 1
  sim_matrix <- dot_products / norm_products
  
  # Handle any NaN values (shouldn't occur with above fix, but be safe)
  sim_matrix[is.nan(sim_matrix)] <- 0
  
  # Ensure diagonal is exactly 1 (self-similarity)
  # Even for zero-norm documents, self-similarity is defined as 1
  diag(sim_matrix) <- 1
  
  # Ensure symmetry (should already be symmetric, but numerical errors can occur)
  sim_matrix <- (sim_matrix + t(sim_matrix)) / 2
  
  return(sim_matrix)
}

#' Calculate Jaccard similarity between two token sets
#' 
#' @param tokens1 First token vector
#' @param tokens2 Second token vector
#' @return Jaccard similarity score (0-1)
jaccard_similarity <- function(tokens1, tokens2) {
  set1 <- unique(tokens1)
  set2 <- unique(tokens2)
  
  intersection <- length(intersect(set1, set2))
  union <- length(union(set1, set2))
  
  if (union == 0) return(0)
  
  return(intersection / union)
}

#' Calculate edit distance-based similarity
#' 
#' @param file1 First file path
#' @param file2 Second file path
#' @return Normalized edit distance similarity (0-1)
edit_distance_similarity <- function(file1, file2) {
  tryCatch({
    # Check if files are RMD or QMD
    file1_ext <- tolower(tools::file_ext(file1))
    file2_ext <- tolower(tools::file_ext(file2))
    
    # Extract code if RMD/QMD, otherwise read normally
    if (file1_ext %in% c("rmd", "qmd")) {
      code1 <- extract_code_chunks(file1)
    } else {
      code1 <- paste(readLines(file1, warn = FALSE), collapse = "\n")
    }
    
    if (file2_ext %in% c("rmd", "qmd")) {
      code2 <- extract_code_chunks(file2)
    } else {
      code2 <- paste(readLines(file2, warn = FALSE), collapse = "\n")
    }
    
    # Handle empty code
    if (nchar(code1) == 0 && nchar(code2) == 0) {
      return(1)  # Both empty, consider them identical
    }
    if (nchar(code1) == 0 || nchar(code2) == 0) {
      return(0)  # One empty, one not, consider them completely different
    }
    
    # Calculate Levenshtein distance
    distance <- adist(code1, code2)[1, 1]
    
    # Normalize by max length
    max_len <- max(nchar(code1), nchar(code2))
    if (max_len == 0) return(0)
    
    similarity <- 1 - (distance / max_len)
    return(max(0, similarity))
  }, error = function(e) {
    warning(paste("Error calculating edit distance:", e$message))
    return(0)
  })
}

#' Tree kernel function for comparing sets of trees
#' 
#' Computes normalized tree kernel similarity between two sets of TreeHarp objects.
#' This function handles the entire computation including pairwise kernel calculations
#' and normalization by self-similarities.
#' 
#' @param trees1 A list of TreeHarp objects or a single TreeHarp object
#' @param trees2 A list of TreeHarp objects or a single TreeHarp object
#' @return Normalized kernel similarity score (0-1)
tree_kernel <- function(trees1, trees2) {
  if (!requireNamespace("autoharp", quietly = TRUE)) {
    stop("autoharp package is required for tree kernel computation")
  }
  # Ensure inputs are lists
  if (!is.list(trees1)) {
    trees1 <- list(trees1)
  }
  if (!is.list(trees2)) {
    trees2 <- list(trees2)
  }

  # Handle empty inputs
  if (length(trees1) == 0 || length(trees2) == 0) {
    return(0)
  }

  # Safe wrapper around K2 that enforces a timeout per call
  safe_K2 <- function(t1, t2) {
    setTimeLimit(cpu = K2_TIMEOUT_SECS, elapsed = K2_TIMEOUT_SECS, transient = TRUE)
    on.exit(setTimeLimit(cpu = Inf, elapsed = Inf, transient = FALSE), add = TRUE)
    tryCatch(
      autoharp::K2(t1, t2, verbose = FALSE),
      error = function(e) {
        warning(paste("K2 call failed or timed out:", e$message))
        0
      }
    )
  }

  # Calculate pairwise kernel values and aggregate
  total_kernel <- 0
  for (t1 in trees1) {
    for (t2 in trees2) {
      total_kernel <- total_kernel + safe_K2(t1, t2)
    }
  }

  # Normalize by self-similarities
  norm1_sum <- 0
  for (t1 in trees1) {
    norm1_sum <- norm1_sum + safe_K2(t1, t1)
  }

  norm2_sum <- 0
  for (t2 in trees2) {
    norm2_sum <- norm2_sum + safe_K2(t2, t2)
  }

  # Return normalized similarity
  if (norm1_sum == 0 || norm2_sum == 0) {
    return(0)
  }

  similarity <- total_kernel / sqrt(norm1_sum * norm2_sum)
  return(max(0, min(1, similarity)))
}

#' Calculate AST-based similarity using autoharp's tree structures
#' 
#' @param file1 First file path
#' @param file2 Second file path
#' @return AST similarity score (0-1)
ast_similarity <- function(file1, file2) {
  print(file1)
  print(file2)
  tryCatch({
    # Check if files are RMD or QMD
    file1_ext <- tolower(tools::file_ext(file1))
    file2_ext <- tolower(tools::file_ext(file2))
    
    # Extract code if RMD/QMD, otherwise parse file directly
    if (file1_ext %in% c("rmd", "qmd")) {
      code1 <- extract_code_chunks(file1)
      # Handle empty code chunks (warning already issued by extract_code_chunks)
      if (code1 == "") {
        return(0)
      }
      expr1 <- parse(text = code1, keep.source = FALSE)
    } else {
      expr1 <- parse(file1, keep.source = FALSE)
    }
    
    if (file2_ext %in% c("rmd", "qmd")) {
      code2 <- extract_code_chunks(file2)
      # Handle empty code chunks (warning already issued by extract_code_chunks)
      if (code2 == "") {
        return(0)
      }
      expr2 <- parse(text = code2, keep.source = FALSE)
    } else {
      expr2 <- parse(file2, keep.source = FALSE)
    }
    
    # Try to use autoharp's tree kernel if available
    if (requireNamespace("autoharp", quietly = TRUE)) {
      # Convert each expression to trees and calculate similarities
      # Use TreeHarp constructor with quote_arg=TRUE for language objects
      trees1 <- lapply(seq_along(expr1), function(i) {
        tryCatch({
          autoharp::TreeHarp(expr1[[i]], quote_arg = TRUE)
        }, error = function(e) {
          warning(paste("Failed to create TreeHarp for expression in file1:", e$message))
          NULL
        })
      })
      
      trees2 <- lapply(seq_along(expr2), function(i) {
        tryCatch({
          autoharp::TreeHarp(expr2[[i]], quote_arg = TRUE)
        }, error = function(e) {
          warning(paste("Failed to create TreeHarp for expression in file2:", e$message))
          NULL
        })
      })
      
      # Filter out NULL values (failed conversions)
      trees1 <- Filter(Negate(is.null), trees1)
      trees2 <- Filter(Negate(is.null), trees2)

      # Filter out excessively large trees that can cause K2 to hang
      # Trees with too many nodes lead to exponential computation
      safe_tree_size <- function(tree) {
        tryCatch({
          n <- length(tree)
          !is.null(n) && n <= 500
        }, error = function(e) TRUE)
      }
      trees1 <- Filter(safe_tree_size, trees1)
      trees2 <- Filter(safe_tree_size, trees2)
      
      # Handle case where no valid trees were created
      if (length(trees1) == 0 || length(trees2) == 0) {
        warning("No valid TreeHarp objects created, falling back to token similarity")
        tokens1 <- parse_script_tokens(file1)
        tokens2 <- parse_script_tokens(file2)
        return(cosine_similarity(tokens1, tokens2))
      }
      
      # Limit the number of trees to prevent hanging on large files
      # If we have too many trees, the computation becomes exponentially expensive
      if (length(trees1) > MAX_AST_TREES || length(trees2) > MAX_AST_TREES) {
        warning(sprintf("Too many expressions detected (%d, %d). Limiting to %d trees per file to prevent hanging.",
                       length(trees1), length(trees2), MAX_AST_TREES))
        # Keep first MAX_AST_TREES trees
        # We take the first trees rather than random sampling to ensure deterministic results
        # The first expressions typically contain key initialization and setup code
        if (length(trees1) > MAX_AST_TREES) {
          trees1 <- trees1[1:MAX_AST_TREES]
        }
        if (length(trees2) > MAX_AST_TREES) {
          trees2 <- trees2[1:MAX_AST_TREES]
        }
      }
      
      # Use the improved tree_kernel function which handles sets of trees
      similarity <- tree_kernel(trees1, trees2)
      print(similarity)
      return(similarity)
    } else {
      # Fallback: use token-based similarity
      tokens1 <- parse_script_tokens(file1)
      tokens2 <- parse_script_tokens(file2)
      return(cosine_similarity(tokens1, tokens2))
    }
  }, error = function(e) {
    warning(paste("Error in AST similarity:", e$message))
    # Fallback to token-based
    tokens1 <- parse_script_tokens(file1)
    tokens2 <- parse_script_tokens(file2)
    return(cosine_similarity(tokens1, tokens2))
  })
}

#' calculate pairwise similarity matrix for all scripts
#' 
#' @param file_paths Character vector of file paths
#' @param method Similarity method: "cosine", "jaccard", "edit", "ast"
#' @param progress_callback Optional function to report progress (receives values 0-1)
#' @param ngram_size Size of n-grams to use for cosine similarity (default 1)
#' @param include_actuals Whether to include actual arguments/literals for cosine similarity (default TRUE)
#' @param exclude_library_calls Whether to exclude library function calls (default TRUE)
#' @param n_cores Number of cores for parallel computation (default 1; >1 only effective on Linux/Mac)
#' @return Similarity matrix
calculate_similarity_matrix <- function(file_paths, method = "cosine", 
                                       progress_callback = NULL,
                                       ngram_size = 1,
                                       include_actuals = TRUE,
                                       exclude_library_calls = TRUE,
                                       n_cores = 1) {
  n <- length(file_paths)
  sim_matrix <- matrix(0, nrow = n, ncol = n)
  rownames(sim_matrix) <- basename(file_paths)
  colnames(sim_matrix) <- basename(file_paths)
  
  # Diagonal is always 1 (self-similarity)
  diag(sim_matrix) <- 1
  
  # For cosine similarity, use optimized vectorized matrix computation
  if (method == "cosine") {
    # Pre-parse all tokens with the specified parameters
    if (!is.null(progress_callback)) {
      progress_callback(0.1)
    }
    
    all_tokens <- lapply(file_paths, function(fp) {
      parse_script_tokens(fp, include_actuals = include_actuals, ngram_size = ngram_size, 
                          exclude_library_calls = exclude_library_calls)
    })
    
    if (!is.null(progress_callback)) {
      progress_callback(0.3)
    }
    
    # use vectorized cosine similarity matrix computation
    # much faster than computing pairwise similarities one by one
    sim_matrix <- cosine_similarity_matrix(all_tokens, use_tfidf = TRUE)
    
    # set row and column names
    rownames(sim_matrix) <- basename(file_paths)
    colnames(sim_matrix) <- basename(file_paths)
    
    if (!is.null(progress_callback)) {
      progress_callback(1.0)
    }
    
    return(sim_matrix)
  }
  
  # Determine effective number of cores (parallel only supported on Linux/Mac)
  effective_cores <- if (.Platform$OS.type == "windows") {
    1L
  } else {
    max(1L, as.integer(n_cores))
  }
  
  # For other methods, use pairwise computation via mcmapply
  # Generate all upper-triangle pair indices at once
  combn_ids <- combn(n, 2)
  total_pairs <- ncol(combn_ids)
  
  compute_pair <- function(idx_i, idx_j) {
    switch(method,
      "jaccard" = {
        tokens1 <- parse_script_tokens(file_paths[idx_i])
        tokens2 <- parse_script_tokens(file_paths[idx_j])
        jaccard_similarity(tokens1, tokens2)
      },
      "edit" = {
        edit_distance_similarity(file_paths[idx_i], file_paths[idx_j])
      },
      "ast" = {
        ast_similarity(file_paths[idx_i], file_paths[idx_j])
      },
      stop("Unknown similarity method")
    )
  }
  
  if (effective_cores > 1) {
    # Parallel computation using mcmapply (Linux/Mac only)
    raw_result <- parallel::mcmapply(
      compute_pair,
      idx_i = combn_ids[1, ],
      idx_j = combn_ids[2, ],
      mc.cores = effective_cores
    )
    
    # mcmapply can return a list when errors occur in child processes;
    # coerce to numeric and replace any failures with 0
    similarities <- tryCatch(
      as.numeric(raw_result),
      warning = function(w) {
        warning(paste("Parallel similarity computation had conversion issues:", w$message))
        numeric(total_pairs)
      }
    )
    if (length(similarities) != total_pairs || anyNA(similarities)) {
      warning("Parallel computation returned unexpected results; replacing NA values with 0.")
      similarities <- numeric(total_pairs)
    }
    
    if (!is.null(progress_callback)) {
      progress_callback(1.0)
    }
  } else {
    # Sequential computation with progress reporting
    similarities <- numeric(total_pairs)
    for (k in seq_len(total_pairs)) {
      similarities[k] <- compute_pair(combn_ids[1, k], combn_ids[2, k])
      
      if (!is.null(progress_callback)) {
        progress_callback(k / total_pairs)
      }
    }
  }
  
  # Fill the similarity matrix from the computed values
  for (k in seq_len(total_pairs)) {
    i <- combn_ids[1, k]
    j <- combn_ids[2, k]
    sim_matrix[i, j] <- similarities[k]
    sim_matrix[j, i] <- similarities[k]
  }
  
  return(sim_matrix)
}

#' Convert similarity matrix to distance matrix
#' 
#' @param sim_matrix Similarity matrix (values 0-1)
#' @return Distance matrix
similarity_to_distance <- function(sim_matrix) {
  # Convert similarity (0-1) to distance
  # distance = 1 - similarity ensures 0 distance means identical
  dist_matrix <- 1 - sim_matrix
  
  # Ensure diagonal is exactly 0
  diag(dist_matrix) <- 0
  
  # Ensure symmetry (should already be symmetric)
  dist_matrix <- (dist_matrix + t(dist_matrix)) / 2
  
  return(dist_matrix)
}

#' Perform t-SNE dimensionality reduction
#' 
#' @param dist_matrix Distance matrix
#' @param perplexity t-SNE perplexity parameter
#' @param max_iter Maximum iterations
#' @return Data frame with 2D coordinates
perform_tsne <- function(dist_matrix, perplexity = NULL, max_iter = 1000) {
  if (!requireNamespace("Rtsne", quietly = TRUE)) {
    stop("Package 'Rtsne' is required for t-SNE. Install it with: install.packages('Rtsne')")
  }
  
  n <- nrow(dist_matrix)
  
  # Auto-adjust perplexity if not specified or invalid
  if (is.null(perplexity)) {
    perplexity <- min(30, max(1, floor((n - 1) / 3)))
  } else {
    # Ensure perplexity is valid
    perplexity <- min(perplexity, (n - 1) / 3)
    perplexity <- max(1, perplexity)
  }
  
  # Handle small datasets
  if (n < 4) {
    warning("Dataset too small for t-SNE. Using simple 2D projection.")
    # For very small datasets, use MDS or simple projection
    if (n == 2) {
      return(data.frame(
        X = c(0, dist_matrix[1, 2]),
        Y = c(0, 0),
        Script = rownames(dist_matrix)
      ))
    } else {
      # Use cmdscale for n=3
      mds_result <- cmdscale(as.dist(dist_matrix), k = 2)
      return(data.frame(
        X = mds_result[, 1],
        Y = mds_result[, 2],
        Script = rownames(dist_matrix)
      ))
    }
  }
  
  tryCatch({
    # Set seed for reproducibility
    set.seed(42)
    
    # Run t-SNE
    tsne_result <- Rtsne::Rtsne(
      as.dist(dist_matrix),
      dims = 2,
      perplexity = perplexity,
      max_iter = max_iter,
      is_distance = TRUE,
      check_duplicates = FALSE
    )
    
    result_df <- data.frame(
      X = tsne_result$Y[, 1],
      Y = tsne_result$Y[, 2],
      Script = rownames(dist_matrix)
    )
    
    return(result_df)
  }, error = function(e) {
    warning(paste("t-SNE failed:", e$message, "- falling back to MDS"))
    return(perform_mds(dist_matrix))
  })
}

#' Perform MDS (Multi-Dimensional Scaling)
#' 
#' @param dist_matrix Distance matrix
#' @return Data frame with 2D coordinates
perform_mds <- function(dist_matrix) {
  tryCatch({
    # classical MDS
    mds_result <- cmdscale(as.dist(dist_matrix), k = 2)
    
    result_df <- data.frame(
      X = mds_result[, 1],
      Y = mds_result[, 2],
      Script = rownames(dist_matrix)
    )
    
    return(result_df)
  }, error = function(e) {
    stop(paste("MDS failed:", e$message))
  })
}

#' Identify clusters of similar scripts
#' 
#' @param sim_matrix Similarity matrix
#' @param threshold Similarity threshold for clustering (0-1)
#' @return List of clusters (each cluster is a vector of script indices)
identify_clusters <- function(sim_matrix, threshold = 0.7) {
  n <- nrow(sim_matrix)
  
  # Create adjacency matrix based on threshold
  adj_matrix <- sim_matrix >= threshold
  diag(adj_matrix) <- FALSE  # Remove self-connections
  
  # Find connected components
  visited <- rep(FALSE, n)
  clusters <- list()
  
  for (i in 1:n) {
    if (!visited[i]) {
      # Start new cluster
      cluster <- c(i)
      visited[i] <- TRUE
      queue <- c(i)
      
      while (length(queue) > 0) {
        current <- queue[1]
        queue <- queue[-1]
        
        # Find neighbors
        neighbors <- which(adj_matrix[current, ] & !visited)
        
        if (length(neighbors) > 0) {
          cluster <- c(cluster, neighbors)
          visited[neighbors] <- TRUE
          queue <- c(queue, neighbors)
        }
      }
      
      # Only keep clusters with more than 1 member
      if (length(cluster) > 1) {
        clusters[[length(clusters) + 1]] <- cluster
      }
    }
  }
  
  # Assign cluster IDs
  cluster_ids <- rep(0, n)
  for (i in seq_along(clusters)) {
    cluster_ids[clusters[[i]]] <- i
  }
  
  return(list(
    clusters = clusters,
    cluster_ids = cluster_ids,
    labels = rownames(sim_matrix)
  ))
}

#' Get top N most similar pairs
#' 
#' @param sim_matrix Similarity matrix
#' @param n Number of pairs to return
#' @return Data frame with top similar pairs
get_top_similar_pairs <- function(sim_matrix, n = 10) {
  # get upper triangle (excluding diagonal)
  upper_tri_indices <- which(upper.tri(sim_matrix), arr.ind = TRUE)
  
  # get similarity values
  similarities <- sim_matrix[upper_tri_indices]
  
  # sort by similarity (descending)
  sorted_indices <- order(similarities, decreasing = TRUE)
  
  # take top n
  top_n <- min(n, length(sorted_indices))
  top_indices <- sorted_indices[1:top_n]
  
  result <- data.frame(
    Script1 = rownames(sim_matrix)[upper_tri_indices[top_indices, 1]],
    Script2 = rownames(sim_matrix)[upper_tri_indices[top_indices, 2]],
    Similarity = round(similarities[top_indices], 4),
    stringsAsFactors = FALSE
  )
  
  return(result)
}

#' Create a heatmap-friendly version of similarity matrix
#' 
#' @param sim_matrix Similarity matrix
#' @return Data frame in long format for heatmap
prepare_heatmap_data <- function(sim_matrix) {
  n <- nrow(sim_matrix)
  
  # Convert to long format
  heatmap_data <- data.frame(
    Script1 = rep(rownames(sim_matrix), each = n),
    Script2 = rep(rownames(sim_matrix), times = n),
    Similarity = as.vector(sim_matrix),
    stringsAsFactors = FALSE
  )
  
  return(heatmap_data)
}

#' calculate summary statistics for similarity matrix
#' 
#' @param sim_matrix Similarity matrix
#' @return List of summary statistics
calculate_similarity_stats <- function(sim_matrix) {
  # Get upper triangle (excluding diagonal)
  upper_tri <- sim_matrix[upper.tri(sim_matrix)]
  
  if (length(upper_tri) == 0) {
    return(list(
      mean = NA,
      median = NA,
      sd = NA,
      min = NA,
      max = NA,
      high_similarity_count = 0,
      medium_similarity_count = 0,
      low_similarity_count = 0
    ))
  }
  
  list(
    mean = mean(upper_tri, na.rm = TRUE),
    median = median(upper_tri, na.rm = TRUE),
    sd = sd(upper_tri, na.rm = TRUE),
    min = min(upper_tri, na.rm = TRUE),
    max = max(upper_tri, na.rm = TRUE),
    high_similarity_count = sum(upper_tri >= 0.7, na.rm = TRUE),
    medium_similarity_count = sum(upper_tri >= 0.4 & upper_tri < 0.7, na.rm = TRUE),
    low_similarity_count = sum(upper_tri < 0.4, na.rm = TRUE)
  )
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = FALSE,
  fig.align = "center"
)

## ----workflow-img, echo=FALSE, eval=TRUE, out.width="90%", fig.cap="The four-phase autoharp grading workflow: Prepare → Distribute → Grade → Review"----
knitr::include_graphics("figs/workflow.png")

## ----install, eval=FALSE------------------------------------------------------
# install.packages("autoharp")

## ----installdev---------------------------------------------------------------
# # install.packages("devtools")
# devtools::install_github("namanlab/autoharp")

## ----load---------------------------------------------------------------------
# library(autoharp)

## ----populate-----------------------------------------------------------------
# soln <- populate_soln_env("solution_template.Rmd", pattern = "test")
# 
# # soln is a list with two elements:
# # $env: the solution environment (contains .rf, .X and other solution objects)
# # $test_fname: path to the extracted test script (code from test chunks)
# str(soln)

## ----render-------------------------------------------------------------------
# result <- render_one(
#   rmd_name      = "student01.Rmd",
#   out_dir       = "output/",
#   knit_root_dir = getwd(),
#   soln_stuff    = soln
# )
# 
# # The result is a one-row data frame
# print(result)

## ----summarise----------------------------------------------------------------
# # Grade all students in a directory
# student_files <- list.files("submissions/", pattern = "\\.Rmd$", full.names = TRUE)
# 
# results_list <- lapply(student_files, function(f) {
#   render_one(rmd_name = f, out_dir = "output/", knit_root_dir = getwd(),
#              soln_stuff = soln)
# })
# 
# all_results <- do.call(rbind, results_list)
# 
# # Print a summary from the log file
# log_summary("output/render_one.log")

## ----lints--------------------------------------------------------------------
# # Count lint violations in a single script
# lint_count <- count_lints_one("student01.R")
# 
# # Count across all submissions
# all_lints <- count_lints_all(
#   file_names = list.files("submissions/", pattern = "\\.R$", full.names = TRUE)
# )
# print(all_lints)

## ----check-rmd----------------------------------------------------------------
# # Check that the submitted file is a valid Rmd
# # Returns TRUE if file has .Rmd extension, YAML header, and R chunks
# rmd_check <- check_rmd(fname = "student01.Rmd")
# print(rmd_check)

## ----grading-app, eval=FALSE--------------------------------------------------
# # Launch the full grading GUI
# shiny::runApp(system.file("shiny/grading_app", package = "autoharp"))

library(shiny)
library(shinymanager)

#For future update of tabs need to change from here
full_tabs = c("lint","html","correctness")
lint_tabs <- tabPanel("Lint Check",uiOutput(outputId = "lint_check"))
html_tabs <- tabPanel("HTML Check",uiOutput(outputId = "html_check"))
correctness_tabs <- tabPanel("Correctness check",tableOutput(outputId = "corr_check"))
list1 <- list(lint_tabs,html_tabs,correctness_tabs)
#Change up to here by adding new tabs to the list

index_match = match(tabs , full_tabs)
list_of_tabs <- list1[index_match]
internalwrapper <- function(...){
  tabsetPanel(..., id = NULL, selected = NULL, type ="tabs")
}
panelOutput <- do.call("internalwrapper",list_of_tabs)


ui <- fluidPage(
  titlePanel(title = app_title),
  
  fluidRow(
    column(4, wellPanel(
      selectInput(
        inputId = "selectfileweek",
        label = h4("Select tutorial number:"),
        choices = soln_choices
      ),
      fileInput(
        inputId = "fileupload",
        label = h4("Upload your solution:"),
        #accept = c(".Rmd", ".rmd", ".R", ".r"),
        width = '100%',
        multiple = FALSE
      ),
      actionButton("goButton", "Check solution!"),
      hr(),
      shiny::tags$small('autoharp solution checker, 2020.', br(),
                 'Found a bug? Report it',
                 #a(href='https://github.com/singator/autoharp/issues',
                 a(href='mailto:vik.gopal@nus.edu.sg','here.'))
      #style = "background: #ECF1F1;"
    )
    ),
    column(8, panelOutput)
  )
)

ui <- secure_app(ui, enable_admin = TRUE)

server <- function(input, output, session) {
  
  res_auth <- secure_server(
    check_credentials = check_credentials("st2137.sqlite", db_key)
  )
    
  lint_df <- reactive({
    input$goButton
    isolate({
      if(is.null(input$fileupload)){
        return(NULL)
      } else {
        file1 <- input$fileupload
        all_lints <- lintr::lint(file1$datapath, linters= lint_list)
        if(length(all_lints) > 0) {
          ll <- sapply(all_lints, function(x) x$line_number)
          mm <- sapply(all_lints, function(x) x$message)
          l_df <- data.frame(ll, mm, stringsAsFactors = FALSE)
          colnames(l_df) <- c("Line", "Message")
          l_df2 <- renderTable(l_df)
        } else {
          l_df2 <- h5(br(), "No lints found. Nice job!")
        }
      }
    })
    l_df2
  })
  
  sess_tmp_dir  <- tempfile("rmd_out")
  dir.create(sess_tmp_dir)
  # print(sess_tmp_dir)
  
  session$onSessionEnded(function() {
    unlink(sess_tmp_dir, TRUE)
  })
  
  output$lint_check <- renderUI({
    fluidPage(lint_df())
  })
  
  htmlUI <- reactive({
    input$goButton
    isolate({
      if(is.null(input$fileupload)) {
        return(NULL)
      } else {
        file1 <- input$fileupload
        #unlink(sess_tmp_dir, recursive = TRUE)
        
        progress <- shiny::Progress$new(session, min=1, max=10)
        on.exit(progress$close())
        progress$set(message="Checking libraries")
        
        lib_used <- get_libraries(file1$datapath)
        lib_used_msg <- paste('Libraries used: ',
                              paste0(lib_used, collapse=", "))
        lib_to_install <- 'Libraries to install: None'
        lib_error <- 'Installation logs: --'
        if(!is.null(lib_used)) {
          
          id <- !(sapply(lib_used, quietly=TRUE, requireNamespace))
          lib_needed <- lib_used[id]
          
          if(length(lib_needed) > 0) {
            
            if(!permission_to_install) {
              lib_error <- paste('Installation logs: Do Not Install')
              lib_to_install <- paste('Need to install:',
                                      paste0(lib_needed, collapse=", "))
              return(list(used=lib_used_msg, install=lib_to_install, 
                          error=lib_error, html=NULL, correctness=NULL))
            }
            
            lib_to_install <- paste('Need to install:',
                                    paste0(lib_needed, collapse=", "))
            showNotification("Installing packages, please wait a while..")
            progress$set(message="Installing libraries", value=2)
            
            try_install <- tryCatch(utils::install.packages(lib_needed,
                                                            dependencies = TRUE),
                                    error = function(e) return(e))
            #warning = function(w) return(w))
            #if("warning" %in% class(try_install) ||
            if("error" %in% class(try_install)) {
              lib_error <- paste('Installation logs:', conditionMessage(try_install))
              return(list(used=lib_used_msg, install=lib_to_install, error=lib_error,
                          html=NULL, correctness=NULL))
              #return(p(h5(lib_used_msg), h5(lib_to_install),
              #         h5(lib_error), hr()))
            } else {
              lib_error <- paste('Installation logs: All OK!')
            }
          }
        }
        soln_to_use <- input$selectfileweek
        
        progress$set(message="Rendering file", value=4)
        try_html <- tryCatch(render_one(file1$datapath, out_dir = sess_tmp_dir, 
                                        knit_root_dir = knit_wd, 
                                        max_time_per_run = max_time,
                                        soln_stuff = soln_out_all[[soln_to_use]]), 
                             error = function(e) return(e))
        progress$set(message="Almost done.. ", value=8)
        return(list(used=lib_used_msg, install=lib_to_install, error=lib_error, 
                    html=try_html))
      }
    })
  })
  
  output$html_check <- renderUI({
    if(is.null(htmlUI()$html)){
      return(p(h5(htmlUI()$used), h5(htmlUI()$install), h5(htmlUI()$error), hr()))
    } else {
  	if(!("data.frame" %in% class(htmlUI()$html))) {
        log_out <- log_summary(file.path(sess_tmp_dir, "render_one.log"))
  	  render_error <- paste('Render logs:', utils::tail(log_out$error_message, 1))
        return(p(h6(htmlUI()$used), h6(htmlUI()$install), h6(htmlUI()$error), 
                 hr(), h6(render_error)))
      }
      if(htmlUI()$html$run_status[1] == "FAIL") {
        log_out <- log_summary(file.path(sess_tmp_dir, "render_one.log"))
        render_error <- paste('Render logs:', utils::tail(log_out$error_message, 1))
        
        return(p(h6(htmlUI()$used), h6(htmlUI()$install), h6(htmlUI()$error), 
                 hr(), h6(render_error)))
      }
      addResourcePath("test", sess_tmp_dir)
      return(p(h6(htmlUI()$used), h6(htmlUI()$install), h6(htmlUI()$error), 
               hr(), 
               shiny::tags$iframe(src="test/0.html", height="1000", width="800", frameborder="0")))
    }
  })
  
  
  summary_df <- reactive({
    input$goButton
    isolate({
      if(is.null(input$fileupload)){ # see if can change to the file select instead
        return(NULL)
      } else {
        soln_to_use <- input$selectfileweek
        summary_df2 <- chunk_out_all[[soln_to_use]]
      }
    })
    summary_df2
  })
  
  
  output$corr_check <- renderUI({
    fluidPage(
      fluidRow(
        renderTable({
          if(is.null(htmlUI())){
            return(NULL)
          }
          if("data.frame" %in% class(htmlUI()$html)){
            if(ncol(htmlUI()$html) == 3) {
              return( htmlUI()$html[,3])
            } else {
              return(htmlUI()$html[,-corr_cols_to_drop])
            }
          }
          #return(NULL)
        })
      ),
      fluidRow(
        fluidPage(summary_df())
      )
    )
    
  })
  
}

# Create Shiny app ----
shinyApp(ui = ui, server = server)

# Student Script 8 - Matrix Operations

# Create matrices
A <- matrix(c(1, 2, 3, 4, 5, 6), nrow = 2, ncol = 3, byrow = TRUE)
B <- matrix(c(7, 8, 9, 10, 11, 12), nrow = 3, ncol = 2, byrow = TRUE)

# Matrix multiplication
C <- A %*% B
print("Matrix multiplication result:")
print(C)

# Transpose
A_transpose <- t(A)
print("Transpose of A:")
print(A_transpose)

# Create a square matrix
D <- matrix(c(4, 2, 2, 3), nrow = 2, ncol = 2)

# Calculate determinant
det_D <- det(D)
print(paste("Determinant of D:", det_D))

# Calculate inverse
D_inverse <- solve(D)
print("Inverse of D:")
print(D_inverse)

# Eigenvalues and eigenvectors
eigen_result <- eigen(D)
print("Eigenvalues:")
print(eigen_result$values)
print("Eigenvectors:")
print(eigen_result$vectors)

# Student Script 1 - Linear Regression Analysis

# Load data
data(mtcars)

# Fit linear model
model <- lm(mpg ~ wt + hp, data = mtcars)

# Get summary
summary(model)

# Make predictions
predictions <- predict(model, mtcars)

# Calculate residuals
residuals <- residuals(model)

# Plot results
plot(mtcars$wt, mtcars$mpg, main = "MPG vs Weight",
     xlab = "Weight", ylab = "MPG", pch = 19, col = "blue")
abline(model, col = "red", lwd = 2)

## Database creation and access
##
## Use this script to create an SQLite database. For the db passphrase, use the 
## environment variable that appears in R/globals.R.
## 
## The environment variable should be stored in .Renviron
## Currently, the name used is AUTOHARP_TUNER_DB_KEY
library(shinymanager)

credentials <- data.frame(
  user = c("st2137-user", "st2137-admin"),
  password = c("***", "***"), # please set your own password here!!!
  admin = c(FALSE, TRUE),
  stringsAsFactors = FALSE
)
create_db(credentials_data = credentials,
          sqlite_path = "st2137.sqlite",
          passphrase = Sys.getenv("AUTOHARP_TUNER_DB_KEY")
)

# Student Script 5 - Statistical Tests

# Load data
data(iris)

# T-test comparing two species
setosa_data <- iris$Sepal.Length[iris$Species == "setosa"]
versicolor_data <- iris$Sepal.Length[iris$Species == "versicolor"]

t_test_result <- t.test(setosa_data, versicolor_data)
print(t_test_result)

# ANOVA for all species
anova_result <- aov(Sepal.Length ~ Species, data = iris)
summary(anova_result)

# Correlation test
cor_test_result <- cor.test(iris$Sepal.Length, iris$Petal.Length)
print(cor_test_result)

# Chi-square test (create categorical data)
size_cat <- cut(iris$Sepal.Length, breaks = 3, labels = c("Small", "Medium", "Large"))
contingency_table <- table(size_cat, iris$Species)
chi_test <- chisq.test(contingency_table)
print(chi_test)

library(shiny)
library(DT)
library(Rtsne)
library(stringr)
library(plotly)

# check if shinyFiles is available
# if (!requireNamespace("shinyFiles", quietly = TRUE)) {
#   message("Note: shinyFiles package not available. Folder selection buttons will not be functional.")
#   SHINYFILES_AVAILABLE <- FALSE
# } else {
#   library(shinyFiles)
#   SHINYFILES_AVAILABLE <- TRUE
# }

# source("helpers.R")
SHINYFILES_AVAILABLE <- TRUE

# helper function to check if a path is absolute
is_absolute_path <- function(path) {
  grepl("^([A-Za-z]:|/|\\\\\\\\|//)", path)
}

# helper function to normalize a path
normalize_path_safe <- function(path) {
  if (dir.exists(path)) {
    normalizePath(path, winslash = "/", mustWork = TRUE)
  } else {
    normalizePath(path, winslash = "/", mustWork = FALSE)
  }
}

ui <- fluidPage(
  tags$head(
    tags$link(rel = "stylesheet", type = "text/css", href = "styles.css")
  ),
  
  div(class = "main-header",
    h2(icon("project-diagram"), " Script Similarity Analysis"),
    p("Explore similarity between student scripts (R, RMD, QMD) using t-SNE and MDS visualization")
  ),
  
  fluidRow(
    # left sidebar: Configuration
    column(
      width = 3,
      div(class = "sidebar-panel",
        h4(icon("cog"), " Configuration"),
        
        textInput(
          inputId = "scripts_folder",
          label = "Path to Scripts Folder:",
          value = "sol",
          placeholder = "sol",
          width = "100%"
        ),
        
        # add folder selection button if shinyFiles is available
        if (SHINYFILES_AVAILABLE) {
          shinyFiles::shinyDirButton("scripts_folder_select", "Browse...", 
                       "Select scripts folder",
                       class = "btn-secondary",
                       style = "width: 100%; margin-top: -10px;")
        } else {
          NULL
        },
        
        div(class = "divider"),
        
        h5(icon("ruler"), " Similarity Method"),
        selectInput(
          inputId = "similarity_method",
          label = NULL,
          choices = c(
            "Cosine (Token-based)" = "cosine",
            "Jaccard (Set-based)" = "jaccard",
            "Edit Distance" = "edit"
            # ,"AST (Tree-based)" = "ast"
          ),
          selected = "cosine",
          width = "100%"
        ),
        
        # cosine-specific options
        conditionalPanel(
          condition = "input.similarity_method == 'cosine'",
          sliderInput(
            inputId = "ngram_size",
            label = "N-gram Size:",
            min = 1,
            max = 5,
            value = 1,
            step = 1,
            width = "100%"
          ),
          checkboxInput(
            inputId = "include_actuals",
            label = "Include actual arguments/literals",
            value = TRUE,
            width = "100%"
          ),
          checkboxInput(
            inputId = "exclude_library_calls",
            label = "Exclude library function calls",
            value = TRUE,
            width = "100%"
          ),
          helpText("N-gram size: 1=unigrams, 2=bigrams, etc. Include actuals: whether to include literal values in token analysis. Exclude library calls: removes function calls from loaded packages to focus on code structure.")
        ),
        
        h5(icon("chart-scatter"), " Visualization Method"),
        selectInput(
          inputId = "viz_method",
          label = NULL,
          choices = c(
            "t-SNE" = "tsne",
            "MDS (Multi-Dimensional Scaling)" = "mds"
          ),
          selected = "tsne",
          width = "100%"
        ),
        
        # t-SNE specific options
        conditionalPanel(
          condition = "input.viz_method == 'tsne'",
          sliderInput(
            inputId = "tsne_perplexity",
            label = "t-SNE Perplexity:",
            min = 1,
            max = 50,
            value = 5,
            step = 1,
            width = "100%"
          )
        ),
        
        div(class = "divider"),
        
        h5(icon("filter"), " Clustering Options"),
        sliderInput(
          inputId = "cluster_threshold",
          label = "Similarity Threshold:",
          min = 0.5,
          max = 0.95,
          value = 0.7,
          step = 0.05,
          width = "100%"
        ),
        
        div(class = "divider"),
        
        h5(icon("microchip"), " Performance"),
        numericInput(
          inputId = "n_cores",
          label = "Number of Cores:",
          value = 1,
          min = 1,
          max = max(1L, parallel::detectCores(logical = FALSE), na.rm = TRUE),
          step = 1,
          width = "100%"
        ),
        helpText("Using multiple cores speeds up Jaccard and Edit Distance methods. Only effective on Linux and Mac (ignored on Windows)."),
        
        div(class = "divider"),
        
        actionButton("run_analysis", "Run Similarity Analysis", 
                    class = "btn-primary", 
                    style = "width: 100%;",
                    icon = icon("play")),
        
        br(), br(),
        
        downloadButton("download_results", "Download Results",
                      class = "btn-success",
                      style = "width: 100%")
      )
    ),
    
    # main content area
    column(
      width = 9,
      
      # status messages
      uiOutput("analysis_status"),
      
      # tabbed interface
      tabsetPanel(
        id = "main_tabs",
        type = "tabs",
        
        # Tab 1: Similarity Analysis
        tabPanel(
          title = "Similarity Analysis",
          icon = icon("chart-line"),
          value = "similarity_tab",
          
          br(),
          
          # Summary statistics
          div(class = "panel-section",
            h4(icon("chart-bar"), " Summary Statistics"),
            uiOutput("summary_stats")
          ),
          
          # Main visualization
          div(class = "panel-section",
            h4(icon("project-diagram"), " Similarity Visualization"),
            p("2D projection of script similarities. Points closer together indicate more similar scripts. Click on points to view scripts (multiple files can be opened)."),
            plotlyOutput("similarity_plot", height = "500px", width = "100%")
          ),
          
          # Similarity matrix heatmap
          div(class = "panel-section",
            h4(icon("th"), " Similarity Matrix Heatmap"),
            plotOutput("heatmap_plot", height = "500px", width = "100%")
          ),
          
          # Top similar pairs
          div(class = "panel-section",
            h4(icon("list-ol"), " Top Similar Script Pairs"),
            p("Scripts with highest similarity scores."),
            DTOutput("similar_pairs_table")
          ),
          
          # Detected clusters
          div(class = "panel-section",
            h4(icon("layer-group"), " Detected Similarity Clusters"),
            p(paste("Clusters of scripts with similarity above the threshold.",
                    "These may indicate potential plagiarism or similar approaches.")),
            uiOutput("clusters_info")
          ),
          
          # Full similarity matrix
          div(class = "panel-section",
            h4(icon("table"), " Complete Similarity Matrix"),
            DTOutput("similarity_matrix_table")
          )
        ),
        
        # Tab 2: Diff View
        tabPanel(
          title = "Diff View",
          icon = icon("columns"),
          value = "diff_tab",
          
          br(),
          
          # Diff view controls
          div(class = "panel-section",
            h4(icon("columns"), " Side-by-Side Script Comparison"),
            p("Select two scripts to compare them line-by-line."),
            
            fluidRow(
              column(6,
                selectInput(
                  inputId = "diff_file1",
                  label = "First Script:",
                  choices = NULL,
                  width = "100%"
                )
              ),
              column(6,
                selectInput(
                  inputId = "diff_file2",
                  label = "Second Script:",
                  choices = NULL,
                  width = "100%"
                )
              )
            ),
            
            fluidRow(
              column(12,
                actionButton("run_diff", "Compare Scripts", 
                           class = "btn-primary",
                           icon = icon("exchange-alt"))
              )
            )
          ),
          
          # Diff output
          div(class = "panel-section",
            h4(icon("file-code"), " Line-by-Line Comparison"),
            uiOutput("diff_output")
          )
        )
      )
    )
  )
)

server <- function(input, output, session) {
  
  # Helper function to generate safe button IDs from file keys
  generate_btn_id <- function(file_key) {
    paste0("close_", gsub("[^a-zA-Z0-9]", "_", file_key))
  }
  
  # Reactive values to store state
  rv <- reactiveValues(
    script_files = NULL,
    similarity_matrix = NULL,
    viz_coords = NULL,
    clusters = NULL,
    processing = FALSE,
    stats = NULL,
    open_files = list(),  # Store multiple open file contents
    file_observers = list()  # Track observers to clean them up
  )
  
  # Setup shinyFiles folder chooser if available
  if (SHINYFILES_AVAILABLE) {
    roots <- if (.Platform$OS.type == "windows") {
      shinyFiles::getVolumes()()
    } else {
      c(root = "/", home = path.expand("~"))
    }
    
    shinyFiles::shinyDirChoose(input, "scripts_folder_select", roots = roots, session = session)
    
    observeEvent(input$scripts_folder_select, {
      if (!is.integer(input$scripts_folder_select)) {
        selected_path <- shinyFiles::parseDirPath(roots, input$scripts_folder_select)
        if (length(selected_path) > 0) {
          updateTextInput(session, "scripts_folder", value = selected_path)
        }
      }
    })
  }
  
  # Load script files when folder path is provided
  observeEvent(input$scripts_folder, {
    req(input$scripts_folder)
    
    folder_path <- input$scripts_folder
    
    # Normalize path
    folder_path_normalized <- tryCatch({
      if (is_absolute_path(folder_path)) {
        normalize_path_safe(folder_path)
      } else {
        cwd <- normalizePath(getwd(), winslash = "/", mustWork = TRUE)
        candidate_path <- file.path(cwd, folder_path)
        normalize_path_safe(candidate_path)
      }
    }, error = function(e) {
      folder_path
    })
    
    if (dir.exists(folder_path_normalized)) {
      # Get all R, Rmd, and QMD files
      files <- list.files(folder_path_normalized, 
                          pattern = "\\.(R|r|Rmd|rmd|qmd|Qmd)$", 
                          full.names = TRUE)
      
      if (length(files) > 0) {
        rv$script_files <- files
        showNotification(paste("Found", length(files), "script files"), type = "message")
      } else {
        rv$script_files <- NULL
        showNotification("No R, Rmd, or QMD files found in the specified folder", type = "warning")
      }
    } else {
      rv$script_files <- NULL
      showNotification("Folder does not exist", type = "error")
    }
  })
  
  # Run similarity analysis
  observeEvent(input$run_analysis, {
    req(rv$script_files)
    req(input$similarity_method)
    req(input$viz_method)
    
    if (length(rv$script_files) < 2) {
      showNotification("Need at least 2 scripts for similarity analysis", type = "error")
      return()
    }
    
    rv$processing <- TRUE
    
    withProgress(message = "Analyzing script similarities...", value = 0, {
      
      tryCatch({
        # Step 1: Calculate similarity matrix
        setProgress(0.05, detail = "Calculating pairwise similarities...")
        
        # Get cosine-specific parameters (with defaults for other methods)
        ngram_size <- if (!is.null(input$ngram_size)) input$ngram_size else 1
        include_actuals <- if (!is.null(input$include_actuals)) input$include_actuals else TRUE
        exclude_library_calls <- if (!is.null(input$exclude_library_calls)) input$exclude_library_calls else TRUE
        n_cores <- if (!is.null(input$n_cores)) input$n_cores else 1
        
        sim_matrix <- calculate_similarity_matrix(
          rv$script_files,
          method = input$similarity_method,
          progress_callback = function(progress) {
            # Use setProgress for absolute positioning to avoid cumulative overshoot
            setProgress(0.05 + 0.55 * progress,
                        detail = paste("Processing pairs:", round(progress * 100), "%"))
          },
          ngram_size = ngram_size,
          include_actuals = include_actuals,
          exclude_library_calls = exclude_library_calls,
          n_cores = n_cores
        )
        
        rv$similarity_matrix <- sim_matrix
        
        # Step 2: Calculate statistics
        setProgress(0.65, detail = "Computing statistics...")
        rv$stats <- calculate_similarity_stats(sim_matrix)
        
        # Step 3: Identify clusters
        setProgress(0.75, detail = "Identifying clusters...")
        rv$clusters <- identify_clusters(sim_matrix, threshold = input$cluster_threshold)
        
        # Step 4: Perform dimensionality reduction
        setProgress(0.85, detail = paste("Running", toupper(input$viz_method), "..."))
        
        dist_matrix <- similarity_to_distance(sim_matrix)
        
        if (input$viz_method == "tsne") {
          perplexity <- input$tsne_perplexity
          rv$viz_coords <- perform_tsne(dist_matrix, perplexity = perplexity)
        } else {
          rv$viz_coords <- perform_mds(dist_matrix)
        }
        
        # Add cluster information to coordinates
        if (!is.null(rv$clusters$cluster_ids)) {
          rv$viz_coords$Cluster <- factor(rv$clusters$cluster_ids)
        }
        
        # Add file paths for click functionality
        rv$viz_coords$FilePath <- rv$script_files[match(rv$viz_coords$Script, basename(rv$script_files))]
        
        setProgress(1, detail = "Complete!")
        
        showNotification("Similarity analysis complete!", type = "message")
        
      }, error = function(e) {
        showNotification(paste("Error during analysis:", e$message), type = "error")
        print(e)
      })
      
      rv$processing <- FALSE
    })
  })
  
  # Status output
  output$analysis_status <- renderUI({
    if (rv$processing) {
      tags$div(
        class = "status-alert status-info",
        icon("spinner", class = "fa-spin"),
        " Processing script similarities..."
      )
    } else if (!is.null(rv$similarity_matrix)) {
      n_scripts <- nrow(rv$similarity_matrix)
      n_clusters <- length(rv$clusters$clusters)
      
      tags$div(
        class = "status-alert status-success",
        icon("check-circle"),
        paste(" Analysis complete:", n_scripts, "scripts analyzed,", 
              n_clusters, "similarity clusters detected")
      )
    } else if (!is.null(rv$script_files)) {
      tags$div(
        class = "status-alert status-info",
        icon("info-circle"),
        paste(" Ready to analyze", length(rv$script_files), 
              "scripts. Click 'Run Similarity Analysis' to begin.")
      )
    }
  })
  
  # Summary statistics
  output$summary_stats <- renderUI({
    req(rv$stats)
    
    stats <- rv$stats
    
    fluidRow(
      column(4,
        div(class = "stat-box",
          h5("Mean Similarity"),
          p(round(stats$mean, 3))
        )
      ),
      column(4,
        div(class = "stat-box",
          h5("Median Similarity"),
          p(round(stats$median, 3))
        )
      ),
      column(4,
        div(class = "stat-box",
          h5("Std Dev"),
          p(round(stats$sd, 3))
        )
      ),
      column(4,
        div(class = "stat-box",
          h5("High Similarity (≥0.7)"),
          p(stats$high_similarity_count)
        )
      ),
      column(4,
        div(class = "stat-box",
          h5("Medium Similarity (0.4-0.7)"),
          p(stats$medium_similarity_count)
        )
      ),
      column(4,
        div(class = "stat-box",
          h5("Low Similarity (<0.4)"),
          p(stats$low_similarity_count)
        )
      )
    )
  })
  
  # Similarity visualization plot
  output$similarity_plot <- renderPlotly({
    req(rv$viz_coords)
    
    coords <- rv$viz_coords
    
    # Determine colors based on clusters
    if ("Cluster" %in% names(coords) && length(unique(coords$Cluster)) > 1) {
      # Color by cluster
      cluster_levels <- unique(coords$Cluster)
      colors <- rainbow(length(cluster_levels), s = 0.6, v = 0.8)
      point_colors <- colors[as.numeric(coords$Cluster)]
      
      # Identify singleton clusters (cluster 0)
      singleton_mask <- coords$Cluster == 0
      point_colors[singleton_mask] <- "#cccccc"
    } else {
      # Single color if no clusters
      point_colors <- "#667eea"
    }
    
    # Create plotly plot
    p <- plot_ly(coords, 
            x = ~X, 
            y = ~Y, 
            text = ~Script,
            customdata = ~FilePath,
            type = 'scatter',
            mode = 'markers+text',
            marker = list(size = 12, color = point_colors, line = list(color = 'white', width = 1)),
            textposition = 'top center',
            textfont = list(size = 10),
            hovertemplate = paste('<b>%{text}</b><br>',
                                  'Click to view script<br>',
                                  '<extra></extra>'),
            source = "similarity_plot") %>%
      layout(
        title = list(
          text = paste(toupper(input$viz_method), "Projection of Script Similarities"),
          font = list(size = 16, weight = 'bold')
        ),
        xaxis = list(title = "Dimension 1", zeroline = FALSE),
        yaxis = list(title = "Dimension 2", zeroline = FALSE),
        hovermode = 'closest',
        showlegend = FALSE
      )
    
    event_register(p, 'plotly_click')
    p
  })
  
  # handle click events on the scatterplot
  observeEvent(event_data("plotly_click", source = "similarity_plot"), {
    click_data <- event_data("plotly_click", source = "similarity_plot")
    if (!is.null(click_data)) {
      # Get the point index (plotly uses 0-based indexing)
      point_index <- click_data$pointNumber + 1
      
      # Get the corresponding file path from the reactive value
      coords <- rv$viz_coords
      
      if (point_index <= nrow(coords)) {
        script_path <- coords$FilePath[point_index]
        script_name <- coords$Script[point_index]
        
        # Read the script content
        script_content <- tryCatch({
          paste(readLines(script_path, warn = FALSE), collapse = "\n")
        }, error = function(e) {
          paste("Error reading file:", e$message)
        })
        
        # Add to open files list (avoid duplicates)
        file_key <- script_name
        if (!(file_key %in% names(rv$open_files))) {
          rv$open_files[[file_key]] <- list(
            name = script_name,
            path = script_path,
            content = script_content
          )
        }
        
        # Show modal dialog with all open files
        show_open_files_modal()
      }
    }
  })
  
  # Helper function to display all open files in a modal
  show_open_files_modal <- function() {
    if (length(rv$open_files) == 0) return(NULL)
    
    # Create tabs for each open file
    file_tabs <- lapply(names(rv$open_files), function(file_key) {
      file_info <- rv$open_files[[file_key]]
      tabPanel(
        title = file_info$name,
        value = file_key,
        br(),
        tags$div(
          style = "position: relative;",
          actionButton(
            generate_btn_id(file_key),
            "Close this file",
            class = "btn-secondary btn-sm",
            style = "margin-bottom: 10px;"
          ),
          tags$pre(
            style = "max-height: 450px; overflow-y: auto; background-color: #f5f5f5; padding: 15px; border-radius: 5px; font-family: 'Courier New', monospace; font-size: 12px;",
            file_info$content
          )
        )
      )
    })
    
    # Show modal dialog with tabbed interface
    showModal(modalDialog(
      title = paste("Viewing", length(rv$open_files), "file(s) - Click on more points to add files"),
      do.call(tabsetPanel, c(list(id = "file_tabs"), file_tabs)),
      footer = tagList(
        actionButton("clear_all_files", "Close All Files", class = "btn-secondary"),
        modalButton("Done")
      ),
      size = "l",
      easyClose = FALSE
    ))
  }
  
  # Observer for close buttons - use a single observer with conditional logic
  observeEvent(rv$open_files, {
    # Clean up old observers with error handling
    lapply(rv$file_observers, function(obs) {
      tryCatch({
        if (!is.null(obs)) obs$destroy()
      }, error = function(e) {
        # Silently handle observer destruction errors
        NULL
      })
    })
    rv$file_observers <- list()
    
    # create new observers for current files
    if (length(rv$open_files) > 0) {
      rv$file_observers <- lapply(names(rv$open_files), function(file_key) {
        btn_id <- generate_btn_id(file_key)
        observeEvent(input[[btn_id]], {
          # Remove this file from open_files
          rv$open_files[[file_key]] <- NULL
          
          # If there are still files open, refresh the modal
          if (length(rv$open_files) > 0) {
            show_open_files_modal()
          } else {
            # Close the modal if no files left
            removeModal()
          }
        }, ignoreInit = TRUE, ignoreNULL = TRUE)
      })
    }
  }, ignoreNULL = FALSE)
  
  # Observer for clear all files button
  observeEvent(input$clear_all_files, {
    rv$open_files <- list()
    removeModal()
  }, ignoreInit = TRUE)
  
  # heatmap plot
  output$heatmap_plot <- renderPlot({
    req(rv$similarity_matrix)
    
    sim_matrix <- rv$similarity_matrix
    
    # Perform hierarchical clustering
    # Convert similarity to distance for clustering
    dist_matrix_for_hclust <- as.dist(1 - sim_matrix)
    hc <- hclust(dist_matrix_for_hclust, method = "complete")
    
    # Reorder the matrix based on clustering
    order_idx <- hc$order
    sim_matrix_ordered <- sim_matrix[order_idx, order_idx]
    
    # Create heatmap
    par(mar = c(8, 8, 2, 2))
    
    # Define color palette
    color_palette <- colorRampPalette(c("#fff5f0", "#fee0d2", "#fcbba1", 
                                       "#fc9272", "#fb6a4a", "#ef3b2c",
                                       "#cb181d", "#99000d"))(100)
    
    image(1:nrow(sim_matrix_ordered), 1:ncol(sim_matrix_ordered), sim_matrix_ordered,
          col = color_palette,
          xlab = "", ylab = "",
          axes = FALSE,
          main = "Pairwise Similarity Matrix (Hierarchically Clustered)",
          cex.main = 1.2,
          font.main = 2)
    
    # Add axes with reordered labels
    axis(1, at = 1:nrow(sim_matrix_ordered), labels = rownames(sim_matrix_ordered), 
         las = 2, cex.axis = 0.7)
    axis(2, at = 1:ncol(sim_matrix_ordered), labels = colnames(sim_matrix_ordered), 
         las = 2, cex.axis = 0.7)
    
    # Add color scale legend
    legend_values <- seq(0, 1, length.out = 5)
    legend("bottomright",
           legend = format(legend_values, digits = 2),
           fill = color_palette[seq(1, 100, length.out = 5)],
           title = "Similarity",
           cex = 0.8,
           bg = "white")
  })
  
  # top similar pairs table
  output$similar_pairs_table <- renderDT({
    req(rv$similarity_matrix)
    
    pairs <- get_top_similar_pairs(rv$similarity_matrix, n = 20)
    
    datatable(
      pairs,
      options = list(
        pageLength = 10,
        scrollX = TRUE,
        dom = 'Bfrtip',
        ordering = TRUE
      ),
      rownames = FALSE,
      class = 'cell-border stripe hover'
    ) %>%
      DT::formatStyle(
        'Similarity',
        background = DT::styleColorBar(c(0, 1), '#667eea'),
        backgroundSize = '100% 90%',
        backgroundRepeat = 'no-repeat',
        backgroundPosition = 'center'
      )
  })
  
  # Clusters information
  output$clusters_info <- renderUI({
    req(rv$clusters)
    
    clusters <- rv$clusters$clusters
    
    if (length(clusters) == 0) {
      return(tags$div(
        class = "status-alert status-info",
        icon("info-circle"),
        " No clusters detected with the current threshold. Try lowering the similarity threshold."
      ))
    }
    
    cluster_items <- lapply(seq_along(clusters), function(i) {
      cluster_scripts <- rv$clusters$labels[clusters[[i]]]
      
      tags$div(
        class = "stat-box",
        tags$h5(paste("Cluster", i, "-", length(cluster_scripts), "scripts")),
        tags$p(paste(cluster_scripts, collapse = ", "),
               style = "font-size: 14px; font-weight: normal;")
      )
    })
    
    do.call(tagList, cluster_items)
  })
  
  # full similarity matrix table
  output$similarity_matrix_table <- renderDT({
    req(rv$similarity_matrix)
    
    # Convert matrix to data frame
    sim_df <- as.data.frame(round(rv$similarity_matrix, 4))
    sim_df <- cbind(Script = rownames(sim_df), sim_df)
    
    datatable(
      sim_df,
      options = list(
        pageLength = 10,
        scrollX = TRUE,
        scrollY = "400px",
        dom = 'Bfrtip',
        ordering = TRUE
      ),
      rownames = FALSE,
      class = 'cell-border stripe hover'
    ) %>%
      DT::formatStyle(
        columns = 2:ncol(sim_df),
        background = DT::styleColorBar(c(0, 1), '#667eea'),
        backgroundSize = '100% 90%',
        backgroundRepeat = 'no-repeat',
        backgroundPosition = 'center'
      )
  })
  
  # Download handler
  output$download_results <- downloadHandler(
    filename = function() {
      paste0("similarity_analysis_", format(Sys.time(), "%Y%m%d_%H%M%S"), ".csv")
    },
    content = function(file) {
      req(rv$similarity_matrix)
      
      # Create comprehensive results
      sim_matrix <- rv$similarity_matrix
      
      # Get all pairwise similarities
      n <- nrow(sim_matrix)
      results_list <- list()
      
      for (i in 1:(n-1)) {
        for (j in (i+1):n) {
          results_list[[length(results_list) + 1]] <- data.frame(
            Script1 = rownames(sim_matrix)[i],
            Script2 = rownames(sim_matrix)[j],
            Similarity = sim_matrix[i, j],
            Method = input$similarity_method,
            stringsAsFactors = FALSE
          )
        }
      }
      
      results_df <- do.call(rbind, results_list)
      results_df <- results_df[order(results_df$Similarity, decreasing = TRUE), ]
      
      write.csv(results_df, file, row.names = FALSE)
    }
  )
  
  # Update file selectors when scripts are loaded
  observeEvent(rv$script_files, {
    if (!is.null(rv$script_files) && length(rv$script_files) > 0) {
      file_choices <- setNames(rv$script_files, basename(rv$script_files))
      
      updateSelectInput(session, "diff_file1", 
                       choices = file_choices,
                       selected = rv$script_files[1])
      
      updateSelectInput(session, "diff_file2", 
                       choices = file_choices,
                       selected = if(length(rv$script_files) > 1) rv$script_files[2] else rv$script_files[1])
    }
  })
  
  # Diff view logic
  observeEvent(input$run_diff, {
    req(input$diff_file1)
    req(input$diff_file2)
    
    file1 <- input$diff_file1
    file2 <- input$diff_file2
    
    if (file1 == file2) {
      showNotification("Please select two different files to compare", type = "warning")
      return()
    }
    
    tryCatch({
      # Read files
      lines1 <- readLines(file1, warn = FALSE)
      lines2 <- readLines(file2, warn = FALSE)
      
      # Create diff output
      output$diff_output <- renderUI({
        create_diff_view(lines1, lines2, basename(file1), basename(file2))
      })
      
      showNotification("Diff comparison complete", type = "message")
      
    }, error = function(e) {
      showNotification(paste("Error creating diff:", e$message), type = "error")
    })
  })
}

#' Create a side-by-side diff view
#' 
#' @param lines1 Lines from first file
#' @param lines2 Lines from second file
#' @param name1 Name of first file
#' @param name2 Name of second file
#' @return HTML UI elements for diff view
create_diff_view <- function(lines1, lines2, name1, name2) {
  # Perform simple line-by-line diff
  max_lines <- max(length(lines1), length(lines2))
  
  # Pad shorter file with empty lines
  if (length(lines1) < max_lines) {
    lines1 <- c(lines1, rep("", max_lines - length(lines1)))
  }
  if (length(lines2) < max_lines) {
    lines2 <- c(lines2, rep("", max_lines - length(lines2)))
  }
  
  # Build diff HTML
  diff_html <- tags$div(
    style = "display: flex; width: 100%; font-family: 'Courier New', monospace; font-size: 12px;",
    
    # Left panel (file 1)
    tags$div(
      style = "flex: 1; border-right: 2px solid #ccc; overflow-x: auto;",
      tags$div(
        style = "background-color: #e3f2fd; padding: 10px; font-weight: bold; border-bottom: 1px solid #ccc;",
        name1
      ),
      tags$div(
        id = "diff-scroll-left",
        style = "max-height: 600px; overflow-y: auto;",
        lapply(seq_along(lines1), function(i) {
          line1 <- lines1[i]
          line2 <- lines2[i]
          is_different <- line1 != line2
          
          bg_color <- if (is_different) {
            if (line1 == "") "#ffebee" else if (line2 == "") "#e8f5e9" else "#fff9c4"
          } else {
            "#ffffff"
          }
          
          tags$div(
            style = sprintf("padding: 2px 10px; background-color: %s; border-bottom: 1px solid #f0f0f0;", bg_color),
            tags$span(
              style = "display: inline-block; width: 40px; color: #999; text-align: right; margin-right: 10px;",
              i
            ),
            tags$span(
              style = "white-space: pre-wrap; word-break: break-all;",
              if (line1 == "") "\u00a0" else line1  # Non-breaking space for empty lines
            )
          )
        })
      )
    ),
    
    # Right panel (file 2)
    tags$div(
      style = "flex: 1; overflow-x: auto;",
      tags$div(
        style = "background-color: #e3f2fd; padding: 10px; font-weight: bold; border-bottom: 1px solid #ccc;",
        name2
      ),
      tags$div(
        id = "diff-scroll-right",
        style = "max-height: 600px; overflow-y: auto;",
        lapply(seq_along(lines2), function(i) {
          line1 <- lines1[i]
          line2 <- lines2[i]
          is_different <- line1 != line2
          
          bg_color <- if (is_different) {
            if (line2 == "") "#ffebee" else if (line1 == "") "#e8f5e9" else "#fff9c4"
          } else {
            "#ffffff"
          }
          
          tags$div(
            style = sprintf("padding: 2px 10px; background-color: %s; border-bottom: 1px solid #f0f0f0;", bg_color),
            tags$span(
              style = "display: inline-block; width: 40px; color: #999; text-align: right; margin-right: 10px;",
              i
            ),
            tags$span(
              style = "white-space: pre-wrap; word-break: break-all;",
              if (line2 == "") "\u00a0" else line2  # Non-breaking space for empty lines
            )
          )
        })
      )
    )
  )
  
  # Add legend
  legend <- tags$div(
    style = "margin-top: 20px; padding: 10px; background-color: #f5f5f5; border-radius: 5px;",
    tags$h5("Color Legend:"),
    tags$div(
      style = "display: flex; gap: 20px;",
      tags$div(
        tags$span(style = "display: inline-block; width: 20px; height: 20px; background-color: #fff9c4; border: 1px solid #ccc; vertical-align: middle;"),
        " Modified lines"
      ),
      tags$div(
        tags$span(style = "display: inline-block; width: 20px; height: 20px; background-color: #e8f5e9; border: 1px solid #ccc; vertical-align: middle;"),
        " Added lines"
      ),
      tags$div(
        tags$span(style = "display: inline-block; width: 20px; height: 20px; background-color: #ffebee; border: 1px solid #ccc; vertical-align: middle;"),
        " Removed lines"
      )
    )
  )
  
  # JavaScript to synchronise the scroll positions of the two panels
  sync_scroll_js <- tags$script(HTML(
    "(function() {
      var syncing = false;
      function syncScroll(source, target) {
        source.addEventListener('scroll', function() {
          if (!syncing) {
            syncing = true;
            target.scrollTop = source.scrollTop;
            syncing = false;
          }
        });
      }
      // Panels are rendered asynchronously; poll until both elements are in the DOM
      var attempts = 0;
      function attachSync() {
        var left  = document.getElementById('diff-scroll-left');
        var right = document.getElementById('diff-scroll-right');
        if (left && right) {
          syncScroll(left, right);
          syncScroll(right, left);
        } else if (attempts < 50) {
          attempts++;
          setTimeout(attachSync, 100);
        }
      }
      attachSync();
    })();"
  ))
  
  tagList(diff_html, legend, sync_scroll_js)
}

# Run the application
shinyApp(ui = ui, server = server)

# Student Script 3 - Data Manipulation with dplyr

library(dplyr)

# Load data
data(iris)

# Filter and summarize
result <- iris %>%
  filter(Species == "setosa") %>%
  group_by(Species) %>%
  summarize(
    mean_sepal_length = mean(Sepal.Length),
    mean_sepal_width = mean(Sepal.Width),
    count = n()
  )

# Display results
print(result)

# Create a new column
iris_modified <- iris %>%
  mutate(
    petal_ratio = Petal.Length / Petal.Width,
    size_category = ifelse(Sepal.Length > 6, "Large", "Small")
  )

# Summary statistics
summary(iris_modified)

library(shiny)
library(DT)
library(base64enc)
library(rmarkdown)
library(autoharp)
library(shinyFiles)

# Check if shinyFiles is available
# if (!requireNamespace("shinyFiles", quietly = TRUE)) {
#   message("Note: shinyFiles package not available. Folder selection buttons will not be functional.")
#   SHINYFILES_AVAILABLE <- FALSE
# } else {
#   library(shinyFiles)
#   SHINYFILES_AVAILABLE <- TRUE
# }

SHINYFILES_AVAILABLE <- TRUE
source(file.path("R", "helpers.R"))

# Helper function to check if a path is absolute
# Returns TRUE if the path starts with:
#   - Windows drive letter: C:, D:, etc.
#   - Unix absolute path: /
#   - UNC path: \\server or //server
is_absolute_path <- function(path) {
  # Pattern breakdown:
  # [A-Za-z]: - Windows drive letter (C:, D:, etc.)
  # / - Unix absolute path starting with /
  # \\\\\\\\ - UNC path starting with \\ (escaped for R)
  # // - UNC path starting with // (forward slash variant)
  grepl("^([A-Za-z]:|/|\\\\\\\\|//)", path)
}

# Helper function to normalize a path based on whether it exists
normalize_path_safe <- function(path) {
  if (dir.exists(path)) {
    normalizePath(path, winslash = "/", mustWork = TRUE)
  } else {
    normalizePath(path, winslash = "/", mustWork = FALSE)
  }
}


ui <- fluidPage(
  tags$head(
    tags$link(rel = "stylesheet", type = "text/css", href = "styles.css"),
    tags$script(HTML("
      // Function to open plot modal from button with data attributes
      function openPlotModalFromButton(button) {
        var imgSrc = button.getAttribute('data-img-src');
        var plotTitle = button.getAttribute('data-plot-title');
        var modal = document.getElementById('plotModal');
        var modalImg = document.getElementById('modalPlotImg');
        var modalTitle = document.getElementById('modalPlotTitle');
        modalImg.src = imgSrc;
        modalTitle.textContent = plotTitle;
        modal.classList.add('active');
      }
      
      // Function to close plot modal
      function closePlotModal() {
        var modal = document.getElementById('plotModal');
        modal.classList.remove('active');
      }
      
      // Move the modal to document.body on load so it escapes any
      // Shiny/Bootstrap stacking contexts that clip fixed elements
      document.addEventListener('DOMContentLoaded', function() {
        var modal = document.getElementById('plotModal');
        if (modal && modal.parentNode !== document.body) {
          document.body.appendChild(modal);
        }
        if (modal) {
          modal.addEventListener('click', function(e) {
            if (e.target === modal) {
              closePlotModal();
            }
          });
        }
      });
      
      // Close modal with Escape key
      document.addEventListener('keydown', function(e) {
        if (e.key === 'Escape') {
          closePlotModal();
        }
      });
    "))
  ),
  div(class = "main-header",
      h2("Autoharp Instructor Grading App"),
      p("Automated grading of R/Rmd/QMD student submissions")
  ),
  
  # Main content with tabs
  tabsetPanel(
    id = "main_tabs",
    type = "tabs",
    
    # Tab 0: Session Management
    tabPanel(
      "Session",
      value = "session",
      
      br(),
      
      fluidRow(
        column(
          width = 6,
          div(class = "panel-section",
              h4(icon("save"), " Load Previous Session"),
              p("Resume a previous grading session by uploading a session file."),
              
              fileInput(
                inputId = "session_file",
                label = "Upload Session File (.rds):",
                accept = c(".rds"),
                width = "100%"
              ),
              
              uiOutput("session_status"),
              
              actionButton("load_session", "Load Session", 
                           class = "btn-primary", 
                           icon = icon("upload"))
          )
        ),
        
        column(
          width = 6,
          div(class = "panel-section",
              h4(icon("info-circle"), " Current Session Info"),
              
              uiOutput("current_session_info"),
              
              br(),
              
              actionButton("save_session_manual", "Save Session Now", 
                           class = "btn-success", 
                           icon = icon("save")),
              
              p(class = "text-muted small", style = "margin-top: 10px;",
                "Sessions are auto-saved after grading operations. Use this to save manually.")
          )
        )
      )
    ),
    
    # Tab 1: Object Testing
    tabPanel(
      "Object Testing",
      value = "testing",
      
      br(),
      
      fluidRow(
        # Left sidebar - Setup
        column(
          width = 3,
          div(class = "sidebar-panel",
              h4(icon("cog"), " Setup"),
              
              fileInput(
                inputId = "solution_file",
                label = "Upload Solution Template (RMD/QMD):",
                accept = c(".Rmd", ".rmd", ".qmd", ".Qmd"),
                width = "100%"
              ),
              
              textInput(
                inputId = "student_folder",
                label = "Path to Student Scripts Folder:",
                placeholder = "../comprehensive_tests/soln_dir",
                width = "100%"
              ),
              
              # Add folder selection button if shinyFiles is available
              if (SHINYFILES_AVAILABLE) {
                shinyFiles::shinyDirButton("student_folder_select", "Browse...", 
                                           "Select student scripts folder",
                                           class = "btn-secondary",
                                           style = "width: 100%; margin-top: -10px;")
              } else {
                NULL
              },
              
              textInput(
                inputId = "output_folder",
                label = "Output Directory:",
                value = "grading_output",
                placeholder = "grading_output",
                width = "100%"
              ),
              
              p(class = "text-muted small", 
                "Directory for plots, results, and session files."),
              
              div(class = "divider"),
              
              h5(icon("list-check"), " Objects to Test"),
              div(class = "object-checkbox-group",
                  uiOutput("object_selector")
              ),
              
              div(class = "divider"),
              
              actionButton("run_grading", "Run Automated Grading", 
                           class = "btn-primary", 
                           style = "width: 100%;",
                           icon = icon("play")),
              
              br(), br(),
              
              downloadButton("download_csv", "Download Results as CSV",
                             class = "btn-success",
                             style = "width: 100%")
          )
        ),
        
        # Main content area
        column(
          width = 9,
          
          # Status messages
          uiOutput("grading_status"),
          
          # Summary Results
          div(class = "panel-section",
              h4(icon("chart-bar"), " Summary Results"),
              p("Summary of test results by student and object."),
              DTOutput("grading_results")
          ),
          
          div(class = "panel-section",
              h4(icon("list-alt"), " Detailed Results"),
              p("Detailed view of failed tests."),
              DTOutput("detailed_results")
          )
        )
      )
    ),
    
    # Tab 2: Template Download
    tabPanel(
      "Template Download",
      value = "download",
      
      br(),
      
      fluidRow(
        column(
          width = 3,
          div(class = "sidebar-panel",
              h4(icon("info-circle"), " Instructions"),
              
              p("This tab generates an autoharp-compatible solution template with test chunks."),
              
              tags$ol(
                tags$li("Upload a solution RMD file in the 'Object Testing' tab"),
                tags$li("Select which objects to include in tests"),
                tags$li("Click 'Generate Template' to create the autoharp version"),
                tags$li("Download the generated template")
              ),
              
              div(class = "divider"),
              
              h5(icon("cogs"), " Options"),
              
              textInput(
                inputId = "test_prefix",
                label = "Test chunk prefix:",
                value = "test",
                width = "100%"
              ),
              
              checkboxInput(
                inputId = "include_all_tests",
                label = "Include all comprehensive tests",
                value = TRUE
              ),
              
              div(class = "divider"),
              
              actionButton("generate_template", "Generate Template",
                           class = "btn-primary",
                           style = "width: 100%;",
                           icon = icon("magic")),
              
              br(), br(),
              
              downloadButton("download_template", "Download Template",
                             class = "btn-success",
                             style = "width: 100%")
          )
        ),
        
        # Main content - Preview
        column(
          width = 9,
          
          div(class = "panel-section",
              h4(icon("file-code"), " Generated Template Preview"),
              p("Preview of the autoharp-compatible solution template with test chunks."),
              
              uiOutput("template_status"),
              
              div(
                style = "background: #1e1e1e; color: #d4d4d4; padding: 20px; border-radius: 8px; font-family: 'Consolas', 'Monaco', monospace; font-size: 13px; max-height: 600px; overflow-y: auto; white-space: pre-wrap;",
                verbatimTextOutput("template_preview")
              )
          ),
          
          # tests documentation
          div(class = "panel-section",
              h4(icon("book"), " Generated Tests Documentation"),
              p("List of all tests that will be generated for each object type."),
              uiOutput("tests_documentation")
          )
        )
      )
    ),
    
    # Tab 3: Autoharp Grading
    tabPanel(
      "Autoharp Grading",
      value = "autoharp",
      
      br(),
      
      fluidRow(
        column(
          width = 3,
          div(class = "sidebar-panel",
              h4(icon("cogs"), " Autoharp Setup"),
              
              p("Use the autoharp render_one function to grade student scripts against a template."),
              
              div(class = "divider"),
              
              h5(icon("cog"), " Configuration"),
              
              # Template file upload
              fileInput(
                inputId = "autoharp_template_file",
                label = "Upload Template File (RMD/QMD):",
                accept = c(".Rmd", ".rmd", ".qmd", ".Qmd"),
                width = "100%"
              ),
              
              textInput(
                inputId = "autoharp_student_folder",
                label = "Path to Student Scripts Folder:",
                value = "sol",
                placeholder = "sol",
                width = "100%"
              ),
              
              # Add folder selection button if shinyFiles is available
              if (SHINYFILES_AVAILABLE) {
                shinyFiles::shinyDirButton("autoharp_student_folder_select", "Browse...", 
                                           "Select student scripts folder",
                                           class = "btn-secondary",
                                           style = "width: 100%; margin-top: -10px;")
              } else {
                NULL
              },
              
              textInput(
                inputId = "autoharp_output_folder",
                label = "Output Directory:",
                value = "student_out",
                placeholder = "student_out",
                width = "100%"
              ),
              
              p(class = "text-muted small", 
                "Directory for render_one output and session files."),
              
              # test pattern input
              textInput(
                inputId = "test_pattern",
                label = "Test Chunk Pattern:",
                value = "test",
                placeholder = "test",
                width = "100%"
              ),
              
              p(class = "text-muted small", 
                "Pattern to identify test chunks in template (default: 'test')."),
              
              # Checkbox for permission to install packages
              checkboxInput(
                inputId = "permission_to_install",
                label = "Allow package installation during grading",
                value = FALSE,
                width = "100%"
              ),
              
              p(class = "text-muted small", 
                "If checked, render_one will install missing packages automatically."),
              
              div(class = "divider"),
              
              actionButton("run_autoharp_grading", "Run Autoharp Grading", 
                           class = "btn-primary", 
                           style = "width: 100%;",
                           icon = icon("play")),
              
              br(), br(),
              
              downloadButton("download_autoharp_excel", "Download Results as Excel",
                             class = "btn-success",
                             style = "width: 100%")
          )
        ),
        
        # Main content area
        column(
          width = 9,
          
          # Status messages
          uiOutput("autoharp_status"),
          
          # Comprehensive Autoharp Results with Plot Grades
          div(class = "panel-section",
              h4(icon("table"), " Comprehensive Grading Results"),
              p("Combined results from autoharp render_one and manual plot grades."),
              DTOutput("autoharp_results")
          )
        )
      )
    ),
    
    # Tab 4: Plot Grading
    tabPanel(
      "Plot Grading",
      value = "plots",
      
      br(),
      
      fluidRow(
        column(
          width = 3,
          div(class = "sidebar-panel",
              h4(icon("user"), " Student Selection"),
              
              uiOutput("student_selector"),
              
              br(),
              
              fluidRow(
                column(6, actionButton("prev_student", icon("arrow-left"), 
                                       class = "btn-secondary",
                                       style = "width: 100%;")),
                column(6, actionButton("next_student", icon("arrow-right"), 
                                       class = "btn-secondary",
                                       style = "width: 100%;"))
              ),
              
              div(class = "divider"),
              
              h5(icon("star"), " Grading"),
              
              # Grade input
              numericInput("plot_grade", "Grade (0-100):", 
                           value = 0, min = 0, max = 100,
                           width = "100%"),
              
              textAreaInput("plot_comments", "Comments:", 
                            rows = 3, 
                            placeholder = "Optional feedback...",
                            width = "100%"),
              
              actionButton("save_grade", "Save Grade", 
                           class = "btn-success",
                           style = "width: 100%;",
                           icon = icon("save")),
              
              div(class = "divider"),
              
              h5(icon("tasks"), " Progress"),
              uiOutput("grading_progress")
          )
        ),
        
        # Main content area
        column(
          width = 9,
          
          # Student Plots
          div(class = "panel-section",
              h4(icon("image"), " Student Plots"),
              uiOutput("plot_display")
          ),
          
          # Test Results for current student
          div(class = "panel-section",
              h4(icon("clipboard-check"), " Test Results for Selected Student"),
              p("View all test results (built-in object checks) for the currently selected student."),
              DTOutput("student_test_results")
          ),
          
          # Plot grades summary
          div(class = "panel-section",
              h4(icon("table"), " Plot Grades Summary"),
              DTOutput("plot_grades_table")
          )
        )
      )
    )
  ),
  
  # Plot Modal - Global level to ensure proper z-index stacking
  tags$div(
    id = "plotModal",
    class = "plot-modal",
    tags$div(
      class = "plot-modal-content",
      tags$button(
        class = "plot-modal-close",
        onclick = "closePlotModal()",
        HTML("&times;")
      ),
      tags$h4(id = "modalPlotTitle", style = "margin-bottom: 15px; color: #495057;"),
      tags$img(id = "modalPlotImg", src = "", alt = "Expanded Plot")
    )
  )
)

# server def
server <- function(input, output, session) {
  
  # Constants for knitr temp directories and default figure path.
  # These are used when knitting solution files to avoid creating stray files
  # in the app's working directory.
  KNITR_DEFAULT_FIG_PATH <- "figure/"
  KNITR_SOLN_FIG_DIR     <- file.path(tempdir(), "soln_figs")
  KNITR_SOLN_FIG_RESTORE <- file.path(tempdir(), "soln_figs_restore")
  
  # reactive values to store state
  rv <- reactiveValues(
    solution_env = NULL,
    solution_objects = NULL,
    solution_file_content = NULL,
    solution_file_extension = NULL,  # store original file extension
    student_files = NULL,
    grading_results = NULL,
    plot_grades = NULL,
    current_student_index = 1,
    student_plots = list(),
    processing = FALSE,
    generated_template = NULL,
    # autoharp grading state
    autoharp_template_path = NULL,
    autoharp_student_files = NULL,
    autoharp_results = NULL,
    autoharp_processing = FALSE,
    # folder synchronization flag to prevent circular updates
    sync_in_progress = FALSE,
    # session management
    session_id = NULL,
    session_loaded = FALSE,
    restoring_session = FALSE,
    advancing_student = FALSE,
    last_save_time = NULL
  )
  
  # helper function to save current student's grade/comments into rv$plot_grades.
  save_current_grade <- function() {
    # skip during session restoration or auto-advance: inputs have stale values
    if (isTRUE(rv$restoring_session) || isTRUE(rv$advancing_student)) {
      return(invisible())
    }
    if (!is.null(rv$plot_grades) && !is.null(rv$current_student_index)) {
      if (rv$current_student_index > 0 && rv$current_student_index <= nrow(rv$plot_grades)) {
        grade_val <- input$plot_grade
        comment_val <- input$plot_comments
        rv$plot_grades$grade[rv$current_student_index] <- grade_val
        rv$plot_grades$comments[rv$current_student_index] <- if (is.null(comment_val)) "" else comment_val
        # Handle NA values to prevent "missing value where TRUE/FALSE needed"
        has_grade <- !is.null(grade_val) && !is.na(grade_val) && grade_val > 0
        has_comments <- !is.null(comment_val) && nchar(trimws(comment_val)) > 0
        if (has_grade || has_comments) {
          rv$plot_grades$graded[rv$current_student_index] <- TRUE
        }
      }
    }
  }
  
  # helper function to save session state.
  # When out_dir is provided (e.g. from autoharp grading), sessions are saved
  # there.  Otherwise falls back to the Object Testing output folder so there
  # is a consistent default location.
  save_session_state <- function(out_dir = NULL) {
    tryCatch({
      # Use the explicitly provided out_dir first, then the Object Testing
      # output folder, then the hard-coded default.
      output_dir <- if (!is.null(out_dir) && !is.na(out_dir) && out_dir != "") {
        out_dir
      } else if (!is.null(input$output_folder) && !is.na(input$output_folder) && input$output_folder != "") {
        input$output_folder
      } else {
        "grading_output"
      }
      
      # Create output directory if it doesn't exist
      if (!dir.exists(output_dir)) {
        dir.create(output_dir, recursive = TRUE)
      }
      
      # Create session subdirectory
      session_dir <- file.path(output_dir, "sessions")
      if (!dir.exists(session_dir)) {
        dir.create(session_dir, recursive = TRUE)
      }
      
      # Generate session ID if not exists
      if (is.null(rv$session_id)) {
        rv$session_id <- format(Sys.time(), "session_%Y%m%d_%H%M%S")
      }
      
      # Prepare session data
      session_data <- list(
        session_id = rv$session_id,
        timestamp = Sys.time(),
        solution_objects = rv$solution_objects,
        solution_file_content = rv$solution_file_content,
        solution_file_extension = rv$solution_file_extension,
        student_files = rv$student_files,
        grading_results = rv$grading_results,
        plot_grades = rv$plot_grades,
        current_student_index = rv$current_student_index,
        student_plots = rv$student_plots,
        generated_template = rv$generated_template,
        autoharp_template_path = rv$autoharp_template_path,
        autoharp_student_files = rv$autoharp_student_files,
        autoharp_results = rv$autoharp_results,
        # Save input values
        student_folder = input$student_folder,
        output_folder = output_dir,
        autoharp_student_folder = input$autoharp_student_folder,
        autoharp_output_folder = input$autoharp_output_folder,
        test_pattern = input$test_pattern
      )
      
      # Save to file
      session_file <- file.path(session_dir, paste0(rv$session_id, ".rds"))
      saveRDS(session_data, session_file)
      rv$last_save_time <- Sys.time()
      
      return(list(success = TRUE, file = session_file))
    }, error = function(e) {
      return(list(success = FALSE, error = e$message))
    })
  }
  
  # Helper function to load session state
  load_session_state <- function(session_file) {
    tryCatch({
      session_data <- readRDS(session_file)
      
      # Guard: prevent save_current_grade and student_folder observer from
      # clobbering restored data while we are restoring reactive values
      rv$restoring_session <- TRUE
      rv$session_loaded <- TRUE
      
      # Restore reactive values
      rv$session_id <- session_data$session_id
      rv$solution_objects <- session_data$solution_objects
      rv$solution_file_content <- session_data$solution_file_content
      rv$solution_file_extension <- session_data$solution_file_extension
      rv$student_files <- session_data$student_files
      rv$grading_results <- session_data$grading_results
      rv$plot_grades <- session_data$plot_grades
      rv$current_student_index <- session_data$current_student_index
      rv$student_plots <- session_data$student_plots
      rv$generated_template <- session_data$generated_template
      rv$autoharp_template_path <- session_data$autoharp_template_path
      rv$autoharp_student_files <- session_data$autoharp_student_files
      rv$autoharp_results <- session_data$autoharp_results
      
      # Restore input values (these trigger observers asynchronously)
      updateTextInput(session, "student_folder", value = session_data$student_folder)
      updateTextInput(session, "output_folder", value = session_data$output_folder)
      updateTextInput(session, "autoharp_student_folder", value = session_data$autoharp_student_folder)
      updateTextInput(session, "autoharp_output_folder", value = session_data$autoharp_output_folder)
      updateTextInput(session, "test_pattern", value = session_data$test_pattern)
      
      # Restore the grade/comment inputs for the current student
      if (!is.null(session_data$plot_grades) && !is.null(session_data$current_student_index)) {
        idx <- session_data$current_student_index
        if (idx > 0 && idx <= nrow(session_data$plot_grades)) {
          saved_grade <- session_data$plot_grades$grade[idx]
          saved_comment <- session_data$plot_grades$comments[idx]
          updateNumericInput(session, "plot_grade",
                             value = if (is.na(saved_grade)) 0 else saved_grade)
          updateTextAreaInput(session, "plot_comments",
                              value = if (is.null(saved_comment)) "" else saved_comment)
        }
      }
      
      # Restore solution environment if solution was loaded
      if (!is.null(rv$solution_file_content)) {
        temp_env <- new.env()
        temp_file <- tempfile(fileext = rv$solution_file_extension)
        writeLines(rv$solution_file_content, temp_file)
        
        # Direct knitr output to a tempfile and redirect fig.path to avoid
        # creating stray files in the app's working directory.
        temp_md <- tempfile(fileext = ".md")
        old_fig_path <- knitr::opts_chunk$get("fig.path")
        knitr::opts_chunk$set(fig.path = file.path(KNITR_SOLN_FIG_RESTORE, "fig-"))
        tryCatch({
          knitr::knit(temp_file, output = temp_md, envir = temp_env, quiet = TRUE)
        }, finally = {
          knitr::opts_chunk$set(
            fig.path = if (!is.null(old_fig_path)) old_fig_path else KNITR_DEFAULT_FIG_PATH)
          if (file.exists(temp_md)) unlink(temp_md)
        })
        
        rv$solution_env <- temp_env
        unlink(temp_file)
      }
      
      # Clear the restoring flag after a short delay so that any
      # observers triggered by the updateTextInput calls above still
      # see the flag.  session$onFlushed runs after the current flush
      # cycle completes, i.e. after all pending observers have fired.
      session$onFlushed(function() {
        rv$restoring_session <- FALSE
      }, once = TRUE)
      
      return(list(success = TRUE, session_id = session_data$session_id, timestamp = session_data$timestamp))
    }, error = function(e) {
      return(list(success = FALSE, error = e$message))
    })
  }
  
  # Setup shinyFiles folder choosers if available
  if (SHINYFILES_AVAILABLE) {
    # Allow browsing from root directory (/) for Unix-like systems
    # or from all available volumes for Windows
    roots <- if (.Platform$OS.type == "windows") {
      # getVolumes() returns a function that, when called, returns the volumes
      # This double call pattern is by design in the shinyFiles package
      shinyFiles::getVolumes()()
    } else {
      c(root = "/", home = path.expand("~"))
    }
    
    # Setup folder choosers
    shinyFiles::shinyDirChoose(input, "student_folder_select", roots = roots, session = session)
    shinyFiles::shinyDirChoose(input, "autoharp_student_folder_select", roots = roots, session = session)
    
    # Observer for Object Testing tab folder selection
    observeEvent(input$student_folder_select, {
      if (!is.integer(input$student_folder_select)) {
        selected_path <- shinyFiles::parseDirPath(roots, input$student_folder_select)
        if (length(selected_path) > 0) {
          updateTextInput(session, "student_folder", value = selected_path)
          # Also update the Autoharp Grading tab folder input
          updateTextInput(session, "autoharp_student_folder", value = selected_path)
        }
      }
    })
    
    # Observer for Autoharp Grading tab folder selection
    observeEvent(input$autoharp_student_folder_select, {
      if (!is.integer(input$autoharp_student_folder_select)) {
        selected_path <- shinyFiles::parseDirPath(roots, input$autoharp_student_folder_select)
        if (length(selected_path) > 0) {
          updateTextInput(session, "autoharp_student_folder", value = selected_path)
        }
      }
    })
  }
  
  # Session management observers
  
  # Load session button
  observeEvent(input$load_session, {
    req(input$session_file)
    
    result <- load_session_state(input$session_file$datapath)
    
    if (result$success) {
      showNotification(
        paste0("Session loaded successfully! ID: ", result$session_id),
        type = "message",
        duration = 5
      )
    } else {
      showNotification(
        paste0("Error loading session: ", result$error),
        type = "error",
        duration = 10
      )
    }
  })
  
  # Manual save session button
  observeEvent(input$save_session_manual, {
    result <- save_session_state()
    
    if (result$success) {
      showNotification(
        paste0("Session saved to: ", result$file),
        type = "message",
        duration = 5
      )
    } else {
      showNotification(
        paste0("Error saving session: ", result$error),
        type = "error",
        duration = 10
      )
    }
  })
  
  # Session status output
  output$session_status <- renderUI({
    if (!is.null(input$session_file)) {
      tags$div(
        class = "alert alert-info",
        icon("info-circle"),
        paste(" Selected:", input$session_file$name)
      )
    }
  })
  
  # Current session info output
  output$current_session_info <- renderUI({
    if (!is.null(rv$session_id)) {
      tags$div(
        tags$p(tags$strong("Session ID: "), rv$session_id),
        if (!is.null(rv$last_save_time)) {
          tags$p(tags$strong("Last Saved: "), format(rv$last_save_time, "%Y-%m-%d %H:%M:%S"))
        } else {
          tags$p(tags$strong("Last Saved: "), "Not yet saved")
        },
        tags$p(tags$strong("Students: "), if (!is.null(rv$student_files)) length(rv$student_files) else 0),
        tags$p(tags$strong("Graded: "), if (!is.null(rv$plot_grades)) sum(rv$plot_grades$graded) else 0)
      )
    } else {
      tags$div(
        class = "alert alert-secondary",
        icon("info-circle"),
        " No active session. Start grading to create a new session."
      )
    }
  })
  
  # load soln template when uploaded
  observeEvent(input$solution_file, {
    req(input$solution_file)
    
    tryCatch({
      # create temp env and knit solution
      temp_env <- new.env()
      temp_file <- input$solution_file$datapath
      
      # Store the original file extension (detect from filename)
      # Preserve the original case-sensitive extension (.qmd or .Rmd)
      file_ext_lower <- tolower(tools::file_ext(input$solution_file$name))
      if (file_ext_lower == "qmd") {
        rv$solution_file_extension <- ".qmd"
      } else {
        # Default to .Rmd for RMD files or any other extension
        # (already validated by fileInput accept parameter)
        rv$solution_file_extension <- ".Rmd"
      }
      
      # read the original file content for template generation
      rv$solution_file_content <- readLines(temp_file, warn = FALSE)
      
      # Knit the solution to populate the environment.
      # Direct output to a tempfile and redirect fig.path to avoid creating
      # stray .md files or figure/ folders in the app's working directory.
      temp_md <- tempfile(fileext = ".md")
      old_fig_path <- knitr::opts_chunk$get("fig.path")
      knitr::opts_chunk$set(fig.path = file.path(KNITR_SOLN_FIG_DIR, "fig-"))
      tryCatch({
        knitr::knit(temp_file, output = temp_md, quiet = TRUE, envir = temp_env)
      }, finally = {
        knitr::opts_chunk$set(
          fig.path = if (!is.null(old_fig_path)) old_fig_path else KNITR_DEFAULT_FIG_PATH)
        if (file.exists(temp_md)) unlink(temp_md)
      })
      
      rv$solution_env <- temp_env
      
      # get all objects created in the solution (excluding internal ones)
      all_objects <- ls(temp_env, all.names = FALSE)
      # filter out functions that start with dots and internal objects
      rv$solution_objects <- all_objects[!grepl("^\\.", all_objects)]
      
      showNotification("Solution template loaded successfully!", type = "message")
    }, error = function(e) {
      showNotification(paste("Error loading solution:", e$message), type = "error")
    })
  })
  
  # render object selector based on solution objects
  output$object_selector <- renderUI({
    req(rv$solution_objects)
    
    checkboxGroupInput(
      inputId = "selected_objects",
      label = "Select Objects to Test:",
      choices = rv$solution_objects,
      selected = rv$solution_objects
    )
  })
  
  # load student files when folder path is provided
  observeEvent(input$student_folder, {
    req(input$student_folder)
    
    # if a session was just loaded, skip the entire observer.
    # load_session_state already restored student_files, plot_grades, etc.
    # the updateTextInput that triggered this observer was only meant to
    # sync the UI text; not to re-scan the folder and wipe state.
    if (isTRUE(rv$session_loaded)) {
      rv$session_loaded <- FALSE
      return(invisible())
    }
    
    folder_path <- input$student_folder
    
    # sync the folder path to Autoharp Grading tab if it's different.
    # Keep rv$sync_in_progress TRUE until the autoharp_student_folder observer
    # fires and clears it, so the autoharp observer can detect and skip the
    # programmatic update (avoiding a double notification).
    if (!rv$sync_in_progress &&
        !is.null(input$autoharp_student_folder) &&
        input$autoharp_student_folder != folder_path) {
      rv$sync_in_progress <- TRUE
      updateTextInput(session, "autoharp_student_folder", value = folder_path)
      # Do NOT reset rv$sync_in_progress here; the autoharp_student_folder
      # observer will clear it when it fires.
    }
    
    if (dir.exists(folder_path)) {
      # get all R, Rmd, and QMD files
      files <- list.files(folder_path,
                          pattern = "\\.(R|r|Rmd|rmd|qmd|Qmd)$",
                          full.names = TRUE)
      
      if (length(files) > 0) {
        rv$student_files <- files
        
        # fresh folder selection: initialize empty plot grades
        rv$plot_grades <- data.frame(
          student_file = basename(files),
          grade = rep(NA_real_, length(files)),
          comments = rep("", length(files)),
          graded = rep(FALSE, length(files)),
          stringsAsFactors = FALSE
        )
        
        showNotification(paste("Found", length(files), "student files"), type = "message")
      } else {
        showNotification("No R, Rmd, or QMD files found in the specified folder", type = "warning")
      }
    } else {
      showNotification("Folder does not exist", type = "error")
    }
  })
  
  # Mirror of the student_folder observer for the Autoharp Grading tab.
  # Shows the "Found xxx student files" notification when the autoharp student
  # folder path is entered directly (not via sync from the Object Testing tab).
  observeEvent(input$autoharp_student_folder, {
    req(input$autoharp_student_folder)
    
    # Skip if a session is being restored (updateTextInput already set the value)
    if (isTRUE(rv$restoring_session)) return(invisible())
    
    # If the Object Testing tab synced this input programmatically, clear the
    # flag and skip - the Object Testing observer already showed a notification.
    if (isTRUE(rv$sync_in_progress)) {
      rv$sync_in_progress <- FALSE
      return(invisible())
    }
    
    folder_path <- input$autoharp_student_folder
    
    if (dir.exists(folder_path)) {
      # Look for Rmd/QMD files (autoharp only processes these)
      files <- list.files(folder_path,
                          pattern = "\\.(R|Rmd|rmd|qmd|Qmd)$",
                          full.names = TRUE)
      
      if (length(files) > 0) {
        rv$autoharp_student_files <- files
        showNotification(paste("Found", length(files), "student files"), type = "message")
      } else {
        showNotification("No Rmd or QMD files found in the specified folder", type = "warning")
      }
    } else {
      showNotification("Folder does not exist", type = "error")
    }
  })
  
  # run automated grading
  observeEvent(input$run_grading, {
    req(rv$solution_env)
    req(rv$student_files)
    req(input$selected_objects)
    
    rv$processing <- TRUE
    
    # create progress indicator
    withProgress(message = "Grading students...", value = 0, {
      
      results_list <- list()
      student_plots_list <- list()
      n_students <- length(rv$student_files)
      
      # Get output directory from input or use default
      output_dir <- if (!is.null(input$output_folder) && !is.na(input$output_folder) && input$output_folder != "") {
        input$output_folder
      } else {
        "grading_output"
      }
      
      # Create output directory if it doesn't exist
      if (!dir.exists(output_dir)) {
        dir.create(output_dir, recursive = TRUE)
      }
      
      # create a plots output directory
      plots_output_dir <- file.path(output_dir, "student_plots")
      if (!dir.exists(plots_output_dir)) {
        dir.create(plots_output_dir, recursive = TRUE)
      }
      
      for (i in seq_along(rv$student_files)) {
        student_file <- rv$student_files[i]
        student_name <- basename(student_file)
        student_base <- tools::file_path_sans_ext(student_name)
        
        incProgress(1/n_students, detail = paste("Processing", student_name))
        
        # create student env
        student_env <- new.env()
        
        tryCatch({
          if (grepl("\\.(Rmd|rmd|qmd|Qmd)$", student_file)) {
            # for Rmd and QMD files, use knitr::knit to generate plots in a known location
            # create student-specific output directory for plots
            student_plot_dir <- file.path(plots_output_dir, student_base)
            if (!dir.exists(student_plot_dir)) {
              dir.create(student_plot_dir, recursive = TRUE)
            }
            
            # Snapshot knitr chunk options so they can be restored regardless of
            # whether knitting succeeds or fails (tryCatch finally block).
            old_opts <- list(
              fig.path = knitr::opts_chunk$get("fig.path"),
              dev = knitr::opts_chunk$get("dev"),
              error = knitr::opts_chunk$get("error")
            )
            
            # Redirect figures to the student-specific directory and allow
            # chunks to continue past errors.
            knitr::opts_chunk$set(
              fig.path = file.path(student_plot_dir, "fig-"),
              dev = "png",
              error = TRUE  # Continue knitting even if a chunk has errors
            )
            
            # knit to temporary file but keep figure output in student_plot_dir
            temp_output <- tempfile(fileext = ".md")
            tryCatch({
              knitr::knit(student_file, output = temp_output, quiet = TRUE, envir = student_env)
            }, finally = {
              # Always restore knitr options and remove the temp md file
              knitr::opts_chunk$set(
                fig.path = if (!is.null(old_opts$fig.path)) old_opts$fig.path else KNITR_DEFAULT_FIG_PATH,
                dev = if (!is.null(old_opts$dev)) old_opts$dev else "png",
                error = if (!is.null(old_opts$error)) old_opts$error else FALSE
              )
              if (file.exists(temp_output)) unlink(temp_output)
            })
          } else {
            # wrap each expression in try-catch by reading and evaluating separately
            source_with_trycatch(student_file, student_env)
          }
          
          # run tests on selected objects
          student_results <- run_object_tests(
            student_env = student_env,
            solution_env = rv$solution_env,
            objects_to_test = input$selected_objects
          )
          
          student_results$student_file <- student_name
          results_list[[i]] <- student_results
          
          # collect any plots generated from the student plot directory
          student_plots_list[[student_name]] <- collect_student_plots(student_file, plots_output_dir)
          
        }, error = function(e) {
          # If there's an error, record it
          error_result <- data.frame(
            student_file = student_name,
            object = "ERROR",
            test = "Execution",
            result = "FAIL",
            details = e$message,
            stringsAsFactors = FALSE
          )
          results_list[[i]] <- error_result
        })
      }
      
      # combine all results
      rv$grading_results <- bind_test_results(results_list)
      rv$student_plots <- student_plots_list
      rv$processing <- FALSE
    })
    
    # Auto-save session after grading
    save_result <- save_session_state()
    if (save_result$success) {
      showNotification(
        paste0("Grading complete! Session auto-saved to: ", basename(save_result$file)),
        type = "message"
      )
    } else {
      showNotification("Grading complete!", type = "message")
    }
  })
  
  # grading status output
  output$grading_status <- renderUI({
    if (rv$processing) {
      tags$div(
        class = "alert alert-info",
        "Processing student submissions..."
      )
    } else if (!is.null(rv$grading_results)) {
      n_students <- length(unique(rv$grading_results$student_file))
      n_tests <- nrow(rv$grading_results)
      n_passed <- sum(rv$grading_results$result == "PASS", na.rm = TRUE)
      
      tags$div(
        class = "alert alert-success",
        paste("Completed:", n_students, "students,", n_tests, "tests run,", 
              n_passed, "passed")
      )
    }
  })
  
  # grading results table (Summary)
  output$grading_results <- renderDT({
    req(rv$grading_results)
    
    # create a summary table using base R
    results <- rv$grading_results
    
    # asggregate by student_file and object
    summary_list <- by(results, list(results$student_file, results$object), function(x) {
      data.frame(
        Student = unique(x$student_file),
        Object = unique(x$object),
        `Tests Run` = nrow(x),
        Passed = sum(x$result == "PASS", na.rm = TRUE),
        Failed = sum(x$result == "FAIL", na.rm = TRUE),
        stringsAsFactors = FALSE,
        check.names = FALSE
      )
    })
    
    summary_df <- do.call(rbind, summary_list)
    rownames(summary_df) <- NULL
    
    summary_df$`Pass Rate` <- paste0(round(summary_df$Passed / summary_df$`Tests Run` * 100, 1), "%")
    
    datatable(
      summary_df,
      options = list(
        pageLength = 10,
        scrollX = TRUE,
        dom = 'Bfrtip',
        ordering = TRUE
      ),
      rownames = FALSE,
      class = 'cell-border stripe hover'
    )
  })
  
  # detailed test results
  output$detailed_results <- renderDT({
    req(rv$grading_results)
    
    failed <- rv$grading_results[rv$grading_results$result == "FAIL", 
                                 c("student_file", "object", "test", "details")]
    
    if (nrow(failed) > 0) {
      colnames(failed) <- c("Student", "Object", "Test", "Details")
      datatable(
        failed,
        options = list(
          pageLength = 10,
          scrollX = TRUE,
          scrollY = "300px",
          dom = 'Bfrtip',
          ordering = TRUE
        ),
        rownames = FALSE,
        class = 'cell-border stripe hover'
      )
    } else {
      # Return empty datatable with message
      datatable(
        data.frame(Message = "No failed tests! All tests passed."),
        options = list(dom = 't'),
        rownames = FALSE
      )
    }
  })
  
  # download handler for CSV
  output$download_csv <- downloadHandler(
    filename = function() {
      paste0("grading_results_", format(Sys.time(), "%Y%m%d_%H%M%S"), ".csv")
    },
    content = function(file) {
      req(rv$grading_results)
      
      # combine object test results with plot grades using base R
      results <- rv$grading_results
      
      # agg by student_file
      summary_list <- by(results, results$student_file, function(x) {
        data.frame(
          student_file = unique(x$student_file),
          total_tests = nrow(x),
          tests_passed = sum(x$result == "PASS", na.rm = TRUE),
          tests_failed = sum(x$result == "FAIL", na.rm = TRUE),
          stringsAsFactors = FALSE
        )
      })
      
      object_summary <- do.call(rbind, summary_list)
      rownames(object_summary) <- NULL
      object_summary$object_score <- round(object_summary$tests_passed / object_summary$total_tests * 100, 2)
      
      if (!is.null(rv$plot_grades)) {
        # merge with plot grades
        plot_subset <- rv$plot_grades[, c("student_file", "grade", "comments")]
        names(plot_subset)[2:3] <- c("plot_grade", "plot_comments")
        
        final_results <- merge(object_summary, plot_subset, 
                               by = "student_file", all.x = TRUE)
      } else {
        final_results <- object_summary
      }
      
      write.csv(final_results, file, row.names = FALSE)
    }
  )
  
  # stu selector for plot grading
  output$student_selector <- renderUI({
    req(rv$student_files)
    
    selectInput(
      inputId = "current_student",
      label = "Select Student:",
      choices = setNames(
        seq_along(rv$student_files),
        basename(rv$student_files)
      ),
      selected = rv$current_student_index
    )
  })
  
  # update current student index when selector changes
  observeEvent(input$current_student, {
    # save current student's grade/comments before switching
    # save_current_grade checks advancing_student/restoring_session internally
    save_current_grade()
    
    rv$current_student_index <- as.integer(input$current_student)
    
    # clear the advancing flag here: this is the observer that the flag
    # was protecting against. Now that it has fired, it's safe to clear
    if (isTRUE(rv$advancing_student)) {
      rv$advancing_student <- FALSE
    }
  })
  
  # update grade inputs when student changes
  observeEvent(rv$current_student_index, {
    req(rv$plot_grades)
    req(rv$current_student_index)
    
    # Bounds checking
    if (rv$current_student_index > 0 && rv$current_student_index <= nrow(rv$plot_grades)) {
      # Get saved values for current student
      current_grade <- rv$plot_grades$grade[rv$current_student_index]
      current_comments <- rv$plot_grades$comments[rv$current_student_index]
      
      # Update inputs to show saved values
      updateNumericInput(session, "plot_grade", value = current_grade)
      updateTextAreaInput(session, "plot_comments", value = current_comments)
    }
  })
  
  # prev student button
  observeEvent(input$prev_student, {
    if (rv$current_student_index > 1) {
      # Save current student's grade/comments before switching
      save_current_grade()
      
      rv$advancing_student <- TRUE
      rv$current_student_index <- rv$current_student_index - 1
      updateSelectInput(session, "current_student",
                        selected = rv$current_student_index)
    }
  })
  
  # next student button
  observeEvent(input$next_student, {
    req(rv$student_files)
    if (rv$current_student_index < length(rv$student_files)) {
      # Save current student's grade/comments before switching
      save_current_grade()
      
      rv$advancing_student <- TRUE
      rv$current_student_index <- rv$current_student_index + 1
      updateSelectInput(session, "current_student",
                        selected = rv$current_student_index)
    }
  })
  
  # plot display
  output$plot_display <- renderUI({
    req(rv$student_files)
    req(rv$current_student_index)
    
    student_name <- basename(rv$student_files[rv$current_student_index])
    
    if (!is.null(rv$student_plots) && student_name %in% names(rv$student_plots)) {
      plots <- rv$student_plots[[student_name]]
      
      if (length(plots) > 0) {
        # display all plots for this student using base64 encoding
        plot_tags <- lapply(seq_along(plots), function(i) {
          plot_path <- plots[i]
          if (file.exists(plot_path)) {
            # read file and encode as base64
            img_data <- base64enc::base64encode(plot_path)
            # determine MIME type
            ext <- tolower(tools::file_ext(plot_path))
            mime_type <- switch(ext,
                                "png" = "image/png",
                                "jpg" = "image/jpeg",
                                "jpeg" = "image/jpeg",
                                "gif" = "image/gif",
                                "svg" = "image/svg+xml",
                                "image/png")
            img_src <- paste0("data:", mime_type, ";base64,", img_data)
            plot_title <- paste("Plot", i)
            # Create unique ID for this plot's data
            plot_id <- paste0("plot_", i, "_", rv$current_student_index)
            
            tags$div(
              class = "plot-container",
              tags$button(
                class = "expand-icon",
                `data-img-src` = img_src,
                `data-plot-title` = plot_title,
                onclick = "openPlotModalFromButton(this)",
                title = "Expand plot",
                shiny::icon("expand")
              ),
              tags$img(src = img_src, style = "max-width: 300px; height: auto;"),
              tags$p(plot_title, style = "text-align: center; margin-top: 5px; font-weight: bold;")
            )
          } else {
            tags$div(
              class = "alert alert-warning",
              paste("Plot file not found:", plot_path)
            )
          }
        })
        
        do.call(tagList, plot_tags)
      } else {
        tags$div(
          class = "alert alert-info",
          tags$i(class = "fa fa-info-circle"),
          " No plots found for this student. Run grading first to generate plots."
        )
      }
    } else {
      tags$div(
        class = "alert alert-warning",
        tags$i(class = "fa fa-exclamation-triangle"),
        " Run automated grading first to collect student plots."
      )
    }
  })
  
  # Student test results for Plot Grading tab
  output$student_test_results <- renderDT({
    req(rv$grading_results)
    req(rv$student_files)
    req(rv$current_student_index)
    
    student_name <- basename(rv$student_files[rv$current_student_index])
    
    # filter results for current student
    student_results <- rv$grading_results[rv$grading_results$student_file == student_name, ]
    
    if (nrow(student_results) > 0) {
      # select and rename columns for display
      display_df <- student_results[, c("object", "test", "result", "details")]
      colnames(display_df) <- c("Object", "Test", "Result", "Details")
      
      datatable(
        display_df,
        options = list(
          pageLength = 10,
          scrollX = TRUE,
          scrollY = "300px",
          dom = 'Bfrtip',
          ordering = TRUE
        ),
        rownames = FALSE,
        class = 'cell-border stripe hover'
      ) %>%
        DT::formatStyle(
          'Result',
          backgroundColor = DT::styleEqual(
            c('PASS', 'FAIL', 'SKIP', 'INFO'),
            c('#d4edda', '#f8d7da', '#fff3cd', '#d1ecf1')
          )
        )
    } else {
      datatable(
        data.frame(Message = "No test results available for this student. Run automated grading first."),
        options = list(dom = 't'),
        rownames = FALSE
      )
    }
  })
  
  # save plot grade
  observeEvent(input$save_grade, {
    req(rv$plot_grades)
    req(rv$current_student_index)
    
    grade_val <- input$plot_grade
    comment_val <- input$plot_comments
    
    # safely store grade (may be NA from empty numeric input)
    rv$plot_grades$grade[rv$current_student_index] <- grade_val
    rv$plot_grades$comments[rv$current_student_index] <- if (is.null(comment_val)) "" else comment_val
    
    # mark as graded: the user explicitly clicked "Save Grade"
    rv$plot_grades$graded[rv$current_student_index] <- TRUE
    
    message("[save_grade] Student ", rv$current_student_index,
            " (", rv$plot_grades$student_file[rv$current_student_index], ")",
            " grade=", grade_val, " comments='", comment_val, "'",
            " graded=TRUE")
    message("[save_grade] plot_grades snapshot: ")
    message(paste(capture.output(print(rv$plot_grades)), collapse = "\n"))
    
    # Auto-save session to disk immediately
    save_result <- save_session_state()
    if (save_result$success) {
      showNotification("Grade saved and session auto-saved!", type = "message")
    } else {
      showNotification("Grade saved!", type = "message")
    }
    
    # Auto-advance to next student.
    # Set advancing_student flag to prevent save_current_grade() from firing
    # during the student switch (inputs still show old student's values).
    if (rv$current_student_index < nrow(rv$plot_grades)) {
      rv$advancing_student <- TRUE
      rv$current_student_index <- rv$current_student_index + 1
      updateSelectInput(session, "current_student",
                        selected = rv$current_student_index)
    }
  })
  
  # grading progress
  output$grading_progress <- renderUI({
    req(rv$plot_grades)
    
    n_graded <- sum(rv$plot_grades$graded)
    n_total <- nrow(rv$plot_grades)
    
    tags$div(
      tags$h6("Grading Progress"),
      tags$div(
        class = "progress",
        tags$div(
          class = "progress-bar",
          role = "progressbar",
          style = paste0("width: ", round(n_graded/n_total * 100), "%;"),
          paste0(n_graded, " / ", n_total)
        )
      )
    )
  })
  
  # plot grades table
  output$plot_grades_table <- renderDT({
    req(rv$plot_grades)
    
    graded_subset <- rv$plot_grades[rv$plot_grades$graded, 
                                    c("student_file", "grade", "comments")]
    
    if (nrow(graded_subset) > 0) {
      colnames(graded_subset) <- c("Student", "Grade", "Comments")
      datatable(
        graded_subset,
        options = list(
          pageLength = 10,
          scrollX = TRUE,
          dom = 'Bfrtip',
          ordering = TRUE
        ),
        rownames = FALSE,
        class = 'cell-border stripe hover'
      )
    } else {
      datatable(
        data.frame(Message = "No plots have been graded yet."),
        options = list(dom = 't'),
        rownames = FALSE
      )
    }
  })
  
  # gen template with autoharp test chunks
  observeEvent(input$generate_template, {
    req(rv$solution_env)
    req(rv$solution_objects)
    req(input$selected_objects)
    
    tryCatch({
      # Generate the autoharp-compatible template
      template <- generate_autoharp_template(
        original_content = rv$solution_file_content,
        solution_env = rv$solution_env,
        objects_to_test = input$selected_objects,
        test_prefix = input$test_prefix,
        include_all_tests = input$include_all_tests
      )
      
      rv$generated_template <- template
      showNotification("Template generated successfully!", type = "message")
    }, error = function(e) {
      showNotification(paste("Error generating template:", e$message), type = "error")
    })
  })
  
  # template status
  output$template_status <- renderUI({
    if (is.null(rv$solution_env)) {
      tags$div(class = "status-alert status-warning",
               icon("exclamation-triangle"),
               " Please upload a solution template in the 'Object Testing' tab first.")
    } else if (is.null(rv$generated_template)) {
      tags$div(class = "status-alert status-info",
               icon("info-circle"),
               " Click 'Generate Template' to create an autoharp-compatible template.")
    } else {
      tags$div(class = "status-alert status-success",
               icon("check-circle"),
               " Template generated successfully! You can preview it below and download it.")
    }
  })
  
  # template preview
  output$template_preview <- renderText({
    if (!is.null(rv$generated_template)) {
      paste(rv$generated_template, collapse = "\n")
    } else {
      "# Template will appear here after generation..."
    }
  })
  
  # tests documentation
  output$tests_documentation <- renderUI({
    tags$div(
      tags$h5("Numeric Vectors"),
      tags$ul(
        tags$li(tags$strong("length"), " - Compares length of vectors"),
        tags$li(tags$strong("mean"), " - Compares mean values (tolerance: 1e-6)"),
        tags$li(tags$strong("sd"), " - Compares standard deviation (tolerance: 1e-6)"),
        tags$li(tags$strong("min"), " - Compares minimum values"),
        tags$li(tags$strong("max"), " - Compares maximum values"),
        tags$li(tags$strong("sum"), " - Compares sum of values"),
        tags$li(tags$strong("na_count"), " - Compares count of NA values"),
        tags$li(tags$strong("identical"), " - Checks if all values match exactly")
      ),
      
      tags$h5("Character Vectors"),
      tags$ul(
        tags$li(tags$strong("length"), " - Compares length of vectors"),
        tags$li(tags$strong("identical"), " - Checks if all values match exactly"),
        tags$li(tags$strong("unique_count"), " - Compares count of unique values"),
        tags$li(tags$strong("nchar_total"), " - Compares total character count")
      ),
      
      tags$h5("Data Frames"),
      tags$ul(
        tags$li(tags$strong("nrow"), " - Compares number of rows"),
        tags$li(tags$strong("ncol"), " - Compares number of columns"),
        tags$li(tags$strong("colnames"), " - Checks if column names match"),
        tags$li(tags$strong("column_types"), " - Checks if column types match"),
        tags$li(tags$strong("column_means"), " - Compares means of numeric columns"),
        tags$li(tags$strong("column_sds"), " - Compares SDs of numeric columns")
      ),
      
      tags$h5("Matrices"),
      tags$ul(
        tags$li(tags$strong("nrow"), " - Compares number of rows"),
        tags$li(tags$strong("ncol"), " - Compares number of columns"),
        tags$li(tags$strong("rownames"), " - Checks if row names match"),
        tags$li(tags$strong("colnames"), " - Checks if column names match"),
        tags$li(tags$strong("mean"), " - Compares overall mean"),
        tags$li(tags$strong("max_diff"), " - Checks maximum element difference")
      ),
      
      tags$h5("Lists"),
      tags$ul(
        tags$li(tags$strong("length"), " - Compares number of elements"),
        tags$li(tags$strong("names"), " - Checks if element names match"),
        tags$li(tags$strong("element_tests"), " - Tests each named element")
      ),
      
      tags$h5("Functions"),
      tags$ul(
        tags$li(tags$strong("num_args"), " - Compares number of arguments"),
        tags$li(tags$strong("arg_names"), " - Checks if argument names match")
      ),
      
      tags$h5("Factors"),
      tags$ul(
        tags$li(tags$strong("length"), " - Compares length"),
        tags$li(tags$strong("nlevels"), " - Compares number of levels"),
        tags$li(tags$strong("levels"), " - Checks if levels match"),
        tags$li(tags$strong("level_order"), " - Checks if level order matches"),
        tags$li(tags$strong("frequencies"), " - Compares level frequencies")
      )
    )
  })
  
  # download template handler
  output$download_template <- downloadHandler(
    filename = function() {
      # Use the original file extension if available, otherwise default to .Rmd
      extension <- if (!is.null(rv$solution_file_extension)) rv$solution_file_extension else ".Rmd"
      paste0("solution_template_autoharp_", format(Sys.time(), "%Y%m%d_%H%M%S"), extension)
    },
    content = function(file) {
      req(rv$generated_template)
      writeLines(rv$generated_template, file)
    }
  )
  
  # =========================================
  # Tab 4: Autoharp Grading Server Logic
  # =========================================
  
  # run autoharp grading with user-selected template and folder
  observeEvent(input$run_autoharp_grading, {
    
    rv$autoharp_processing <- TRUE
    
    # Create progress indicator
    withProgress(message = "Running autoharp grading...", value = 0, {
      
      tryCatch({
        # get template path from user input or use default
        if (!is.null(input$autoharp_template_file)) {
          template_path <- input$autoharp_template_file$datapath
          
          # validate file extension (case-insensitive by converting to lowercase)
          # handles .Rmd, .rmd, .RMD, .qmd, .Qmd, .QMD, etc.
          if (!grepl("\\.(rmd|qmd)$", tolower(input$autoharp_template_file$name))) {
            stop("Template file must be an RMD or QMD file (.Rmd, .rmd, .qmd, or .Qmd)")
          }
          
          # validate file exists and is readable
          if (!file.exists(template_path) || file.access(template_path, mode = 4) != 0) {
            stop("Template file cannot be read")
          }
        } else {
          # use default template file if no file uploaded (RMD format for backward compatibility)
          template_path <- "solution_template_autoharp_20251218_121348.Rmd"
        }
        
        # check if template exists
        if (!file.exists(template_path)) {
          stop(paste("Template file not found:", template_path))
        }
        
        # get pattern from input, default to "test"
        pattern <- input$test_pattern
        if (is.null(pattern) || pattern == "") {
          pattern <- "test"
        }
        
        # get student folder from user input or use default
        sol_folder <- input$autoharp_student_folder
        if (is.null(sol_folder) || sol_folder == "") {
          sol_folder <- "sol"
        }
        
        # Normalize the path (handle both absolute and relative paths)
        sol_folder_normalized <- tryCatch({
          if (is_absolute_path(sol_folder)) {
            # Absolute path - normalize it directly
            normalize_path_safe(sol_folder)
          } else {
            # Relative path - construct relative to cwd
            cwd <- normalizePath(getwd(), winslash = "/", mustWork = TRUE)
            candidate_path <- file.path(cwd, sol_folder)
            normalize_path_safe(candidate_path)
          }
        }, error = function(e) {
          stop(paste("Invalid folder path:", e$message))
        })
        
        # use normalized path
        sol_folder <- sol_folder_normalized
        
        if (!dir.exists(sol_folder)) {
          stop(paste("Student scripts folder not found:", sol_folder))
        }
        
        student_files <- list.files(sol_folder, 
                                    pattern = "\\.(R|Rmd|rmd|qmd|Qmd)$", 
                                    full.names = TRUE)
        
        if (length(student_files) == 0) {
          stop(paste("No Rmd or QMD files found in folder:", sol_folder))
        }
        
        incProgress(0.1, detail = paste("Found", length(student_files), "student files"))
        
        # Get output directory from input or use default
        out_dir <- if (!is.null(input$autoharp_output_folder) && !is.na(input$autoharp_output_folder) && input$autoharp_output_folder != "") {
          input$autoharp_output_folder
        } else {
          "student_out"
        }

        # create output directory for autoharp
        if (!dir.exists(out_dir)) {
          dir.create(out_dir, recursive = TRUE)
        }

        # Convert out_dir to absolute path so it works regardless of knit_root_dir
        out_dir <- normalizePath(out_dir, winslash = "/", mustWork = TRUE)
        
        # pop solution environment using autoharp function
        incProgress(0.1, detail = "Loading template...")
        # print(student_files[1])
        s_env <- tryCatch({
          populate_soln_env(template_path, pattern = pattern,
                            knit_root_dir = 
                            dirname(normalizePath(student_files[1], mustWork = TRUE)))
        }, error = function(e) {
          stop(paste("Failed to process autoharp template:", e$message,
                     "\nPlease ensure the template contains valid autoharp test chunks with pattern:", pattern))
        })
        
        n_students <- length(student_files)
        
        # run render_one on all student scripts using lapply
        incProgress(0.1, detail = "Processing student scripts...")
        
        # helper function to create consistent error dataframe
        create_error_result <- function(student_file, error_msg) {
          data.frame(
            fname = student_file,
            time_stamp = as.character(Sys.time()),
            run_status = "FAIL",
            run_time = NA_real_,
            run_mem = NA_real_,
            error_message = error_msg,
            stringsAsFactors = FALSE
          )
        }
        
        # track progress more precisely
        progress_per_student <- 0.6 / n_students
        
        corr_out <- lapply(seq_along(student_files), function(i) {
          student_file <- student_files[i]
          incProgress(progress_per_student, detail = paste("Processing", basename(student_file)))

          # Use the student file's directory as the working directory for knitting
          # This ensures relative paths in student scripts (e.g., "../data/file.csv") work correctly
          student_dir <- dirname(student_file)

          tryCatch({
            result <- render_one(
              rmd_name = student_file,
              out_dir = out_dir,
              knit_root_dir = student_dir,
              soln_stuff = s_env,
              permission_to_install = input$permission_to_install
            )
            # Ensure consistent columns by adding error_message as NA
            if (!"error_message" %in% names(result)) {
              result$error_message <- NA_character_
            }
            # Ensure we only return one row per student (take first row if multiple)
            if (nrow(result) > 1) {
              result <- result[1, , drop = FALSE]
            }
            result
          }, error = function(e) {
            # Return error result with consistent structure
            create_error_result(student_file, e$message)
          })
        })
        
        # combine results into a single dataframe by normalizing columns
        # handles cases where different results may have different columns
        all_cols <- unique(unlist(lapply(corr_out, names)))
        normalized_results <- lapply(corr_out, function(df) {
          missing_cols <- setdiff(all_cols, names(df))
          for (col in missing_cols) {
            df[[col]] <- NA
          }
          df[, all_cols, drop = FALSE]
        })
        combined_results <- do.call("rbind", normalized_results)

        # Deduplicate: ensure only one row per student file
        # Keep only the first occurrence of each fname
        if (!is.null(combined_results) && nrow(combined_results) > 0) {
          combined_results <- combined_results[!duplicated(combined_results$fname), , drop = FALSE]
        }
        rv$autoharp_results <- combined_results
        rv$autoharp_student_files <- student_files
        
        # Populate student_files and plot_grades for Plot Grading tab if not already set
        # from Object Testing workflow
        if (is.null(rv$student_files)) {
          rv$student_files <- student_files
          rv$plot_grades <- data.frame(
            student_file = basename(student_files),
            grade = rep(NA_real_, length(student_files)),
            comments = rep("", length(student_files)),
            graded = rep(FALSE, length(student_files)),
            stringsAsFactors = FALSE
          )
        }
        
        # Collect plots from autoharp output directory for Plot Grading tab
        incProgress(0.05, detail = "Collecting student plots...")
        autoharp_plots <- list()
        for (sf in student_files) {
          sname <- basename(sf)
          autoharp_plots[[sname]] <- collect_student_plots(sf, out_dir)
        }
        # Merge with any existing plots from Object Testing (don't overwrite)
        for (sname in names(autoharp_plots)) {
          if (length(autoharp_plots[[sname]]) > 0 &&
              length(rv$student_plots[[sname]]) == 0) {
            rv$student_plots[[sname]] <- autoharp_plots[[sname]]
          }
        }
        
        incProgress(0.05, detail = "Complete!")
        
        # Auto-save session after autoharp grading, using the autoharp output
        # directory (out_dir) so the session file lives alongside the grading
        # results rather than in the Object Testing output folder.
        save_result <- save_session_state(out_dir = out_dir)
        if (save_result$success) {
          showNotification(
            paste0("Autoharp grading complete! Session auto-saved to: ", basename(save_result$file)),
            type = "message"
          )
        } else {
          showNotification("Autoharp grading complete!", type = "message")
        }
        
      }, error = function(e) {
        showNotification(paste("Error during autoharp grading:", e$message), type = "error")
      })
      
      rv$autoharp_processing <- FALSE
    })
  })
  
  # autoharp status output
  output$autoharp_status <- renderUI({
    if (rv$autoharp_processing) {
      tags$div(
        class = "alert alert-info",
        "Processing student submissions with autoharp..."
      )
    } else if (!is.null(rv$autoharp_results)) {
      # Count unique student files (in case there are duplicate rows)
      n_students <- length(unique(rv$autoharp_results$fname))
      n_success <- sum(rv$autoharp_results$run_status == "SUCCESS", na.rm = TRUE)
      
      tags$div(
        class = "alert alert-success",
        paste("Completed:", n_students, "students processed,", 
              n_success, "successful,", n_students - n_success, "failed")
      )
    } else {
      tags$div(
        class = "alert alert-info",
        icon("info-circle"),
        " Upload a template file and specify the student scripts folder, then click 'Run Autoharp Grading'."
      )
    }
  })
  
  # autoharp results table with plot grades
  output$autoharp_results <- renderDT({
    req(rv$autoharp_results)
    
    results <- rv$autoharp_results
    
    # Create comprehensive display dataframe
    display_df <- data.frame(
      Student = basename(results$fname),
      Status = results$run_status,
      Timestamp = results$time_stamp,
      stringsAsFactors = FALSE
    )
    
    # Add run_time and run_mem if available
    if ("run_time" %in% names(results)) {
      display_df$`Run Time (s)` <- round(results$run_time, 2)
    }
    if ("run_mem" %in% names(results)) {
      display_df$`Memory Used` <- results$run_mem
    }
    
    # Add any correctness check columns (they have specific patterns)
    corr_cols <- names(results)[!names(results) %in% c("fname", "time_stamp", "run_status", "run_time", "run_mem", "error", "error_message")]
    for (col in corr_cols) {
      display_df[[col]] <- results[[col]]
    }
    
    # Add plot grades if available
    if (!is.null(rv$plot_grades)) {
      # Match students between autoharp results and plot grades
      # Extract basename for consistent matching
      plot_grades_df <- rv$plot_grades[rv$plot_grades$graded, c("student_file", "grade", "comments")]
      
      if (nrow(plot_grades_df) > 0) {
        # Ensure consistent basename matching
        plot_grades_df$student_file <- basename(plot_grades_df$student_file)
        
        # Merge by student name
        display_df <- merge(
          display_df, 
          plot_grades_df, 
          by.x = "Student", 
          by.y = "student_file", 
          all.x = TRUE
        )
        names(display_df)[names(display_df) == "grade"] <- "Plot_Grade"
        names(display_df)[names(display_df) == "comments"] <- "Plot_Comments"
      }
    }
    
    datatable(
      display_df,
      options = list(
        pageLength = 10,
        scrollX = TRUE,
        dom = 'Bfrtip',
        ordering = TRUE
      ),
      rownames = FALSE,
      class = 'cell-border stripe hover'
    ) %>%
      DT::formatStyle(
        'Status',
        backgroundColor = DT::styleEqual(
          c('SUCCESS', 'FAIL', 'UNKNOWN'),
          c('#d4edda', '#f8d7da', '#fff3cd')
        )
      )
  })
  
  # Download handler for Excel file with two sheets
  output$download_autoharp_excel <- downloadHandler(
    filename = function() {
      paste0("autoharp_grading_results_", format(Sys.time(), "%Y%m%d_%H%M%S"), ".xlsx")
    },
    content = function(file) {
      req(rv$autoharp_results)
      
      # Check for openxlsx package with user-friendly error handling
      if (!requireNamespace("openxlsx", quietly = TRUE)) {
        showNotification(
          "Package 'openxlsx' is required for Excel export but is not installed.\nPlease install it with: install.packages('openxlsx')",
          type = "error",
          duration = 10
        )
        stop("openxlsx package not available")
      }
      
      results <- rv$autoharp_results
      
      # Create comprehensive grading results dataframe
      sheet1_df <- data.frame(
        student_file = basename(results$fname),
        autoharp_status = results$run_status,
        timestamp = results$time_stamp,
        stringsAsFactors = FALSE
      )
      
      # Add all available columns from autoharp results
      for (col in names(results)) {
        if (!col %in% c("fname", "time_stamp", "run_status")) {
          sheet1_df[[col]] <- results[[col]]
        }
      }
      
      # Create workbook
      wb <- openxlsx::createWorkbook()
      
      # Add first sheet: Grading Results
      openxlsx::addWorksheet(wb, "Grading Results")
      openxlsx::writeData(wb, "Grading Results", sheet1_df)
      
      # Add second sheet: Plot Grades/Comments
      if (!is.null(rv$plot_grades)) {
        sheet2_df <- rv$plot_grades[, c("student_file", "grade", "comments", "graded")]
        names(sheet2_df) <- c("Student", "Plot Grade", "Comments", "Graded")
        
        openxlsx::addWorksheet(wb, "Plot Grades")
        openxlsx::writeData(wb, "Plot Grades", sheet2_df)
      } else {
        # Create empty sheet with message
        openxlsx::addWorksheet(wb, "Plot Grades")
        openxlsx::writeData(wb, "Plot Grades", 
                            data.frame(Message = "No plot grades available. Use Tab 2 to grade plots."))
      }
      
      # Save workbook
      openxlsx::saveWorkbook(wb, file, overwrite = TRUE)
    }
  )
}

# Run the application
shinyApp(ui = ui, server = server)


# devtools::load_all()
# 
# s_env <- populate_soln_env("solution_template_autoharp.Rmd", pattern="test")
# stud_script_names <- c("sol/student_01.rmd", "sol/student_02.rmd")
# 
# # run autoharp function "render_one" on student scripts.
# corr_out <- lapply(stud_script_names, render_one, out_dir = "student_out",
#                    knit_root_dir = getwd(), soln_stuff = s_env)
# a = do.call("rbind", corr_out)

# Student Script 6 - Custom Functions

# Define a function to calculate factorial
factorial_calc <- function(n) {
  if (n < 0) {
    stop("Input must be non-negative")
  }
  if (n == 0 || n == 1) {
    return(1)
  }
  result <- 1
  for (i in 2:n) {
    result <- result * i
  }
  return(result)
}

# Test the function
test_values <- c(5, 0, 10)
for (val in test_values) {
  cat("Factorial of", val, "is", factorial_calc(val), "\n")
}

# Define a function to compute Fibonacci sequence
fibonacci <- function(n) {
  if (n <= 0) {
    return(numeric(0))
  }
  if (n == 1) {
    return(1)
  }
  fib_seq <- numeric(n)
  fib_seq[1] <- 1
  fib_seq[2] <- 1
  if (n > 2) {
    for (i in 3:n) {
      fib_seq[i] <- fib_seq[i-1] + fib_seq[i-2]
    }
  }
  return(fib_seq)
}

# Test Fibonacci
print(fibonacci(10))

## (01) Invalid Rmd file (no chunks)
file_contents <- c("---",
                   "output: html_document",
                   "---",
                   "",
                   "Hello!")
writeLines(file_contents, "test01.Rmd")

## (02) Invalid Rmd file (no yaml header)

file_contents <- c("## no yaml header, file extension correct",
                   "x <- 1")
writeLines(file_contents, "test02.Rmd")

## (03) Check libraries (valid Rmd file)
file_contents <- c("---",
		"title: \"Sample\"",
		"output: html_document",
		"---",
		"",
		"```{r setup, include=FALSE}",
		"knitr::opts_chunk$set(echo = TRUE)",
		"```",
		"",
		"## R Markdown",
		"testing",
		"```{r cars}",
		"library(dplyr)",
		"summary(cars)",
		" x <- rnorm(100)",
		"```",
		"",
		"## Including Plots",
		"",
		"You can also embed plots, for example:",
		"",
		"```{r pressure, echo=FALSE}",
		"plot(pressure)",
		"```")
writeLines(file_contents, "test03.Rmd")

## (04) valid solution template
file_contents <- c("---",
		"title: \"Solution template\"",
		"output: html_document",
		"---",
		"",
		"## R Markdown",
		"testing",
		"```{r cars}",
		"library(dplyr)",
		"summary(cars)",
		" x <- rnorm(100)",
		"```",
		"",
		"## Test chunks",
		"```{r test_01, autoharp.objs=c(\"x\"), autoharp.scalars=c(\"len_x\")}",
		"len_x <- (length(x) == length(.x))",
		"```")
writeLines(file_contents, "test04.Rmd")


## Solution template
file_contents <- c("---",
            "title: \"test template\"",
            "output: html_document",
            "---",
            "",
            "```{r}",
            "X <- 1:10",
            "Y <- LETTERS[1:5]",
            "Z <- 10:15",
            "```",
            "",
            "",
            "```{r test02, autoharp.objs=c(\"X\", \"Y\", \"Z\"), autoharp.scalars=c(\"Xmean\", \"Zmean\")}",
            "Xmean <- mean(X)",
            "Zmean <- mean(Z)",
            "```",
            "",
            "```{r test03, autoharp.scalars=c(\"Y1\")}",
            "Y1 <- Y[3]",
            "```")
writeLines(file_contents, "soln_template.Rmd")

## student file 1
file_contents <- c( "X <- 1:10",
            "Y <- LETTERS[1:5]",
            "Z <- 10:15")
writeLines(file_contents, "s1.R")

## student file 2
file_contents <- c( "X <- 1:10",
            "Z <- 10:15")
writeLines(file_contents, "s2.R")

## student file 3
file_contents <- c( "X <- 1:10", "Y <- 1:10")
writeLines(file_contents, "s3.R")

## student file 4
file_contents <- c( "X <- 1:10", 
                    "Y <- LETTERS[1:10]",
                    "X + Y")
writeLines(file_contents, "s4.R")

#' An R expression as a tree.
#'
#' This class is used to represent a \emph{single} R expression as a tree.
#'
#' @slot adjList The adjacency list of the tree. The list must be named. The
#'   nodes should be labelled in Breadth-First Order. The first component must
#'   be the root of the tree. Leaves of the tree should be NULL elements.
#' @slot nodeTypes A data frame describing the type of node. The columns in the 
#'   data frame will be derived from the expression used to instantiate the 
#'   object. The column names will be id (node id), name, call_status, 
#'   formal_arg and depth. This slot can be left missing (i.e., populated with 
#'   NA). This latter feature is useful when we just wish to test something out.
#'   
#'   This slot is only populated automatically when an R expression is provided 
#'   as \code{lang_obj} and \code{quote_arg} is TRUE.
#' @slot repr A string representation of the tree. This will be printed when the 
#'   show method of TreeHarp is called.
#' @slot call The language object that was used to construct the tree (if it 
#' was). If the object was constructed from a list/matrix, this will be NA.
#'
#' @importFrom methods setClass new setValidity as show slot
#' @name TreeHarp-class
#' @rdname TreeHarp-class
#' @details The following validity checks are conducted on the object:
#' \enumerate{
#' \item Is the graph connected? If no, the object is invalid.
#' \item Are there cycles? If yes, the object is invalid.
#' \item Are the nodes labelled in a BFS ordering? If not, the object is not 
#' valid.
#' }
#' 
#' @export 
#'
#' @examples 
#' l1 <- list(a=c(2,3), b=NULL, c=NULL)
#' # directly using new() 
#' treeharp1 <- new("TreeHarp", adjList = l1, nodeTypes = NA)
#' 
#' # using one of the constructor methods (for lists)
#' treeharp2 <- TreeHarp(l1)
#' 
#' # using the constructor for matrices.
#' m1 <- matrix(0L, 3, 3)
#' dimnames(m1) <- list(letters[1:3], letters[1:3])
#' m1[1, ] <- c(0, 1L, 1L)
#' m1[, 1] <- c(0, 1L, 1L)
#' treeharp3 <- TreeHarp(m1)
#' 
#' # Supplying a language object to get the same tree (with nodeTypes 
#' # populated)
#' ex1 <- quote(a(b,c))
#' TreeHarp(ex1, TRUE)
#' 
TreeHarp <- setClass("TreeHarp",
                     slots = c(adjList = "list", nodeTypes = "ANY", 
                               repr="character", call="ANY"),
                     prototype = list(adjList = list(a = NULL), nodeTypes = NA,
                                      repr="", call=NA)
)

#' @return Constructors return an object of class TreeHarp.
#' @rdname TreeHarp-class
#' @export
setGeneric("TreeHarp", function(lang_obj, quote_arg, ...) standardGeneric("TreeHarp"),
           signature = "quote_arg")

#' @describeIn TreeHarp A constructor for TreeHarp.
#'
#' Converts either adjacency list or matrix into a TreeHarp object.
#'
#' @param lang_obj This should be an adjacency list for a tree (not a graph), or the
#' adjacency matrix of a tree, or the expression to be parsed. If it is a list,
#' only child nodes should be indicated (see the examples).
#' @param quote_arg If this argument is missing or FALSE, then the class of 
#' \code{lang_obj} will be evaluated, and, if it is either a list or matrix, 
#' the TreeHarp object will be returned. 
#' 
#' If this argument is TRUE, the \code{lang_obj} argument will be quoted and a 
#' parse tree for the expression will be computed and used as the tree.
#' @param ... Unused at the moment.
#'
#' @importFrom rlang enexpr
#'
#' @export
setMethod("TreeHarp", signature=c(quote_arg="logical"),
          function(lang_obj, quote_arg, ...) {
            if(!quote_arg){
              if("list" %in% class(lang_obj)){
                out <- list_2_tree(lang_obj)
                return(out)
              } else if("matrix" %in% class(lang_obj)){
                out <- mat_2_tree(lang_obj)
                return(out)
              } else {
                stop("Not sure what to do with this.\n")
              }
            } else {
              # ex1 <- enexpr(lang_obj)
              env1 <- new.env()
              lang_2_tree(lang_obj, 0, env1)
              out_list <- to_BFS(env1$adj_list, env1$node_info)
              out <- new("TreeHarp", adjList = out_list[[1]],
                         nodeTypes = out_list[[2]], repr=deparse(lang_obj),
                         call=lang_obj)
              return(out)
            }
})

#' @describeIn TreeHarp A constructor for TreeHarp.
#' 
#' Converts language object into a TreeHarp object.
#' 
#' @export
setMethod("TreeHarp", signature=c(quote_arg="missing"),
          function(lang_obj, quote_arg, ...) {
            TreeHarp(lang_obj, FALSE)
          })

list_2_tree <- function(x) {
  # check all components named.
  name_lens <- nchar(names(x))
  nonzero_lens <- sum(name_lens > 0)
  if(nonzero_lens != length(x))
    stop("Not all nodes named!")

  # check at least one null.
  null_id <- sapply(x, is.null)
  number_null_ids <- sum(null_id)
  if(number_null_ids == 0){
    stop("At least one component in the list should be NULL.")
  }

  # convert to integer (if numeric)
  for(ii in 1:length(x)){
    tmp <- x[[ii]][1]
    if(!is.null(tmp) && !is.integer(tmp)){
      x[[ii]] <- as.integer(x[[ii]])
    }
  }
  new("TreeHarp", adjList = x, nodeTypes = NA, 
      repr=paste0(names(x), collapse="-"), call=NA)
}


mat_2_tree <- function(x) {
   # check named, symmetric
   if(!isSymmetric(x)){
     stop("non-symmetric matrix.")
   }
   nn <- dimnames(x)
   if(!identical(nn[[1]], nn[[2]])){
     stop("node names do not match.")
   }
   name_lens <- nchar(nn[[1]])
   nonzero_lens <- sum(name_lens > 0)
   if(nonzero_lens != nrow(x))
     stop("Not all nodes named!")

   # convert to adj list.
   adj_list <- matrix_2_adj_list(x)
   new("TreeHarp", adjList = adj_list, nodeTypes = NA, 
       repr=paste0(nn[[1]], collapse="-"), call=NA)
  }

setValidity("TreeHarp", function(object) {
  # is it connected?
  adj_list <- object@adjList

  # if it is a single leaf node, return TRUE.
  if(length(adj_list) == 1 && is.null(adj_list[[1]])){
    return(TRUE)
  }

  if(!is_connected(adj_list)){
    return("Graph is not connected.")
  }
  # is it in BFS order?
  ll <- get_levels(adj_list)
  if(!identical(ll, sort(ll))){
    return("Not in BFS ordering.")
  }

  # does it contain cycles?
  adj_mat <- adj_list_2_matrix(adj_list)
  e1 <- new.env()
  e1$visited <- rep(FALSE, length(adj_list))
  if(is_cyclic_r(adj_mat, 1, -1, e1)){
    return("Graph contains cycles.")
  }

  return(TRUE)
})

#' Forestharp helpers
#'
#' Example of functions that can be \emph{directly used on TreeHarp objects
#' individually,} and on forestharp objects via \code{\link{fapply}}.
#'
#' @param th A TreeHarp object.
#'
#' @details These are examples of functions that be called on a list of TreeHarp
#'   objects, which we refer to as a forestharp object. Such objects are not
#'   formally defined yet, but can be created using
#'   \code{\link{rmd_to_forestharp}} or using \code{\link{join_treeharps}}.
#'
#' @return On their own, each of these functions should return a scalar or a
#' 1-dimensional array. When called with \code{\link{fapply}}, the scalar
#' numerical values can be combined (by taking the sum, any other provided
#' combiner function).
#'
#' The ultimate idea is that fapply should return a single feature for each rmd
#' file that it is called upon.
#'
#' @examples
#' # Dummy trees
#' th1 <- TreeHarp(quote(X <- rnorm(10, mean=0.9, sd=4)), TRUE)
#' th2 <- TreeHarp(quote(Y <- rbeta(10, shape1=3, shape2=5)), TRUE)
#' th3 <- TreeHarp(quote(fn1 <- function(x) x + 2), TRUE)
#' th4 <- TreeHarp(quote(df1 <- mutate(df1, new_col=2*old_col)), TRUE)
#'
#' # Run helpers
#' count_self_fn(th3)
#' count_fn_call(th4, pkg_name="dplyr")
#' count_fn_call(th1, pattern="^r.*")
#'
#' @name forestharp-helpers
NULL

#' @describeIn forestharp-helpers Counts the number of self-defined functions.
#'
#' This helper counts the number of self-defined functions. It excludes
#' lambda functions. It returns an integer scalar.
#'
#' As long as the function \code{function} was called and assigned, it will be
#' counted.
#'
#' @export
count_self_fn <- function(th) {
  #adj_list <- get_adj_list(th)
  node_types <- get_node_types(th)

  fn_defs <- dplyr::filter(node_types, .data$name == "function",
                           .data$call_status == TRUE)
  if(nrow(fn_defs) == 0)
    return(0L)

  fn_def_ids <- fn_defs$id
  parent_ids <- sapply(fn_def_ids, get_parent_id, x=th)

  sum(node_types$name[parent_ids] %in% c("=", "<-"))
}

#' @describeIn forestharp-helpers Counts the number of anonymous functions.
#'
#' Counts the number of anonymous functions, typically used in sapply, etc. It
#' returns an integer scalar. As long as the function \code{function} was
#' called but \emph{not} assigned, it will be counted here.
#'
#' @export
count_lam_fn <- function(th) {
  #adj_list <- get_adj_list(th)
  node_types <- get_node_types(th)

  fn_defs <- dplyr::filter(node_types, .data$name == "function",
                           .data$call_status == TRUE)
  if(nrow(fn_defs) == 0)
    return(0L)

  fn_def_ids <- fn_defs$id
  parent_ids <- sapply(fn_def_ids, get_parent_id, x = th)

  sum(!node_types$name[parent_ids] %in% c("=", "<-"))
}

#' @describeIn forestharp-helpers Counts the number of function calls that match a pattern.
#'
#' This helper counts the number of function calls that match a pattern. It
#' returns a count, i.e. an integer vector of length 1.
#'
#' If \code{pkg_name} is provided instead of \code{pattern}, then this function
#' counts the number of function calls from that package.
#'
#' @param pattern A regular expression to pick up function names.
#' @param pkg_name The name of a package to match functions with. This should
#' be an exact match for the package name. The package should be attached for
#' this to work. In order to avoid picking up duplicate names, for instance
#' \code{tolower} is a function in base R and in ggplot2, run
#' \code{\link{get_libraries}} on the file as well, and match against it.
#'
#' @export
count_fn_call <- function(th, pattern, pkg_name) {
  #adj_list <- get_adj_list(th)
  node_types <- get_node_types(th)
  
  # modify node_types for presence of magrittr pipe operator
  # check for "%<%" or "%>%"
  # For "%<%", 
  # - extract left hand node
  # - if no children, this must be a call; change call status to TRUE
  # For "%>%", 
  # - extract right hand node
  # - if no children, this must be a call; change call status to TRUE
  # Without this block, count_fn_call() misses out "View" on 
  # expressions such as x %>% View.
  pipe_matches <- match(node_types$name, c("%<%", "%>%"))
  pipe_ids <- which(!is.na(pipe_matches))
  if(length(pipe_ids) > 0){
    for(ii in pipe_ids) {
      if(node_types$name[ii] == "%<%"){
        left_child_id <- get_child_ids(th, ii)[1]
        children_of_left <- get_child_ids(th, left_child_id)
        if(is.null(children_of_left)) {
          node_types$call_status[left_child_id] <- TRUE
        }
      } else {
        # "%>%"
        right_child_id <- get_child_ids(th, ii)[2]
        children_of_right <- get_child_ids(th, right_child_id)
        if(is.null(children_of_right)) {
          node_types$call_status[right_child_id] <- TRUE
        }
      }
    }
  }
  
  if(!missing(pattern) && !missing(pkg_name)){
    stop("Only one of pattern or pkg_name should be supplied.")
  }

  if(!missing(pattern)) {
    fn_calls <- dplyr::filter(node_types, stringr::str_detect(.data$name, pattern),
                              .data$call_status == TRUE)
    return(nrow(fn_calls))
  }
  if(!missing(pkg_name)){
    #requireNamespace(pkg_name)
    fn_list <- ls(getNamespace(pkg_name))
    fn_calls <- dplyr::filter(node_types, .data$call_status == TRUE)

    return(sum(fn_calls$name %in% fn_list))
  }
}

#' @describeIn forestharp-helpers Extracts function calls as a string.
#'
#' Extracts the function calls that match a pattern. It returns a character
#' vector. Remember to set \code{combine = FALSE} when calling
#' \code{\link{fapply}} with it.
#'
#' @export
extract_fn_call <- function(th, pattern, pkg_name) {
  # adj_list <- get_adj_list(th)
  node_types <- get_node_types(th)
  if(!missing(pattern) && !missing(pkg_name)){
    stop("Only one of pattern or pkg_name should be supplied.")
  }

  if(!missing(pattern)) {
    fn_calls <- dplyr::filter(node_types, stringr::str_detect(.data$name, pattern),
                              .data$call_status == TRUE)
    return(fn_calls$name)
  }
  if(!missing(pkg_name)){
    #requireNamespace(pkg_name)
    fn_list <- ls(getNamespace(pkg_name))
    fn_calls <- dplyr::filter(node_types, .data$call_status == TRUE)

    return(fn_calls$name[fn_calls$name %in% fn_list])
  }
}

#' @describeIn forestharp-helpers Extracts function formal arguments called.
#'
#' Extracts the function \emph{formal} arguments from functions with a given
#' name. The name must match the function name exactly. This returns a character
#' vector or NULL, if no formal arguments are used.
#'
#' @param fn_name The (exact) function name.
#'
#' @export
extract_formal_args <- function(th, fn_name) {
  # adj_list <- get_adj_list(th)
  node_types <- get_node_types(th)

  fn_calls <- dplyr::filter(node_types, .data$name == fn_name,
                            .data$call_status == TRUE)
  fn_call_ids <- fn_calls$id
  child_ids <- lapply(fn_call_ids, get_child_ids, x=th)
  if(length(child_ids) == 0L) {
    return(NULL)
  } else {
    child_ids <- unlist(child_ids)
  }
  fn_args <- dplyr::filter(node_types, .data$id %in% child_ids,
                           .data$formal_arg == TRUE)
  if(nrow(fn_args) == 0)
    return(NULL)
  return(fn_args$name)
}

#' @describeIn forestharp-helpers Extracts names of assigned objects
#'
#' Extracts the names of assigned objects. This was written to assist in
#' detecting missed opportunities to use the pipe operator.
#'
#' @export
extract_assigned_objects <- function(th) {
  # adj_list <- get_adj_list(th)
  node_types <- get_node_types(th)

  all_assign_rows <- dplyr::filter(node_types, .data$name %in% c("<-", "="))
  if(nrow(all_assign_rows) == 0)
    return(character(0))
  child_id_list <- lapply(all_assign_rows$id, get_child_ids, x=th)
  first_child_id <- sapply(child_id_list, function(x) x[1])
  #first_child_call_status <- node_types$call_status[first_child_id]
  #browser()
  sub_tree_list <- lapply(first_child_id, function(x) {
    if(node_types$call_status[x]){
      subtree_at(th, x, TRUE)
    } else {
      subtree_at(th, x, FALSE)
    }
  })
#      subtree_at, obj=th,
#                          preserve_call=TRUE)

  sapply(sub_tree_list, function(x) x@repr, USE.NAMES = FALSE)
}

#' @describeIn forestharp-helpers Extracts actual argument names
#'
#' Extracts the actual arguments from an expression, not the formal
#' arguments. It only returns syntactic literals. It should be improved
#' to return the actual arguments for a specified function so that something
#' similar to \code{extract_assigned_objects} could be returned.
#'
#' @param include_assigned_obj A Boolean flag. If TRUE (default), it also returns
#' the name of the assigned object. If FALSE, it drops the name of the assigned 
#' object. The FALSE flag is useful for when checking if the we wish to match the 
#' assigned object from a previous call with the actual argument in this call.
#'
#' @export
extract_actual_args <- function(th, include_assigned_obj=TRUE) {
  node_types <- get_node_types(th)
  
  if(!include_assigned_obj) {
    # browser()
    all_assign_rows <- dplyr::filter(node_types, .data$name %in% c("<-", "="))
    if(NROW(all_assign_rows) > 0){
      top_assign_id <- min(all_assign_rows$id)
      child_ids <- get_child_ids(th, top_assign_id)
      if(!node_types$call_status[child_ids[2]]) {
        return(node_types$name[child_ids[2]])
      }
      th_sub <- subtree_at(th, child_ids[2], TRUE)
      node_types <- get_node_types(th_sub)
    }
    
  }
  
  actual_arg_rows <- dplyr::filter(node_types, !.data$call_status,
                                   !.data$formal_arg)
  if(nrow(actual_arg_rows) == 0)
    return(0L)
  actual_arg_rows$name
}

#' @describeIn forestharp-helpers Counts number of lines in a for loop
#'
#' Extracts the actual arguments from an expression, not the formal
#' First, this function checks if the expression contains a for loop. 
#' If it does, the innermost \code{for} loop is identified. Finally, the 
#' number of expressions below it is counted and returned.
#'
#' @export
count_num_lines_for_loop <- function(th) {
  if(count_fn_call(th, "^for$") == 0){
    return(0)
  }
  
  node_types <- get_node_types(th)
  
  # extract id of innermost "for" loop, and the children of it
  innermost_for_id <- max(which(node_types$name == "for"))
  for_children <- get_child_ids(th, innermost_for_id)
  
  # Get the third child
  th_sub <- subtree_at(th, for_children[3], TRUE)
  
  node_types_sub <- get_node_types(th_sub)
  if(node_types_sub$name[1] == "{"){
    return(length(get_child_ids(th_sub, 1)))
  } else {
    return(1)
  }
  
}


#' @describeIn forestharp-helpers Detects if a vector is being grown.
#'
#' It detects if there is an expression of form: x <- c(x, new_val). This is
#' generally bad programming practice
#'
#' @param count For \code{detect_growing}, this is a logical value that indicates
#' if the number of "grow" expressions should be counted and returned, or if just a
#' logical value should be returned.
#' @param within_for If TRUE, only expresssions within a for loop are included.
#'
#' @export
detect_growing <- function(th, count=FALSE, within_for=FALSE) {
  nt <- get_node_types(th)

  # Get nodes named c/append, get parents
  c_rows <- dplyr::filter(nt, .data$name %in% c("c", "append"))
  if(nrow(c_rows) == 0)  {
    if(count) return(0) else return(FALSE)
  }
  if(within_for) {
    # browser()
    paths_to_root <- lapply(c_rows$id, function(x) path_to_root(th, x))
    keep_ids <- sapply(paths_to_root,
                       function(x) {
                         "for" %in% nt$name[x==1]
                         })
    c_rows <- c_rows[keep_ids,]
    if(nrow(c_rows) == 0)  {
      if(count) return(0) else return(FALSE)
    }
  }

  # Check if parents of c/append is an assignment operator
  c_parents <- sapply(c_rows$id, get_parent_id, x=th)
  c_parents <- Filter(function(x) nt$name[x] %in% c("=", "<-"), c_parents)
  if(length(c_parents) == 0)
    return(FALSE)
  # browser()

  # check if assigned name is in argument to c/append
  # will fail if append called with formal argument
  c_assigned_name <- sapply(c_parents,
                                  function(y){
                                    child_ids <- get_child_ids(th, y)
                                    nt$name[child_ids[1]]
                                    })
  c_children_names <- lapply(c_rows$id,
                             function(y) {
                               child_ids <- get_child_ids(th, y)
                               nt$name[child_ids]
                               })
  detect_out <- mapply(function(x, y) x %in% y,
                       x=c_assigned_name, y=c_children_names)
  if(count){
    return(sum(unname(detect_out)))
  }
  any(detect_out)
}

#' @describeIn forestharp-helpers Detects if a for loop is present within a function
#'
#' It detects if a for loop is present within a function definition.
#'
#' @param fn_name Function name, as a character string
#'
#' @export
detect_for_in_fn_def <- function(th, fn_name) {
  # check if this is an assignment
  nt <- get_node_types(th)
  nt_names <- nt$name
  if((nt_names[1] != "<-") || (nt_names[2] != fn_name) || (nt_names[3] != "function"))
      return(FALSE)

  # look for within the nodes
  for_id <- which(nt_names == "for")
  if(length(for_id) > 0  && nt$depth[for_id] > 2)
      return(TRUE) else return(FALSE)
}

#' @describeIn forestharp-helpers Count use of a function within another.
#'
#' It counts the number of times a function is used within another.
#'
#' @param sub_fn (For count_fn_in_fn), the function to count (to look for within
#' fn_name).
#'
#' @export
count_fn_in_fn <- function(th, fn_name, sub_fn) {
  # check if this is an assignment
  nt <- get_node_types(th)
  nt_names <- nt$name
  if((nt_names[1] != "<-") || (nt_names[2] != fn_name) || (nt_names[3] != "function"))
      return(0)

  # look for within the nodes
  sub_fn_count <- count_fn_call(th, pattern=sub_fn)
  sub_fn_count
}

#' @describeIn forestharp-helpers Detect for loop to call a function
#'
#' Checks if a function has been called within a for loop.
#'
#' @export
detect_fn_call_in_for <- function(th, fn_name) {
  for_loop_indicator <- count_fn_call(th, pattern="for")
  send_ltr_indicator <- count_fn_call(th, pattern=fn_name)
  if((for_loop_indicator == 0) || (send_ltr_indicator == 0)){
    return(FALSE)
  }

  nt <- get_node_types(th)
  send_ids <- which(nt$name == fn_name)
  paths_to_parent <- lapply(send_ids, path_to_root, th=th)
  parent_names <- lapply(paths_to_parent, function(x) nt$name[which(x == 1)])
  for_in_parent <- lapply(parent_names, function(x) "for" %in% x)
  return(any(unlist(for_in_parent)))
}

#' @describeIn forestharp-helpers Extract names of functions defined by user.
#'
#' Extracts names of user-defined functions. They may not all look nice, because
#' sum functions may be anonymous functions. This function needs to be improved.
#'
#' @export
extract_self_fn <- function(th) {
    fn_def <- count_self_fn(th)
    if(fn_def==0){
        return(NULL)
    }
    node_types <- get_node_types(th)
    out_tree <- subtree_at(th, 2, node_types$call_status[2])
    out_tree@repr
}

#' @describeIn forestharp-helpers Was a function called with a particular argument?
#'
#' Checks if a function was called with a particular argument, which could be
#' the formal or actual one. The immediate child of the function call node is
#' checked.
#'
#' @param arg The argument to check for within fn_name (as a character string).
#'
#' @export
detect_fn_arg <- function(th, fn_name, arg) {
  nt <- get_node_types(th)
  nt_names <- nt$name

  fn_ids <- which(nt_names == fn_name)
  if(length(fn_ids) == 0){
    return(FALSE)
  }
  arg_ids <- lapply(fn_ids, get_child_ids, x=th)
  check_args <- sapply(arg_ids, function(x) arg %in% nt_names[x])
  any(check_args)
}

#' @describeIn forestharp-helpers Was a nested "for" loop called anywhere within the code?
#'
#' Checks if a nested for-loop was called anywhere within the code. This returns
#' a logical scalar for each TreeHarp object given.
#'
#' @export
detect_nested_for <- function(th) {
  node_types <- get_node_types(th)
  for_ids <- which(node_types$name == "for")
  if(length(for_ids) == 0){
    return(FALSE)
  } else {
    for_count_along_branch <- vapply(for_ids,
      function(x) sum(node_types$name[which(path_to_root(th, x) == 1)] == "for"),
      FUN.VALUE=1L)
  }
  any(for_count_along_branch > 1)
}

# Test constructors

test_that("Constructors: from list", {
  expect_s4_class(TreeHarp(th1), "TreeHarp")
  expect_s4_class(TreeHarp(th2), "TreeHarp")
  expect_s4_class(TreeHarp(th3), "TreeHarp")
  expect_s4_class(TreeHarp(th00), "TreeHarp")
  expect_error(TreeHarp(th8))
  expect_error(TreeHarp(th5))

  expect_error(TreeHarp(th4), regexp = "not connected")
  expect_error(TreeHarp(th7), regexp = "in BFS ordering")
  expect_error(TreeHarp(th7a), regexp = "in BFS ordering")
})

test_that("Constructors: From matrix",{
  expect_s4_class(TreeHarp(m1), "TreeHarp")
  expect_error(TreeHarp(m1a), regexp = "Not all nodes named")
})

knitr_problem <- quote(knitr::opts_chunk$set(echo = TRUE))
slot_dollar_mix <- quote(y <- l$ll$mmm@the(x = 3))
fml_problem <- quote(~x)
assignment_problem <- quote(x$xx$xxx <- g(x))

test_that("Constructors: From language",{
  expect_s4_class(TreeHarp(ex1, TRUE), "TreeHarp")
  expect_s4_class(TreeHarp(ex2, TRUE), "TreeHarp")
  expect_s4_class(TreeHarp(fml_problem, TRUE), "TreeHarp")
  expect_error(TreeHarp(quote(x <- 1)))

  nT <- get_node_types(TreeHarp(ex4, TRUE))
  expect_setequal(nT$name, c("<-", "x", "NA"))

  nT <- get_node_types(TreeHarp(knitr_problem, TRUE))
  expect_setequal(nT$name, c("$", "::", "set", "knitr", "opts_chunk", "echo", "TRUE"))

  nT <- get_node_types(TreeHarp(slot_dollar_mix, TRUE))
  expect_setequal(nT$name, c("<-", "y", "@", "$", "the", "$", "mmm",
                             "x", "l", "ll", "3") )

  expect_true(extract_assigned_objects(TreeHarp(assignment_problem, TRUE))[1] != "$")

})

e1 <- quote(View %<% x)
e2 <- quote(x %>% View())
e3 <- quote(df %>% filter(x == 1) %>% View)

# No need to worry for native pipe operations! Try this:
# e4 <- quote(x |> View)

f1 <- TreeHarp(e1, TRUE)
f2 <- TreeHarp(e2, TRUE)
f3 <- TreeHarp(e3, TRUE)

test_that("count_fn_call: pipe fixes", {
  expect_equal(count_fn_call(f3, pattern="^[vV]iew"), 1L)
  expect_equal(count_fn_call(f1, pattern="^[vV]iew"), 1L)
  expect_equal(count_fn_call(f2, pattern="^[vV]iew"), 1L)
})

## this should go under zzz.R, but I cant create a new R file, the building part will give error

.onLoad <- function(libname, pkgname) {
  #ahook_list <- autoharp_hooks[names(autoharp_hooks) != "autoharp.set"]
  #knitr::knit_hooks$set(ahook_list)
  knitr::knit_hooks$set(autoharp_hooks)
}

.onUnload <- function(libpath) {
  knitr::knit_hooks$restore()
  #knitr::knit_hooks$delete("autoharp.objs")
  #knitr::knit_hooks$delete("autoharp.scalars")
}

.onDetach <- function(libpath) {
  knitr::knit_hooks$restore()
  #knitr::knit_hooks$delete("autoharp.objs")
  #knitr::knit_hooks$delete("autoharp.scalars")
}

#' Get the children node ids
#'
#' This function retrieves the child node ids of a given node from an
#' adjacency list of a tree.
#'
#' @param adj_list The adjacency list of the tree.
#' @param at_node The node whose children should be extracted.
#'
#' @details
#' Remember that the list has to be for a tree, not a general graph. Please
#' see other help pages for more specifications.
#'
#' This is a low-level function, used within the S4 class TreeHarp. It is not
#' generally meant for use by the user.
#'
#' @return A vector of integers specifying the children of that particular node.
#' If the node is a leaf, it returns NULL.
get_child_ids2 <- function(adj_list, at_node){
  adj_list[[at_node]]
}

#' Get the parent node id
#'
#' This function retrieves the parent node id of a given node from an
#' adjacency list of a tree.
#'
#' @param adj_list The adjacency list of the tree.
#' @param at_node The node whose parent should be extracted.
#'
#' @details
#' Remember that the list has to be for a tree, not a general graph. Please
#' see other help pages for more specifications.
#'
#' This is a low-level function, used within the S4 class TreeHarp. It is not
#' generally meant for use by the user.
#'
#' @return A integer of length 1 should be returned for all nodes except the
#' root. For the latter, the function will return NULL.
#'
#' @details
#' If there are nodes that have more than one parent, then a warning is issued.
get_parent_id2 <- function(adj_list, at_node) {
  if(at_node == 1){
    return(NULL)
  }
  ids <- which(sapply(adj_list, function(x) at_node %in% x))
  if(length(ids) == 0){
    stop("Node does not exist")
  }
  if(length(ids) > 1){
    warning("More than 1 parent! Is it really a tree?")
  }
  unname(ids)
}

#' Extract a sub-tree.
#'
#' Extracts a sub-tree rooted at a particular node.
#'
#' @param obj An object of class TreeHarp
#' @param at_node The root of the new sub-tree. An integer, not a label, that
#' corresponds to BFS indexing of the tree.
#' @param preserve_call A logical value that indicates if a sub-call should be 
#' extracted. This might be slower, but it allows you to evaluate it later.
#'
#' @return An object of class TreeHarp.
#' @export
#' @importFrom methods is
#' 
#' @details This is meant for internal use, so the nodeTypes slot is silently
#' dropped, unless preserve_call is set to TRUE
#'
#' @examples
#' th3 <- list(a= c(2L,3L,4L), b=NULL, c=c(5L, 6L), d=7L, e=NULL, f=NULL, g=NULL)
#' subtree_at(TreeHarp(th3), 3)
#' st <- subtree_at(TreeHarp(th3), 4)
#' plot(st)
subtree_at <- function(obj, at_node, preserve_call=FALSE) {
  if(!is(obj, "TreeHarp")) {  
    stop("object should be of class TreeHarp")
  }
  if(!preserve_call) {
    th_obj <- get_adj_list(obj)
  
    nodes_list <- at_node
    last_done <- 0
    while(last_done < length(nodes_list)){
      nodes_list <- append(nodes_list, 
                           get_child_ids(th_obj, nodes_list[last_done+1]))
      last_done <- last_done + 1
    }
    # re-order nodes
    rep_vals <- 1:length(nodes_list)
    names(rep_vals) <- nodes_list
  
    sub_list <- th_obj[nodes_list]
    for(ii in 1:length(sub_list)) {
      if(is.null(sub_list[[ii]][1])){
        next
      } else{
        sub_list[[ii]] <- dplyr::recode(sub_list[[ii]], !!!rep_vals)
      }
    }
    return(TreeHarp(sub_list))
  } else {
    if(is(obj@call, "logical")){
      stop("TreeHarp should be from a language object.")
    }
    th_call <- obj@call
    node_types <- get_node_types(obj)
    if(!node_types$call_status[at_node]) {
      stop("The new root node should be a call.")
    }
    rec_id <- get_recursive_index(obj, at_node)
    #browser()
    t1 <- TreeHarp(th_call[[rec_id]], TRUE)
    return(t1)
  }
}

#' Convert adjacency list to a matrix
#'
#' Converts a list that represents a tree into a binary matrix.
#'
#' @param adj_list The adjacency list of the tree.
#'
#' @return A symmetric matrix of 1's and 0's, with 1 in entry (i,j) representing
#' an edge between the two vertices.
#'
#' @details
#' Remember that the list has to be for a tree, not a general graph. Please
#' see other help pages for more specifications.
#'
#' This is a low-level function, used within the S4 class TreeHarp. It is not
#' generally meant for use by the user.
#'
#' It works by filling up the upper diagonal of the matrix before reflecting it.
adj_list_2_matrix <- function(adj_list) {
  out_mat <- matrix(0L, nrow=length(adj_list), ncol=length(adj_list))
  node_names <- names(adj_list)
  dimnames(out_mat) <- list(node_names, node_names)

  for(ii in 1:nrow(out_mat)) {
    if(is.null(adj_list[[ii]][1])){
      next
    }
    out_mat[ii, adj_list[[ii]]] <- 1L
  }
  out_mat[lower.tri(out_mat)] <- t(out_mat)[lower.tri(t(out_mat))]
  if(!isSymmetric.matrix(out_mat))
    stop("Non-symmetrix matrix created!")
  out_mat
}

#' Convert adjacency matrix to a list.
#'
#' Converts a binary matrix that represents a tree into an adjacency list.
#'
#' @param mat  A symmetric matrix of 1's and 0's, with 1 in entry (i,j)
#'   representing an edge between the two vertices.
#'
#' @return The adjacency list of the tree.
#'
#' @details
#' Remember that the list that is finally output is for a tree, not a general
#' graph. Please see other help pages for more specifications.
#'
#' The input matrix should be BFS ordered. The adjacency list only notes the
#' child node(s) of a particular node. If a matrix denotes multiple parents,
#' it will not be picked up.
#'
#' This is a low-level function, used within the S4 class TreeHarp. It is not
#' generally meant for use by the user.
#'
matrix_2_adj_list <- function(mat) {
  node_names <- colnames(mat)
  out_list <- vector("list", length(node_names))
  names(out_list) <- node_names

  for(ii in seq_along(out_list)) {
    id <- which(mat[ii,] == 1)
    id <- id[id > ii]
    if(length(id) > 0){
      out_list[[ii]] <- unname(id)
    }
  }
  out_list
}

#' Checks if a graph is connected.
#'
#' A tree is a graph that is connected but does not have any cycles. This
#' function checks if a provided adjacency list is connected.
#'
#' @param adj_list The adjacency list of the tree.
#' @param root The root node to start checking from. This defaults to the
#' first node in the adjacency list.
#'
#' @details This function is used as one of the validity checks within the
#' definition of the TreeHarp class. It is a low-level function, not really
#' meant for the general user of the package. Hence it is not exported.
#'
#' The nodes are traversed in a BFS order. The function could actually be
#' combined with \link{is_cyclic_r}, but it is kept separate for modularity
#' reasons.
#'
#' An alternative was to convert the list to an adjacency matrix and check
#' for a row and column of zeros.
#'
#' @return The function returns a TRUE if the graph is connected and FALSE
#' otherwise.
is_connected <- function(adj_list, root=1) {
  nodes_list <- root
  last_done <- 0

  while(last_done < length(nodes_list)){
    nodes_list <- append(nodes_list,
                         get_child_ids(adj_list, nodes_list[last_done+1]))
    last_done <- last_done + 1
  }

  nodes_list <- sort(unique(nodes_list))
  if(length(nodes_list) == length(adj_list))
    return(TRUE)
  else return(FALSE)
}

#' Checks if a graph contains any cycles.
#'
#' A tree is a graph that is connected but does not have any cycles. This
#' function checks if a provided adjacency matrix contains cycles.
#'
#' @param adj_mat  A symmetric matrix of 1's and 0's, with 1 in entry (i,j)
#'   representing an edge between the two vertices.
#' @param node_v The node to begin searching for cycles from. An integer.
#' @param parent_node The parent node of node_v. Also an integer. Use -1 if
#' you are starting from node 1. This is in fact the default.
#' @param visited_env An environment containing a logical vector indicating which
#' nodes have already been visited. The vector has to be named "visited".
#' See the details.
#'
#' The function works by traversing all the nodes, in a BFS order. If it finds
#' a node has a parent that has already been visited, it concludes that there
#' is a cycle.
#'
#' The function is recursive, and has to update the vector of visisted nodes
#' within each call. Hence the visited vector is stored in an environment
#' that is passed along. It will return an error if no such environment is
#' provided. It is a very specific input that the function requires, and this
#' is another reason that this function is not exported.
#'
#' This function is used within the validity checks for the S4 class. It is
#' not exported for the user.
#'
#' @return A logical value indicating if the graph contains cycles.
is_cyclic_r <- function(adj_mat, node_v, parent_node=-1, visited_env) {
  if(!exists("visited", envir=visited_env)){
    stop("visited vector does not exist!")
  }
  visited_env$visited[node_v] <- TRUE
  nbr_ids <- which(adj_mat[node_v, ] == 1)
  for(jj in nbr_ids){
    if(!visited_env$visited[jj]){
      if(is_cyclic_r(adj_mat, jj, node_v, visited_env))
        return(TRUE)
    }else{
      if(jj != parent_node)
        return(TRUE)
    }
  }
  return(FALSE)
}

#' Generate the next sub-tree.
#'
#' This generates the next sub-tree in the enumeration list.
#'
#' @param obj An object of class TreeHarp.
#' @param char_arr A vector of 1's and 0's indicating which nodes to keep. The
#' vector should have length equal to the number of nodes in obj.
#'
#' @details Need to reference the paper. This generates the next sub-tree,
#' rooted at the root node of this tree. It will generate singletons on it's
#' own. It has to be used within a loop to do that.
#'
#' @return A vector of 1's and 0's, which denotes the next sub-tree in the list.
#' @export
#' @seealso \code{\link{generate_all_subtrees}}
#' 
#' @examples 
#' th1 <- TreeHarp(list(a=c(2,3), b=NULL, c=NULL))
#' get_next_subtree(th1, c(1,0,0))
#' get_next_subtree(th1, c(1,1,0))
#'
get_next_subtree <- function(obj, char_arr) {
  th_obj <- get_adj_list(obj)
  # error-checking
  if(length(char_arr) != length(th_obj)){
    stop("char_arr should be a 0-1 vector length node.")
  }
  if(all(char_arr == 1)) {
    #message("No sub-trees left!")
    return(NULL)
  }
  #
  for(ii in length(char_arr):1){
    if(char_arr[ii] == 1) {
      char_arr[ii] <- 0
    } else {
      # print("OK")
      # get parent node
      parent_id <- get_parent_id(th_obj, ii)
      # return(parent_id)
      if(char_arr[parent_id] == 1) {
        char_arr[ii] <- 1
        return(char_arr)
      } else {
        next
      }
    }
  }
  char_arr
}

#' Generate all subtrees from a tree.
#'
#' This routines generates all subtrees rooted at the root node for a particular
#' tree.
#'
#' @param th An object of class TreeHarp.
#'
#' @return A 0-1 matrix with n rows and m columns. n is the number of sub-trees
#' rooted at the root node of th. m is the number of nodes in this given tree.
#' The leading column will be a 1 for all the rows.
#'
#' @seealso \code{\link{get_next_subtree}}
#' @references 
#' \emph{Listing and counting subtrees of a tree}, F Ruskey, \emph{SIAM Journal on Computing}, 1981
#' 
#' @export
#' @examples
#' th1 <- TreeHarp(list(a=c(2,3), b=NULL, c=NULL))
#' generate_all_subtrees(th1)
generate_all_subtrees <- function(th){
  num_nodes <- length(th)

  init_tree <- rep(0, num_nodes)
  init_tree[1] <- 1
  all_trees <- init_tree

  while(!is.null(init_tree)){
    init_tree <- get_next_subtree(th, init_tree)
    all_trees <- rbind(all_trees, init_tree, deparse.level = 0L)
  }

  all_trees
}

#' Carve out branches to form a new tree.
#'
#' This functions keeps only the indicated nodes, returning a new sub-tree.
#'
#' @param obj An object of class TreeHarp.
#' @param char_arr A vector of 1's and 0's indicating which nodes to keep. The
#' vector should have length equal to the number of nodes in obj.
#'
#' @return An object of class TreeHarp.
#' @details This returns an error if the sub-tree does not define a new tree.
#' @export
#' @importFrom methods is
#'
#' @examples
#' th3 <- list(a= c(2L,3L,4L), b=NULL, c=c(5L, 6L), d=7L, e=NULL, f=NULL, g=NULL)
#' carve_subtree(TreeHarp(th3), c(1,0,0,0,0,0,0))
#' st <- subtree_at(TreeHarp(th3), 4)
#' plot(st)
carve_subtree <- function(obj, char_arr) {
  th_obj <- get_adj_list(obj)
  th_mat <- adj_list_2_matrix(th_obj)
  out_mat <- th_mat[which(char_arr == 1), which(char_arr == 1), drop=FALSE]
  #TreeHarp(out_mat)
  nT <- get_node_types(obj)
  if( is(nT, "data.frame") ){
    nT <- nT[which(char_arr==1), ]
    nT$id <- 1:nrow(nT)
    out_obj <-  tryCatch(new("TreeHarp", adjList = matrix_2_adj_list(out_mat),
                             nodeTypes = nT),
                         error = function(e) return(e))
  } else {
    out_obj <-  tryCatch(TreeHarp(out_mat),
                         error = function(e) return(e))
  }

  if(is(out_obj, "error")){
    stop("Selected nodes do not define a tree.")
  }
  return(out_obj)
  # out_mat
  # mat_2_treeharp(out_mat)
}

#' Obtains the node levels from a tree.
#'
#' This function obtains the node levels from a tree.
#'
#' @param adj_list The adjacency list of the tree.
#'
#' @details This function is used to check if the specification of the tree is
#' in BFS order. If that is indeed the case, the levels of each node should be
#' sorted.
#'
#' This function is not exported for the general user.
#'
#' @return It returns a vector of integers. The length of this vector will be
#' the number of nodes in the tree. The root is at level 1, the next is at level
#'  2, and so on.
#'
get_levels <- function(adj_list) {
  num_nodes <- length(adj_list)
  # if it is a single leaf node, return TRUE.
  if(num_nodes == 1 && is.null(adj_list[[1]])){
    return(1L)
  }

  level_vec <- rep(NA, num_nodes)
  level_vec[1] <- 1
  for(jj in 2:num_nodes){
    steps <- 1
    anc <- get_parent_id(adj_list, jj)
    while(is.na(level_vec[anc])){
      anc <- get_parent_id(adj_list, anc)
      steps <- 1 + steps
    }
    level_vec[jj] <- steps + level_vec[anc]
  }
  level_vec
}

#' Checks if a tree is rooted at a node of another tree.
#'
#' This function checks if a given tree is a sub-tree of another tree at a
#' particular node.
#'
#' @param x An object of class TreeHarp.
#' @param y An object of class TreeHarp.
#' @param at_node An integer, corresponding to a node in object y. The sub-tree
#' of y, rooted at at_node, is compared to x.
#'
#' @details
#' Here's how it works: The sub-tree of y, rooted at at_node is first
#' extracted. The tree x is then compared to this. If x is a sub-tree of it,
#' then this function returns FALSE. Otherwise it returns TRUE.
#'
#' @return A logical value indicating if x is a sub-tree of y, rooted at
#' at_node.
#' @export
#' @importFrom methods is
#'
#' @examples
#' thb1 <- TreeHarp(list(b=2, d=NULL))
#' tha1 <- TreeHarp(list(a=c(2,3), b=4, c = NULL, d=NULL))
#' is_subtree_rooted_at(thb1, tha1, 1) # FALSE
#' is_subtree_rooted_at(thb1, tha1, 2) # TRUE
is_subtree_rooted_at <- function(x, y, at_node) {
  if(  !is(x, "TreeHarp") || !is(y, "TreeHarp")){
    stop("objects should both be of class TreeHarp")
  }
  y_sub <- subtree_at(y, at_node)

  if(length(x) == 1){
    if(names(x)[1] == names(y_sub)[1]){
      return(TRUE)
    } else{
      return(FALSE)
    }
  }

  adj_list_x <- get_adj_list(x)
  levels_x <- get_levels(adj_list_x)
  names_x <- names(x)
  df_x <- data.frame(names=names_x, levels=levels_x,
                     stringsAsFactors = FALSE)

  adj_list_y <- get_adj_list(y_sub)
  levels_y <- get_levels(adj_list_y)
  names_y <- names(y_sub)
  df_y <- data.frame(names=names_y, levels=levels_y,
                     stringsAsFactors = FALSE)

  aj <- dplyr::anti_join(df_x, df_y, by=c("names", "levels"))
  if(nrow(aj) > 0) {
    return(FALSE)
  }
  df_y$id <- 1:nrow(df_y)
  sj <- dplyr::semi_join(df_y, df_x, by=c("names", "levels"))
  char_arr <- rep(0, nrow(df_y))
  char_arr[sj$id] <- 1L

  y_test <- carve_subtree(y_sub, char_arr)
  return(identical(x@adjList, y_test@adjList))
}


#' Prune a tree up to a specified depth.
#' 
#' Prunes a tree up to a depth specified by a set of node names.
#'
#' @param th A TreeHarp object.
#' @param names_to_keep  The node names to keep in the pruned tree.
#' 
#' @details This is a seldom used function. It works in this way. Given a set 
#' of node names, it identifies the node with the greatest depth in that set.
#' The function then returns the sub-tree, that contains all the nodes with a
#' depth smaller than or equal to that depth. If the node types slot is not NA,
#' then that data frame is filtered and returned too.
#' 
#' Take a look at the examples for a clearer picture.
#'
#' @return An object of class TreeHarp.
#' @export
#' @seealso \code{\link{carve_subtree}}, \code{\link{path_to_root}}, 
#' \code{\link{carve_mst}}
#'
#' @examples
#' ex1 <- quote(x <- f(y, g(5)))
#' th1 <- TreeHarp(ex1, TRUE)
#' s1 <- prune_depth(th1, c("f", "y"))
#' s2 <- prune_depth(th1, c("f", "z")) # node not present!
#' plot(s1)
#' plot(s2)
prune_depth <- function(th, names_to_keep) {
  lvls <- get_levels(get_adj_list(th))
  ids <- which(names(th) %in% names_to_keep)
  if(length(ids) == 0){
    return(th)
  }
  last_id <- max(ids)
  keep_depth <- lvls[last_id]
  char_arr <- rep(0, length(lvls))
  char_arr[lvls <= keep_depth] <- 1
  carve_subtree(th, char_arr)
}

#' Extract a path from node to root.
#' 
#' Identifies the nodes on the path from a node up to the root of a TreeHarp
#' object.
#'
#' @param th A TreeHarp object.
#' @param node_num A node number to start tracking upwards from.
#' 
#' @details This function allows the user to identify the branch from a node 
#' up to the root of a tree. 
#'
#' @return A vector of 1's and 0's that can be used to carve out the branch 
#' alone, using \code{\link{carve_subtree}}.
#' @export
#'
#' @examples
#' ex1 <- quote(x <- f(y, g(5)))
#' th1 <- TreeHarp(ex1, TRUE)
#' path_to_root(th1, 5)
path_to_root <- function(th, node_num) {
  adj_list <- get_adj_list(th)
  char_array <- rep(0, length(th))
  char_array[node_num] <- 1

  while(char_array[1] != 1) {
    node_num <- get_parent_id(adj_list, node_num)
    char_array[node_num] <- 1
  }
  char_array
}

#' Carve a Minimal Spanning Tree Out
#' 
#' Given node names, this function retrieves the smallest tree containing at most 
#' those nodes.
#'
#' @param th An object of class TreeHarp.
#' @param node_names A character vector of node names. Nodes outside this set 
#' will not be returned in the tree. It must include the root node name.
#'
#' @return An object of class TreeHarp.
#' @export
#' 
#' @details The function starts from each node specified and works it's way up
#' to the root. If a branch contains nodes outside the list, it is shortened.
#' 
#' In the end, the tree that is returned will try to contain all the named nodes,
#' but if that's not possible some will dropped to ensure a tree is returned, 
#' not a disconnected graph.
#' 
#' @examples
#' ex1 <- quote(x <- f(y, g(5)))
#' th1 <- TreeHarp(ex1, TRUE)
#' carve_mst(th1, c("<-", "x", "f", "5")) ## note: 5 is dropped.
#' carve_mst(th1, c("<-", "x", "f", "y")) 
#' carve_mst(th1, c("<-", "f", "g")) 
#' 
carve_mst <- function(th, node_names) {
  all_names <- names(th)
  if(!all_names[1] %in% node_names){
    stop("root name must be included.")
  }
  keep_these <- which(all_names %in% node_names)
  char_array <- rep(FALSE, length(all_names))

  for(ii in keep_these) {
    char_tmp <- path_to_root(th, ii)
    if(any(!(all_names[which(char_tmp == 1)] %in% node_names))){
      next
    }
    char_array <- (char_array | char_tmp)
  }
  char_array <- ifelse(char_array, 1, 0)
  carve_subtree(th, char_array)
}

#' Keep only branches specified by node numbers
#' 
#' Retains only specific branches, that are identified by their node numbers.
#'
#' @param th  A TreeHarp object.
#' @param branch_nodes An integer vector, specifying the nodes to keep.
#' @param include_lower A logical value - whether or not the lower branches should 
#' also be kept.
#'
#' @return A TreeHarp object.
#' @export
#'
#' @examples
#' ex1 <- quote(x <- f(y, g(5)))
#' th1 <- TreeHarp(ex1, TRUE)
#' keep_branches(th1, 3)
#' keep_branches(th1, 3, include_lower = FALSE)
#' keep_branches(th1, c(2,3), FALSE)
#' keep_branches(th1, c(3, 4), FALSE)

keep_branches <- function(th, branch_nodes, include_lower = TRUE) {
  if(length(branch_nodes) == 1){
    upper_ids <- which(path_to_root(th, branch_nodes) == 1)
  } else {
    paths_to_root <- lapply(branch_nodes, path_to_root, th=th)
    upper_ids <- which(Reduce("|", paths_to_root))
  }

  if(include_lower) {
    nodes_list <- branch_nodes
    last_done <- 0
    adj_list <- get_adj_list(th)

    while(last_done < length(nodes_list)){
      nodes_list <- append(nodes_list,
                           get_child_ids(adj_list, nodes_list[last_done+1]))
      last_done <- last_done + 1
    }

    upper_ids <- unique(c(upper_ids, nodes_list))
  }

  char_arr <- rep(0, length(th))
  char_arr[upper_ids] <- 1L
  carve_subtree(th, char_arr)
}

#' Find the branch that leads from one node to another.
#' 
#' Given two nodes that are on the same path to the root, this function
#' determines the branch that leads to the child node.
#' 
#' @param th A TreeHarp object.
#' @param child_id An integer node id. It corresponds to the node to trace up
#'   from.
#' @param ancestor_id An integer node id. It corresponds to the node to trace 
#' down from.
#' 
#' @details This is used when trying to find a sub-call from a TreeHarp object.
#' It is useful in determining the indices to use when extracting the sub-call.
#'
#' @return An integer that denotes the branch to follow down (from the ancestor)
#' to reach the child.
#' @export
#'
#' @examples
#' ex3 <- quote(x <- f(y = g(3, 4), z=1L))
#' t1 <- TreeHarp(ex3, TRUE)
#' find_branch_num(t1, 8, 3) # should be 1
#' find_branch_num(t1, 5, 3) # should be 2
find_branch_num <- function(th, child_id, ancestor_id) {
  node_to_root <- path_to_root(th, child_id)
  nodes_on_path <- which(node_to_root == 1L)
  
  if(!ancestor_id %in% nodes_on_path) {
    stop("child is not descendant of ancestor.")
  }
  p_id <- get_parent_id(th, child_id)
  ancestors_children <- get_child_ids(th, ancestor_id)
  if(p_id == ancestor_id){
    # direct descendant
    return(which(ancestors_children == child_id))
  }
  
  while(!p_id %in% ancestors_children) {
    p_id <- get_parent_id(th, p_id)
  }
  which(ancestors_children == p_id)
}

#' Obtain an index to extract out a sub-call
#' 
#' Obtains an index that can be used to extract a sub-call from a language object.
#' 
#' @param th A TreeHarp object.
#' @param node_id An integer corresponding to a call within the parse tree 
#' (not a literal, symbol or a formal argument).
#'
#' @return A vector of indices, that can be used (together with "[[") to obtain a 
#' sub-call
#' @importFrom methods is
#' @export
#'
#' @examples
#' ex3 <- quote(x <- f(y = g(3, 4), z=1L))
#' t1 <- TreeHarp(ex3, TRUE)
#' rec_index <- get_recursive_index(t1, 6)
#' ex3[[rec_index]]
#' ex3[[get_recursive_index(t1, 3)]]
get_recursive_index <- function(th, node_id) {
  node_types <- get_node_types(th)
  if( is(node_types, "logical") || (!node_types$call_status[node_id])) {
    return(NA)
  }
  call_status <- node_types$call_status
  nodes_to_root <- which(path_to_root(th, node_id) == 1L)
  
  fns_to_root <- nodes_to_root[call_status[nodes_to_root]]
  parent_fns_to_root <- sapply(fns_to_root[-1], get_parent_call_id, x=th)
  # browser()
  # initialise branch_nums to be empty vector
  # find branch_num, check if parent is "function"
  branch_nums <- vector(mode="integer", length = 0)
  for(nn in 1:length(parent_fns_to_root)){
    tmp_branch_num <- find_branch_num(th, 
                                      child_id = fns_to_root[-1][nn],
                                      ancestor_id = parent_fns_to_root[nn])
    parent_fn_name <- node_types$name[parent_fns_to_root[nn]]
    if(parent_fn_name == "function"){
      # check if node is formal argument or not
      direct_parent_code_id <- get_parent_id(th, fns_to_root[-1][nn])
      if(node_types$formal_arg[direct_parent_code_id]){
        branch_nums <- c(branch_nums, c(2, tmp_branch_num))
      } else {
        branch_nums <- c(branch_nums, 3)
      }
    } else {
      branch_nums <- c(branch_nums, tmp_branch_num + 1)
    }
  }
  # branch_nums <- mapply(find_branch_num, child_id = fns_to_root[-1], 
  #                      ancestor_id = parent_fns_to_root, 
  #                      MoreArgs = list(th=th))
  branch_nums
  #parent_nodes <- sapply(nodes_to_root[-1], get_parent_call_id, x=th)
  #child_nodes <- lapply(parent_nodes, get_child_ids, x=th)
  #child_nodes
}

#' Get Node Id of Parent Call
#' 
#' Get the node id of the parent call for a given node.
#'
#' @param x  A TreeHarp object.
#' @param node_id  The id of the node whose parent call is to be found. An 
#' integer value.
#'
#' @details When we need to go up the parse tree to obtain the function that 
#' called this node, we use this function. It is similar to \code{get_parent_id},
#' except that that function only returns the immediate parent.
#' 
#' It is not useful to call this function when the TreeHarp object is not 
#' constructed from a language object.
#' 
#' Perhaps this function is necessary only because of the way language objects 
#' are represented by the autoharp: formal arguments are included in the 
#' tree representation. When we wish to find the calling function, we have to 
#' walk up the branches till we reach a function call.
#' 
#' @return An integer corresponding to the node id of the calling function.
#' @importFrom methods is
#' @export
#'
#' @seealso \code{\link{get_parent_id}}
#' @examples
#' ex3 <- quote(x <- f(y = g(3, 4), z=1L))
#' t1 <- TreeHarp(ex3, TRUE)
#' 
#' # get the function that calls g:
#' get_parent_call_id(t1, 6) 
#' #contrast with this:
#' get_parent_id(t1, 6)
get_parent_call_id <- function(x, node_id) {
  node_types <- get_node_types(x)
  if( is(node_types, "logical") ){
    return(NA)
  }
  call_status <- node_types$call_status
  p_id <- get_parent_id(x, node_id)
  while(!call_status[p_id]){
    p_id <- get_parent_id(x, p_id)
  }
  p_id
}

test_that("Utility: get children and parents", {
  expect_equal(get_child_ids(th1, 2), NULL)
  expect_equal(get_child_ids(th1, 1), c(2L, 3L))
  expect_equal(get_child_ids(th3, 3), 4L)

  expect_equal(get_parent_id(th1, 1L), NULL)
  expect_equal(get_parent_id(th1, 2L), 1L)
  expect_error(get_parent_id(th1, 6L), regexp = "does not exist")
  expect_warning(get_parent_id(th5, 4L), regexp = "More than 1")
})

test_that("Utility: conversion b/w adjacency list and matrix", {
  expect_equal(adj_list_2_matrix(th1), m1)
  expect_equal(matrix_2_adj_list(m2), th6)
})

test_that("Utility: connected", {
  expect_equal(is_connected(th4), FALSE)
  expect_equal(is_connected(th6), TRUE)
  expect_equal(is_connected(th1), TRUE)
})

test_that("Utility: cyclic", {
  e1$visited <- rep(FALSE, nrow(m2))
  expect_equal(is_cyclic_r(m2, 1, -1, e1), TRUE)
  e1$visited <- rep(FALSE, nrow(m1))
  expect_equal(is_cyclic_r(m1, 1, -1, e1), FALSE)
})

test_that("Utility: obtain levels", {
  expect_identical(get_levels(th7), c(1, 3, 2, 2, 3, 2))
  expect_identical(get_levels(th1), c(1, 2, 2))
  expect_equal(get_levels(th00), 1L)
})

test_that("Carve trees: from char array", {
  expect_s4_class(carve_subtree(tho3a, c(1,0,1,0,1,1,0)), "TreeHarp")
  expect_s4_class(carve_subtree(tho3a, c(0,0,0,0,0,0,1)), "TreeHarp")
  expect_error(carve_subtree(tho3a, c(1,0,1,0,1,1,1)), regexp="define a tree")
})

test_that("Check rooted at", {
  expect_true(is_subtree_rooted_at(tho1, tho2, 1))
  expect_false(is_subtree_rooted_at(tho1, tho2, 2))
})

test_that("Identical trees", {
  expect_true(identical(carve_subtree(tho3a, c(1,1,1,0,0,0,0)), tho1))
  expect_false(identical(carve_subtree(tho1, c(1,0,0)), tho00))
})

ex1_tree <- TreeHarp(ex1, TRUE)
test_that("Path to root", {
  expect_equal(path_to_root(t6, 5), c(1, 0, 1, 0, 1))
  expect_equal(path_to_root(t6, 1), c(1, 0, 0, 0, 0))
  expect_equal(path_to_root(ex1_tree, 6), c(1,0,1,0,1,1,0,0,0))
})

test_that("Generate sub-tree counts", {
  expect_equal(nrow(generate_all_subtrees(t6)), 12)
  expect_equal(nrow(generate_all_subtrees(ex1_tree)), 33)
})

br1 <- keep_branches(ex1_tree, 7)
br2 <- keep_branches(ex1_tree, 7, FALSE)
test_that("Keep branches", {
  expect_equal(names(br1), c("+", "geom_point", "aes", "y", "y1"))
  expect_equal(names(br2), c("+", "geom_point", "aes", "y"))
})

tt3 <- TreeHarp(quote(f1 <- function(a = 4L) {x <- a; 2 -> y; z = 3; y}), 
                TRUE)
tt4 <- TreeHarp(quote(x$g[[34]] <- rnorm(10, mean = g(5, h(34)), sd = 3.4)), TRUE)
ex6 <- quote(function(y, x = g(z)) w <- 1)
tt5 <- TreeHarp(ex6, TRUE)
test_that("Helpers work", {
  expect_setequal(extract_assigned_objects(tt3), c("f1", "x", "y", "z"))
  expect_equal(get_recursive_index(tt3, 7), c(3,3,2))
  expect_equal(get_recursive_index(tt3, 9), c(3,3,4))
  expect_equal(get_recursive_index(tt5, 5), c(2,2))
  expect_equal(get_recursive_index(tt4, 14), c(3,3,3))
  expect_equal(get_recursive_index(tt4, 4), c(2,2))
  expect_equal(get_parent_call_id(tt4, 11), 3)
  expect_equal(get_parent_call_id(tt4, 12), 3)
})

#' autoharp: Semi-Automatic Grading of R and Rmd Scripts
#'
#' The autoharp package provides functions for running and analysing 
#' R script/Rmd submissions from students.
#' 
#' @section Main functions are:
#' \enumerate{
#'   \item \code{\link{populate_soln_env}}
#'   \item \code{\link{render_one}}
#'   \item \code{\link{TreeHarp}}
#' }
#' 
#' The user manuals can be found at \url{https://namanlab.github.io/autoharp/}
#'
#' @name autoharp
"_PACKAGE"

#' Calculate Run-time Statistics
#'
#' This is stand-alone function. It computes the runtime stats
#' without rendering the md/html/pdf file.
#'
#' @param stud_fname  The rmd filename of the student.
#' @param knit_root_dir  The working directory to use when knitting the file.
#' @param return_env A logical value to indicate if the environment from the
#' rmd file should be return. If FALSE, an NA value is returned.
#' 
#' @details This routine is not used within any other function within the 
#' package. Figures are not cleaned or removed.
#'
#' @return A list containing the running time in seconds, the memory used
#' by the final environment in bytes (as a numeric scalar), and the
#' environment object containing all the generated objects from the rmd file.
#' @seealso \code{\link{render_one}}
#' @export
#'
check_runtime <- function(stud_fname, knit_root_dir, return_env=FALSE) {
  # create tempdir
  if(!missing(knit_root_dir)){
    old_root <- knitr::opts_knit$get("root.dir")
    knitr::opts_knit$set(root.dir = knit_root_dir)
  }

  e_stud <- new.env()
  start_time <- Sys.time()
  stud_out <- knitr::knit(stud_fname, quiet=TRUE, envir=e_stud)
  end_time <- Sys.time()
  mem_usage <- env_size(e_stud)
  if(file.exists(stud_out)) {
    unlink(stud_out) # does not clean figures!
  }
  if(!missing(knit_root_dir)){
    knitr::opts_knit$set(root.dir = old_root)
  }
  runtime_secs <- as.numeric(difftime(end_time, start_time, units="secs"))
  out_obj <- list(run_time = runtime_secs,
                  mem_usage = mem_usage,
                  env=NA)
  if(return_env)
    out_obj$env <- e_stud
  return(out_obj)
}

#' Conducts checks before rendering file
#'
#' This routine conducts two checks before rendering the file: presence of 
#' View function, and presence of system() calls. The former is a nuisance,
#' while the latter is possibly nefarious.
#' 
#' @param fname The name of the Rmd/qmd file to check.
#' @param in_place If TRUE, the original file will be overwritten in place.
#' @param new_name If in_place is FALSE, this should be the full path to the 
#' new file to be written (including directory location).
#' @param verbose If TRUE, debugging messages will be printed.
#'
#' @returns Either TRUE or FALSE. If TRUE, all View() lines have been commented,
#' and there are no system calls. If FALSE, it means there are system calls (to 
#' be screened further).
#' @export
#'
render_prechecks <- function(fname, in_place=TRUE, new_name, verbose=FALSE) {
  if(!in_place){
    if(missing(new_name)){
      stop("If in_place is FALSE, new_name cannot be missing.")
    }
  } else {
    new_name <- fname
  }
  f1 <- rmd_to_forestharp(fname, FALSE)
  if(is(f1, "logical")){
    if(verbose) {
      message("Could not run purl on:", fname)
    }
    return(FALSE)
  }
  if(identical(f1, list(forest=NA, line_nums=NA))) {
    if(verbose) {
      message("No expressions could be found in: ", fname)
    }
    return(FALSE)
  }
  
  # Comment out View calls
  contain_view <- unlist(fapply(f1, count_fn_call, 
                                pattern="^[vV]iew$", combine=FALSE))
  if(sum(contain_view) > 0) {
    # browser()
    ids <- f1$line_nums[contain_view  > 0]
    fname_org_lines <- readLines(fname)
    fname_org_lines[ids] <- paste0("# ", fname_org_lines[ids])
    writeLines(fname_org_lines, con=new_name)
    if(verbose){
      message(paste(length(ids), "lines commented."))
    }
  } else {
    if(verbose){
      message("No calls to View()")
    }
  }
  
  # Skip file if system calls present
  contain_system <- unlist(fapply(f1, count_fn_call, 
                                  pattern="^system2{0,1}$", combine=FALSE)) 
  if(sum(contain_system) > 0) {
    if(verbose){
      message("System calls found.. skipping render.")
    }
    return(FALSE)
  } else {
    return(TRUE)
  }
}

#' Extract section text from Rmd
#'
#' @param rmd_name The filename of the Rmd script.
#' @param hdr_pat  The regular expression pattern to pick up the section title.
#' @param ignore_case A boolean - whether or not to ignore case when matching for
#' the section title.
#'
#' @return
#'
#' Returns a character vector containing all the text written in the section
#' that begins with the specified pattern.
#'
#' The pattern should pick up a unique section/sub-section/sub-sub-section.
#' Otherwise, it will stop and raise an error.
#'
#' @export
#'
#' @details
#' The text that is picked up begins with the specified section, and ends with
#' the next string of pound symbols (#)
#'
extract_section_text <- function(rmd_name, hdr_pat, ignore_case=TRUE) {
  all_lines <- extract_non_chunks(rmd_name)
  num_lines <- length(all_lines)

  hdr_lines <- stringr::str_which(all_lines, "^#{1,6} ")
  if(length(hdr_lines) == 0){
    stop("No headings found.")
  } else {
    hdr_lines <- c(hdr_lines, num_lines)
  }

  r1 <- stringr::regex(paste0("^#{1,6} ", hdr_pat), ignore_case = ignore_case)
  match_hdr <- stringr::str_which(all_lines, r1)
  if(length(match_hdr) != 1){
    stop(paste(length(match_hdr), "matches found."))
  } else {
    next_stop <- hdr_lines[hdr_lines > match_hdr][1]
    next_stop <- ifelse(next_stop != hdr_lines[length(hdr_lines)], next_stop-1,
                        next_stop)
    out_lines <- all_lines[match_hdr:next_stop]
  }
  out_lines
}

# wraps a chunk with autoharp.scalars hook with a try expression.
wrap_chunk <- function(chunk) {
  new_chunk <- chunk
  if(stringr::str_detect(chunk[1], "autoharp.scalars") || 
     any(stringr::str_detect(chunk, "^#\\| autoharp\\.scalars:"))
     ){
    chunk_head <- chunk[1]
    chunk_tail <- chunk[length(chunk)]
    chunk_body <- chunk[-c(1, length(chunk))]
    new_chunk <- c(chunk_head, 
                   "try_out <- try({",
                   chunk_body,
                   "})", chunk_tail)
    
  }
  new_chunk
}

#' Returns solution environment and test code from template.
#' 
#' Generates objects for checking solution correctness.
#'
#' @param soln_fname An rmd file containing the checks to be
#'   run on the student solution.
#' @param pattern The pattern that identifies which chunks in the solution are 
#' are testing chunks. If this argument is missing, 
#' the default pattern used is "test".
#' @param knit_root_dir The root directory to use for knitting the rmd file. This
#' argument is optional. If it is missing, it uses the root directory in
#' knitr::opts_knit$get('root.dir').
#' @param render_only A logical value. If this is TRUE, then the solution is
#' run, rendered and returned. Otherwise the rendered html is deleted.
#' @param output The path to the knitted solution md file. This is usually 
#' deleted immediately, but sometimes we may want to keep it. This 
#' argument is passed on to \code{\link[knitr]{knit}}, so please refer to 
#' that page for the warnings about setting this argument when figures are 
#' involved.
#' @param dummy A logical value. If TRUE, then a dummy solution object is created
#' The environment will be empty, and the temporary test script will only have 
#' one line in it: `ah_dummy <- TRUE`. This solution object can be used to render 
#' a solution file without any correctness checks. Of course, one could just knit 
#' it directly, but this is still useful when testing things out.
#'
#' @details  Test code should be written in a chunk that generates scalars 
#' from student objects.
#' 
#' The solution file has to be an Rmd file (not an R script), because it 
#' relies on the autoharp.obj and autoharp.scalars knitr hooks being present.
#' 
#' In addition, if it is required that a solution object is to be tested against
#' the analogous object within the student environment, these objects should be 
#' listed within the autoharp.objs option of a code chunk. These objects will be 
#' copied with the "." prefix.
#' 
#' Here is an overview of how the function works:
#' \enumerate{
#'   \item Knit the solution file to generate the solution (or "correct") 
#'   objects. 
#'   \item Rename these with the "." prefix in the solution environment
#'   object.
#'   \item Extract the lines of test code into a temporary R script.
#'   \item Wrap those chunks that contain autoharp.scalars hook with tryCatch.
#'   \item Return the solution environment and path to the R test script.
#' }
#' 
#' Typically, the next step is to call \code{\link{check_correctness}}.
#' 
#' @return If render_only is FALSE, a list containing 2 components: the
#' environment populated by the solution rmd and the path to an R script
#' containing the test code.
#'
#' If render_only is TRUE, then the output list consists of the 
#' aforementioned environment, and the path to the rendered solution file
#' (html). This option is useful for debugging the solution file.
#' 
#' @export
#' @seealso \code{\link{check_correctness}}, \code{\link{render_one}}
#'
populate_soln_env <- function(soln_fname, pattern, knit_root_dir, 
                              render_only=FALSE, output=NULL, dummy=FALSE) {

  if(dummy) {
      e_soln <- new.env()
      tmp_fname <- tempfile(pattern = "test", fileext = ".R")
      cat("ah_dummy <- TRUE\n", file=tmp_fname)
      return(list(env=e_soln, test_fname = tmp_fname))
  }
  e_soln <- new.env()
  e_soln$.myfilename <- normalizePath(soln_fname)
  if(!missing(knit_root_dir)){
    old_root <- knitr::opts_knit$get("root.dir")
    knitr::opts_knit$set(root.dir = knit_root_dir)
  }
  soln_out <- knitr::knit(soln_fname, quiet=TRUE, envir=e_soln, output=output)
  if(render_only) {
      return(list(env=e_soln, soln_out=soln_out))
  }
  if(file.exists(soln_out))
    unlink(soln_out)

  # Check that all model scalars are generated and present:
  if(exists(".scalars_to_keep", e_soln)){
    scalars_in_soln <- vapply(e_soln$.scalars_to_keep, exists, envir=e_soln, FUN.VALUE=TRUE)
    if(any(!scalars_in_soln)){
      missing <- names(scalars_in_soln)[!scalars_in_soln]
      message("The following scalars are missing from solution environment: ", missing)
      message("Try running populate_soln_env with render_only = TRUE to figure this out.")
    }
  }

  # the chunk labels for test code should be "test"
  if(missing(pattern)) {
    pattern <- "test"
  }

  # get expect_calls
  #tt_list <- get_expect_calls(soln_fname)

  test_code <- extract_chunks(soln_fname, pattern=pattern)
  if(length(test_code) > 0) {
    test_code <- lapply(test_code, wrap_chunk)
    lines_all <- test_code %>% lapply(function(x) x[2:(length(x) - 1)]) %>% unlist
    # if(".scalars_to_keep" %in% names(e_soln)){
    # lines_all <- c(lines_all, " ",
    #                "get_objs <- mget(.scalars_to_keep, ifnotfound=NA)",   
    #                "mapply(base::assign, x=.scalars_to_keep, value=get_objs, MoreArgs = list(envir=.myenv))")
    # } 
  } else {
    stop("No testing code detected.")
    lines_all <- c("NULL")
  }
  
  tmp_fname <- tempfile(pattern = "test_", fileext = ".R")
  writeLines(text = lines_all, con = tmp_fname)

  if(!missing(knit_root_dir)){
    knitr::opts_knit$set(root.dir = old_root)
  }
  
  return(list(env=e_soln, #chunk=test_code,
              test_fname = tmp_fname))#, tt_list =tt_list))
}

if(rmarkdown::pandoc_available()) {
    soln_out <- populate_soln_env("soln_template_01.Rmd") 
    corr_out_1 <- as.data.frame(render_one("example_01.R", "test_output", ".", soln_stuff = soln_out))
    corr_out_4 <- as.data.frame(render_one("example_04.R", "test_output", ".", soln_stuff = soln_out))
    test_that("Rendering1", {
      expect_equal(corr_out_1[1, 7], TRUE)
      expect_equal(corr_out_4[1, 7], FALSE)
    })

    soln_out <- populate_soln_env("soln_template_02.Rmd") 
    corr_out_2 <- as.data.frame(render_one("example_02.R", "test_output", ".",  soln_stuff = soln_out))
    corr_out_3 <- as.data.frame(render_one("example_03.R", "test_output", ".",  soln_stuff = soln_out))
    test_that("Rendering2", {
      expect_equal(corr_out_2[1, "Y1"], "C")
      expect_equal(corr_out_3[1, 8], NA)
    })

}

#' Root a list of trees.
#' 
#' Given a list of trees, this will root them.
#'
#' @param ... A list of Treeharp objects.
#' 
#' @details This function combines TreeHarp objects into a single TreeHarp.
#' The function will root all of them at a node called "script", which is 
#' neither a function call nor an argument nor a symbol. The BFS ordering is 
#' then updated.
#' 
#' Objects that are not of class TreeHarp will be dropped from the list 
#' before the rooting takes place. 
#'
#' @return A TreeHarp object
#' @importFrom methods is
#' @export
#' 
join_treeharps <- function(...) {
  # remove NA trees
  all_trees <- c(...)
  all_trees <- Filter(function(x) is(x, "TreeHarp"), all_trees)
  all_trees_info <- NULL
  
  for(ii in 1:length(all_trees)) {
    tmp <- get_node_types(all_trees[[ii]])
    if(is(tmp, "logical") && is.na(tmp))
      stop("Need node type information..")
    tmp$tree_id <- ii
    all_trees_info <- rbind(all_trees_info, tmp)
  }
  all_trees_info$new_depth <- all_trees_info$depth + 1
  all_trees_info <- rbind(all_trees_info,
                          data.frame(id=0L, name = "script", call_status = FALSE,
                                     formal_arg=FALSE, depth=0, tree_id=0,
                                     new_depth=1))
  all_trees_info <- dplyr::arrange(all_trees_info, .data$new_depth, .data$tree_id) %>%
    dplyr::mutate(new_id = dplyr::row_number())

  new_adj_list <- vector("list", length=nrow(all_trees_info))
  new_adj_list[[1]] <- all_trees_info$new_id[all_trees_info$new_depth == 2]
  names(new_adj_list)[1] <- all_trees_info$name[1]
  for(ii in 2:nrow(all_trees_info)){
    org_tree_id <- all_trees_info$tree_id[ii]
    org_node_id <- all_trees_info$id[ii]
    new_adj_list[ii] <- get_adj_list(all_trees[[org_tree_id]])[org_node_id]
    names(new_adj_list)[ii] <- all_trees_info$name[ii]

    tree_df <- dplyr::filter(all_trees_info, .data$tree_id == org_tree_id)
    old_ids <- tree_df$id
    new_ids <- tree_df$new_id
    names(new_ids) <- old_ids

    if(!is.null(new_adj_list[[ii]])){
      new_adj_list[[ii]] <- dplyr::recode(new_adj_list[[ii]], !!!new_ids)
    }
  }
  new("TreeHarp", adjList = new_adj_list, nodeTypes= all_trees_info)
  #all_trees_info
}

#' Replace Special Characters in File Name
#' 
#' Replaces special characters in the name of an R or Rmd script. 
#'
#' @param dir_name A character string, referring to the directory of Rmd
#' files whose names should be replaced.
#' @param return_df A logical value, indicating if the old and new names should
#' be returned (in a tibble).
#' 
#' @details If a filename contains one of the following special characters 
#' (ignore the quotes here): "\code{[ <>()|\\:&;#?*']}", the  
#' \link[knitr]{knit} function will replace 
#' them with underscores. Hence the filenames in the autoharp input directory 
#' and the output directory will not match, even allowing for the change in 
#' file extension. This will cause problems when we try to run 
#' \code{\link{render_one}} again on the same input directory.
#' 
#' This function renames the files in the input directory by replacing all 
#' special characters there.
#' 
#' The NUS LMS (LumiNUS) introduces parenthesized names or numbers in order 
#' to make filenames unique, so this function is necessary for NUS 
#' instructors.
#'
#' @return A tibble containing the old and new names.
#' @export
#'
replace_sp_chars_filename <- function(dir_name, return_df = TRUE) {
  fnames <- list.files(dir_name)

  new_names <- stringr::str_replace_all(fnames, "[\\[ <>()|\\:&;#?*'\\]]", "_")
  out <- file.rename(from=file.path(dir_name, fnames), to=file.path(dir_name, new_names))
  if(!all(out)) {
    message("Some files could not be renamed..")
  }
  if(return_df) {
    return(tibble::tibble(old_names = fnames, new_names=new_names))
  }
}

#' Removes md Files when no HTML Present
#' 
#' Cleans up the autoharp output directory.
#'
#' @param dir_name The directory containing the files to be cleaned.
#' @param verbose If TRUE, then the files and directories being removed will be 
#' printed.
#' 
#' @details When batch rendering Rmd files, it is inevitable that some files fail. These
#' files would have their knit.md and utf.md present, but they would not have a
#' corresponding html file generated.
#'
#' This function is called for its' side-effect, to remove those lonely md files.
#'
#' If this clean-up is not done, when we try to re-run the files (perhaps with
#' some of the errors fixed), these straggling md files will cause problems. 
#' The most crucial one is that the Rmd files will not be re-knitted, even 
#' though they have been changed.
#'
#' @return No return value.
#'
#' @export
#'
clean_dir <- function(dir_name, verbose=FALSE) {
  all_md_files <- list.files(dir_name, pattern="md$")
  root_names <- sapply(all_md_files, xfun::sans_ext, USE.NAMES = FALSE) %>%
    unique()

  html_ver <- paste0(root_names, ".html")
  to_remove <- !file.exists(file.path(dir_name, html_ver))
  if(any(to_remove)) {
    exts_to_remove <- c(".knit.md", ".utf8.md", "_files")
    all_to_remove <- sapply(root_names[to_remove], paste0, exts_to_remove)
    if(verbose) {
      message("The following files will be removed:")
      message(paste(all_to_remove, collapse="\n"))
    }
    unlink(file.path(dir_name, all_to_remove), recursive = TRUE)
  }
}

#' Extracts the Packages Used in An Rmd File.
#'
#' The input filename could correspond to an R script or an Rmd file.
#'
#' @param fname The Rmd filename or R script.
#'
#' @return A character vector containing the packages used within the Rmd
#' document.
#'
#' @details The file is assumed to be either an R script or an Rmd file.
#' If it is found to be an Rmd file using \link{extract_chunks}, it is purl-ed before
#' libraries are extracted. If it is found to be NOT an Rmd, it is assumed
#' to be an R script and nothing is done to process it.
#' 
#' The file is not parsed, so even text files will work with this function.
#'
#' @export
#'
get_libraries <- function(fname) {
  #all_chunks <- extract_chunks(fname)
  all_chunks <- tryCatch(extract_chunks(fname), error = function(e) return(e))
  if("error" %in% class(all_chunks)){
      return(all_chunks)
  }

  if(!is.null(all_chunks)){
    all_chunks <- Filter(function(x) length(x) > 2, all_chunks)
    all_lines <- lapply(all_chunks, function(x) x[2:(length(x)-1)])
    src_lines <- unlist(all_lines)
  } else {
    src_lines <- readLines(fname, warn=FALSE)
  }
  #src_lines <- readLines(out_file)
  non_comments <- src_lines[!stringr::str_detect(src_lines, "^#")]

  libr <- non_comments %>%
    unlist(use.names = FALSE) %>%
    stringr::str_replace_all("\"","") %>%
    #stringr::str_extract_all("(?<=library\\()[^)]+|[a-z0-9A-Z\\.]+(?=\\:\\:)")
    stringr::str_extract_all("(?<=library\\()[^)]+|[a-z0-9A-Z\\.]+(?=\\:\\:)|(?<=require\\()[^)]+")

  libr <- Filter(length, libr) %>% unlist()
  #unlink(out_file)

  sort(unique(libr))
}

#' Extract chunks that match a pattern.
#' 
#' Extracts chunks whose labels match a pattern from the rmd file.
#' 
#' @param rmd_name  A character string, the name of the rmd file to get the
#' chunks from.
#' @param pattern The pattern to match within the label. In fact, the match is
#' applied to the whole chunk option.
#' 
#' @return A list of character vectors. Each vector contains the chunk from the
#' file. If no pattern is specified, all chunks are returned. Remember that
#' the chunk header and tail are also included in the returned list.
#'
#' @export
#'
extract_chunks <- function(rmd_name, pattern) {
  if(missing(pattern)){
    pattern <- ".*"
  }
  all_lines <- readLines(rmd_name, warn=FALSE)
  kept_chunks <- NULL
  copy_out <- FALSE
  
  for(ii in seq_along(all_lines)){
    cur_line <- all_lines[ii]
    
    if(stringr::str_detect(cur_line, knitr::all_patterns$md$chunk.begin)){
      if(stringr::str_detect(cur_line, pattern)) {
        copy_out <- TRUE
        chunk_lines <- NULL
      }
    }
    
    if(copy_out){
      chunk_lines <- c(chunk_lines, cur_line)
    }
    
    if(stringr::str_detect(cur_line, knitr::all_patterns$md$chunk.end)){
      if(copy_out) {
        copy_out <- FALSE
        kept_chunks <- c(kept_chunks, list(chunk_lines))
      }
    }
    
  }
  
  if(length(kept_chunks) == 0)
    return(NULL)
  else 
    return(kept_chunks)
}

#' Extract non-chunks from an Rmd file.
#' 
#' Extracts non-chunks from an Rmd file.
#' 
#' @param rmd_name  A character string, the name of the rmd file to get the
#' chunks from.
#' @param out_name An output filename, to dump the text to.
#' 
#' @return If out_name is missing, then a character vector is returned. If 
#' outfname is specified, then nothing is returned. The text is written to the 
#' file instead.
#'
#' @export
#'
extract_non_chunks <- function(rmd_name, out_name) {
  all_lines <- readLines(rmd_name, warn=FALSE)
  begin_ids <- which(stringr::str_detect(all_lines, knitr::all_patterns$md$chunk.begin))
  end_ids <- which(stringr::str_detect(all_lines, knitr::all_patterns$md$chunk.end))
  if(length(begin_ids) == 0 || length(end_ids) == 0){
      keep_lines <- all_lines
  } else {
      within_chunks <- mapply(":", begin_ids, end_ids, 
                              SIMPLIFY = FALSE,USE.NAMES = FALSE )
      within_chunks <- unique(unlist(within_chunks))
      
      keep_lines <- all_lines[-within_chunks]
  }

  if(missing(out_name)) {
    return(keep_lines)
  } else {
    writeLines(keep_lines, out_name)
  }
}

#' Copy an object from one env to another.
#'
#' A wrapper function that uses assign and get.
#' 
#' @param from_obj The name of the object to copy. It has to be a string.
#' @param from_env The environment in which the object lives. It has to be an
#' object of class environment.
#' @param to_obj The name of the object to assign it to, in the new environment.
#' Also a string.
#' @param to_env The environment to which the new object is to be assigned. It
#' has to be an object of class environment.
#'
#' @return There is no return value. This function is called for its' side
#' effect.
#' @export
#'
#' @examples
#'
#' e1 <- new.env(); e2 <- new.env()
#' ls(e2)
#' evalq(x <- 1L, e1)
#' copy_e2e("x", e1, "y", e2)
#' ls(e2)
copy_e2e <- function(from_obj, from_env, to_obj, to_env) {
  assign(to_obj, get(from_obj, envir=from_env), envir = to_env)
}

# # run soln, gather objects into e_soln
# # run student, gather objects into e_stud
# # copy e2e
# # extract test_chunk into temp file.
# # execute tests
#
# stud_fname <- "test_many_things_student.Rmd"
# e_stud <- new.env()
# soln_fname <- "test_many_things.Rmd"
# obj_to_copy <- ls(e_soln, pattern="_soln")
# stud_out <- knit(stud_fname, quiet=TRUE, envir=e_stud)
# copy_out <- sapply(obj_to_copy, function(zzz) copy_e2e(zzz, e_soln, zzz, e_stud))
# tmp_fname <- tempfile(pattern="test_", ".", fileext = ".R")
# test_code <- extract_chunks(soln_fname, pattern="test")
# writeLines(test_code[2:(length(test_code)-1)], tmp_fname)
# test_file(tmp_fname, env = e_stud)
#
# #
# returns solution environment and test code chunks (as character vectors)
# knit_hooks set inside (yes)
#

autoharp_hooks <- list(
  autoharp.objs = function(before, options, envir) {
    if(before) {
      ss <- options$autoharp.objs
      sapply(ss, function(x) copy_e2e(x, from_env=envir,
                                      to_obj = paste0(".", x), 
                                      to_env = envir))
    }
  },
  autoharp.scalars = function(before, options, envir) {
    if(before) {
      checkout <- options$autoharp.scalars
      
      if(".scalars_to_keep" %in% names(envir)) {
        xset <- assign(".scalars_to_keep", 
                 c(get(".scalars_to_keep", envir = envir), checkout), envir = envir)
      } else {
        xset <- assign(".scalars_to_keep", checkout, envir = envir)
      }
    }
    invisible(1)
  }#,
  # autoharp.set = function(...) {
  #   dots <- list(...)
  #   ndots <- length(dots)
  #   if(ndots > 0L) {
  #     autoharp_hooks <<- c(autoharp_hooks, dots)
  #   }
  # }
)

test_that("get_libraries", {
  expect_equal(get_libraries("test03.Rmd"), c("dplyr", "knitr"))
})

pressure_chunk <- c("```{r pressure, echo=FALSE}", "plot(pressure)", "```")
test_that("extract_chunks", {
  expect_equal(extract_chunks("test03.Rmd", "pressure")[[1]], pressure_chunk)
})

if(rmarkdown::pandoc_available()){
    soln_out <- populate_soln_env("test04.Rmd", pattern = "test", ".")
    stud_env <- new.env()
    rmarkdown::render("test03.Rmd", clean = TRUE, envir = stud_env)
    .myenv <- stud_env

    test_that("correctness functions", {
      expect_true("x" %in% ls(soln_out$env))
      expect_true(".x" %in% ls(soln_out$env, all.names = TRUE))
      expect_equal(check_correctness(stud_env, soln_out$env, 
                                     soln_out$test_fname)[1,1], TRUE)
    })

    pop_out <- populate_soln_env("soln_template.Rmd", "test", ".")
    e1 <- new.env()
    source("s1.R", local=e1)
    .myenv <- e1
    out1 <- check_correctness(e1, pop_out$env, pop_out$test_fname)

    e2 <- new.env()
    source("s2.R", local=e2)
    .myenv <- e2
    out2 <- check_correctness(e2, pop_out$env, pop_out$test_fname)

    e3 <- new.env()
    source("s3.R", local=e3)
    .myenv <- e3
    out3 <- check_correctness(e3, pop_out$env, pop_out$test_fname)

    test_that("correctness2", {
      expect_equal(out1[1, "Y1"], "C")
      expect_equal(out2[1, 3], NA)
      expect_true(is.na(out3[1, "Zmean"]))
    })
}

# To reset search path by detaching missing paths:

#' Reset search path of current R session
#' 
#' This function is used to detach packages that have been added by a student 
#' script.
#'
#' @param old_path A character vector of package namespaces. This is usually 
#' the output of \code{\link[base]{search}}, run before an R script or Rmd 
#' file is rendered, which could cause the search path to change.
#'
#' @return There is no object returned. This function is called for it's side-
#' effect of altering the search path.
#' 
#' @details When a student script is rendered using \code{\link{render_one}}, 
#' new packages might be added to the search path. These may conflict with the 
#' instructors' search path order, or with subsequent runs of 
#' \code{\link{render_one}} on students. Hence there is a need to reset the 
#' search path before this is done.
#' 
#' This function does not unload namespaces. It only detaches them from the 
#' search path. For a difference between the two, please see Hadley's page.
#' 
#' @export
#' @examples 
#' opath <- search()
#' # Load a package
#' reset_path(opath)
reset_path <- function(old_path) {
    cur_search_path <- search()
    path_mods <- setdiff(cur_search_path, old_path)
    while(length(path_mods) > 0) {
      id <- which(cur_search_path == path_mods[1])
      detach(pos=id[1])
      cur_search_path <- search()
      path_mods <- setdiff(cur_search_path, old_path)
    }
    return(invisible(1L))
}

write_css <- function() {
  as.character(shiny::tags$style(shiny::HTML('
    .zoomDiv {
      opacity: 0;
      position: fixed;
      top: 50%;
      left: 50%;
      z-index: 50;
      transform: translate(-50%, -50%);
      box-shadow: 0px 0px 3000px 800px #888888;
      max-height: 100%;
      overflow: scroll;
    }
    .zoomImg {width: 100%}
    a {font-style: italic}
    figure {
      text-align: center;
      margin: 2%;
      display: inline-block;
      width: 96%;
    }
    .figimg{
      margin: 0;
      display: inline-block;
    }
    figcaption{
      max-width: 90%;
      overflow: hidden;
      margin-left: 5%;
    }
    div{
      vertical-align:middle;
    }
    .imgContainer {
      text-align: center;
      text-indent: 0;
      display: inline-block;
      border: thin silver solid;
    }
    .small {max-width: 18%}
    .medium {max-width: 23%}
    .large {max-width: 32%}
    .imgsmall {width: 45%}
    .imgmedium {width: 30%}
    .imglarge {width: 22%}
    img {width: 100%}
    img:hover {
      transition: .2s;
      -moz-box-shadow: 0 0 20px #ccc;
      -webkit-box-shadow: 0 0 20px #ccc;
      box-shadow: 0 12 12px #ccc;
    }
  ')))
}


write_jscript <- function(){
  as.character(paste0(
    shiny::tags$script(type="text/javascript", 
                src="http://ajax.googleapis.com/ajax/libs/jquery/3.4.1/jquery.min.js"
    ), "\n",
    shiny::tags$script(type="text/javascript", shiny::HTML('
      $(document).ready(function() {
        $("body").prepend("<div class=\\\"zoomDiv\\\"><img src=\\\"\\\" class=\\\"zoomImg\\\"></div>");
        $("img:not(.zoomImg)").click(function() {
          $(".zoomImg").attr("src", $(this).attr("src"));
          $(".zoomDiv").css({opacity: "1", width: "70%"});
        });
        $("img.zoomImg").click(function() {
          $(".zoomDiv").css({opacity: "0", width: "0%"});
        });
      });
    ')), "\n"
  ))
}

#' File lint counter
#' 
#' Count number of lints in one file
#' 
#' @param rmd_file The path to the rmd file to check for lints.
#' @param lint_list List of lints to check for. 
#' @param lint_labels List of labels to name the vector to return.   
#' 
#' @details The function will count the number of lints in a file. The lints to be 
#' checked can be passed as an argument. Else, the default will be used. 
#' The defaults are as follows: 
#' * T_and_F_symbol_linter
#' * line_length_linter
#' * assignment_linter
#' * absolute_path_linter
#' * pipe_continuation_linter
#' Note that labels would also need to be given if the non-default lints 
#' are chosen.
#' 
#' @return Vector containing the lints. 
#' 
#' @export

count_lints_one <- function(rmd_file, lint_list, lint_labels) {
  if (missing(lint_list)) {
    lint_list <- c(lintr::T_and_F_symbol_linter,
                   lintr::line_length_linter,
                   lintr::assignment_linter,
                   lintr::absolute_path_linter,
                   lintr::pipe_continuation_linter)
  }
  
  if (missing(lint_labels)) {
    lint_labels <-  c("T_and_F", "Line_Length", "Assignment", "Abs_Path",
                      "Pipe")
  }
  
  if (length(lint_labels) != length(lint_list)) {
    message("Lint list and labels are not of the same length")
    return(NULL)
  }
  lint_count <- rep(NA, times = length(lint_labels))
  tryCatch(
    {
      lint_count <- suppressWarnings(
        sapply(lint_list, function(x) length(lintr::lint(rmd_file, x))))
    }, 
    error = function(cond) {
      message(paste0("Error with file: ", rmd_file))
    }, 
    finally = {
      names(lint_count) <- lint_labels
    }
  )
  return(lint_count)
}


#' Lint counter
#' 
#' Count number of lints in one folder
#' 
#' @param file_names The path to the rmd files that need to be checked for 
#' lints.
#' @param lint_list List of lints to check for. 
#' @param lint_labels List of labels to name the vector to return.   
#' 
#' @details The function will count the number of lints in a file. The lints to be 
#' checked can be passed as an argument. Else, the default will be used. 
#' The defaults are as follows: 
#' * T_and_F_symbol_linter
#' * line_length_linter
#' * assignment_linter
#' * absolute_path_linter
#' * pipe_continuation_linter
#' 
#' Note that labels would also need to be given if the non-default lints 
#' are chosen.
#' @md
#' 
#' @return Dataframe containing the lints. 
#' 
#' @export

count_lints_all <- function(file_names, lint_list, lint_labels) {
  if (missing(lint_list)) {
    lint_list <- c(lintr::T_and_F_symbol_linter,
                   lintr::line_length_linter,
                   lintr::assignment_linter,
                   lintr::absolute_path_linter,
                   lintr::pipe_continuation_linter)
  }
  
  if (missing(lint_labels)) {
    lint_labels <-  c("T_and_F", "Line_Length", "Assignment", "Abs_Path",
                      "Pipe")
  }
  
  if (length(lint_labels) != length(lint_list)) {
    message("Lint list and labels are not of the same length")
    return(NULL)
  }

  tabulated_lints <- t(sapply(file_names, count_lints_one, 
                              lint_list = lint_list, 
                              lint_labels = lint_labels))

  total <- rowSums(tabulated_lints)
  final_df <- cbind(file_names, tabulated_lints, total)
  final_df <- as.data.frame(final_df)
  rownames(final_df) <- NULL
  return(final_df)
}

# library(autoharp)
#
# T1 <- TreeHarp(list(a=c(2,3), b=NULL, c=4, d=NULL))
# T2 <- TreeHarp(list(a=c(2,3), b=NULL, d=NULL))
# T3 <- TreeHarp(list(a=c(2,3), b=NULL, c=NULL))
# #plot(T2)

#' Compute tree similarity
#'
#' Computes similarity between two trees (non-recursively)
#'
#' @param t1 A TreeHarp object
#' @param t2 Anothe TreeHarp object.
#' @param norm A logical value, indicating if the kernel function should be
#' normalised, to account for different tree lengths.
#' @param ... Unused arguments, reserved for mcmapply
#' 
#' @return A numerical value between 0 and 1 (if normed).
tree_sim <- function(t1, t2, norm=FALSE, ...) {
  Kval <- K_nr(t1, t2, ...)
  if(norm){
    Kval <- Kval/sqrt(K_nr(t1,t1,...)*K_nr(t2,t2, ...))
  }
  Kval
}

K <- function(t1, t2) {
  names1 <- names(t1)
  adj1 <- get_adj_list(t1)

  names2 <- names(t2)
  adj2 <- get_adj_list(t2)

  l1 <- length(t1)
  l2 <- length(t2)

  # Fill up matrix as much as possible.
  out_mat <- matrix(NA_integer_, nrow=l1, ncol=l2)
  for(i in seq_along(names1))
    for(j in seq_along(names2)) {
      if(names1[i] != names2[j]) {
        out_mat[i, j] <- 0L
      } else {
        if(is.null(adj1[[i]]) || is.null(adj2[[j]]))
          out_mat[i,j] <- 1L
      }
    }
  # assign to a new environment
  e1 <- new.env()
  assign("out_mat", out_mat, envir = e1)
  remaining <- which(is.na(out_mat), arr.ind = TRUE)
  if(nrow(remaining) > 0){
    for(jj in nrow(remaining):1){
      S_r(remaining[jj,"row"], t1, remaining[jj, "col"], t2, e1)
    }
  }
  sum(e1$out_mat)
}

# Recursive method
S_r <- function(n1, t1, n2, t2, mat_env) {
  tmp <- mat_env$out_mat[n1, n2]
  if(!is.na(tmp)){
    return(tmp)
  }
  nn1 <- get_child_ids(t1, n1)
  cn1_names <- names(t1)[nn1]
  t1_df <- data.frame(ids=nn1, names=cn1_names,
                      stringsAsFactors = FALSE)

  nn2 <- get_child_ids(t2, n2)
  cn2_names <- names(t2)[nn2]
  t2_df <- data.frame(ids=nn2, names=cn2_names,
                      stringsAsFactors = FALSE)

  common_names <- dplyr::inner_join(t1_df, t2_df, by="names")
  if(nrow(common_names) == 0) {
    mat_env$out_mat[n1, n2] <- 1
    return(1)
  }
  tmp_prod <- 1
  for(ii in seq_along(common_names$ids.x)){
    tmp_prod <- tmp_prod*(1 + S_r(common_names$ids.x[ii], t1,
                                  common_names$ids.y[ii], t2, mat_env))
  }
  mat_env$out_mat[n1, n2] <- tmp_prod
  return(tmp_prod)

}

# Brute force by working on only the minimal spanning tree.
S_nr2 <- function(n1, t1, n2, t2, verbose=FALSE) {
  # extract subtrees
  if(n1 == 1)
    s1 <- t1 else
    s1 <- subtree_at(t1, n1)

  if(n2 == 1)
    s2 <- t2 else
    s2 <- subtree_at(t2, n2)

  # find common nodes
  s1_names <- names(s1)
  s2_names <- names(s2)
  if(names(s1)[1] != names(s2)[1])
    return(0)
  common_names <- base::intersect(s1_names, s2_names)
  if(length(common_names) == 0)
    return(0)

  #common_names
  if(verbose) {
    nnodes_1 <- sum(s1_names %in% common_names)
    nnodes_2 <- sum(s2_names %in% common_names)
    message(paste("tree 1 has", nnodes_1, "in common.\n", sep=" "))
    message(paste("tree 2 has", nnodes_2, "in common.\n", sep=" "))
  }

  # prune trees
  #ss1 <- prune_depth(s1, common_names)
  ss1 <- carve_mst(s1, common_names)
  #ss2 <- prune_depth(s2, common_names)
  ss2 <- carve_mst(s2, common_names)

  # # generate all trees
  if(verbose){
    message("Generating all sub-trees of 1.\n")
  }
  s1_trees <- generate_all_subtrees(ss1)
  s1_mat <- apply(s1_trees, 1, function(x) as(carve_subtree(obj=ss1, x),
                                              "matrix"))
  # s1_mat <- apply(s1_trees, 1, carve_subtree, obj=ss1)
  if(verbose){
    message("Removing duplicates for tree 1.\n")
  }
  s1_dup_ids <- duplicated(s1_mat)
  s1_mat <- s1_mat[!s1_dup_ids]

  if(verbose){
    message("Generating all sub-trees of 2.\n")
  }
  s2_trees <- generate_all_subtrees(ss2)
  s2_mat <- apply(s2_trees, 1, function(x) as(carve_subtree(obj=ss2, x),
                                              "matrix"))
  #s2_mat <- apply(s2_trees, 1, carve_subtree, obj=ss2)
  if(verbose){
    message("Removing duplicates for tree 2.\n")
  }
  s2_dup_ids <- duplicated(s2_mat)
  s2_mat <- s2_mat[!s2_dup_ids]

  # # expand.grid, filter
  cross_df <- expand.grid(t1=1:length(s1_mat), t2=1:length(s2_mat))

  # call mapply or mcmapply
  if(verbose){
    message("Cross-comparing sub-trees.\n")
  }
  equality_ids <- mapply(function(x, y) identical(s1_mat[[x]], s2_mat[[y]]),
                         x=cross_df$t1, y=cross_df$t2)
  sum(equality_ids)
}

# Blind brute force
S_nr <- function(n1, t1, n2, t2, verbose=FALSE) {
  # extract subtrees
  s1 <- subtree_at(t1, n1)
  s2 <- subtree_at(t2, n2)

  # generate all trees
  # remove duplicates
  if(verbose){
    message("Generating all sub-trees of 1.\n")
  }
  s1_trees <- generate_all_subtrees(s1)
  s1_mat <- apply(s1_trees, 1, function(x) as(carve_subtree(obj=s1, x),
                                                 "matrix"))
  if(verbose){
    message("Removing duplicates for tree 1.\n")
  }
  s1_dup_ids <- duplicated(s1_mat)
  s1_mat <- s1_mat[!s1_dup_ids]

  if(verbose){
    message("Generating all sub-trees of 2.\n")
  }
  s2_trees <- generate_all_subtrees(s2)
  s2_mat <- apply(s2_trees, 1, function(x) as(carve_subtree(obj=s2, x),
                                                 "matrix"))
  if(verbose){
    message("Removing duplicates for tree 2.\n")
  }
  s2_dup_ids <- duplicated(s2_mat)
  s2_mat <- s2_mat[!s2_dup_ids]

  # expand.grid, filter
  cross_df <- expand.grid(t1=1:length(s1_mat), t2=1:length(s2_mat))

  # call mapply or mcmapply
  equality_ids <- mapply(function(x, y) identical(s1_mat[[x]], s2_mat[[y]]),
                         x=cross_df$t1, y=cross_df$t2)
  sum(equality_ids)
}

K_nr <- function(t1, t2, ...) {
  names1 <- names(t1)
  adj1 <- get_adj_list(t1)

  names2 <- names(t2)
  adj2 <- get_adj_list(t2)

  l1 <- length(t1)
  l2 <- length(t2)

  # Fill up matrix as much as possible.
  out_mat <- matrix(NA_integer_, nrow=l1, ncol=l2)
  for(i in seq_along(names1))
    for(j in seq_along(names2)) {
      if(names1[i] != names2[j]) {
        out_mat[i, j] <- 0L
      } else {
        if(is.null(adj1[[i]]) || is.null(adj2[[j]]))
          out_mat[i,j] <- 1L
      }
    }
  # assign to a new environment
  remaining <- which(is.na(out_mat), arr.ind = TRUE)
  if(nrow(remaining) > 0){
    out <- mapply(function(x, y) S_nr(x, t1, y, t2, ...),
                  x=remaining[,"row"], y=remaining[,"col"])
  }
  out_mat[remaining] <- out
  #out_mat
  #remaining
  sum(out_mat)
}

#' Compute tree similarity
#'
#' @param t1 A TreeHarp object.
#' @param t2 A TreeHarp object.
#' @param verbose A logical value, indicating if the output should be verbose.
#'
#' @return An integer, that counts the number of sub-trees in common between 
#' the two trees. Please see the reference papers for more information.
#' 
#' @details As far as possible, this function tries to do things recursively.
#' It sets up a n x m matrix and fills up as much as it can. Then it uses 
#' recursive relationships to fill in the rest. When it cannot, it uses
#' \code{\link{generate_all_subtrees}} to generate and count common subtrees.
#' 
#' @references 
#' 
#' \enumerate{
#' \item 
#' \emph{Convolution kernels for natural language}, M Collins and N Duffy, 
#' \emph{Advances in neural information processing systems}, 2002.
#' 
#' \item 
#' \emph{Convolution kernels on discrete structures}, D Haussler, 
#' \emph{Technical report, Department of Computer Science, UC Santa Cruz}, 1999.
#' }
#'
#' @export
#' @examples
#' tree1 <- TreeHarp(quote(x <- 1), TRUE)
#' tree2 <- TreeHarp(quote(y <- 1), TRUE)
#' K2(tree1, tree2, TRUE)
#' 
K2 <- function(t1, t2, verbose=FALSE) {
  names1 <- names(t1)
  adj1 <- get_adj_list(t1)

  names2 <- names(t2)
  adj2 <- get_adj_list(t2)

  l1 <- length(t1)
  l2 <- length(t2)

  if(verbose){
    message("Filling up matrix.\n")
  }
  # Fill up matrix as much as possible.
  out_mat <- matrix(NA_integer_, nrow=l1, ncol=l2)
  for(i in seq_along(names1))
    for(j in seq_along(names2)) {
      if(names1[i] != names2[j]) {
        out_mat[i, j] <- 0L
      } else {
        if(is.null(adj1[[i]]) || is.null(adj2[[j]]))
          out_mat[i,j] <- 1L
      }
    }
  # return(out_mat) # debug
  # assign to a new environment
  e1 <- new.env()
  assign("out_mat", out_mat, envir = e1)
  remaining <- which(is.na(out_mat), arr.ind = TRUE)

  if(verbose){
    message(paste0(nrow(remaining), " entries remaining.\n"))
  }
  # return(remaining) # debug

  if(nrow(remaining) > 0){
    for(jj in nrow(remaining):1){
      S_r2(remaining[jj,"row"], t1, remaining[jj, "col"], t2, e1, verbose)
    }
  }
  sum(e1$out_mat)
}

# Recursive method
S_r2 <- function(n1, t1, n2, t2, mat_env, verbose=FALSE) {
  tmp <- mat_env$out_mat[n1, n2]
  if(!is.na(tmp)){
    return(tmp)
  }
  tmp_prod <- 1

  nn1 <- get_child_ids(t1, n1)
  cn1_names <- names(t1)[nn1]
  t1_df <- data.frame(ids=nn1, names=cn1_names,
                      stringsAsFactors = FALSE)
  nn2 <- get_child_ids(t2, n2)
  cn2_names <- names(t2)[nn2]
  t2_df <- data.frame(ids=nn2, names=cn2_names,
                      stringsAsFactors = FALSE)
  dups_found <- (anyDuplicated(cn1_names) || anyDuplicated(cn2_names))
  # browser()

  if(dups_found) {
    if(verbose){
      message("Duplicate children nodes found.\n")
    }
    if(anyDuplicated(cn1_names)) {
      dup_names1 <- cn1_names[duplicated(cn1_names)]
    } else {
      dup_names1 <- cn1_names
    }
    if(anyDuplicated(cn2_names)) {
      dup_names2 <- cn2_names[duplicated(cn2_names)]
    } else {
      dup_names2 <- cn2_names
    }
    dup_names12 <- unique(base::intersect(dup_names1, dup_names2))
    if(length(dup_names12) > 0) {
      subtree_br1 <- subtree_at(t1, n1)
      new_dup1 <- base::intersect(get_child_ids(subtree_br1, 1),
                                  which(names(subtree_br1) %in% dup_names12))
      subtree_br1  <- keep_branches(subtree_br1, new_dup1)
      t1_df <- dplyr::filter(t1_df, !(names %in% dup_names12))
  
      subtree_br2 <- subtree_at(t2, n2)
      new_dup2 <- base::intersect(get_child_ids(subtree_br2, 1),
                                  which(names(subtree_br2) %in% dup_names12))
      subtree_br2  <- keep_branches(subtree_br2, new_dup2)
      t2_df <- dplyr::filter(t2_df, !(names %in% dup_names12))
  
      tmp_prod <- S_nr2(1, subtree_br1, 1, subtree_br2, verbose)
      # browser()
    }
  }

  common_names <- dplyr::inner_join(t1_df, t2_df, by="names")
  if(nrow(common_names) == 0) {
    mat_env$out_mat[n1, n2] <- tmp_prod
    return(tmp_prod)
  }
  for(ii in seq_along(common_names$ids.x)){
    tmp_prod <- tmp_prod*(1 + S_r2(common_names$ids.x[ii], t1,
                                  common_names$ids.y[ii], t2, mat_env))
  }
  mat_env$out_mat[n1, n2] <- tmp_prod
  return(tmp_prod)
}

#' Computes Jaccard Index
#' 
#' Computes the Jaccard index between two trees.
#'
#' @param th1 A TreeHarp object.
#' @param th2 A TreeHarp object.
#' @param weighted A logical value, indicating if the weighted Jaccard 
#' similarity should be computed.
#' 
#' @details The unweighted form is just the cardinality of the intersection 
#' of the two sets of tokens, divided by the union of the two sets.
#' 
#' The weighted form is described on the WIkipedia page:
#' https://en.wikipedia.org/wiki/Jaccard_index#Weighted_Jaccard_similarity_and_distance
#'
#' @return A real number between 0 and 1.
#' @export
jaccard_treeharp <- function(th1, th2, weighted=FALSE) {
  tokens1 <- names(th1)
  tokens2 <- names(th2)
  token_union <- union(tokens1, tokens2)
  
  if(!weighted) {
    denom <- length(token_union)
    numer <- length(base::intersect(tokens1, tokens2))
  } else {
    count1 <- sapply(token_union, function(x) sum(x == tokens1))
    count2 <- sapply(token_union, function(x) sum(x == tokens2))
    numer <- sum(pmin(count1, count2))
    denom <- sum(pmax(count1, count2))
  }
  numer/denom
}

test_that("Tree Kernel: K2", {
  expect_equal(K2(t2, t2, FALSE), 6)
  expect_equal(K2(t2, t6, FALSE), 7)
  
  expect_equal(K2(t6, t6, FALSE), 17)
  expect_equal(K2(t7, t7, FALSE), 13)
  expect_equal(K2(t5, t5, FALSE), 11)
  
  expect_equal(K2(t4, t4, FALSE), 7)
  
  expect_equal(K2(t4, t2, FALSE), 4)
})


test_that("Tree Kernel: Jaccard", {
  expect_equal(jaccard_treeharp(t2, t6, FALSE), 0.75)
  expect_equal(jaccard_treeharp(t2, t6, TRUE), 0.6)
})

#' Count tokens in R/Rmd
#'
#' Count the individual tokens. Part of the NLP analysis process.
#'
#' @param fname The Rmd or R file name.
#' @param include_actuals Whether actual arguments/literals should be included.
#' If this is FALSE, then only calls and formal arguments will be used in
#' the count.
#'
#' @return A tibble. The tibble will contain a the frequency count for all 
#' tokens present in the student script.
#' @export
#'
#'
rmd_to_token_count <- function(fname, include_actuals=TRUE) {
  # get fharp object
  fh <- rmd_to_forestharp(fname, FALSE)$forest

  # remove NA entries
  fh <- Filter(function(x) is(x, "TreeHarp"), fh)

  all_dfs <- lapply(fh, get_node_types)

  # combine all dataframes
  raw_tokens <- dplyr::bind_rows(all_dfs)

  if(!include_actuals) {
    raw_tokens <- dplyr::filter(raw_tokens,
                         .data$call_status == TRUE |  .data$formal_arg == TRUE)
  }

  summary_count <- dplyr::group_by(raw_tokens, .data$name) %>% 
    dplyr::count() %>%
    dplyr::ungroup() %>%
    dplyr::mutate(fname = basename(fname)) %>%
    dplyr::rename(token = .data$name)

  summary_count
}

#### from rmd/r to forest-harp

#' Convert to TreeHarp objects
#'
#' Reads in an Rmd file or an R script and converts it to a list of TreeHarp
#' objects.
#'
#' @param fname The filename that is to be read in.
#' @param verbose If TRUE, then messages will be printed to enable debugging.
#' dropped.
#'
#' @return There are three possible return outcomes: (1) NA, indicating that 
#' \code{\link[knitr]{purl}} was unable to extract the R expressions from the 
#' Rmd/qmd file. (2) A list with two components, named \code{forest} and
#' \code{line_nums}, each being NA. (3) A list with two components, named as in 
#' (2), containing the TreeHarp objects and a vector of line numbers.
#' 
#' If there are syntax errors in the file, the expressions may not be parsed
#' correctly. The \code{verbose} option will reflect this. 
#'
#' @details If an Rmd or qmd file is supplied as input, it is first converted
#' to an R script before \code{\link[lintr]{get_source_expressions}} is used 
#' to try to parse the expressions. These expressions are then matched to the
#' original file to retrieve line numbers. Unlike previous versions of this 
#' function, line numbers will always be returned.
#'
#' Expressions that could not be parsed will be returned as NA.
#' 
#' Line numbers are extracted using \code{\link[lintr]{get_source_expressions}} from
#' the lintr package.
#' 
#' @importFrom methods is
#' @importFrom utils tail
#' @export
#' @seealso \code{\link{fapply}}, \code{\link{extract_chunks}},
#' \code{\link{extract_chunks}}, \code{\link[lintr]{get_source_expressions}}
rmd_to_forestharp <- function(fname, verbose=FALSE) {
  # purl the file with doc 0L
  # delete the temp file
  file_extension <- stringr::str_to_lower(xfun::file_ext(fname))
  if(file_extension %in% c("rmd", "qmd")){
    if(verbose){
      message("[R|q]md found.")
    }
    tmp_fname <- tempfile("purl_out", fileext = ".R")
    out_fname <- tryCatch({
      knitr::purl(input=fname, output=tmp_fname, documentation = 0L, quiet=TRUE)
    }, error = function(e) return(e))
    if(is(out_fname, "error")){
      message("Error in purl call to ", fname)
      return(NA)
    }
  } else {
    if(verbose){
      message("R script found. Skipping purl..")
    }
    out_fname <- fname
  }
  
  src_expr <- lintr::get_source_expressions(filename = out_fname)
  num_expr <- length(src_expr$expressions) - 1
  if(num_expr == 0) {
    message("No expressions in ", fname)
    return(list(forest=NA, line_nums=NA))
  }
  if(verbose){
    message(num_expr, " expressions found.")
  }
  tree_list <- list()
  expr_found <- integer(0)
  jj <- 1
  # browser()
  if(file_extension %in% c("rmd", "qmd")){ 
    unlink(tmp_fname)
  }
  
  if(verbose){
    message("Parsing expressions, converting to trees...")
  }
  for(ii in 1:num_expr){
    # remove comment lines
    parsed_c <- src_expr$expressions[[ii]]$parsed_content
    if(parsed_c$token[1] == "COMMENT" && NROW(parsed_c) == 1){
      next
    }
    ex1 <- tryCatch(str2lang(src_expr$expressions[[ii]]$content),
                    error = function(e) return(e))
    if(!is(ex1, "error")){
      tree_list[[jj]] <- autoharp::TreeHarp(ex1, TRUE)
      expr_found[jj] <- ii
      jj <- jj + 1
    } else{
      if(verbose){
        message(src_expr$expressions[[ii]]$content, " could not be parsed.")
      }
    }
  }
  if(length(tree_list) == 0){
    message("No expressions could be parsed in ", fname)
    return(list(forest=NA, line_nums=NA))
  }
  
  if(verbose){
    message("Matching expressions with lines in original file..")
  }
  # browser()
  orig_r_lines <- readLines(fname, warn=FALSE)
  search_start <- 1
  line_nums <- integer(0)
  for(ii in 1:length(tree_list)){
    
    match_output <- match_code_lines(src_expr$expressions[[expr_found[ii]]]$lines,
                                     orig_r_lines, search_start)
    if(is.na(match_output[1])) {
      if(verbose){
        message("No match for expression!\nSyntax errors may be present in: ", 
                src_expr$expressions[[expr_found[ii]]]$lines, "\n")
      }
      line_nums[ii] <- NA_integer_
    } else {
      line_nums[ii] <- match_output[1]
      search_start <- tail(match_output, n=1) + 1
    }
  }
  if(verbose){
    message("Done!")
  }
  
  remove_ids <- which(is.na(line_nums))
  if(length(remove_ids) > 0){
    tree_list <- tree_list[-remove_ids]
    line_nums <- line_nums[-remove_ids]
  }
  list(forest=tree_list, line_nums=line_nums)
}

match_code_lines <- function(code_lines, code_table, start_search=1) {
  table_len <- length(code_table)
  code_len <- length(code_lines)
  unnamed_code_lines <- unname(code_lines)
  if(start_search > table_len - code_len + 1) {
    return(NA)
  }
  match_found <- FALSE
  for(jj in start_search:(table_len - code_len + 1)) {
    if(identical(unnamed_code_lines, code_table[jj:(jj+code_len-1)])) {
      match_found <- TRUE
      restricted_match_ids <- jj:(jj+code_len-1)
      break
    }
  }
  if(match_found){
    return(restricted_match_ids)
  } else{
    return(NA)
  }
}

#### forest-harp apply

#' Apply a function to a forest of trees.
#'
#' A convenience function, for applying a function to many trees.
#'
#' @param fharp The output of rmd_to_forestharp. It could also just be a 
#' list of TreeHarp objects.
#' @param TFUN A function that works on a single TreeHarp and returns an output. 
#' See \link{forestharp-helpers} for examples.
#' @param combine A logical value that indicates if the output from all function
#' applications should be combined.
#' @param combiner_fn A function to use to combine the individual output from 
#' each tree into a single scalar for each forest. It should handle NA 
#' values in the input vector or list. If it is missing, it defaults to 
#' sum, with na.rm=TRUE.
#' @param ... Additional arguments to be passed on to TFUN.
#' 
#' @details The input is simply a list of TreeHarp objects. First, the TFUN 
#' function is lapply-ed to each TreeHarp item, resulting in either a list, 
#' or a vector with possible NA elements.
#' 
#' The combiner function should be aware of this sort of output, and summarise
#' the list or vector accordingly, handling NA's and returning a \emph{scalar}.
#' 
#' If you need to create a partial function out of a forestharp helper, 
#' use an anonymous function, as shown in the examples below.
#'
#' @return A vector, list or a single value. If TFUN returned an error for a 
#' particular TreeHarp, that component in the list or vector would be NA. This 
#' input vector or list will then be combined by combiner_fn.
#' @importFrom methods is
#' @export
#' 
#' @examples 
#' 
#' ex1 <- quote(X <- rnorm(10, mean=0.9, sd=4))
#' ex2 <- quote(Y <- rbeta(10, shape1=3, shape2=5))
#' f1 <- lapply(c(ex1, ex2), TreeHarp, quote_arg=TRUE) 
#' 
#' # returns all function calls that begin with "r", like rnorm and rbeta.
#' # calls are returned as a list.
#' fapply(f1, extract_fn_call, combine =FALSE, pattern="^r.*")
#' 
#' # list is catenated.
#' fapply(f1, extract_fn_call, combine =TRUE, pattern="^r.*", 
#'        combiner_fn = function(x) {paste0(unlist(x), collapse=",")})
#'
fapply <-  function(fharp, TFUN, combine = TRUE, combiner_fn, ...) {
  if("line_nums" %in% names(fharp)){
    fharp <- fharp$forest
  }
  
  f_wrap <- function(x, ...) {
    if(!is(x, "TreeHarp")){
      return(NA)
    }
    # trycatch it here.
    tryCatch(TFUN(x, ...), error = function(e) return(NA))
  }
  out_list <- lapply(fharp, f_wrap, ...)
  if(combine) {
    if(missing(combiner_fn)){
      return(sum(unlist(out_list), na.rm = TRUE))
    } else {
      out <- combiner_fn(out_list)
      return(out)
    }
  } else {
    return(out_list)
  }
}

#' Identify if use of pipe operator can improve a code section
#'
#' @param fname A filename - either a Rmd/qmd or R script.
#' @param window_len A window length to analyse. 
#' 
#' @details A rolling window approach is used here. For each 
#' expression, the assigned object is extracted (if any). Subsequent lines are 
#' checked to see if this object appears as an actual argument.
#' 
#' Window length two means that only one more line is checked.
#'
#' @returns A numeric vector corresponding to start of the window to be reviewed.
#' If no lines are found, NULL is returned.
#' 
#' @export
#'
pipe_can_improve_this <- function(fname, window_len = 2) {
  # need line numbers
  fh <- rmd_to_forestharp(fname, verbose=FALSE)
  lines_found <- NULL
  
  fh_len <- length(fh$line_nums)
  for(i in 1:(fh_len - window_len + 1)) {
    node_types <- get_node_types(fh$forest[[i]])
    
    # check for assignment
    is_assignment <- node_types$name[1] %in% c("<-", "=")
    if(!is_assignment) 
      next
    
    # if assignment, 
      # extract assigned obj
    assigned_objs <- extract_assigned_objects(fh$forest[[i]])
    actual_args_window <- lapply(fh$forest[(i+1):(i+window_len-1)], 
                                 extract_actual_args,
                                 include_assigned_obj=FALSE)
    actual_args_window <- unique(unlist(actual_args_window))
    if(any(!is.na(match(assigned_objs, actual_args_window)))) {
      #browser()
      lines_found <- c(lines_found, fh$line_nums[i])
    }
    
  }
  
  # return identified lines
  lines_found
}

#' Checks if two expressions differ only in terms of actual arguments
#'
#' @param th1 A TreeHarp object.
#' @param th2 A TreeHarp object.
#' 
#' @details From the node types, only calls and formal arguments are retained.
#' If these are identical, then a 1 is returned. Otherwise 0 is returned.
#'
#' @returns Either 1 or 0. 1 means that the two expressions differ only in their 
#' actual arguments. 
#' 
#' @export
#'
#' @examples
#' ex1 <- TreeHarp(quote(X <- rnorm(10, mean=0.9, sd=4)), TRUE)
#' ex2 <- TreeHarp(quote(Y <- rnorm(20, mean=9, sd=4)), TRUE)
#' only_actual_args_differ(ex1, ex2)
only_actual_args_differ <- function(th1, th2) {
  nodes1_minus_aa <- get_node_types(th1) %>% 
    dplyr::filter(.data$call_status | .data$formal_arg)
  nodes2_minus_aa <- get_node_types(th2) %>% 
    dplyr::filter(.data$call_status | .data$formal_arg)
  
  if(identical(nodes1_minus_aa, nodes2_minus_aa)) {
    return(1L)
  } else {
    return(0L)
  }
}


#' Identifies if an apply function can improve the code
#'
#' @param fname A filename: either Rmd/qmd or R.
#' @param window_size A window size to analyse.
#' 
#' @details First, all library() calls and read() calls are removed. Next,
#' \code{only_actual_args_differ} is run on all pairwise expressions. The 
#' matrix is searched for blocks of 1's along the diagonal.
#' 
#'
#' @returns A numeric vector corresponding to windows for review is returned. If 
#' none are found, then NULL is returned.
#' @export
#'
apply_can_improve_this <- function(fname, window_size=2) {
  # filter to remove "library" and "read" calls
  f1 <- rmd_to_forestharp(fname, verbose=FALSE)
  remove_these <-  which(fapply(f1, count_fn_call, 
                                pattern="(^library$)|(^read.*)", 
                                combiner_fn = function(x) unlist(x)) > 0)
  if(length(remove_these) > 0) {
    f1a <- list(forest = f1$forest[-remove_these], 
                line_nums = f1$line_nums[-remove_these])
  } else {
    f1a <- f1
  }
  
  only_actual_args_differ_v <- Vectorize(only_actual_args_differ, 
                                         vectorize.args = c("th1", "th2"))
  
  sim_mat <- outer(f1a$forest, f1a$forest, only_actual_args_differ_v) 
  
  window_slider <- matrix(1L, nrow=window_size, ncol=window_size)
  line_nums <- NULL
  for(i in 1:(NROW(sim_mat)-window_size +1)) {
    if(identical(window_slider, sim_mat[i:(i+window_size-1),i:(i+window_size-1)])){
      line_nums <- c(line_nums, f1a$line_nums[i])
    }
  }
  line_nums
}

#' Generate a dataframe from the log file.
#'
#' @param log_file The name of the log file generated from 
#' \code{\link{render_one}}.
#'
#' @return The function returns a dataframe summarising the details in the
#' log file.
#' 
#' @details This provides a table view of the log file, which is updated 
#' in a more natural format by simply concatenating new updates. The output 
#' of this function makes it easier to group entries by filename, time, or
#' status, or even error message.
#' 
#' The output table does not contain correctness output. It only contains the 
#' columns name, timestamp, status (SUCCESS/FAIL), error message, 
#' number of libraries used and number of libraries installed.
#'
#' @seealso \code{\link{render_one}}
#' 
#' @importFrom stats na.omit
#'
#' @export
#'
log_summary <- function(log_file) {
  file_details <- readLines(log_file,encoding = "UTF-8")

  front_blocks <- which(stringr::str_detect(file_details, "^---render_one"))
  end_blocks <- c(front_blocks[-1] - 1, length(file_details))
  render_blocks <- mapply(c, front_blocks, end_blocks, SIMPLIFY = FALSE)
  output <- list()
  for(each_block in render_blocks) {
    curr_block <- file_details[seq(each_block[1], each_block[2])]


    timestart <- stringr::str_extract(curr_block[1], "(?<=at )[^+]+(?=\\-\\-\\-)")
    front_chunk <- which(stringr::str_detect(curr_block, "Starting on"))
    end_chunk <- c(front_chunk[-1] - 2, length(curr_block))
    indiv_chunk <- mapply(c, front_chunk, end_chunk, SIMPLIFY = FALSE)


    df <- lapply(indiv_chunk, function(x) {
      # name, timestamp, status, error_msg, num_libr_used, num_libr_install
      # vector of 6
      indiv_entry <- rep(NA, 6)
      this_chunk <- curr_block[seq(x[1], x[2])]
      indiv_entry[1] <- stringr::str_extract(this_chunk, "(?<=Starting on ).+") %>% na.omit
      indiv_entry[2] <- timestart

      if(diff(x) + 1 >= 3) {
        indiv_entry[3] <- stringr::str_extract(this_chunk, "FAIL|SUCCESS|UNKNOWN") %>% na.omit

        indiv_entry[5] <- stringr::str_extract(this_chunk, "(?<=Libraries used:).+") %>%
          trimws %>% na.omit %>% stringr::str_split(",") %>% sapply(function(x) nchar(x) > 0) %>%
          sum(na.rm = TRUE)

        need_install <- stringr::str_extract(this_chunk, "(?<=Need to install:).+") %>%
          trimws %>% na.omit %>% stringr::str_split(",") %>% sapply(function(x) nchar(x) > 0)
        if(length(need_install) == 0L)
          indiv_entry[6] <- 0
        else
          indiv_entry[6] <- need_install %>% sum(na.rm = TRUE)
          #sum(na.rm = TRUE)
      } else {
        if(length(this_chunk[-1] > 0)){
          indiv_entry[3] <- this_chunk[-1]
        } else {
          indiv_entry[3] <- "UNKNOWN"
          indiv_entry[4:6] <- NA
        }
      }
      if(indiv_entry[3] == "FAIL") {
        indiv_entry[4] <- this_chunk[(which(stringr::str_detect(this_chunk, "FAIL"))+1):length(this_chunk)] %>%
          paste(sep = "", collapse = "")
      }

      indiv_entry
    })

    output <- c(output, df)
  }

  output <- data.frame(matrix(unlist(output), nrow = length(output), byrow = TRUE))
  colnames(output) <- c("name", "timestamp", "status",
                        "error_message", "num_libr_used", "num_libr_install")
  output
}

#a <- log_summary("../t11.log")
#b <- readLines("../t11.log", encoding = "UTF-8")
#c <- log_summary("../a.log")
#d <- readLines("../a.log", encoding = "UTF-8")

match_ttout_tfile <- function(ttout, tfileout) {
  ttout_test_names <- sapply(ttout, function(x) x$test, USE.NAMES=FALSE)
  all_results <- NULL
  
  for(ii in seq_along(tfileout)){
    tfileout_test_name <- names(tfileout)[ii]
    ttout_results <- rep(NA_character_, length(tfileout[[ii]]))
    names_ttout_results <- paste(tfileout_test_name, tfileout[[ii]], sep=".")
    names_ttout_results <- stringr::str_replace_all(names_ttout_results,  '[ ,()"]', '') 
    names(ttout_results) <- names_ttout_results
    
    ttout_id <- which(ttout_test_names == tfileout_test_name)
    if(length(ttout_id) != 0){
      ttout_results_srcref <- vapply(ttout[[ttout_id]]$results, 
                              function(x) paste0(as.character(x$srcref),
                              collapse=""),
                              FUN.VALUE = "character")
      #print(ttout_results_srcref)
      
      for(jj in seq_along(tfileout[[ii]])){
        ttout_res_id <- which(stringr::str_replace_all(ttout_results_srcref, " ", "") == stringr::str_replace_all(tfileout[[ii]][jj], " ", ""))
        if(length(ttout_res_id) > 0) {
          # take the first check result if there's more than one. e.g. with setequal
          if(length(ttout_res_id) > 1){
              ttout_res_id <- ttout_res_id[1] 
          }
          ttout_res_class <- class(ttout[[ttout_id]]$results[[ttout_res_id]])[1]
          ttout_results[jj] <- ttout_res_class
          # print(tfileout[[ii]][jj])
        }
      }
      # print(ttout[[ttout_id]]$test)
    }
    all_results <- c(all_results, ttout_results)
      
  }
  all_results
}

#' Check correctness of student solution rmd.
#'
#' This will run unit tests on the students' rmd file.
#'
#' @param e_stud The environment containing the output objects from running the
#' studnent Rmd file.
#' @param e_soln The environment containing the objects from the solution
#' template. It will probably contain objects with the suffix "_soln". These
#' will be tested against the versions generated by the student.
#' @param test_fname The R script containing the test chunks.
#' 
#' @details Prior to calling this, \code{\link{populate_soln_env}} should 
#' already have been called on the solution template, and the student 
#' file should already have been knitted in order to generate the students' 
#' objects. Of course, one could generate the test script independent of 
#' \code{\link{populate_soln_env}}, but the solution environment that contains 
#' objects with a "_soln" suffix is also needed.
#' 
#' The student environment, solution environment, test file and the list of 
#' tests and expectations are the inputs to this function.
#'
#' @return A data frame with one row, and the number of columns equal to the 
#' number of tests run plus the number of scalars to keep.
#' @export
#' 
#' @seealso \code{\link{populate_soln_env}}, \code{\link{render_one}}
#'
check_correctness <- function(e_stud, e_soln, test_fname) {

  obj_to_copy <- ls(e_soln, pattern="^\\.", all.names = TRUE)
  obj_to_copy <- setdiff(obj_to_copy, ".myfilename")
  copy_out <- sapply(obj_to_copy,
                     function(zzz) copy_e2e(zzz, e_soln, zzz, e_stud))
  # running test_file
  source(test_fname, local=e_stud)

  #test_output <- testthat::test_file(test_fname, env = e_stud, reporter = "silent")

#  if(length(test_list) > 0){
#    tt_parsed <- match_ttout_tfile(test_output, test_list)
#    #tt_parsed <- stringr::str_replace_all(tt_parsed, "expectation_", "")
#    tt_parsed <- data.frame(as.list(tt_parsed), stringsAsFactors=FALSE)
#  } else {
#    tt_parsed <- NULL
#  }
  #browser()

#  # # running ah_check_scalars
   if(".scalars_to_keep" %in% names(e_stud)) {
     scalars_out <- lapply(get(".scalars_to_keep", envir = e_stud), 
                           function(x) {
                             get0(x, envir = e_stud, ifnotfound=NA,
                                  mode=mode(get(x, e_soln)))
                           })
     scalars_out <- lapply(scalars_out, 
                    function(x){ 
                        if(is.atomic(x) && (length(x) == 1)){
                            return(x)
                        } else {
                            return(NA)
                        }
                    })
     names(scalars_out) <- e_stud$.scalars_to_keep
     scalars_out <- data.frame(scalars_out, stringsAsFactors = FALSE)
#     #outcome <- c(outcome, scalars_out)
#     outcome2 <- data.frame(t(scalars_out), stringsAsFactors = FALSE)
#     names(outcome2) <- e_stud$.scalars_to_keep
#     outcome <- dplyr::bind_cols(outcome, outcome2)
   } else {
     scalars_out <- NULL
   }
  
#  # combine if necessary
#  if(!is.null(tt_parsed) && !is.null(scalars_out)) {
#    out <- cbind(tt_parsed, scalars_out)
#  } else if(!is.null(tt_parsed)){
#    out <- tt_parsed
#  } else if(!is.null(scalars_out)) {
#    out <- scalars_out
#  } else {
#    out <- NULL
#  }
  scalars_out
#  return(outcome)
}

#library(codetools)
# node_id is the CALLING node.

#' Convert language object to tree.
#'
#' A recursive function for converting a language object to treeharp.
#'
#' @param lang_obj A language object.
#' @param node_id  The calling node to this language object. This should only 
#' be greater than 0 if the \code{ni_env} already contains a partial adjacency 
#' list and corresponding node information. This will happen when this function 
#' is called recursively.
#' @param ni_env  An environment to store the adjacency list and node information.
#'
#' @importFrom rlang is_symbol is_syntactic_literal call_args call_args_names
#' @importFrom rlang call_name is_pairlist is_na is_formula
#' @return Nothing
#' 
#' @details This function is used by TreeHarp constructors. It should not have
#' to be called by a user. It works by bulding up an adjacency list and node 
#' node information data frame within the supplied environment.
#' 
#' @export
#' @examples 
#' e1 <- new.env() 
#' lang_2_tree(quote(X <- 1), 0, e1) 
#' e1$adj_list 
#' e1$node_info

lang_2_tree <- function(lang_obj, node_id, ni_env) {
  #if(is.null(lang_obj)) {
  #  print("NULL!")
  #  return(NULL)
  #}

  if(is_symbol(lang_obj) || is_syntactic_literal(lang_obj)){
    #cat("It's a symbol or literal!\n")
    id_depth <- get_next_depth_id(parent_node_id = node_id, ni_env)
    if(is.null(lang_obj)){
      sym_name <- "NULL"
    } else if (rlang::is_na(lang_obj)){
      sym_name <- "NA"
    } else {
      sym_name <- as.character(lang_obj)
    }
    rbind_to_nodes_info(id_depth$id, sym_name, FALSE,
                        FALSE, id_depth$depth, ni_env)
    update_adj_list("new_node", node_id = id_depth$id,  node_name = sym_name,
                    env_ni = ni_env)
    if(id_depth$id > 1) { # need to check this branch!!
      update_adj_list("add_child", node_id = node_id, child_node = id_depth$id,
                      env_ni = ni_env)
    }
  } else if(is_pairlist(lang_obj)) {
    #cat("It's a pairlist.\n")
    lang_list <- as.list(lang_obj)
    formal_args <- names(lang_list)
    if(length(formal_args) > 0) {
      for(ii in seq_along(formal_args)) {
        # ii <- 1
        id_depth <- get_next_depth_id(parent_node_id = node_id, ni_env)
        rbind_to_nodes_info(id_depth$id, formal_args[ii], FALSE, TRUE,
                            id_depth$depth, ni_env)
        update_adj_list("new_node", node_id = id_depth$id,
                        node_name = formal_args[ii], env_ni = ni_env)
        if(id_depth$id > 1) { # need to check this branch!!
          update_adj_list("add_child", node_id = node_id, child_node = id_depth$id,
                          env_ni = ni_env)
        }
        #browser()
        #if(as.character(lang_list[[ii]]) != "") {
        if(is.null(lang_list[[ii]])) {
          lang_2_tree(quote(NULL), id_depth$id, ni_env)
        }else if(as.character(lang_list[[ii]])[1] != "") {
          lang_2_tree(lang_obj[[ii]], id_depth$id, ni_env)
        }
      }
    }
    }   else {
    #cat("It's a call!\n")
      # rlang functiosn call_ do not work well when the object is a one-sided
      # formula. When the call name is NULL, we also need to process that
      # first.
      if(is_formula(lang_obj) && length(lang_obj) == 2) {
        arg_num <- 1
        arg_names <- ""
        call_name1 <- "~"
        call_args1 <- lang_obj[2]
      } else {
        arg_num <- length(call_args(lang_obj))
        arg_names <-  call_args_names(lang_obj)
        call_name1 <- call_name(lang_obj)
        call_args1 <- call_args(lang_obj)
      }
    #
    # for call itself:
    id_depth <- get_next_depth_id(parent_node_id = node_id, ni_env)
    ### This is a fall-back fix if call_name returns NULL:
    # if(is.null(call_name1)){
    #   call_name1 <- as_label(lang_obj[[1]])
    # }
    if(is.null(call_name1)) {
      # browser()
      lang_2_tree(lang_obj[[1]], node_id, ni_env)
      fn_node_id <- max(ni_env$node_info$id)
    } else {
      if(call_name1 == "function") {
        arg_num <- arg_num - 1
      }
      rbind_to_nodes_info(id_depth$id, call_name1, TRUE, FALSE, id_depth$depth,
                          ni_env)
      update_adj_list("new_node", node_id = id_depth$id,  node_name = call_name1,
                      env_ni = ni_env)
      if(id_depth$id > 1) { # need to check this branch!!
        update_adj_list("add_child", node_id = node_id, child_node = id_depth$id,
                        env_ni = ni_env)
      }
      fn_node_id <- id_depth$id
    }
    if(arg_num > 0) {
      for(ii in 1:arg_num) {
        if(nchar(arg_names[ii]) > 0){
          id_depth <- get_next_depth_id(parent_node_id = fn_node_id, ni_env)
          rbind_to_nodes_info(id_depth$id, arg_names[ii], FALSE, TRUE,
                              id_depth$depth, ni_env)
          update_adj_list("new_node", node_id = id_depth$id,
                          node_name = arg_names[ii],  env_ni = ni_env)
          update_adj_list("add_child", node_id = fn_node_id,
                          child_node = id_depth$id, env_ni = ni_env)
          # lang_2_tree(call_args(lang_obj)[[ii]], fn_node_id+1, ni_env)
          lang_2_tree(call_args1[[ii]], id_depth$id, ni_env)
          # lang_2_tree
        } else {
          # lang_2_tree
          # cat("recurse\n")
          lang_2_tree(call_args1[[ii]], fn_node_id, ni_env)
        }
      }
    }
  }
}

unlink("test01.Rmd")
unlink("test02.Rmd")
unlink("test03.Rmd")
unlink("test03.html")
unlink("test04.Rmd")
unlink("soln_template.Rmd")
unlink("s1.R")
unlink("s2.R")
unlink("s3.R")
unlink("s4.R")
unlink("render_all.log")
unlink("test_output", TRUE, TRUE)

# Valid adjacency lists for TreeHarp:
th1 <- list(a= c(2L,3L), b=NULL, c=NULL)
th2 <- list(a= c(2L,3L), b=4L, c=NULL, d=NULL)
th3 <- list(a= c(2L,3L), b=NULL, c=4L, d=NULL)
th3a <- list(a= c(2L,3L,4L), b=NULL, c=c(5L, 6L), d=7L, e=NULL, f=NULL, g=NULL)
th00 <- list(leaf=NULL)

# Invalid adjacency list (disconnected):
th4 <- list(a=c(2L,3L), b=NULL, c=NULL, d=NULL)
# Invalid adjacency list (cyclic, and each edge represented twice):
th5 <- list(a=c(2L,3L), b=c(1L,4L), c=c(1L,4L), d=c(2L,3L))
# Invalid adjacency list (cyclic):
th6 <- list(a=c(2L,3L), b=4L, c=4L, d=NULL)
# Invalid adjacency list (not in BFS order):
th7 <- list(a = c(3L,4L,6L), f = NULL, c = c(2L,5L), b=NULL, e = NULL, d = NULL)
th7a <- list(a=c(2L, 4L), b=NULL, d=NULL, c=3L)
# Invalid adjacency list (unlabelled):
th8 <- list(c(2,3), b=NULL, c=NULL)

# Valid adjacency matrix:
m1 <- matrix(0L, 3, 3)
dimnames(m1) <- list(letters[1:3], letters[1:3])
m1[1, ] <- c(0, 1L, 1L)
m1[, 1] <- c(0, 1L, 1L)
# Invalid adjacency matrix:
m1a <- m1
dimnames(m1a) <- NULL

# Invalid adjacency matrix (contains cycle):
m2 <- matrix(c(0, 1, 1, 0,
               1, 0, 0, 1,
               1, 0, 0, 1,
               0, 1, 1, 0), nrow=4, byrow = TRUE,
             dimnames = list(letters[1:4], letters[1:4]))

e1 <- new.env()

## Language objects

ex1 <- quote(ggplot(data) + geom_point(aes(x=x1, y=y1)))
ex2 <- quote(x)
ex3 <- quote(x <- 1)
ex4 <- quote(x <- NA)

#' Calculates the Total Memory Used
#'
#' This function uses the utils package to compute the total
#' amount of memory used by objects in an environment.
#'
#' @param env The environment whose size is to be computed.
#'
#' @return The size in bytes, as a numeric value (scalar).
#' @export
#' @details The names are wrapped in backticks. Otherwise, non-syntactic
#' names will cause problems. This function is used within
#' \code{\link{render_one}} as part of the runtime stats assessment.
#'
#' @examples
#'
#' e1 <- new.env()
#' env_size(e1)
#' evalq(x <- 1:10000L, e1)
#' env_size(e1)

env_size <- function(env) {
  obj_names <- ls(envir = env, all.names = TRUE)
  if (length(obj_names) == 0L) return(0)
  objs <- mget(obj_names, envir = env, inherits = FALSE)
  size_byte <- sum(vapply(objs, utils::object.size, numeric(1))) 
  size_byte
}

#' Generate a html of thumbnails
#'
#' @param out_dir The directory in which student html files and the figures 
#' are kept.
#' @param html_fname The name of the master html file which will contain all
#' thumbnails. This file will be created in \code{out_dir}.
#' @param html_title The title tag of the master html page. This will be displayed 
#' on top of the output html page.
#' @param anonymise If TRUE, the original filenames will be replaced with
#'        inocuous numbers. If FALSE, the original filenames will be retained.
#'
#' @details After running \code{\link{render_one}} on a set of R/Rmd files 
#' in a directory, this function helps to consolidate them for review.
#' 
#' The output folder contains all the generated html files, images and a log 
#' file. This function will extract the images from each html file and display 
#' them as thumbnails on a new html page, with links to all individual files. 
#' 
#' @return The function returns nothing, but it should create a html page 
#' of thumbnails of all the images that students plotted, along with links to 
#' their individual pages.
#' @export
generate_thumbnails <- function(out_dir, html_fname, html_title,
                                anonymise=FALSE) {
  out_html <- file.path(out_dir, html_fname)
  # start from scratch again. 
  if(file.exists(out_html)) {
    unlink(out_html)
  } 
  overall_con <- file(out_html, open='wt')
  #cat("<!DOCTYPE html>", "<html>", "<head>", file=overall_con, sep='\n')
  #cat(as.character(tags$meta(charset="UTF-8")), file=overall_con, sep='\n')
  cat(write_css(), write_jscript(), file=overall_con, sep='\n')
  cat(as.character(shiny::tags$title(html_title)), file=overall_con, sep='\n')
  #cat("</head>", file=overall_con, sep='\n')
  
  cat("<body>", file=overall_con, sep='\n')
  cat(as.character(shiny::tags$h1(html_title)), file=overall_con, sep='\n')
  
  # removes the thumbnail html itself.
  h_files <- list.files(out_dir, "*html") %>% 
    setdiff(html_fname)
  
  i <- 1
  
  for(fname in h_files){
    root_fname <- xfun::sans_ext(fname)
    try_out_files <- file.path(out_dir, paste0(root_fname, "_files"))
    
    divsize <- c("small", "imgsmall")
    all_imgs <- NULL
    if(dir.exists(try_out_files)) {
      img_fnames <- list.files(try_out_files, recursive = TRUE)
      sq_imgs <- ceiling(sqrt(length(img_fnames)))
      
      if(sq_imgs < 3) divsize <- c("small", "imgsmall")
      else if(sq_imgs < 4) divsize <- c("medium", "imgmedium")
      else if(sq_imgs >= 4) divsize <- c("large", "imglarge")
      
      all_imgs <- file.path(basename(try_out_files), img_fnames) %>%
        lapply(function(x) 
          shiny::tags$figure(class = "figimg", class = divsize[2],
          shiny::tags$img(src = x, title = x, alt = x))) %>%
        lapply(as.character) %>% unlist() %>% paste(collapse = "\n")
    }

    cat(as.character(
      shiny::tags$div(class = "imgContainer", class = divsize[1],
        shiny::tags$figure(
          shiny::HTML(all_imgs),
          shiny::tags$figcaption(
            shiny::tags$a(href=fname, title=fname, 
                   ifelse(anonymise, paste0("File_Num_", sprintf("%03d", i)), 
                          root_fname))
        ))
      )), file=overall_con, sep="\n")
    i <- i + 1
  }
  cat("</body>", "</html>", file=overall_con,sep='\n')
  close(overall_con)
}

#' Run a single Rmd file through autoharp.
#'
#' Renders the specified file, and collates run time, static and 
#' correctness checks.
#'
#' @param rmd_name The path to the file to be rendered and checked.
#' @param out_dir The directory to store all the html output, md output, and
#' figures.
#' @param knit_root_dir The working directory while knitting the file.
#' @param log_name A character string, denoting the log file name. It defaults to
#' "render_one.log". If this file is already present in the directory, this
#' function will append to it.
#' @param soln_stuff This is a list, with components env, test_fname, and
#' tt_list. This object is the output of \code{\link{populate_soln_env}}. Set
#' this to be NA if you wish to skip correctness checks, and only do rendering.
#' @param max_time_per_run The maximum time to wait before aborting the
#' rendering of a particular file.
#' @param permission_to_install If TRUE, then the function will try to install 
#' any packages needed. By default, this is FALSE.
#'
#' @details The log file contains a record of the libraries used by the student, and if
#' any new libraries needed to be installed. The status will be one of SUCCESS,
#' FAIL or UNKNOWN.
#'
#' @importFrom parallel makePSOCKcluster clusterEvalQ clusterCall stopCluster clusterExport
#' @importFrom rmarkdown render
#' @return A data frame with one row for each file in the input directory.
#' @seealso \code{\link{populate_soln_env}}, \code{\link{check_correctness}}
#' @export
#'
render_one <- function(rmd_name, out_dir, knit_root_dir, log_name, soln_stuff,
                       max_time_per_run = 120, permission_to_install=FALSE) {
    # check if out_dir exists, if not, create it.
    if(!dir.exists(out_dir)) {
        dir.create(out_dir)
    }
    
    # Initialise log file.
    if(missing(log_name)){
        log_name <- "render_one.log"
    }
    out_log <- file.path(out_dir, log_name)
    out_file <- file(out_log, 'a+b')
    time_stamp <- as.character(Sys.time())
    cat('---', file=out_file, append=TRUE)
    cat('render_one commenced at', time_stamp, file=out_file, append=TRUE)
    cat('---\n', file=out_file, append=TRUE)
    cat('\nStarting on', rmd_name, '\n', file=out_file, append=TRUE)

#   Spin up a worker.
    worker_proc <- makePSOCKcluster(1)

    # Clean up steps
    on.exit({
        close.connection(out_file)
        stopCluster(worker_proc)
    })

    # Check solution stuff
    if(length(soln_stuff) == 2) {
      if(!setequal(names(soln_stuff), c("env", "test_fname"))) {
          stop("Incorrect solution object.")
      }
      check_corr_flag <- TRUE
    } else if(length(soln_stuff) == 1 && is.na(soln_stuff)){
      check_corr_flag <- FALSE
    } else {
      stop("Incorrect solution object.")
    }

    # Remove related files from out_dir
    root_name <- xfun::sans_ext(basename(rmd_name))
    unlink(paste0(file.path(out_dir, root_name), "*"))
    
    # prepare out_df
    out_df <- dplyr::bind_cols(fname = rmd_name, time_stamp = time_stamp)

    # Check for libraries
    lib_used <- get_libraries(rmd_name)
    if("error" %in% class(lib_used)) {
      cat('Libraries used: \n', file=out_file, append=TRUE)
      status <- "FAIL"
      cat(rmd_name, ":", status, "\n", file=out_file, append=TRUE)
      cat(conditionMessage(lib_used), '\n', file=out_file, append=TRUE)
      out_df <- dplyr::bind_cols(out_df, run_status = status)
      return(out_df)
    }
    cat('Libraries used:', paste0(lib_used, collapse=","), '\n', file=out_file,
        append=TRUE)
    if(!is.null(lib_used)) {
      id <- !(sapply(lib_used, quietly=TRUE, requireNamespace))
      lib_needed <- lib_used[id]
      if(length(lib_needed) > 0) {
        cat('Need to install:', paste0(lib_needed, collapse=","), '\n',
            file=out_file, append=TRUE)
        if(permission_to_install){
          try_install <- tryCatch(utils::install.packages(lib_needed, dependencies = TRUE),
                                error = function(e) return(e))
        } else {
            try_install <- simpleError("do not install")
        }
        if("error" %in% class(try_install)) {
          status <- "FAIL"
          cat(rmd_name, ":", status, "\n", file=out_file, append=TRUE)
          cat(conditionMessage(try_install), '\n', file=out_file, append=TRUE)
          out_df <- dplyr::bind_cols(out_df, run_status = status)
          return(out_df)
        }
      }
    }
    
    # Perform pre-checks
    pre_check_status <- render_prechecks(rmd_name)
    if(!pre_check_status) {
      status <- "FAIL"
      cat(rmd_name, ":", status, "\n", file=out_file, append=TRUE)
      cat("Precheck failed.. system( ) calls present or purl failed.", '\n', file=out_file, append=TRUE)
      out_df <- dplyr::bind_cols(out_df, run_status = status)
      return(out_df)
    }

    .myenv <- new.env()
    .myenv$.myfilename <- normalizePath(rmd_name)
    start_time <- Sys.time()

    clusterExport(worker_proc, c(".myenv", "rmd_name", "out_dir", "soln_stuff",
                                 "knit_root_dir", "max_time_per_run"), 
                  sys.frame(sys.nframe()))

    # set the limit on the worker process
    try_setLimit <- clusterEvalQ(worker_proc, 
                                 setTimeLimit(cpu=max_time_per_run, elapsed=max_time_per_run))
    clusterEvalQ(worker_proc, 
        try_out <- tryCatch(rmarkdown::render(rmd_name, output_format = 'html_document', 
                            output_dir = out_dir, clean=FALSE, quiet=TRUE, 
                            intermediates_dir = out_dir, knit_root_dir = knit_root_dir, 
                            envir = .myenv), error = function(e) e))
    try_out <- clusterCall(worker_proc, get, "try_out")[[1]]
    end_time <- Sys.time()

    if(class(try_out)[1] == "character") {
      if(stringr::str_sub(try_out, -4) == "html") {
        status <- "SUCCESS"
        .myenv <- clusterCall(worker_proc, get, ".myenv")[[1]]
      }
      else
        status <- "UNKNOWN"
    } else if("error" %in% class(try_out)) {
      status <- "FAIL"
      cat(rmd_name, ":", status, "\n", file=out_file, append=TRUE)
      cat(conditionMessage(try_out), '\n', file=out_file, append=TRUE)
      out_df <- dplyr::bind_cols(out_df, run_status = status)
      return(out_df)
    } else {
      status <- "UNKNOWN"
    }
    cat(rmd_name, ":", status, "\n", file=out_file, append=TRUE)
    run_time <-  round(as.numeric(difftime(end_time, start_time, units="secs")), 4)
    run_mem <- env_size(.myenv)
    # run correctness check on the worker process.
    # - use clusterEvalQ, NOT clusterCall
    # - Need to load autoharp on the worker in order to call check_correctness.
    if(check_corr_flag) {
        try_setLimit <- clusterEvalQ(worker_proc, 
                                     setTimeLimit(cpu=max_time_per_run, elapsed=max_time_per_run))
        load_ah <- clusterEvalQ(worker_proc, library(autoharp))
        clusterEvalQ(worker_proc, 
                      corr_out <- check_correctness(.myenv, soln_stuff$env, 
                                                    soln_stuff$test_fname))
        
        corr_out <- clusterCall(worker_proc, get, "corr_out")[[1]]
        
        if("error" %in% class(corr_out)){
          status <- "FAIL"
          cat(rmd_name, ":", status, "\n", file=out_file, append=TRUE)
          cat(conditionMessage(corr_out), '\n', file=out_file, append=TRUE)
          out_df <- dplyr::bind_cols(out_df, run_status = status)
          return(out_df)
        }
    } else {
        corr_out <- NULL
    }

    out_df <- dplyr::bind_cols(out_df, run_status = status, run_time = run_time, 
                               run_mem = run_mem, corr_out)
    out_df
}

tho1 <- TreeHarp(th1)
tho2 <- TreeHarp(th2)
tho3 <- TreeHarp(th3)
tho3a <- TreeHarp(th3a)
tho00 <- TreeHarp(th00)

l1 <- list(a=c(2,3), b=NULL, c=4, d=NULL)
l2 <- list(a=c(2,3), b=NULL, d=NULL)
l3 <- list(a=c(2,3), b=NULL, c=NULL)

l4 <- list(a=c(2,3), b=NULL, b=NULL)
l5 <- list(a=c(2,3), b=NULL, b=4, c=NULL)
l6 <- list(a=c(2,3,4), b=NULL, b=5, d=NULL, c=NULL)
l7 <- list(a=c(2,3,4), b=NULL, b=NULL, b=NULL)

t1 <- TreeHarp(l1)
t2 <- TreeHarp(l2)
t3 <- TreeHarp(l3)

t4 <- TreeHarp(l4)
t5 <- TreeHarp(l5)
t6 <- TreeHarp(l6)
t7 <- TreeHarp(l7)

#' Function to rearrage nodes in BFS
#'
#' @param adj_list The output of lang_2_tree.
#' @param node_info The output of lang_2_tree.
#'
#' @details This function is for an internal TreeHarp constructor use. It is
#' not exported.
#'
#' @return An adjacency list and nodes info data frame in BFS order.
#'
#' @importFrom rlang .data
to_BFS <- function(adj_list, node_info) {
  n_nodes <- nrow(node_info)
  node_info <- dplyr::arrange(node_info, .data$depth)
  old_ids <- node_info$id
  node_info$id <- 1:n_nodes

  new_adj_list <- vector(mode="list", length=n_nodes)
  for(ii in 1:n_nodes) {
    new_adj_list[ii] <- adj_list[old_ids[ii]]
    names(new_adj_list)[ii] <- names(adj_list)[old_ids[ii]]
  }

  rep_vals <- 1:n_nodes
  names(rep_vals) <- old_ids

  for(ii in 1:n_nodes) {
    if(is.null(new_adj_list[[ii]][1])){
      next
    } else{
      new_adj_list[[ii]] <- dplyr::recode(new_adj_list[[ii]], !!!rep_vals)
    }
  }

  list(new_adj_list, node_info)
}

#' @include treeharp.R th_getter-length.R
NULL

#' TreeHarp Cast a TreeHarp to Matrix.
#'
#' Convert a treeharp object to an adjacency matrix.
#'
#' @param from  A treeharp object.
#'
#' @name as.matrix
#' @return A matrix.
#'
setAs("TreeHarp", "matrix",
      function(from) {
        mat <- adj_list_2_matrix(from@adjList)
        mat
      })

#' Check if a File is Rmd
#'
#' Checks if a file actually is an Rmd file.
#'
#' @param fname A character string. It is the name of the student submission
#' file.
#' @param verbose A logical value that prints messages if a non-rmd file
#' is found.
#'
#' @details It runs three checks. First, it checks for the file extension to be
#' Rmd or rmd or any such variant. Second, it checks for a YAML header at the
#' beginning of file. Finally, it checks if there is at least one properly
#' defined R chunk within the file.
#'
#' @export
#'
#' @seealso \code{\link{get_libraries}}
#'
#' @return The function will return TRUE if all the (3) checks pass, and FALSE
#' otherwise.
#'
check_rmd <- function(fname, verbose = TRUE) {
  fname_lower <- stringr::str_to_lower(fname)
  if(!stringr::str_detect(fname_lower, "\\.rmd$")){
    if(verbose) {
      message(paste(fname, "File extension is not Rmd.", sep=": "))
    }
    return(FALSE)
  }
  file_text <- readLines(fname, warn=FALSE)

  # YAML header should be at the beginning of the file.
  yaml_pos <- stringr::str_which(file_text, "^---$")
  if(length(yaml_pos) == 0 || yaml_pos[1] != 1) {
    if(verbose) {
      message(paste(fname, "YAML header not properly defined.", sep=": "))
    }
    return(FALSE)
  }

  # No. of starts should equal number of ends, and should be
  # positive.
  begin_ids <- which(stringr::str_detect(file_text, knitr::all_patterns$md$chunk.begin))
  end_ids <- which(stringr::str_detect(file_text, knitr::all_patterns$md$chunk.end))
  l_rchunk_starts <- length(begin_ids)
  l_rchunk_ends <- length(end_ids)
  
  if(l_rchunk_starts == 0 || l_rchunk_ends == 0 || 
     l_rchunk_ends != l_rchunk_starts) {
    if(verbose) {
      message(paste(fname, "No proper R chunks found.", sep=": "))
    }
    return(FALSE)
  }

  return(TRUE)
}

#' @include treeharp.R
NULL

#' @describeIn TreeHarp To get the length of a tree.
#'
#' The length of the tree refers to the number of nodes in the tree.
#'
#' @param x A Treeharp object.
#'
#' @return \code{length}: An integer of length 1.
#' @export
setMethod("length", signature=c(x="TreeHarp"),
          function(x) {
            return(length(x@adjList))
            })

#' @describeIn TreeHarp To print a tree representation.
#'
#' A string representation of a TreeHarp object.
#' 
#' @param object A TreeHarp object.
#'
#' @return \code{print}: Returns NULL. It prints a string representation of a 
#' TreeHarp object.
#' @export
setMethod("show",
    signature(object = "TreeHarp"),
    function (object) 
    {
            cat(object@repr, "\n")
            return(invisible(NULL))
    }
)

#' @describeIn TreeHarp To get tree labels
#'
#' This function returns the node labels of the tree.
#'
#' @return \code{names}: A character vector with length equal to the number of 
#' nodes.
#' @export
setMethod("names", signature=c(x="TreeHarp"),
          function(x) {
            return(names(x@adjList))
            })

#' Generic for Getting Parent Node Id.
#'
#' The generic method definition for getting parent node id.
#'
#' @param x An object of class TreeHarp or an adjacency list.
#' @param node_num An integer, length 1. This the node whose parent we are
#' after. If node_num is equal to 1, then NULL is returned because that should
#' be the root node.
#'
#' @return An integer, indicating the parent node.
#' @export
setGeneric("get_parent_id", function(x, node_num) standardGeneric("get_parent_id"),
           signature = "x"
           )

#' @describeIn get_parent_id Obtain parent node id.
#'
#' Extracts parent id of a node from a TreeHarp object.
#'
#' @export
#' @seealso \code{\link{get_child_ids}}
setMethod("get_parent_id", signature=c(x="TreeHarp"),
          function(x, node_num) {
            get_parent_id2(x@adjList, node_num)
            })

#' @describeIn get_parent_id Obtain parent node id.
#'
#' Extracts parent id of a node from an adjacency list object.
#'
#' @export
setMethod("get_parent_id", signature=c(x="list"),
          function(x, node_num) {
            get_parent_id2(x, node_num)
            })

#' Generic for Getting Child Node Ids
#'
#' The generic method definition for getting child node ids.
#'
#' @param x An object of class TreeHarp.
#' @param node_num An integer, length 1. This the node whose children we are
#' after. If the specified node is a leaf, the NULL is returned.
#'
#' @return An integer vector, indicating the children node ids.
#' @seealso \code{\link{get_parent_id}}
#' @export
setGeneric("get_child_ids", function(x, node_num) standardGeneric("get_child_ids"),
           signature = "x"
           )

#' @describeIn get_child_ids Obtain child nodes.
#'
#' Allows user to extract the child nodes from a specified node from TreeHarp
#' object.
#'
#' @export
setMethod("get_child_ids", signature=c(x="TreeHarp"),
          function(x, node_num) {
            get_child_ids2(x@adjList, node_num)
            })

#' @describeIn get_child_ids Obtain child nodes.
#'
#' Allows user to extract the child nodes from a specified node from an 
#' adjacency list.
#'
#' @export
setMethod("get_child_ids", signature=c(x="list"),
          function(x, node_num) {
            get_child_ids2(x, node_num)
            })

# #' To check if two trees are equal.
# #'
# #' This function checks if two trees are equal.
# #'
# #' @param x A Treeharp object.
# #' @param y A Treeharp object.
# #' @param num.eq See original
# #' @param single.NA See original
# #' @param attrib.as.set See original
# #' @param ignore.bytecode See original
# #' @param ignore.environment See original
# #' @param ignore.srcref See original
# #'
# #' @return A logical indicating if two TreeHarp objects are equal.
# #' @export
# setMethod("identical",
#     signature(x = "TreeHarp", y = "TreeHarp"),
#     function (x, y, num.eq = TRUE, single.NA = TRUE, attrib.as.set = TRUE,
#         ignore.bytecode = TRUE, ignore.environment = FALSE, ignore.srcref = TRUE)
#     {
#             x_list <- adjList(x)
#             y_list <- adjList(y)
#             return(identical(x_list, y_list))
#     }
# )

#' Generic for Getting Adjacency List
#'
#' The generic method definition for getting adjacency list from a
#' TreeHarp object.
#'
#' @param x An object of class TreeHarp.
#' @param ... Unused arguments, for now.
#'
#' @return The adjacency list for a TreeHarp object.
#' @export
setGeneric("get_adj_list", function(x, ...) standardGeneric("get_adj_list"),
           signature = "x"
           )

#' @describeIn get_adj_list A getter.
#'
#' Allows user to extract the adjacency list of a treeharp object.
#'
#' @export
setMethod("get_adj_list", signature=c(x="TreeHarp"),
          function(x, ...) {
            return(x@adjList)
            })

#' Generic for Getting Node Types
#'
#' The generic method definition for getting node types from a
#' TreeHarp object.
#'
#' @param x An object of class TreeHarp.
#' @param ... Unused arguments, for now.
#'
#' @return A data frame containing the node types for a TreeHarp object. If the 
#' slot is empty, NA is returned.
#' @export
setGeneric("get_node_types", function(x, ...) standardGeneric("get_node_types"),
           signature = "x"
           )

#' @describeIn get_node_types A getter.
#'
#' Allows user to extract the node types of a treeharp object.
#'
#' @export
setMethod("get_node_types", signature=c(x="TreeHarp"),
          function(x, ...) {
            return(x@nodeTypes)
            })

#' TreeHarp Plotting TreeHarp Objects
#'
#' A plot method for visualising treeharp objects.
#'
#' @param x An object of class TreeHarp.
#' @param y Unused.
#' @param ... Additional arguments passed on to plot.igraph().
#'
#' @details The treeharp object is converted to an igraph object before it
#' is plotted.
#'
#' @return Returns NULL, invisibly.
#' @export
setMethod("plot", signature=c("TreeHarp"),
          function(x, y, ...){
            # convert to igraph
            ig <- igraph::graph_from_adj_list(get_adj_list(x), mode="all",
                                              duplicate=FALSE)
            ll <- names(x@adjList)
            igraph::plot.igraph(ig,
                                layout=igraph::layout_as_tree(ig, root=1),
                                vertex.label=ll, ...)
          })

#' Update node information.
#' 
#' Updates the node information regarding an R expression.
#'
#' @param id The id of the node to be added. This should be an integer
#' of length 1.
#' @param name The name of the node.
#' @param call_status Is the language object a call or a symbol/literal? This 
#' should a logical value.
#' @param formal_arg Is the language object a formal argument or not? This 
#' should be a logical value.
#' @param depth An integer indicating the depth of this language object in 
#' the parse tree.
#' @param env_ni An environment object, possibly containing a data frame with
#' columns id, name, call_status, formal_arg and depth.
#'
#' @return TRUE is returned invisibly. 
#' 
#' @details This is for internal use. It may be removed from user-view soon!
rbind_to_nodes_info <- function(id, name, call_status, formal_arg, depth, 
                                env_ni) {
  tmp_df <- data.frame(id = id, name = name, call_status = call_status, 
                       formal_arg = formal_arg, depth = depth, 
                       stringsAsFactors = FALSE)
  if(!exists("node_info", env_ni)){
    assign("node_info", tmp_df, env_ni)
  } else {
    old_parent <- parent.env(env_ni)
    parent.env(env_ni) <- sys.frame(sys.nframe()) # set parent to be evaluation env.
    evalq(node_info <- rbind(node_info, tmp_df), envir = env_ni)
    # reset parent
    parent.env(env_ni) <- old_parent
  }
  return(invisible(TRUE))
}

#' Get the id and depth of a child node.
#'
#' From the parent's depth and the last labelled node, we obtain the node id 
#' and depth of a child.
#' 
#' @param parent_node_id The id of the parent node we are considering.
#' @param env_ni An environment object, possibly containing a data frame with
#' columns id, name, call_status, arg_type and depth.
#'
#' @return A list containing the id and depth of the next node. 
#' @details This is for internal use. It may be removed from user-view soon!
get_next_depth_id <- function(parent_node_id, env_ni) {
  if(!exists("node_info", env_ni)){
    if(parent_node_id != 0){
      stop("If no prior node information exists, the caller node should be 0.")
    }
    next_id <- 1L
    next_depth <- 1L
  } else {
    if(!parent_node_id %in% env_ni$node_info$id) {
      stop("caller node id does not exist in node info table.")
    }
    next_id <- max(env_ni$node_info$id) + 1L
    next_depth <- 
      env_ni$node_info$depth[env_ni$node_info$id == parent_node_id] + 1L
  }
  return(list(id=next_id, depth=next_depth))
}

#' Update adjacency list.
#' 
#' Updates the adjacency list for an R expression parse tree.
#'
#' @param update_type This should be either "new_node" or "add_child". If it 
#' is a new node, an empty list component is added. If it is add child, 
#' then child_node should be provided too.
#' @param node_id An integer.
#' @param node_name The name of the new node to be added. This must be provided 
#' if the update_type is "new_node".
#' @param child_node An integer.
#' @param env_ni An environment object, possibly containing an adjacency list 
#' that will later be used to construct a TreeHarp object.
#'
#' @return An invisible TRUE is returned.
#' 
#' 
#' @details This is for internal use. It may be removed from user-view soon!
update_adj_list <- function(update_type=c("new_node", "add_child"), 
                            node_id, node_name, child_node, env_ni) {
  if(!exists("adj_list", env_ni)){
    if(update_type == "add_child"){
      stop("No existing nodes; can't add child.")
    }
    adj_list <- list(NULL)
    names(adj_list) <- node_name
    assign("adj_list", adj_list, env_ni)
  } else {
    old_parent <- parent.env(env_ni)
    parent.env(env_ni) <- sys.frame(sys.nframe()) # set parent to be evaluation env.
    if(update_type == "new_node"){
      evalq(adj_list[node_id] <- list(NULL), envir = env_ni)   
      evalq(names(adj_list)[node_id] <- node_name, envir = env_ni)   
    } else {
      evalq(adj_list[[node_id]] <- append(adj_list[[node_id]], 
                                          child_node), envir=env_ni)
    }
    parent.env(env_ni) <- old_parent
  }
  return(invisible(TRUE))
}

---
title: "Student 02 Testing"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

## Problem 1: Basic Calculations

Students should compute the mean and standard deviation of a vector.

```{r solution_objects}
# Solution objects that students should create
X <- 1:28
Y <- c(5, 10, 15, 20, 25)
Z <- rnorm(100, mean = 10, sd = 3)
```

## Problem 2: Custom Function

Students should create a function that computes the square of a number.

```{r solution_function}
# Solution function
square_number <- function(n) {
  return(n*n + 2)
}

# Expected result
expected_result <- square_number(5)
```


## Problem 3: Data Manipulation

Students should filter a data frame.

```{r solution_data}
# Solution data manipulation
original_df <- data.frame(
  name = c("Alice", "Bob", "Charlie", "Diana"),
  age = c(25, 30, 35, 28),
  score = c(85, 92, 78, 95)
)

filtered_df <- original_df[original_df$score > 80, ]



library(tidyverse)

ggplot(filtered_df) +
  geom_point(aes(x = score, y = age)) +
  theme_minimal()
```

---
title: "Example QMD 01 - Correct"
format: html
---

This is a simple example that should match the solution template.

```{r}
# Numeric vectors
x <- 1:10
y <- 20:11  # Different order but same set
z <- c(2.5, 5.0, 7.5, 10.0)

# Character vector
names_vec <- c("Alice", "Bob", "Charlie")

# Data frame
students_df <- data.frame(
  name = c("Alice", "Bob", "Charlie", "Diana"),
  age = c(22, 23, 21, 24),
  score = c(85, 92, 78, 95)
)

# Matrix
m <- matrix(1:12, nrow = 3, ncol = 4)

# List
my_list <- list(
  numbers = 1:5,
  letters = c("a", "b", "c"),
  flag = TRUE
)

# Factor
grades <- factor(c("A", "B", "A", "C", "B"), levels = c("A", "B", "C"))
```

---
title: "Example Workflow: autoharp"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Example Workflow: autoharp}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(autoharp)
```

# Introduction

The purpose of this package is to assist an instructor in grading R codes.  The
following diagram presents an overview of how the autoharp works:

<div align="center">
  <img src="figs/instructor_overview.png" alt="autoharp framework" width="500">
</div>

To use the autoharp for grading, the instructor needs to prepare a solution
template. The autoharp will help to execute the code in each student script,
test the objects in them, and return a convenient dataframe, with one row per
student script.

## Description of Files

Thus, we begin with a simple example. We describe the assignment, solution file,
and two student scripts in the subsequent sections. Following this, we
demonstrate how the package can help.

The files needed to run this example can be found in the `examples` directory
of the `autoharp` package.

```{r ex01_files, echo=TRUE}
list.files(system.file("examples", "example-01", package="autoharp"))
```

This folder contains: 

* 1 x question paper (`sample_questions_01.Rmd`), 
* 1 x solution template (`soln_template_01.Rmd`), and 
* 2 x sample student scripts (`qn01_scr_01.R`, `qn01_scr_02.R`), which are in the 
folder `student_scripts`.

Imagine that `sample_questions_01.Rmd` is a worksheet that you assign to your
students. The question contains the following stem:

---

Consider the following probability density function (pdf) over the support 
$(0,1)$:

\begin{equation*}
f(x) = 
\begin{cases}
4x^3 & \text{if } 0 < x < 1 \\
0 & \text{otherwise}
\end{cases}
\end{equation*}

```{r ex_fn, echo=FALSE}
rf <- function(n) {
  U <- runif(n)
  X <- U^(1/4)
  X
}
```

Write a function called `rf`, that generates i.i.d observations from this pdf.
It should take in exactly one argument, $n$, that determines how many random
variates to return. For instance:

```{r run_fn}
set.seed(33)
rf(n = 5)
```

Now generate 10,000 random variates from this pdf and store them in a vector
named `X`. Your script *must* generate a function named `rf` and a vector named
`X`.

---

## Student Scripts

Consider the two student submissions, that are included with the package.
Student 1 (`qn01_scr_01.R`) has a model solution. However student 2
(`qn01_scr_02.R`) has made a few mistakes: 

  * A `for` loop has been used within the function, 
  * The algorithm itself is wrong, and 
  * The length of the `X` vector is wrong.

# Elements of the Framework

## The Question Paper

The question paper details what the students need to create within their
submission, which could be a plain R script, or an Rmd file. The required
objects could be any R object, such as a data frame, a vector, a list, or a
function. The question paper should clearly state the name of the object(s) and
their key attributes.

For instance, if a function is to be created, the question paper should specify
it's name, number of arguments, names of formal arguments and the return value.

## The Solution Key or Template

The solution template can be an Rmd or a quarto file. It is where you specify
the things that should be checked about the student script. First of all, it
should generate the correct versions of the objects. These "model" objects can
then be used to check against student-created objects.

The autoharp package defines two knitr hooks, for use within the solution
template:

* `autoharp.objs`: This is a character vector of names of objects that are going 
  to be used as the "model" solution. They could be correct values of something 
  the students were asked to evaluate from their data, or a dataset, or a function.
* `autoharp.scalars`: These are character vectors of objects that will be generated
  within the chunk. These are objects that will possibly be generated using 
  student-created objects. For example, it could be the output from running a 
  student-written function.

To make things concrete, let's take a look at the lines 19 to 22 from the 
`soln_template_01.Rmd`:

````markdown
`r ''````{r test01, autoharp.objs=c("rf", "X"), autoharp.scalars=c("lenX", "lenfn")}
lenX <- (length(X) == length(.X))
lenfn <- (length(fn_fmls(rf)) == length(fn_fmls(.rf)))
```
````

The first hook communicates to the autoharp that the objects `rf` and `X`
(created in the previous chunk) are to be used as reference objects - we may wish
to compare the student versions with these later on. In preparation, the
autoharp duplicates them as `.rf` and `.X`.

The second one informs the autoharp that the subsequent code, when run on 
student-created `X` and `rf`, should yield two values: `lenX` and `lenfn`. These
should be returned as part of the "correctness checks" for each student.

Here is another example of a chunk that will return autoharp scalars:

````markdown
`r ''````{r testxx, autoharp.scalars=c("max_X", "min_X")}
max_X <- max(X)
min_X <- min(X)
```
````

In short, these chunks contain normal R code that utilise the objects created by
students. However, they could also contain autoharp code that analyses the
structure of student R code. For instance, the following `autoharp`-specific
code would extract the number of calls made to `mutate` in the student script:

````markdown
`r ''````{r testxxx, autoharp.scalars=c("f1", "mutate_count")}
f1 <- rmd_to_forestharp(.myfilename)
mutate_count <- fapply(f1, count_fn_call, combine = TRUE, pattern="mutate")
```
````

`rmd_to_forestharp`, `count_fn_call` and `fapply` are `autoharp` functions. The
variable `.myfilename` is hard-coded to contain the path to the current student
script. It allows the autoharp to access the student file from within the
solution script.

# How the Elements Work Together

Here is a visual depiction of how the elements work together, alongside autoharp
functions.

<div align="center">
  <img src="figs/instructor_workflow_diagram.png" alt="autoharp process" width="500">
</div>

1. First, the solution environment has to be populated. This is where the
   `populate_soln_env` from autoharp comes into play. The inputs to this function
   are 

   * the solution script, 
   * a pattern that identifies which chunks are test chunks, and  
   * the directory to knit the solution script in. 

   The return object from this function is a list of length 2,
   containing the solution environment, and a path to the solution script. 
    
2. The next step is to render the student script or Rmd into a html file. This is
   done within the `render_one` function of autoharp. Once the student file has been
   successfully rendered, the objects it generates are stored in the **student_environment**. 

   Rendering of student files is carried out in a separate R process, so that paths 
   are reset for each student. At this point, run-time statistics would already 
   have been generated. 
    
3. The next step is to run the correctness check. The "model" objects from the 
   **solution  environment** are then copied into the student environment. Remember,
   these should not conflict with what is in the student environment, because they 
   would have a period in their prefix. For instance, the student environment 
   will now contain `.X` (from solution template) and `X` (from student).

4. Correctness is assessed by running the temporary solution script (from step 1) within
   the student environment. The `autoharp.scalars` are then appended to the runtime 
   statistics to generate a data frame with one row for each student script.

# Package Output 

Here is how the autoharp package can be used to assess the scripts.

```{r motivating_example, eval=FALSE, echo=TRUE}
soln_fname <- system.file("examples", "example-01", "soln_template_01.Rmd", 
			                    package="autoharp")
temp_dir <- tempdir()
s_env <- populate_soln_env(soln_fname, pattern = "test", knit_root_dir = temp_dir)
stud_script_names <- list.files(system.file("examples", "example-01", 
					                                  "student_scripts", package="autoharp"), 
                                full.names = TRUE) 

corr_out <- lapply(stud_script_names, render_one, out_dir = temp_dir, 
                   knit_root_dir = temp_dir, soln_stuff = s_env)

do.call("rbind", corr_out)
```

---
title: "TreeHarp S4 class"
output: 
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{TreeHarp S4 class}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r setup}
library(autoharp)
```

# Introduction

Within `autoharp`, the TreeHarp S4 class is used to represent an R
expression. It can then be manipulated in several ways in order to perform
static code analysis of student submissions.

The `lintr` package does an excellent job of parsing R code,
but it provides too much detail for the simpler tasks that `autoharp`
carries out. It is still used in the constructor, but some of the parsed output
from `lintr` are dropped. For instance, the `(` parentheses are dropped. 

To understand the elements in a TreeHarp, let us consider the use of a simple
expression. Suppose we fit a linear model to variables in a dataset. To create
a TreeHarp object from an expression, we provide the expression together with
the `quote = TRUE` argument. This is important because method dispatch is
performed based on that second argument, not the first! If we were to dispatch
on the first, R would evalate the expression in order to check its class - thus
destroying the expression we intended to capture.

```{r th-example-1, echo=TRUE}
tree1 <- TreeHarp(quote(lm(y ~ x1 + x2, data=mydata)), TRUE)
```

TreeHarp objects have an associated plot method for visualisation of the
expression. This method relies on the plotting functions from the `igraph`
package.  The full set of parameter options from `igraph` can be utilised when
plotting TreeHarp objects. Figure 1 displays the visualisation of the `tree1`
object created earlier.

```{r th-example-1-plot, echo=TRUE, fig.align='center', collapse=TRUE, fig.cap='Example TreeHarp object'}
opar <- par(mar=c(0,0,0,0))
plot(tree1, vertex.size=25, asp=0.6, vertex.color="gray", vertex.frame.color=NA)
par(opar)
```

# Slots

There are 4 slots in a TreeHarp object. The only required one for valid
instantiation is the adjList.

## adjList

```{r th1-adjlist}
slot(tree1, "adjList")
```

This slot contains an adjacency list that represents the tree structure of the
code. Nodes in a tree are labelled in Breadth-First Search (BFS) order. Thus
the root node has id 1, and does not appear in the adjacency list. To avoid
redundancy, the TreeHarp convention is to list each edge only once, as a child.
Here's what we mean: node 2 in the example above is a neighbour of node 1 and
4, but it only appears under node 1. It does not appear as an adjacent node of
node 4 because it is not child of node 4. Terminal nodes (leafs) have a NULL
entry in the list.

## nodeTypes

If the TreeHarp object was constructed from an R language object, this slot will
be automatically populated. To identify node types, functions from
[rlang](https://cran.r-project.org/package=rlang) are applied to sub-expressions defined by nodes
recursively. Each node is then identified as either:

* a function call, or 
* a formal argument.

The `nodeTypes` slot stores the information in a data frame with one row per
node. The columns are:

1.  *id* (node id). The root node has id 1.
2.  *name*. The name of the node.
3.  *call_status*. A TRUE/FALSE column indicating if the node was a call or not.
4.  *formal_arg*. If the node is not a call, then this column will indicate if it 
    is a formal argument or not. If it is not a call and not a formal argument, it 
    is a symbol representing an R object - we call this an actual argument.
5.  *depth*. This is the depth of the node in the tree. The root of the tree has 
    depth 1.

autoharp provides a getter function to retrieve the node types easily.

```{r th1-nodetypes, echo=TRUE}
get_node_types(tree1)
```

## call

The call slot stores the original expression that was used to construct the
TreeHarp object, just in case it needs to be executed later.

```{r ex1_call}
slot(tree1, "call")
```

## repr

The repr slot contains a string representation of the object. If the original
TreeHarp object has been modified, then it may not be a proper R expression, so
this slot stores the best representation of it. This slot is used when the
object is printed, or when `show` is called on the S4 object.

```{r ex1_repr}
tree1
```

# TreeHarp Methods

As we have already demonstrated, the plot method exists for this class. It
relies on the tree layout of igraph package, but additional arguments can be
used to customise the plot. For instance, we could use colour to distinguish
between calls and non-call nodes:

```{r th-example-2-plot, echo=TRUE, fig.align='center', fig.cap='TreeHarp object with colored nodes'}
opar <- par(mar=c(0,0,0,0))
plot(tree1, vertex.size=25, asp=0.6, vertex.color=tree1@nodeTypes$call_status)
par(opar)
```

These are the other S4 methods defined for the TreeHarp class:

* `length`: returns the number of nodes.
* `names`: returns the node names.
* `get_parent_id`: returns the parent id of a node.
* `get_child_ids`: returns the ids of the children of a node.
* `get_node_types`: returns the nodeTypes slot from a TreeHarp object.
* `get_adj_list`: returns the adjacency list slot from a TreeHarp object.

---
title: "Solution Template (02)"
output: html_document
---
# Summary Output

This is a summary of the tests conducted in this solution template.

1.  `run_status`: SUCCESS indicates that I am able to run your code - 
    there are no syntax errors, and path settings are correct.
2.  `last.X`: This entry displays the last value of the `X` vector in your code.
    An NA indicates that the vector was not of the correct length, or that the 
    vector is not present.

```{r test00}
library(testthat)
library(autoharp)
```

```{r corr_solutions, echo=FALSE}
X <- c(1L, 3L, 5L, 7L,  9L, 11L)
mean_X <- mean(X)
```

```{r test02, autoharp.scalars=c("last.X")}
last.X <- X[length(X)]
```

---
title: "Key Concepts"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Key Concepts}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
bibliography: references.bib
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Working Directories

As part of the grading process, student scripts are converted into HTML
documents. It is easier for the instructor to review this format, if necessary,
since it incorporates both the code and output in a single document. The
autoharp uses the `knitr` package [@knitr1] to achieve this goal. The process
of rendering an R script (or Rmd/qmd) into an HTML file is referred to as *knitting*.

As part of the assignments, student scripts may need to read in local datasets,
or to write to output files. Hence it is important to consider the
working directory when knitting a file. The working directory is the default
location that R looks when it is asked to load a file or save output. When a
file is knitted, the working directory is typically the location of the file.
However, this may not be ideal when grading hundreds of student scripts, across
multiple assignments. The instructor would want to keep their own files separate
from student output. To achieve this, the autoharp provides the
functionality to specify the working directory when knitting a file, and to
specify the directory where output should be written.

The only impositions that the package prescribes are:

*   students should mirror the instructor set-up with respect to data and output 
    files.
*   students should use relative paths in all their scripts.

Relative paths are file paths that refer to files *relative* to the working directory.
The alternative to relative paths are absolute paths, where a path is hard-coded. 
However, as pointed out in section 6.4 of "R for Data Science" (@r4ds), the use of absolute 
paths hinders sharing. Thus *one should only ever use relative paths in scripts*.

# Environments

The autoharp relies heavily on the concept of environments within R. As
described in chapter 7 (@wickham2019advanced), an environment is simply 
a data structure that associates a set of names to a set of values. It is the key 
structure in how R implements lexical scoping. Every environment (except the empty 
environment) also contains a pointer to a parent environment. When a function is 
executed, it first looks within it's execution environment for objects required. If 
an object cannot be found, the functions looks to the enclosing parent, and so on 
and so forth, until the object is found. If the object cannot be found, an error 
occurs.

The autoharp creates a separate environment for each student's script.
Objects created when rendering the student scripts are compared against model
objects, which are created in a solution environment when the solution
template is rendered.

# Forked R Processes

The `parallel` package allows a user to create forked R processes from a 
session. A forked session is a separate R session. It is primarily used for running
parallel code. However, the autoharp uses it to sandbox the execution of 
student scripts.

For each student script, `render_one` creates a new fork, within which the script 
is rendered. The objects generated by the student code are stored in an environment
within the forked process. Subsequently, test chunks are executed by the forked 
process within the student environment. 

The above implementation ensures that objects in the global environment (where
the instructor executes `render_one`) and packages loaded by the instructor do
not interfere with the student code. Suppose, on the other hand, that the
student scripts were executed in the same R process as the grading code. Then,
if the student had forgotten to load certain packages that their code needed,
but the instructor session had loaded those packages. Instead of failing to
render (which is what should happen in this case), the rendering would be
successful.

# `knitr` Hooks

As described in chapter 13 (@rmarkdown3), a chunk hook is a function that is
triggered by a chunk option, when the value of this chunk option is not `NULL`.
When defining a chunk hook, one can specify if the hook is to be run before or
after the execution of the subsequent chunk. 

The autoharp defines two hooks:

* `autoharp.objs`: The input to this hook must be a character vector. When the 
  input is not `NULL`, the chunk hook will create a copy of each object specified
  in the character vector, pre-prended with a period. For instance, if the 
  user specifies `autoharp.objs = "X"`, an object named `.X` will be created in the 
  execution environment of the Rmd file.
  
  The names of model solution objects in the solution environment begin with a period.
* `autoharp.scalars`: Once again, the input to this hook must be a character 
  vector. When not `NULL`, this character vector will be appended to a vector in
  the solution environment named `.scalars_to_keep`. 
  
  This serves as a reference vector. When the test chunks are executed in the student
  solution environment, the objects specified in `.scalars_to_keep` will be copied 
  out as columns in the correctness dataframe.

# References

---
title: "Shiny App: Similarity Checks"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Shiny App: Similarity Checks}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

The similarity app is an instructor-facing [shiny](https://shiny.posit.co/) web
application for analyzing code similarity across a batch of student
submissions. While the solution checker helps individual students validate
their work before submission, the similarity app helps instructors review an
entire class at once. It can identify clusters of suspiciously similar scripts,
visualize how different students approached the same problem, and provide a
side-by-side comparison of any two submissions. This makes the application
useful both for detecting potential plagiarism and for understanding the range
of solutions students produce.

To run the similarity app locally:

```{r similarity-app-run, eval=FALSE, echo=TRUE}
library(shiny)
runApp(system.file(file.path('shiny', 'similarity_app', 'app.R'),
                   package = 'autoharp'))
```

When the application launches, the instructor is presented with a configuration
panel on the left and a results area on the right. The configuration panel
allows the instructor to specify the folder containing student scripts, choose
a similarity method and visualization technique, and set a threshold for
cluster detection.

<div align="center">
  <img src="figs/similarity_app_ui.png" alt="similarity app UI" width="500">
</div>

The application supports three methods for computing pairwise similarity:

1.  **Cosine similarity (token-based).** This is the default and recommended
    method for most use cases. It parses each script into tokens using
    \pkg{autoharp}'s `rmd_to_token_count()` function, constructs TF-IDF
    weighted frequency vectors, and computes cosine similarity between them.
    The method is fast and robust to formatting differences. Instructors can
    configure n-gram size (from unigrams to 5-grams), choose whether to include
    literal values in the token set, and optionally exclude function calls from
    loaded packages to focus on the structure of student-written code.

2.  **Jaccard similarity (set-based).** This method compares the unique tokens
    present in each script using the Jaccard index (intersection over union).
    Unlike cosine similarity, it ignores how often tokens appear and focuses
    only on whether they appear at all. This can be useful when detecting
    shared vocabulary matters more than shared patterns of use.

3.  **Edit distance.** This method computes the Levenshtein edit distance between
    the raw code strings, normalized by the length of the longer script. It is
    the strictest method, sensitive to whitespace, variable names, and
    formatting. It works best for detecting near-verbatim copies.

After selecting a similarity method, the instructor chooses a visualization
technique. The application offers t-SNE (t-Distributed Stochastic Neighbor
Embedding) via [Rtsne](https://cran.r-project.org/package=Rtsne), which excels at revealing local
clusters, and MDS (Multi-Dimensional Scaling), which provides a more stable,
deterministic projection that preserves global distances. The resulting 2D
scatter plot is rendered using
[plotly](https://cran.r-project.org/package=plotly), making it
interactive: instructors can hover over points to see script names and click to
view the full script content.

<div align="center">
  <img src="figs/similarity_app_scatter.png" alt="similarity app scatter" width="500">
</div>

The application also generates a hierarchically clustered heatmap showing all
pairwise similarity scores. This visualization makes it easy to spot blocks of
high similarity and understand the overall distribution of scores across the
class.

<div align="center">
  <img src="figs/similarity_app_heatmap.png" alt="similarity app heatmap" width="500">
</div>

Below the visualizations, the application presents several summary panels:

*   *Summary statistics.* Mean, median, and standard deviation of pairwise
    similarities, along with counts of high-similarity (0.7 and above),
    medium-similarity (0.4 to 0.7), and low-similarity (below 0.4) pairs.
*   *Top similar pairs.* A sortable \CRANpkg{DT} [@DT] table listing the script
    pairs with highest similarity scores, allowing instructors to quickly
    identify submissions that warrant closer review.
*   *Detected clusters.* Groups of scripts whose pairwise similarities all
    exceed the configured threshold. Each cluster is listed with its member
    scripts, making it easy to identify groups that may have collaborated or
    copied.
*   *Full similarity matrix.* A searchable table containing all pairwise
    similarity scores, which can be downloaded as a CSV file for record-keeping
    or further analysis.

The application also includes a dedicated Diff View tab for detailed comparison
of any two scripts. Instructors select two files from dropdown menus and click
"Compare Scripts" to see a side-by-side view with synchronized scrolling. Lines
are color-coded to highlight differences: yellow for modified lines, green for
additions, and red for deletions.

Here is an outline of the files in the similarity app directory:

```{r similarity-app-files, echo=TRUE}
list.files(system.file("shiny", "similarity_app", package = "autoharp"))
```

*   `app.R`: Contains the UI and server logic for the application.
*   `R/helpers.R`: Helper functions for token parsing, similarity computation,
    dimensionality reduction, and cluster detection.
*   `www/styles.css`: Custom CSS for styling the application interface.
*   `README.md`: Documentation with usage instructions and guidance on choosing
    between similarity methods.
*   `sol/`: A folder containing example scripts for demonstration purposes.

The `R/helpers.R` file implements the core computational functions. For cosine
similarity, it uses a vectorized approach that constructs a document-term
matrix and computes all pairwise similarities in a single matrix operation,
making it efficient even for large classes. For Jaccard and edit distance,
which require pairwise computation, the app supports parallel processing on
Linux and macOS through the `n_cores` parameter, using `parallel::mcmapply()`
to distribute the workload across multiple CPU cores.

Interpreting similarity scores requires context. A score of 1.0 indicates
identical scripts. Scores between 0.7 and 0.99 suggest high similarity that may
warrant investigation. Scores between 0.4 and 0.69 typically indicate similar
approaches or shared patterns, which may be expected when students solve the
same problem. Scores below 0.4 generally indicate distinct implementations. The
similarity threshold slider (defaulting to 0.7) controls how aggressively the
app groups scripts into clusters; a higher threshold produces fewer, tighter
clusters.

---
title: "Getting Started with autoharp"
description: >
  Learn how to install autoharp and use it to grade your first R or
  R Markdown student submission in minutes.
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Getting Started with autoharp}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
  collapse  = TRUE,
  comment   = "#>",
  eval      = FALSE,
  fig.align = "center"
)
```

## What is autoharp?

`autoharp` is an R package for **semi-automatic grading** of R and R Markdown
(Rmd/qmd) scripts. It was designed at the National University of Singapore to
handle the practical challenges of grading programming assignments at scale:

- Students submit **complete, reproducible documents** (not isolated code snippets)
- Instructors need to check both *output correctness* and *code quality*
- The same workflow must work for 10 or 1,000 submissions
- Some checks must go beyond outputs: e.g., "did the student avoid using a `for` loop?"

`autoharp` achieves this through four complementary layers:

| Layer | What it Checks |
|-------|---------------|
| **Output correctness** | Objects match the reference solution (typed, tolerance-aware) |
| **Static code analysis** | AST structure: e.g., no `for` loops, correct function signature |
| **Runtime profiling** | Execution time and peak memory usage per submission |
| **Code style (lint)** | `lintr`-based style violations count |

---

## The Big Picture: Four-Phase Workflow

```{r workflow-img, echo=FALSE, eval=TRUE, out.width="90%", fig.cap="The four-phase autoharp grading workflow: Prepare → Distribute → Grade → Review"}
knitr::include_graphics("figs/workflow.png")
```

The typical `autoharp` workflow has four phases:

1. **Prepare**: Write a question sheet and a *solution template* (an Rmd with
   special `autoharp.objs`/`autoharp.scalars` chunk options)
2. **Distribute**: Share the question PDF and a blank student template
3. **Grade**: Run `populate_soln_env()` then `render_one()` per student;
   each runs in a **sandboxed R process**
4. **Review**: Inspect logs, thumbnail galleries, and the Grading App

---

## Installation

You can install **autoharp** from CRAN with:

```{r install, eval=FALSE}
install.packages("autoharp")
```

You can also install the development version from GitHub:

```{r installdev}
# install.packages("devtools")
devtools::install_github("namanlab/autoharp")
```

Then load the package:

```{r load}
library(autoharp)
```

---

## Core Concepts

### Solution Templates

The heart of `autoharp` is the **solution template** - an R Markdown file
that does two things simultaneously:

1. Contains the **reference solution** to the problem
2. Defines **test code** using special knitr chunk options

Two special chunk options are used in **test chunks** (chunks whose names match a pattern, default "test"):

| Chunk Option | Purpose |
|--------|---------|
| `autoharp.objs` | Lists object names to copy from the solution environment so tests can compare student objects against solution objects |
| `autoharp.scalars` | Lists scalar variable names that will be created by the test code; these become columns in the output data frame |

> **Why Rmd?** Grading complete documents (rather than isolated snippets)
> ensures students practice good scientific computing habits: their entire
> analysis must render cleanly.

### The Grading Pipeline

```
Solution Template (.Rmd)
        │
        ▼
populate_soln_env()  --> Solution Environment + Test Script
        │
        ▼
render_one(student.Rmd)  --> Grading Results Data Frame
        │
        ▼
log_summary()  --> Summary Report
```

Each `render_one()` call:
1. Launches a **fresh, sandboxed R process** (via `parallel::makePSOCKcluster`)
2. Checks for forbidden calls (`system()`, `setwd()`, etc.)
3. Knits the student's Rmd with `autoharp` hooks active
4. Runs the test script in the student's environment
5. Returns a data frame with status, runtime, memory, and test results

---

## Motivating Example

### The Problem

Suppose you assign students the following problem:

> Write a function `rf(n)` that generates `n` random variates from the
> density $f(x) = 4x^3$, $0 < x < 1$. Use the inverse transform method.
> Then create a vector `X` of 10,000 variates using `rf()`.

### Step 1: Create a Solution Template

Create `solution_template.Rmd`:

````{verbatim}
---
title: "Solution Template"
output: html_document
---


```{r solution}
rf <- function(n) {
  u <- runif(n)
  u^(1/4)  # inverse CDF of f(x) = 4x^3
}
X <- rf(10000)
```

```{r test01, autoharp.objs=c("rf", "X"), autoharp.scalars=c("lenX", "lenfn", "meanX", "sdX")}

# Compare student objects with solution objects
lenX <- (length(X) == length(.X))
lenfn <- (length(formals(rf)) == length(formals(.rf)))

# Check statistical properties
meanX <- (abs(mean(X) - 0.8) < 0.02)
sdX <- (abs(sd(X) - 0.163) < 0.02)
```

````

### Step 2: Populate the Solution Environment

```{r populate}
soln <- populate_soln_env("solution_template.Rmd", pattern = "test")

# soln is a list with two elements:
# $env: the solution environment (contains .rf, .X and other solution objects)
# $test_fname: path to the extracted test script (code from test chunks)
str(soln)
```

### Step 3: Grade a Student Submission

Suppose a student submits `student01.Rmd`:

```{r render}
result <- render_one(
  rmd_name      = "student01.Rmd",
  out_dir       = "output/",
  knit_root_dir = getwd(),
  soln_stuff    = soln
)

# The result is a one-row data frame
print(result)
```

The output data frame contains:

| Column | Description |
|--------|-------------|
| `fname` | Student's filename |
| `time_stamp` | Timestamp of the grading run |
| `run_status` | `"SUCCESS"`, `"FAIL"`, or `"UNKNOWN"` |
| `run_time` | Total execution time (seconds) |
| `run_mem` | Memory usage of the environment |
| `lenX`, `lenfn`, ... | Scalar test results defined in `autoharp.scalars` |

### Step 4: Summarise Results Across All Students

```{r summarise}
# Grade all students in a directory
student_files <- list.files("submissions/", pattern = "\\.Rmd$", full.names = TRUE)

results_list <- lapply(student_files, function(f) {
  render_one(rmd_name = f, out_dir = "output/", knit_root_dir = getwd(),
             soln_stuff = soln)
})

all_results <- do.call(rbind, results_list)

# Print a summary from the log file
log_summary("output/render_one.log")
```

---

## Checking Code Style

`autoharp` integrates with the `lintr` package to count style violations:

```{r lints}
# Count lint violations in a single script
lint_count <- count_lints_one("student01.R")

# Count across all submissions
all_lints <- count_lints_all(
  file_names = list.files("submissions/", pattern = "\\.R$", full.names = TRUE)
)
print(all_lints)
```

You can run lint checks separately from the main grading pipeline using these functions.

---

## Checking Rmd Structure

For R Markdown submissions, verify that the file is a valid Rmd:

```{r check-rmd}
# Check that the submitted file is a valid Rmd
# Returns TRUE if file has .Rmd extension, YAML header, and R chunks
rmd_check <- check_rmd(fname = "student01.Rmd")
print(rmd_check)
```

---

## Interactive Grading with Shiny Apps

For large classes, the **Grading App** provides a browser-based interface
that wraps the entire workflow:

```{r grading-app, eval=FALSE}
# Launch the full grading GUI
shiny::runApp(system.file("shiny/grading_app", package = "autoharp"))
```

The Grading App has five tabs:

1. **Session**: start/resume a grading session. 
2. **Object Testing**: auto-generate and run correctness tests. 
3. **Template**: generate a solution template from a script. 
4. **Autoharp**: run `render_one()` for all submissions with progress tracking. 
5. **Plots**: review student plots side-by-side with the solution. 

See the [Shiny Apps Guide](https://namanlab.github.io/autoharp/articles/shiny_grading.html) for full details.

---

## Next Steps

- **[TreeHarp Tutorial](https://namanlab.github.io/autoharp/articles/treeharp_class.html)**: learn AST-based static
  analysis: detect `for` loops, check function signatures, and more
- **[Complete Workflow](https://namanlab.github.io/autoharp/articles/demo_workflow.html)**: a detailed end-to-end walkthrough
  with a real data analysis assignment
- **[Shiny Apps](https://namanlab.github.io/autoharp/articles/shiny_grading.html)**: use the Grading App, Similarity App, and Student Solution Checker
- **[Function Reference](https://namanlab.github.io/autoharp/reference/index.html)**: full API documentation

---

## Getting Help

- 📋 [GitHub Issues](https://github.com/namanlab/autoharp/issues): bug reports and feature requests
- 📖 [Full Documentation](https://namanlab.github.io/autoharp/): pkgdown site

---
title: "Shiny App: Grading"
output:
  rmarkdown::html_vignette:
    toc: true
    toc_depth: 2
vignette: >
  %\VignetteIndexEntry{Shiny App: Grading}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

# Introduction

The grading app is an instructor-facing [shiny](https://shiny.posit.co/) web
application that provides a comprehensive environment for evaluating student
submissions. Unlike the similarity app, which focuses on comparing submissions
to each other, the grading app centers on assessing correctness against a
solution template. It combines automated object-level testing with manual plot
grading, session persistence for interrupted workflows, and template generation
for use with autoharp's `render_one()` function. To run the grading app locally:

```{r grading-app-run, eval=FALSE, echo=TRUE}
library(shiny)
runApp(system.file(file.path('shiny', 'grading_app', 'app.R'),
                   package = 'autoharp'))
```

The application organizes its functionality across five tabs: Session
management for saving and resuming work, Object Testing for automated
correctness checks, Template Download for generating autoharp-compatible
templates, Autoharp Grading for running `render_one` workflows, and Plot
Grading for manual assessment of visualizations.

## Session Management

A key design principle of the grading app is workflow persistence. Grading
large classes often spans multiple sessions, and losing progress due to
application restarts or system interruptions can be frustrating. The Session
tab addresses this by providing save and resume functionality.
Sessions are stored as RDS files containing the complete application state:
solution objects, student file lists, grading results, plot grades with
comments, and the current position in the grading workflow. The application
auto-saves after significant operations (completing automated grading, saving
plot grades, finishing autoharp grading), and instructors can trigger manual
saves at any time.

<div align="center">
  <img src="figs/grading_app_session.png" alt="grading app session" width="500">
</div>

To resume a previous session, the instructor uploads the saved RDS file and
clicks "Load Session." The application restores all state, including input
field values and the current student being graded, allowing work to continue
exactly where it left off.

## Automated Object Testing

The Object Testing tab implements type-aware correctness checking without
requiring instructors to write test code. The instructor uploads a solution
template (an R Markdown or Quarto file containing correct implementations) and
specifies the folder containing student submissions. The application then
executes both the solution and each student script in isolated environments,
comparing the objects they produce.

<div align="center">
  <img src="figs/grading_app_object_testing.png" alt="grading app object testing" width="500">
</div>

The key innovation in this tab is the type-based test generation. When the
solution template is executed, the application inspects each created object and
determines its type. Based on this type information, it generates appropriate
tests automatically.


## Template Generation

While the Object Testing tab provides immediate feedback, instructors may want
to use autoharp's full `render_one()` workflow for more detailed grading.
The Template Download tab bridges these approaches by generating
autoharp-compatible solution templates with properly formatted test chunks.

<div align="center">
  <img src="figs/grading_app_template.png" alt="grading app template generation" width="500">
</div>

The template generator works by analyzing the solution environment and creating
test chunks that mirror the automated tests from the Object Testing tab. Each
chunk includes the `autoharp.objs` option (specifying which objects to extract
from student code) and the `autoharp.scalars` option (specifying which test
results to record). The generated code follows \pkg{autoharp}'s conventions,
using dot-prefixed names for solution objects (e.g., `.X` for the solution's
`X` object). Instructors can preview the generated template before downloading, allowing
them to verify the test structure and make any necessary adjustments. The
downloaded template can then be used directly with `populate_soln_env()` and
`render_one()`.

## Autoharp Grading Integration

The Autoharp Grading tab provides a graphical interface for running
autoharp's core grading workflow. Instead of writing R scripts to call
`populate_soln_env()` and `render_one()`, instructors upload a template file
and specify the student scripts folder through the interface.

<div align="center">
  <img src="figs/grading_app_autoharp.png" alt="grading app grading" width="500">
</div>

The tab offers several configuration options:

*   *Test chunk pattern.* The pattern used to identify test chunks in the
    template (defaulting to "test"). This corresponds to the `pattern` argument
    of `populate_soln_env()`.
*   *Permission to install packages.* Whether `render_one()` should
    automatically install packages that student scripts require but are not
    present on the grading machine.
*   *Output directory.* Where rendered HTML files, logs, and session files are
    stored.

After grading completes, results appear in a comprehensive table showing run
status (SUCCESS or FAIL), execution time, memory usage, and all correctness
check scalars defined in the template. If plot grades have been assigned in the
Plot Grading tab, they appear alongside the automated results. 
The tab includes an Excel export feature that creates a workbook with two
sheets: one containing all `render_one()` results and another containing plot
grades and comments. This format facilitates integration with course management
systems and grade books.

## Manual Plot Grading

The Plot Grading tab addresses a challenge that automated systems cannot fully
solve: evaluating the quality of data visualizations. While code correctness
can be verified programmatically, assessing whether a plot effectively
communicates its message requires human judgment.

<div align="center">
  <img src="figs/grading_app_plots.png" alt="grading app plot grading" width="500">
</div>

When automated grading runs (either through the Object Testing tab or the
Autoharp Grading tab), the application collects all plots generated by each
student script. These plots are stored in a structured directory hierarchy
within the output folder.

The Plot Grading interface allows instructors to navigate through students
sequentially (using Previous/Next buttons or a dropdown selector), view all
plots for the current student, and assign numeric grades with optional
comments. Clicking a plot thumbnail opens an expanded view in a modal dialog,
making it easier to examine details. The tab also displays a summary of
automated test results for the current student, providing context when
assessing plots. For example, if a student's data processing produced incorrect
values, this context helps explain why their visualization may look different
from expectations.  A progress bar tracks grading completion, and all grades
are auto-saved to the session file after each entry. This combination of
automated collection, easy navigation, and persistent storage streamlines what
would otherwise be a tedious manual process.

---
title: "Solution Template (01)"
output: html_document
---
```{r test00}
library(autoharp)
library(rlang)
```

```{r corr_solutions, echo=FALSE}
rf <- function(n) {
  U <- runif(n)
  X <- U^(1/4)
  X
}
X <- rf(1e4)
```

```{r test01, autoharp.objs=c("rf", "X"), autoharp.scalars=c("lenX", "lenfn"), echo=FALSE}
lenX <- (length(X) == length(.X))
lenfn <- (length(fn_fmls(rf)) == length(fn_fmls(.rf)))
```

```{r test02, autoharp.scalars=c("mean.X", "sd.X")}
mean.X <- mean(X)
sd.X <- sd(X)
```

```{r test03, autoharp.scalars=c("for_loop")}
f1 <- rmd_to_forestharp(.myfilename)
for_loop <- fapply(f1, detect_for_in_fn_def, fn_name = "rf", combine=TRUE, 
                   combiner_fn = function(x) any(unlist(x)))
```

---
title: "Student 01 Testing"
output: html_document
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(echo = TRUE)
```

## Problem 1: Basic Calculations

Students should compute the mean and standard deviation of a vector.

```{r solution_objects}
# Solution objects that students should create
X <- 1:20
Y <- c(5, 10, 15, 20, 25)
Z <- rnorm(100, mean = 10, sd = 3)
```

## Problem 2: Custom Function

Students should create a function that computes the square of a number.

```{r solution_function}
# Solution function
square_number <- function(n) {
  return(n*n)
}

# Expected result
expected_result <- square_number(5)
```


## Problem 3: Data Manipulation

Students should filter a data frame.

```{r solution_data}
# Solution data manipulation
original_df <- data.frame(
  name = c("Alice", "Bob", "Charlie", "Diana"),
  age = c(25, 30, 35, 28),
  score = c(85, 92, 78, 95)
)

filtered_df <- original_df[original_df$score > 80, ]
```



```{r solution_plot}
library(tidyverse)

ggplot(filtered_df) +
  geom_point(aes(x = age, y = score)) +
  theme_dark()


par(mfrow = c(2, 2))

# Panel 1: Line plot
t <- seq(0, 2*pi, length.out = 100)
plot(t, sin(t), type = "l", col = "blue", main = "Sine Wave")

# Panel 2: Points
plot(rnorm(50), rnorm(50), pch = 16, col = "red", main = "Random Points")

# Panel 3: Bar plot
barplot(c(3, 7, 4, 9, 2), names.arg = LETTERS[1:5], col = "purple", main = "Bar Chart")

# Panel 4: Pie chart
pie(c(30, 25, 20, 15, 10), 
    labels = c("A: 30%", "B: 25%", "C: 20%", "D: 15%", "E: 10%"),
    main = "Pie Chart")
```

---
title: "Sample Questions 01"
output: html_document
---

Consider the following probability density function (pdf) over the support 
$(0,1)$:

\begin{equation*}
f(x) = 
\begin{cases}
4x^3 & \text{if } 0 < x < 1 \\
0 & \text{otherwise}
\end{cases}
\end{equation*}

```{r echo=FALSE}
rf <- function(n) {
  U <- runif(n)
  X <- U^(1/4)
  X
}
```

Write a function called `rf`, that generates i.i.d observations from this pdf.
It should take in exactly one argument, $n$, that determines how many random
variates to return. For instance:

```{r}
set.seed(33)
rf(n = 5)
```

Now generate 10,000 random variates from this pdf and store them in a vector 
named `X`.

Your script must generate a function named `rf` and a vector named `X`.

---
title: "test01"
output: html_document
---

```{r}
X <- 1:10
Y <- LETTERS[1:5]
Z <- 10:15
```


```{r test02, autoharp.objs=c("X", "Y", "Z"),autoharp.scalars=c("Xmean", "Zmean")}
Xmean <- mean(X)
Zmean <- mean(Z)
```

```{r test03, autoharp.scalars=c("Y1")}
Y1 <- Y[3] # Note that it is character here.
```

---
title: "Solution template 01"
output: html_document
---

```{r generate_objects}
x <- 1:10
y <- 11:20
```


```{r test_01, autoharp.objs=c("x", "y"), autoharp.scalars=c("means_check", "sets_check")}

means_check <- mean(x) == mean(.x)
sets_check <- setequal(y, 20:11)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pre_checks.R
\name{render_prechecks}
\alias{render_prechecks}
\title{Conducts checks before rendering file}
\usage{
render_prechecks(fname, in_place = TRUE, new_name, verbose = FALSE)
}
\arguments{
\item{fname}{The name of the Rmd/qmd file to check.}

\item{in_place}{If TRUE, the original file will be overwritten in place.}

\item{new_name}{If in_place is FALSE, this should be the full path to the 
new file to be written (including directory location).}

\item{verbose}{If TRUE, debugging messages will be printed.}
}
\value{
Either TRUE or FALSE. If TRUE, all View() lines have been commented,
and there are no system calls. If FALSE, it means there are system calls (to 
be screened further).
}
\description{
This routine conducts two checks before rendering the file: presence of 
View function, and presence of system() calls. The former is a nuisance,
while the latter is possibly nefarious.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{carve_subtree}
\alias{carve_subtree}
\title{Carve out branches to form a new tree.}
\usage{
carve_subtree(obj, char_arr)
}
\arguments{
\item{obj}{An object of class TreeHarp.}

\item{char_arr}{A vector of 1's and 0's indicating which nodes to keep. The
vector should have length equal to the number of nodes in obj.}
}
\value{
An object of class TreeHarp.
}
\description{
This functions keeps only the indicated nodes, returning a new sub-tree.
}
\details{
This returns an error if the sub-tree does not define a new tree.
}
\examples{
th3 <- list(a= c(2L,3L,4L), b=NULL, c=c(5L, 6L), d=7L, e=NULL, f=NULL, g=NULL)
carve_subtree(TreeHarp(th3), c(1,0,0,0,0,0,0))
st <- subtree_at(TreeHarp(th3), 4)
plot(st)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{get_levels}
\alias{get_levels}
\title{Obtains the node levels from a tree.}
\usage{
get_levels(adj_list)
}
\arguments{
\item{adj_list}{The adjacency list of the tree.}
}
\value{
It returns a vector of integers. The length of this vector will be
the number of nodes in the tree. The root is at level 1, the next is at level
 2, and so on.
}
\description{
This function obtains the node levels from a tree.
}
\details{
This function is used to check if the specification of the tree is
in BFS order. If that is indeed the case, the levels of each node should be
sorted.

This function is not exported for the general user.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{carve_mst}
\alias{carve_mst}
\title{Carve a Minimal Spanning Tree Out}
\usage{
carve_mst(th, node_names)
}
\arguments{
\item{th}{An object of class TreeHarp.}

\item{node_names}{A character vector of node names. Nodes outside this set 
will not be returned in the tree. It must include the root node name.}
}
\value{
An object of class TreeHarp.
}
\description{
Given node names, this function retrieves the smallest tree containing at most 
those nodes.
}
\details{
The function starts from each node specified and works it's way up
to the root. If a branch contains nodes outside the list, it is shortened.

In the end, the tree that is returned will try to contain all the named nodes,
but if that's not possible some will dropped to ensure a tree is returned, 
not a disconnected graph.
}
\examples{
ex1 <- quote(x <- f(y, g(5)))
th1 <- TreeHarp(ex1, TRUE)
carve_mst(th1, c("<-", "x", "f", "5")) ## note: 5 is dropped.
carve_mst(th1, c("<-", "x", "f", "y")) 
carve_mst(th1, c("<-", "f", "g")) 

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_runtime.R
\name{check_runtime}
\alias{check_runtime}
\title{Calculate Run-time Statistics}
\usage{
check_runtime(stud_fname, knit_root_dir, return_env = FALSE)
}
\arguments{
\item{stud_fname}{The rmd filename of the student.}

\item{knit_root_dir}{The working directory to use when knitting the file.}

\item{return_env}{A logical value to indicate if the environment from the
rmd file should be return. If FALSE, an NA value is returned.}
}
\value{
A list containing the running time in seconds, the memory used
by the final environment in bytes (as a numeric scalar), and the
environment object containing all the generated objects from the rmd file.
}
\description{
This is stand-alone function. It computes the runtime stats
without rendering the md/html/pdf file.
}
\details{
This routine is not used within any other function within the 
package. Figures are not cleaned or removed.
}
\seealso{
\code{\link{render_one}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/th_getter-length.R
\name{get_node_types}
\alias{get_node_types}
\alias{get_node_types,TreeHarp-method}
\title{Generic for Getting Node Types}
\usage{
get_node_types(x, ...)

\S4method{get_node_types}{TreeHarp}(x, ...)
}
\arguments{
\item{x}{An object of class TreeHarp.}

\item{...}{Unused arguments, for now.}
}
\value{
A data frame containing the node types for a TreeHarp object. If the 
slot is empty, NA is returned.
}
\description{
The generic method definition for getting node types from a
TreeHarp object.
}
\section{Methods (by class)}{
\itemize{
\item \code{get_node_types(TreeHarp)}: A getter.

Allows user to extract the node types of a treeharp object.

}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forestharp.R
\name{pipe_can_improve_this}
\alias{pipe_can_improve_this}
\title{Identify if use of pipe operator can improve a code section}
\usage{
pipe_can_improve_this(fname, window_len = 2)
}
\arguments{
\item{fname}{A filename - either a Rmd/qmd or R script.}

\item{window_len}{A window length to analyse.}
}
\value{
A numeric vector corresponding to start of the window to be reviewed.
If no lines are found, NULL is returned.
}
\description{
Identify if use of pipe operator can improve a code section
}
\details{
A rolling window approach is used here. For each 
expression, the assigned object is extracted (if any). Subsequent lines are 
checked to see if this object appears as an actual argument.

Window length two means that only one more line is checked.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forestharp.R
\name{only_actual_args_differ}
\alias{only_actual_args_differ}
\title{Checks if two expressions differ only in terms of actual arguments}
\usage{
only_actual_args_differ(th1, th2)
}
\arguments{
\item{th1}{A TreeHarp object.}

\item{th2}{A TreeHarp object.}
}
\value{
Either 1 or 0. 1 means that the two expressions differ only in their 
actual arguments.
}
\description{
Checks if two expressions differ only in terms of actual arguments
}
\details{
From the node types, only calls and formal arguments are retained.
If these are identical, then a 1 is returned. Otherwise 0 is returned.
}
\examples{
ex1 <- TreeHarp(quote(X <- rnorm(10, mean=0.9, sd=4)), TRUE)
ex2 <- TreeHarp(quote(Y <- rnorm(20, mean=9, sd=4)), TRUE)
only_actual_args_differ(ex1, ex2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{generate_all_subtrees}
\alias{generate_all_subtrees}
\title{Generate all subtrees from a tree.}
\usage{
generate_all_subtrees(th)
}
\arguments{
\item{th}{An object of class TreeHarp.}
}
\value{
A 0-1 matrix with n rows and m columns. n is the number of sub-trees
rooted at the root node of th. m is the number of nodes in this given tree.
The leading column will be a 1 for all the rows.
}
\description{
This routines generates all subtrees rooted at the root node for a particular
tree.
}
\examples{
th1 <- TreeHarp(list(a=c(2,3), b=NULL, c=NULL))
generate_all_subtrees(th1)
}
\references{
\emph{Listing and counting subtrees of a tree}, F Ruskey, \emph{SIAM Journal on Computing}, 1981
}
\seealso{
\code{\link{get_next_subtree}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/th_getter-length.R
\name{plot,TreeHarp-method}
\alias{plot,TreeHarp-method}
\title{TreeHarp Plotting TreeHarp Objects}
\usage{
\S4method{plot}{TreeHarp}(x, y, ...)
}
\arguments{
\item{x}{An object of class TreeHarp.}

\item{y}{Unused.}

\item{...}{Additional arguments passed on to plot.igraph().}
}
\value{
Returns NULL, invisibly.
}
\description{
A plot method for visualising treeharp objects.
}
\details{
The treeharp object is converted to an igraph object before it
is plotted.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lang_2_tree_helpers.R
\name{update_adj_list}
\alias{update_adj_list}
\title{Update adjacency list.}
\usage{
update_adj_list(
  update_type = c("new_node", "add_child"),
  node_id,
  node_name,
  child_node,
  env_ni
)
}
\arguments{
\item{update_type}{This should be either "new_node" or "add_child". If it 
is a new node, an empty list component is added. If it is add child, 
then child_node should be provided too.}

\item{node_id}{An integer.}

\item{node_name}{The name of the new node to be added. This must be provided 
if the update_type is "new_node".}

\item{child_node}{An integer.}

\item{env_ni}{An environment object, possibly containing an adjacency list 
that will later be used to construct a TreeHarp object.}
}
\value{
An invisible TRUE is returned.
}
\description{
Updates the adjacency list for an R expression parse tree.
}
\details{
This is for internal use. It may be removed from user-view soon!
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_one.R
\name{render_one}
\alias{render_one}
\title{Run a single Rmd file through autoharp.}
\usage{
render_one(
  rmd_name,
  out_dir,
  knit_root_dir,
  log_name,
  soln_stuff,
  max_time_per_run = 120,
  permission_to_install = FALSE
)
}
\arguments{
\item{rmd_name}{The path to the file to be rendered and checked.}

\item{out_dir}{The directory to store all the html output, md output, and
figures.}

\item{knit_root_dir}{The working directory while knitting the file.}

\item{log_name}{A character string, denoting the log file name. It defaults to
"render_one.log". If this file is already present in the directory, this
function will append to it.}

\item{soln_stuff}{This is a list, with components env, test_fname, and
tt_list. This object is the output of \code{\link{populate_soln_env}}. Set
this to be NA if you wish to skip correctness checks, and only do rendering.}

\item{max_time_per_run}{The maximum time to wait before aborting the
rendering of a particular file.}

\item{permission_to_install}{If TRUE, then the function will try to install 
any packages needed. By default, this is FALSE.}
}
\value{
A data frame with one row for each file in the input directory.
}
\description{
Renders the specified file, and collates run time, static and 
correctness checks.
}
\details{
The log file contains a record of the libraries used by the student, and if
any new libraries needed to be installed. The status will be one of SUCCESS,
FAIL or UNKNOWN.
}
\seealso{
\code{\link{populate_soln_env}}, \code{\link{check_correctness}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/env_size.R
\name{env_size}
\alias{env_size}
\title{Calculates the Total Memory Used}
\usage{
env_size(env)
}
\arguments{
\item{env}{The environment whose size is to be computed.}
}
\value{
The size in bytes, as a numeric value (scalar).
}
\description{
This function uses the utils package to compute the total
amount of memory used by objects in an environment.
}
\details{
The names are wrapped in backticks. Otherwise, non-syntactic
names will cause problems. This function is used within
\code{\link{render_one}} as part of the runtime stats assessment.
}
\examples{

e1 <- new.env()
env_size(e1)
evalq(x <- 1:10000L, e1)
env_size(e1)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{get_parent_id2}
\alias{get_parent_id2}
\title{Get the parent node id}
\usage{
get_parent_id2(adj_list, at_node)
}
\arguments{
\item{adj_list}{The adjacency list of the tree.}

\item{at_node}{The node whose parent should be extracted.}
}
\value{
A integer of length 1 should be returned for all nodes except the
root. For the latter, the function will return NULL.
}
\description{
This function retrieves the parent node id of a given node from an
adjacency list of a tree.
}
\details{
Remember that the list has to be for a tree, not a general graph. Please
see other help pages for more specifications.

This is a low-level function, used within the S4 class TreeHarp. It is not
generally meant for use by the user.

If there are nodes that have more than one parent, then a warning is issued.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reset_path.R
\name{reset_path}
\alias{reset_path}
\title{Reset search path of current R session}
\usage{
reset_path(old_path)
}
\arguments{
\item{old_path}{A character vector of package namespaces. This is usually 
the output of \code{\link[base]{search}}, run before an R script or Rmd 
file is rendered, which could cause the search path to change.}
}
\value{
There is no object returned. This function is called for it's side-
effect of altering the search path.
}
\description{
This function is used to detach packages that have been added by a student 
script.
}
\details{
When a student script is rendered using \code{\link{render_one}}, 
new packages might be added to the search path. These may conflict with the 
instructors' search path order, or with subsequent runs of 
\code{\link{render_one}} on students. Hence there is a need to reset the 
search path before this is done.

This function does not unload namespaces. It only detaches them from the 
search path. For a difference between the two, please see Hadley's page.
}
\examples{
opath <- search()
# Load a package
reset_path(opath)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{is_subtree_rooted_at}
\alias{is_subtree_rooted_at}
\title{Checks if a tree is rooted at a node of another tree.}
\usage{
is_subtree_rooted_at(x, y, at_node)
}
\arguments{
\item{x}{An object of class TreeHarp.}

\item{y}{An object of class TreeHarp.}

\item{at_node}{An integer, corresponding to a node in object y. The sub-tree
of y, rooted at at_node, is compared to x.}
}
\value{
A logical value indicating if x is a sub-tree of y, rooted at
at_node.
}
\description{
This function checks if a given tree is a sub-tree of another tree at a
particular node.
}
\details{
Here's how it works: The sub-tree of y, rooted at at_node is first
extracted. The tree x is then compared to this. If x is a sub-tree of it,
then this function returns FALSE. Otherwise it returns TRUE.
}
\examples{
thb1 <- TreeHarp(list(b=2, d=NULL))
tha1 <- TreeHarp(list(a=c(2,3), b=4, c = NULL, d=NULL))
is_subtree_rooted_at(thb1, tha1, 1) # FALSE
is_subtree_rooted_at(thb1, tha1, 2) # TRUE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nlp_related.R
\name{rmd_to_token_count}
\alias{rmd_to_token_count}
\title{Count tokens in R/Rmd}
\usage{
rmd_to_token_count(fname, include_actuals = TRUE)
}
\arguments{
\item{fname}{The Rmd or R file name.}

\item{include_actuals}{Whether actual arguments/literals should be included.
If this is FALSE, then only calls and formal arguments will be used in
the count.}
}
\value{
A tibble. The tibble will contain a the frequency count for all 
tokens present in the student script.
}
\description{
Count the individual tokens. Part of the NLP analysis process.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/populate_soln_env.R
\name{populate_soln_env}
\alias{populate_soln_env}
\title{Returns solution environment and test code from template.}
\usage{
populate_soln_env(
  soln_fname,
  pattern,
  knit_root_dir,
  render_only = FALSE,
  output = NULL,
  dummy = FALSE
)
}
\arguments{
\item{soln_fname}{An rmd file containing the checks to be
run on the student solution.}

\item{pattern}{The pattern that identifies which chunks in the solution are 
are testing chunks. If this argument is missing, 
the default pattern used is "test".}

\item{knit_root_dir}{The root directory to use for knitting the rmd file. This
argument is optional. If it is missing, it uses the root directory in
knitr::opts_knit$get('root.dir').}

\item{render_only}{A logical value. If this is TRUE, then the solution is
run, rendered and returned. Otherwise the rendered html is deleted.}

\item{output}{The path to the knitted solution md file. This is usually 
deleted immediately, but sometimes we may want to keep it. This 
argument is passed on to \code{\link[knitr]{knit}}, so please refer to 
that page for the warnings about setting this argument when figures are 
involved.}

\item{dummy}{A logical value. If TRUE, then a dummy solution object is created
The environment will be empty, and the temporary test script will only have 
one line in it: `ah_dummy <- TRUE`. This solution object can be used to render 
a solution file without any correctness checks. Of course, one could just knit 
it directly, but this is still useful when testing things out.}
}
\value{
If render_only is FALSE, a list containing 2 components: the
environment populated by the solution rmd and the path to an R script
containing the test code.

If render_only is TRUE, then the output list consists of the 
aforementioned environment, and the path to the rendered solution file
(html). This option is useful for debugging the solution file.
}
\description{
Generates objects for checking solution correctness.
}
\details{
Test code should be written in a chunk that generates scalars 
from student objects.

The solution file has to be an Rmd file (not an R script), because it 
relies on the autoharp.obj and autoharp.scalars knitr hooks being present.

In addition, if it is required that a solution object is to be tested against
the analogous object within the student environment, these objects should be 
listed within the autoharp.objs option of a code chunk. These objects will be 
copied with the "." prefix.

Here is an overview of how the function works:
\enumerate{
  \item Knit the solution file to generate the solution (or "correct") 
  objects. 
  \item Rename these with the "." prefix in the solution environment
  object.
  \item Extract the lines of test code into a temporary R script.
  \item Wrap those chunks that contain autoharp.scalars hook with tryCatch.
  \item Return the solution environment and path to the R test script.
}

Typically, the next step is to call \code{\link{check_correctness}}.
}
\seealso{
\code{\link{check_correctness}}, \code{\link{render_one}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_rmd.R
\name{check_rmd}
\alias{check_rmd}
\title{Check if a File is Rmd}
\usage{
check_rmd(fname, verbose = TRUE)
}
\arguments{
\item{fname}{A character string. It is the name of the student submission
file.}

\item{verbose}{A logical value that prints messages if a non-rmd file
is found.}
}
\value{
The function will return TRUE if all the (3) checks pass, and FALSE
otherwise.
}
\description{
Checks if a file actually is an Rmd file.
}
\details{
It runs three checks. First, it checks for the file extension to be
Rmd or rmd or any such variant. Second, it checks for a YAML header at the
beginning of file. Finally, it checks if there is at least one properly
defined R chunk within the file.
}
\seealso{
\code{\link{get_libraries}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_kernel.R
\name{K2}
\alias{K2}
\title{Compute tree similarity}
\usage{
K2(t1, t2, verbose = FALSE)
}
\arguments{
\item{t1}{A TreeHarp object.}

\item{t2}{A TreeHarp object.}

\item{verbose}{A logical value, indicating if the output should be verbose.}
}
\value{
An integer, that counts the number of sub-trees in common between 
the two trees. Please see the reference papers for more information.
}
\description{
Compute tree similarity
}
\details{
As far as possible, this function tries to do things recursively.
It sets up a n x m matrix and fills up as much as it can. Then it uses 
recursive relationships to fill in the rest. When it cannot, it uses
\code{\link{generate_all_subtrees}} to generate and count common subtrees.
}
\examples{
tree1 <- TreeHarp(quote(x <- 1), TRUE)
tree2 <- TreeHarp(quote(y <- 1), TRUE)
K2(tree1, tree2, TRUE)

}
\references{
\enumerate{
\item 
\emph{Convolution kernels for natural language}, M Collins and N Duffy, 
\emph{Advances in neural information processing systems}, 2002.

\item 
\emph{Convolution kernels on discrete structures}, D Haussler, 
\emph{Technical report, Department of Computer Science, UC Santa Cruz}, 1999.
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/th_getter-length.R
\name{get_parent_id}
\alias{get_parent_id}
\alias{get_parent_id,TreeHarp-method}
\alias{get_parent_id,list-method}
\title{Generic for Getting Parent Node Id.}
\usage{
get_parent_id(x, node_num)

\S4method{get_parent_id}{TreeHarp}(x, node_num)

\S4method{get_parent_id}{list}(x, node_num)
}
\arguments{
\item{x}{An object of class TreeHarp or an adjacency list.}

\item{node_num}{An integer, length 1. This the node whose parent we are
after. If node_num is equal to 1, then NULL is returned because that should
be the root node.}
}
\value{
An integer, indicating the parent node.
}
\description{
The generic method definition for getting parent node id.
}
\section{Methods (by class)}{
\itemize{
\item \code{get_parent_id(TreeHarp)}: Obtain parent node id.

Extracts parent id of a node from a TreeHarp object.

\item \code{get_parent_id(list)}: Obtain parent node id.

Extracts parent id of a node from an adjacency list object.

}}
\seealso{
\code{\link{get_child_ids}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lang_2_tree.R
\name{lang_2_tree}
\alias{lang_2_tree}
\title{Convert language object to tree.}
\usage{
lang_2_tree(lang_obj, node_id, ni_env)
}
\arguments{
\item{lang_obj}{A language object.}

\item{node_id}{The calling node to this language object. This should only 
be greater than 0 if the \code{ni_env} already contains a partial adjacency 
list and corresponding node information. This will happen when this function 
is called recursively.}

\item{ni_env}{An environment to store the adjacency list and node information.}
}
\value{
Nothing
}
\description{
A recursive function for converting a language object to treeharp.
}
\details{
This function is used by TreeHarp constructors. It should not have
to be called by a user. It works by bulding up an adjacency list and node 
node information data frame within the supplied environment.
}
\examples{
e1 <- new.env() 
lang_2_tree(quote(X <- 1), 0, e1) 
e1$adj_list 
e1$node_info
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forestharp.R
\name{fapply}
\alias{fapply}
\title{Apply a function to a forest of trees.}
\usage{
fapply(fharp, TFUN, combine = TRUE, combiner_fn, ...)
}
\arguments{
\item{fharp}{The output of rmd_to_forestharp. It could also just be a 
list of TreeHarp objects.}

\item{TFUN}{A function that works on a single TreeHarp and returns an output. 
See \link{forestharp-helpers} for examples.}

\item{combine}{A logical value that indicates if the output from all function
applications should be combined.}

\item{combiner_fn}{A function to use to combine the individual output from 
each tree into a single scalar for each forest. It should handle NA 
values in the input vector or list. If it is missing, it defaults to 
sum, with na.rm=TRUE.}

\item{...}{Additional arguments to be passed on to TFUN.}
}
\value{
A vector, list or a single value. If TFUN returned an error for a 
particular TreeHarp, that component in the list or vector would be NA. This 
input vector or list will then be combined by combiner_fn.
}
\description{
A convenience function, for applying a function to many trees.
}
\details{
The input is simply a list of TreeHarp objects. First, the TFUN 
function is lapply-ed to each TreeHarp item, resulting in either a list, 
or a vector with possible NA elements.

The combiner function should be aware of this sort of output, and summarise
the list or vector accordingly, handling NA's and returning a \emph{scalar}.

If you need to create a partial function out of a forestharp helper, 
use an anonymous function, as shown in the examples below.
}
\examples{

ex1 <- quote(X <- rnorm(10, mean=0.9, sd=4))
ex2 <- quote(Y <- rbeta(10, shape1=3, shape2=5))
f1 <- lapply(c(ex1, ex2), TreeHarp, quote_arg=TRUE) 

# returns all function calls that begin with "r", like rnorm and rbeta.
# calls are returned as a list.
fapply(f1, extract_fn_call, combine =FALSE, pattern="^r.*")

# list is catenated.
fapply(f1, extract_fn_call, combine =TRUE, pattern="^r.*", 
       combiner_fn = function(x) {paste0(unlist(x), collapse=",")})

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{get_next_subtree}
\alias{get_next_subtree}
\title{Generate the next sub-tree.}
\usage{
get_next_subtree(obj, char_arr)
}
\arguments{
\item{obj}{An object of class TreeHarp.}

\item{char_arr}{A vector of 1's and 0's indicating which nodes to keep. The
vector should have length equal to the number of nodes in obj.}
}
\value{
A vector of 1's and 0's, which denotes the next sub-tree in the list.
}
\description{
This generates the next sub-tree in the enumeration list.
}
\details{
Need to reference the paper. This generates the next sub-tree,
rooted at the root node of this tree. It will generate singletons on it's
own. It has to be used within a loop to do that.
}
\examples{
th1 <- TreeHarp(list(a=c(2,3), b=NULL, c=NULL))
get_next_subtree(th1, c(1,0,0))
get_next_subtree(th1, c(1,1,0))

}
\seealso{
\code{\link{generate_all_subtrees}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{prune_depth}
\alias{prune_depth}
\title{Prune a tree up to a specified depth.}
\usage{
prune_depth(th, names_to_keep)
}
\arguments{
\item{th}{A TreeHarp object.}

\item{names_to_keep}{The node names to keep in the pruned tree.}
}
\value{
An object of class TreeHarp.
}
\description{
Prunes a tree up to a depth specified by a set of node names.
}
\details{
This is a seldom used function. It works in this way. Given a set 
of node names, it identifies the node with the greatest depth in that set.
The function then returns the sub-tree, that contains all the nodes with a
depth smaller than or equal to that depth. If the node types slot is not NA,
then that data frame is filtered and returned too.

Take a look at the examples for a clearer picture.
}
\examples{
ex1 <- quote(x <- f(y, g(5)))
th1 <- TreeHarp(ex1, TRUE)
s1 <- prune_depth(th1, c("f", "y"))
s2 <- prune_depth(th1, c("f", "z")) # node not present!
plot(s1)
plot(s2)
}
\seealso{
\code{\link{carve_subtree}}, \code{\link{path_to_root}}, 
\code{\link{carve_mst}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/treeharp.R, R/th_getter-length.R
\name{TreeHarp-class}
\alias{TreeHarp-class}
\alias{TreeHarp}
\alias{TreeHarp,logical-method}
\alias{TreeHarp,missing-method}
\alias{length,TreeHarp-method}
\alias{show,TreeHarp-method}
\alias{names,TreeHarp-method}
\title{An R expression as a tree.}
\usage{
TreeHarp(lang_obj, quote_arg, ...)

TreeHarp(lang_obj, quote_arg, ...)

\S4method{TreeHarp}{logical}(lang_obj, quote_arg, ...)

\S4method{TreeHarp}{missing}(lang_obj, quote_arg, ...)

\S4method{length}{TreeHarp}(x)

\S4method{show}{TreeHarp}(object)

\S4method{names}{TreeHarp}(x)
}
\arguments{
\item{lang_obj}{This should be an adjacency list for a tree (not a graph), or the
adjacency matrix of a tree, or the expression to be parsed. If it is a list,
only child nodes should be indicated (see the examples).}

\item{quote_arg}{If this argument is missing or FALSE, then the class of 
\code{lang_obj} will be evaluated, and, if it is either a list or matrix, 
the TreeHarp object will be returned. 

If this argument is TRUE, the \code{lang_obj} argument will be quoted and a 
parse tree for the expression will be computed and used as the tree.}

\item{...}{Unused at the moment.}

\item{x}{A Treeharp object.}

\item{object}{A TreeHarp object.}
}
\value{
Constructors return an object of class TreeHarp.

\code{length}: An integer of length 1.

\code{print}: Returns NULL. It prints a string representation of a 
TreeHarp object.

\code{names}: A character vector with length equal to the number of 
nodes.
}
\description{
This class is used to represent a \emph{single} R expression as a tree.
}
\details{
The following validity checks are conducted on the object:
\enumerate{
\item Is the graph connected? If no, the object is invalid.
\item Are there cycles? If yes, the object is invalid.
\item Are the nodes labelled in a BFS ordering? If not, the object is not 
valid.
}
}
\section{Methods (by generic)}{
\itemize{
\item \code{TreeHarp(logical)}: A constructor for TreeHarp.

Converts either adjacency list or matrix into a TreeHarp object.

\item \code{TreeHarp(missing)}: A constructor for TreeHarp.

Converts language object into a TreeHarp object.

\item \code{length(TreeHarp)}: To get the length of a tree.

The length of the tree refers to the number of nodes in the tree.

\item \code{show(TreeHarp)}: To print a tree representation.

A string representation of a TreeHarp object.

\item \code{names(TreeHarp)}: To get tree labels

This function returns the node labels of the tree.

}}
\section{Slots}{

\describe{
\item{\code{adjList}}{The adjacency list of the tree. The list must be named. The
nodes should be labelled in Breadth-First Order. The first component must
be the root of the tree. Leaves of the tree should be NULL elements.}

\item{\code{nodeTypes}}{A data frame describing the type of node. The columns in the 
data frame will be derived from the expression used to instantiate the 
object. The column names will be id (node id), name, call_status, 
formal_arg and depth. This slot can be left missing (i.e., populated with 
NA). This latter feature is useful when we just wish to test something out.

This slot is only populated automatically when an R expression is provided 
as \code{lang_obj} and \code{quote_arg} is TRUE.}

\item{\code{repr}}{A string representation of the tree. This will be printed when the 
show method of TreeHarp is called.}

\item{\code{call}}{The language object that was used to construct the tree (if it 
was). If the object was constructed from a list/matrix, this will be NA.}
}}

\examples{
l1 <- list(a=c(2,3), b=NULL, c=NULL)
# directly using new() 
treeharp1 <- new("TreeHarp", adjList = l1, nodeTypes = NA)

# using one of the constructor methods (for lists)
treeharp2 <- TreeHarp(l1)

# using the constructor for matrices.
m1 <- matrix(0L, 3, 3)
dimnames(m1) <- list(letters[1:3], letters[1:3])
m1[1, ] <- c(0, 1L, 1L)
m1[, 1] <- c(0, 1L, 1L)
treeharp3 <- TreeHarp(m1)

# Supplying a language object to get the same tree (with nodeTypes 
# populated)
ex1 <- quote(a(b,c))
TreeHarp(ex1, TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{matrix_2_adj_list}
\alias{matrix_2_adj_list}
\title{Convert adjacency matrix to a list.}
\usage{
matrix_2_adj_list(mat)
}
\arguments{
\item{mat}{A symmetric matrix of 1's and 0's, with 1 in entry (i,j)
representing an edge between the two vertices.}
}
\value{
The adjacency list of the tree.
}
\description{
Converts a binary matrix that represents a tree into an adjacency list.
}
\details{
Remember that the list that is finally output is for a tree, not a general
graph. Please see other help pages for more specifications.

The input matrix should be BFS ordered. The adjacency list only notes the
child node(s) of a particular node. If a matrix denotes multiple parents,
it will not be picked up.

This is a low-level function, used within the S4 class TreeHarp. It is not
generally meant for use by the user.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{adj_list_2_matrix}
\alias{adj_list_2_matrix}
\title{Convert adjacency list to a matrix}
\usage{
adj_list_2_matrix(adj_list)
}
\arguments{
\item{adj_list}{The adjacency list of the tree.}
}
\value{
A symmetric matrix of 1's and 0's, with 1 in entry (i,j) representing
an edge between the two vertices.
}
\description{
Converts a list that represents a tree into a binary matrix.
}
\details{
Remember that the list has to be for a tree, not a general graph. Please
see other help pages for more specifications.

This is a low-level function, used within the S4 class TreeHarp. It is not
generally meant for use by the user.

It works by filling up the upper diagonal of the matrix before reflecting it.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forestharp.R
\name{rmd_to_forestharp}
\alias{rmd_to_forestharp}
\title{Convert to TreeHarp objects}
\usage{
rmd_to_forestharp(fname, verbose = FALSE)
}
\arguments{
\item{fname}{The filename that is to be read in.}

\item{verbose}{If TRUE, then messages will be printed to enable debugging.
dropped.}
}
\value{
There are three possible return outcomes: (1) NA, indicating that 
\code{\link[knitr]{purl}} was unable to extract the R expressions from the 
Rmd/qmd file. (2) A list with two components, named \code{forest} and
\code{line_nums}, each being NA. (3) A list with two components, named as in 
(2), containing the TreeHarp objects and a vector of line numbers.

If there are syntax errors in the file, the expressions may not be parsed
correctly. The \code{verbose} option will reflect this.
}
\description{
Reads in an Rmd file or an R script and converts it to a list of TreeHarp
objects.
}
\details{
If an Rmd or qmd file is supplied as input, it is first converted
to an R script before \code{\link[lintr]{get_source_expressions}} is used 
to try to parse the expressions. These expressions are then matched to the
original file to retrieve line numbers. Unlike previous versions of this 
function, line numbers will always be returned.

Expressions that could not be parsed will be returned as NA.

Line numbers are extracted using \code{\link[lintr]{get_source_expressions}} from
the lintr package.
}
\seealso{
\code{\link{fapply}}, \code{\link{extract_chunks}},
\code{\link{extract_chunks}}, \code{\link[lintr]{get_source_expressions}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/autoharp.R
\docType{package}
\name{autoharp}
\alias{autoharp-package}
\alias{autoharp}
\title{autoharp: Semi-Automatic Grading of R and Rmd Scripts}
\description{
The autoharp package provides functions for running and analysing 
R script/Rmd submissions from students.
}
\section{Main functions are}{

\enumerate{
  \item \code{\link{populate_soln_env}}
  \item \code{\link{render_one}}
  \item \code{\link{TreeHarp}}
}

The user manuals can be found at \url{https://namanlab.github.io/autoharp/}
}

\seealso{
Useful links:
\itemize{
  \item \url{https://namanlab.github.io/autoharp/}
}

}
\author{
\strong{Maintainer}: Agrawal Naman \email{naman.a@nus.edu.sg}

Authors:
\itemize{
  \item Vik Gopal \email{vik.gopal@nus.edu.sg}
  \item Samuel Seah \email{samuelseah@u.nus.edu}
  \item Viknesh Jeya Kumar \email{viknesh@u.nus.edu}
  \item Gabriel Ang \email{gabrielang@u.nus.edu}
}

Other contributors:
\itemize{
  \item Yuting Huang \email{yhuang@nus.edu.sg} [contributor]
  \item Ruofan Liu \email{kelseyliu1998@gmail.com} [contributor]
  \item National University of Singapore [copyright holder]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_lints.R
\name{count_lints_all}
\alias{count_lints_all}
\title{Lint counter}
\usage{
count_lints_all(file_names, lint_list, lint_labels)
}
\arguments{
\item{file_names}{The path to the rmd files that need to be checked for
lints.}

\item{lint_list}{List of lints to check for.}

\item{lint_labels}{List of labels to name the vector to return.}
}
\value{
Dataframe containing the lints.
}
\description{
Count number of lints in one folder
}
\details{
The function will count the number of lints in a file. The lints to be
checked can be passed as an argument. Else, the default will be used.
The defaults are as follows:
\itemize{
\item T_and_F_symbol_linter
\item line_length_linter
\item assignment_linter
\item absolute_path_linter
\item pipe_continuation_linter
}

Note that labels would also need to be given if the non-default lints
are chosen.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{find_branch_num}
\alias{find_branch_num}
\title{Find the branch that leads from one node to another.}
\usage{
find_branch_num(th, child_id, ancestor_id)
}
\arguments{
\item{th}{A TreeHarp object.}

\item{child_id}{An integer node id. It corresponds to the node to trace up
from.}

\item{ancestor_id}{An integer node id. It corresponds to the node to trace 
down from.}
}
\value{
An integer that denotes the branch to follow down (from the ancestor)
to reach the child.
}
\description{
Given two nodes that are on the same path to the root, this function
determines the branch that leads to the child node.
}
\details{
This is used when trying to find a sub-call from a TreeHarp object.
It is useful in determining the indices to use when extracting the sub-call.
}
\examples{
ex3 <- quote(x <- f(y = g(3, 4), z=1L))
t1 <- TreeHarp(ex3, TRUE)
find_branch_num(t1, 8, 3) # should be 1
find_branch_num(t1, 5, 3) # should be 2
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{is_cyclic_r}
\alias{is_cyclic_r}
\title{Checks if a graph contains any cycles.}
\usage{
is_cyclic_r(adj_mat, node_v, parent_node = -1, visited_env)
}
\arguments{
\item{adj_mat}{A symmetric matrix of 1's and 0's, with 1 in entry (i,j)
representing an edge between the two vertices.}

\item{node_v}{The node to begin searching for cycles from. An integer.}

\item{parent_node}{The parent node of node_v. Also an integer. Use -1 if
you are starting from node 1. This is in fact the default.}

\item{visited_env}{An environment containing a logical vector indicating which
nodes have already been visited. The vector has to be named "visited".
See the details.

The function works by traversing all the nodes, in a BFS order. If it finds
a node has a parent that has already been visited, it concludes that there
is a cycle.

The function is recursive, and has to update the vector of visisted nodes
within each call. Hence the visited vector is stored in an environment
that is passed along. It will return an error if no such environment is
provided. It is a very specific input that the function requires, and this
is another reason that this function is not exported.

This function is used within the validity checks for the S4 class. It is
not exported for the user.}
}
\value{
A logical value indicating if the graph contains cycles.
}
\description{
A tree is a graph that is connected but does not have any cycles. This
function checks if a provided adjacency matrix contains cycles.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{path_to_root}
\alias{path_to_root}
\title{Extract a path from node to root.}
\usage{
path_to_root(th, node_num)
}
\arguments{
\item{th}{A TreeHarp object.}

\item{node_num}{A node number to start tracking upwards from.}
}
\value{
A vector of 1's and 0's that can be used to carve out the branch 
alone, using \code{\link{carve_subtree}}.
}
\description{
Identifies the nodes on the path from a node up to the root of a TreeHarp
object.
}
\details{
This function allows the user to identify the branch from a node 
up to the root of a tree.
}
\examples{
ex1 <- quote(x <- f(y, g(5)))
th1 <- TreeHarp(ex1, TRUE)
path_to_root(th1, 5)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{extract_chunks}
\alias{extract_chunks}
\title{Extract chunks that match a pattern.}
\usage{
extract_chunks(rmd_name, pattern)
}
\arguments{
\item{rmd_name}{A character string, the name of the rmd file to get the
chunks from.}

\item{pattern}{The pattern to match within the label. In fact, the match is
applied to the whole chunk option.}
}
\value{
A list of character vectors. Each vector contains the chunk from the
file. If no pattern is specified, all chunks are returned. Remember that
the chunk header and tail are also included in the returned list.
}
\description{
Extracts chunks whose labels match a pattern from the rmd file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lang_2_tree_helpers.R
\name{rbind_to_nodes_info}
\alias{rbind_to_nodes_info}
\title{Update node information.}
\usage{
rbind_to_nodes_info(id, name, call_status, formal_arg, depth, env_ni)
}
\arguments{
\item{id}{The id of the node to be added. This should be an integer
of length 1.}

\item{name}{The name of the node.}

\item{call_status}{Is the language object a call or a symbol/literal? This 
should a logical value.}

\item{formal_arg}{Is the language object a formal argument or not? This 
should be a logical value.}

\item{depth}{An integer indicating the depth of this language object in 
the parse tree.}

\item{env_ni}{An environment object, possibly containing a data frame with
columns id, name, call_status, formal_arg and depth.}
}
\value{
TRUE is returned invisibly.
}
\description{
Updates the node information regarding an R expression.
}
\details{
This is for internal use. It may be removed from user-view soon!
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{keep_branches}
\alias{keep_branches}
\title{Keep only branches specified by node numbers}
\usage{
keep_branches(th, branch_nodes, include_lower = TRUE)
}
\arguments{
\item{th}{A TreeHarp object.}

\item{branch_nodes}{An integer vector, specifying the nodes to keep.}

\item{include_lower}{A logical value - whether or not the lower branches should 
also be kept.}
}
\value{
A TreeHarp object.
}
\description{
Retains only specific branches, that are identified by their node numbers.
}
\examples{
ex1 <- quote(x <- f(y, g(5)))
th1 <- TreeHarp(ex1, TRUE)
keep_branches(th1, 3)
keep_branches(th1, 3, include_lower = FALSE)
keep_branches(th1, c(2,3), FALSE)
keep_branches(th1, c(3, 4), FALSE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/join_treeharp.R
\name{join_treeharps}
\alias{join_treeharps}
\title{Root a list of trees.}
\usage{
join_treeharps(...)
}
\arguments{
\item{...}{A list of Treeharp objects.}
}
\value{
A TreeHarp object
}
\description{
Given a list of trees, this will root them.
}
\details{
This function combines TreeHarp objects into a single TreeHarp.
The function will root all of them at a node called "script", which is 
neither a function call nor an argument nor a symbol. The BFS ordering is 
then updated.

Objects that are not of class TreeHarp will be dropped from the list 
before the rooting takes place.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/as.matrix.R
\name{as.matrix}
\alias{as.matrix}
\title{TreeHarp Cast a TreeHarp to Matrix.}
\arguments{
\item{from}{A treeharp object.}
}
\value{
A matrix.
}
\description{
Convert a treeharp object to an adjacency matrix.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{subtree_at}
\alias{subtree_at}
\title{Extract a sub-tree.}
\usage{
subtree_at(obj, at_node, preserve_call = FALSE)
}
\arguments{
\item{obj}{An object of class TreeHarp}

\item{at_node}{The root of the new sub-tree. An integer, not a label, that
corresponds to BFS indexing of the tree.}

\item{preserve_call}{A logical value that indicates if a sub-call should be 
extracted. This might be slower, but it allows you to evaluate it later.}
}
\value{
An object of class TreeHarp.
}
\description{
Extracts a sub-tree rooted at a particular node.
}
\details{
This is meant for internal use, so the nodeTypes slot is silently
dropped, unless preserve_call is set to TRUE
}
\examples{
th3 <- list(a= c(2L,3L,4L), b=NULL, c=c(5L, 6L), d=7L, e=NULL, f=NULL, g=NULL)
subtree_at(TreeHarp(th3), 3)
st <- subtree_at(TreeHarp(th3), 4)
plot(st)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{get_recursive_index}
\alias{get_recursive_index}
\title{Obtain an index to extract out a sub-call}
\usage{
get_recursive_index(th, node_id)
}
\arguments{
\item{th}{A TreeHarp object.}

\item{node_id}{An integer corresponding to a call within the parse tree 
(not a literal, symbol or a formal argument).}
}
\value{
A vector of indices, that can be used (together with "[[") to obtain a 
sub-call
}
\description{
Obtains an index that can be used to extract a sub-call from a language object.
}
\examples{
ex3 <- quote(x <- f(y = g(3, 4), z=1L))
t1 <- TreeHarp(ex3, TRUE)
rec_index <- get_recursive_index(t1, 6)
ex3[[rec_index]]
ex3[[get_recursive_index(t1, 3)]]
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_kernel.R
\name{jaccard_treeharp}
\alias{jaccard_treeharp}
\title{Computes Jaccard Index}
\usage{
jaccard_treeharp(th1, th2, weighted = FALSE)
}
\arguments{
\item{th1}{A TreeHarp object.}

\item{th2}{A TreeHarp object.}

\item{weighted}{A logical value, indicating if the weighted Jaccard 
similarity should be computed.}
}
\value{
A real number between 0 and 1.
}
\description{
Computes the Jaccard index between two trees.
}
\details{
The unweighted form is just the cardinality of the intersection 
of the two sets of tokens, divided by the union of the two sets.

The weighted form is described on the WIkipedia page:
https://en.wikipedia.org/wiki/Jaccard_index#Weighted_Jaccard_similarity_and_distance
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forestharp_helpers.R
\name{forestharp-helpers}
\alias{forestharp-helpers}
\alias{count_self_fn}
\alias{count_lam_fn}
\alias{count_fn_call}
\alias{extract_fn_call}
\alias{extract_formal_args}
\alias{extract_assigned_objects}
\alias{extract_actual_args}
\alias{count_num_lines_for_loop}
\alias{detect_growing}
\alias{detect_for_in_fn_def}
\alias{count_fn_in_fn}
\alias{detect_fn_call_in_for}
\alias{extract_self_fn}
\alias{detect_fn_arg}
\alias{detect_nested_for}
\title{Forestharp helpers}
\usage{
count_self_fn(th)

count_lam_fn(th)

count_fn_call(th, pattern, pkg_name)

extract_fn_call(th, pattern, pkg_name)

extract_formal_args(th, fn_name)

extract_assigned_objects(th)

extract_actual_args(th, include_assigned_obj = TRUE)

count_num_lines_for_loop(th)

detect_growing(th, count = FALSE, within_for = FALSE)

detect_for_in_fn_def(th, fn_name)

count_fn_in_fn(th, fn_name, sub_fn)

detect_fn_call_in_for(th, fn_name)

extract_self_fn(th)

detect_fn_arg(th, fn_name, arg)

detect_nested_for(th)
}
\arguments{
\item{th}{A TreeHarp object.}

\item{pattern}{A regular expression to pick up function names.}

\item{pkg_name}{The name of a package to match functions with. This should
be an exact match for the package name. The package should be attached for
this to work. In order to avoid picking up duplicate names, for instance
\code{tolower} is a function in base R and in ggplot2, run
\code{\link{get_libraries}} on the file as well, and match against it.}

\item{fn_name}{Function name, as a character string}

\item{include_assigned_obj}{A Boolean flag. If TRUE (default), it also returns
the name of the assigned object. If FALSE, it drops the name of the assigned 
object. The FALSE flag is useful for when checking if the we wish to match the 
assigned object from a previous call with the actual argument in this call.}

\item{count}{For \code{detect_growing}, this is a logical value that indicates
if the number of "grow" expressions should be counted and returned, or if just a
logical value should be returned.}

\item{within_for}{If TRUE, only expresssions within a for loop are included.}

\item{sub_fn}{(For count_fn_in_fn), the function to count (to look for within
fn_name).}

\item{arg}{The argument to check for within fn_name (as a character string).}
}
\value{
On their own, each of these functions should return a scalar or a
1-dimensional array. When called with \code{\link{fapply}}, the scalar
numerical values can be combined (by taking the sum, any other provided
combiner function).

The ultimate idea is that fapply should return a single feature for each rmd
file that it is called upon.
}
\description{
Example of functions that can be \emph{directly used on TreeHarp objects
individually,} and on forestharp objects via \code{\link{fapply}}.
}
\details{
These are examples of functions that be called on a list of TreeHarp
  objects, which we refer to as a forestharp object. Such objects are not
  formally defined yet, but can be created using
  \code{\link{rmd_to_forestharp}} or using \code{\link{join_treeharps}}.
}
\section{Functions}{
\itemize{
\item \code{count_self_fn()}: Counts the number of self-defined functions.

This helper counts the number of self-defined functions. It excludes
lambda functions. It returns an integer scalar.

As long as the function \code{function} was called and assigned, it will be
counted.

\item \code{count_lam_fn()}: Counts the number of anonymous functions.

Counts the number of anonymous functions, typically used in sapply, etc. It
returns an integer scalar. As long as the function \code{function} was
called but \emph{not} assigned, it will be counted here.

\item \code{count_fn_call()}: Counts the number of function calls that match a pattern.

This helper counts the number of function calls that match a pattern. It
returns a count, i.e. an integer vector of length 1.

If \code{pkg_name} is provided instead of \code{pattern}, then this function
counts the number of function calls from that package.

\item \code{extract_fn_call()}: Extracts function calls as a string.

Extracts the function calls that match a pattern. It returns a character
vector. Remember to set \code{combine = FALSE} when calling
\code{\link{fapply}} with it.

\item \code{extract_formal_args()}: Extracts function formal arguments called.

Extracts the function \emph{formal} arguments from functions with a given
name. The name must match the function name exactly. This returns a character
vector or NULL, if no formal arguments are used.

\item \code{extract_assigned_objects()}: Extracts names of assigned objects

Extracts the names of assigned objects. This was written to assist in
detecting missed opportunities to use the pipe operator.

\item \code{extract_actual_args()}: Extracts actual argument names

Extracts the actual arguments from an expression, not the formal
arguments. It only returns syntactic literals. It should be improved
to return the actual arguments for a specified function so that something
similar to \code{extract_assigned_objects} could be returned.

\item \code{count_num_lines_for_loop()}: Counts number of lines in a for loop

Extracts the actual arguments from an expression, not the formal
First, this function checks if the expression contains a for loop. 
If it does, the innermost \code{for} loop is identified. Finally, the 
number of expressions below it is counted and returned.

\item \code{detect_growing()}: Detects if a vector is being grown.

It detects if there is an expression of form: x <- c(x, new_val). This is
generally bad programming practice

\item \code{detect_for_in_fn_def()}: Detects if a for loop is present within a function

It detects if a for loop is present within a function definition.

\item \code{count_fn_in_fn()}: Count use of a function within another.

It counts the number of times a function is used within another.

\item \code{detect_fn_call_in_for()}: Detect for loop to call a function

Checks if a function has been called within a for loop.

\item \code{extract_self_fn()}: Extract names of functions defined by user.

Extracts names of user-defined functions. They may not all look nice, because
sum functions may be anonymous functions. This function needs to be improved.

\item \code{detect_fn_arg()}: Was a function called with a particular argument?

Checks if a function was called with a particular argument, which could be
the formal or actual one. The immediate child of the function call node is
checked.

\item \code{detect_nested_for()}: Was a nested "for" loop called anywhere within the code?

Checks if a nested for-loop was called anywhere within the code. This returns
a logical scalar for each TreeHarp object given.

}}
\examples{
# Dummy trees
th1 <- TreeHarp(quote(X <- rnorm(10, mean=0.9, sd=4)), TRUE)
th2 <- TreeHarp(quote(Y <- rbeta(10, shape1=3, shape2=5)), TRUE)
th3 <- TreeHarp(quote(fn1 <- function(x) x + 2), TRUE)
th4 <- TreeHarp(quote(df1 <- mutate(df1, new_col=2*old_col)), TRUE)

# Run helpers
count_self_fn(th3)
count_fn_call(th4, pkg_name="dplyr")
count_fn_call(th1, pattern="^r.*")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{is_connected}
\alias{is_connected}
\title{Checks if a graph is connected.}
\usage{
is_connected(adj_list, root = 1)
}
\arguments{
\item{adj_list}{The adjacency list of the tree.}

\item{root}{The root node to start checking from. This defaults to the
first node in the adjacency list.}
}
\value{
The function returns a TRUE if the graph is connected and FALSE
otherwise.
}
\description{
A tree is a graph that is connected but does not have any cycles. This
function checks if a provided adjacency list is connected.
}
\details{
This function is used as one of the validity checks within the
definition of the TreeHarp class. It is a low-level function, not really
meant for the general user of the package. Hence it is not exported.

The nodes are traversed in a BFS order. The function could actually be
combined with \link{is_cyclic_r}, but it is kept separate for modularity
reasons.

An alternative was to convert the list to an adjacency matrix and check
for a row and column of zeros.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_thumbnails.R
\name{generate_thumbnails}
\alias{generate_thumbnails}
\title{Generate a html of thumbnails}
\usage{
generate_thumbnails(out_dir, html_fname, html_title, anonymise = FALSE)
}
\arguments{
\item{out_dir}{The directory in which student html files and the figures 
are kept.}

\item{html_fname}{The name of the master html file which will contain all
thumbnails. This file will be created in \code{out_dir}.}

\item{html_title}{The title tag of the master html page. This will be displayed 
on top of the output html page.}

\item{anonymise}{If TRUE, the original filenames will be replaced with
inocuous numbers. If FALSE, the original filenames will be retained.}
}
\value{
The function returns nothing, but it should create a html page 
of thumbnails of all the images that students plotted, along with links to 
their individual pages.
}
\description{
Generate a html of thumbnails
}
\details{
After running \code{\link{render_one}} on a set of R/Rmd files 
in a directory, this function helps to consolidate them for review.

The output folder contains all the generated html files, images and a log 
file. This function will extract the images from each html file and display 
them as thumbnails on a new html page, with links to all individual files.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/check_correctness.R
\name{check_correctness}
\alias{check_correctness}
\title{Check correctness of student solution rmd.}
\usage{
check_correctness(e_stud, e_soln, test_fname)
}
\arguments{
\item{e_stud}{The environment containing the output objects from running the
studnent Rmd file.}

\item{e_soln}{The environment containing the objects from the solution
template. It will probably contain objects with the suffix "_soln". These
will be tested against the versions generated by the student.}

\item{test_fname}{The R script containing the test chunks.}
}
\value{
A data frame with one row, and the number of columns equal to the 
number of tests run plus the number of scalars to keep.
}
\description{
This will run unit tests on the students' rmd file.
}
\details{
Prior to calling this, \code{\link{populate_soln_env}} should 
already have been called on the solution template, and the student 
file should already have been knitted in order to generate the students' 
objects. Of course, one could generate the test script independent of 
\code{\link{populate_soln_env}}, but the solution environment that contains 
objects with a "_soln" suffix is also needed.

The student environment, solution environment, test file and the list of 
tests and expectations are the inputs to this function.
}
\seealso{
\code{\link{populate_soln_env}}, \code{\link{render_one}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{copy_e2e}
\alias{copy_e2e}
\title{Copy an object from one env to another.}
\usage{
copy_e2e(from_obj, from_env, to_obj, to_env)
}
\arguments{
\item{from_obj}{The name of the object to copy. It has to be a string.}

\item{from_env}{The environment in which the object lives. It has to be an
object of class environment.}

\item{to_obj}{The name of the object to assign it to, in the new environment.
Also a string.}

\item{to_env}{The environment to which the new object is to be assigned. It
has to be an object of class environment.}
}
\value{
There is no return value. This function is called for its' side
effect.
}
\description{
A wrapper function that uses assign and get.
}
\examples{

e1 <- new.env(); e2 <- new.env()
ls(e2)
evalq(x <- 1L, e1)
copy_e2e("x", e1, "y", e2)
ls(e2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lang_2_tree_helpers.R
\name{get_next_depth_id}
\alias{get_next_depth_id}
\title{Get the id and depth of a child node.}
\usage{
get_next_depth_id(parent_node_id, env_ni)
}
\arguments{
\item{parent_node_id}{The id of the parent node we are considering.}

\item{env_ni}{An environment object, possibly containing a data frame with
columns id, name, call_status, arg_type and depth.}
}
\value{
A list containing the id and depth of the next node.
}
\description{
From the parent's depth and the last labelled node, we obtain the node id 
and depth of a child.
}
\details{
This is for internal use. It may be removed from user-view soon!
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{extract_non_chunks}
\alias{extract_non_chunks}
\title{Extract non-chunks from an Rmd file.}
\usage{
extract_non_chunks(rmd_name, out_name)
}
\arguments{
\item{rmd_name}{A character string, the name of the rmd file to get the
chunks from.}

\item{out_name}{An output filename, to dump the text to.}
}
\value{
If out_name is missing, then a character vector is returned. If 
outfname is specified, then nothing is returned. The text is written to the 
file instead.
}
\description{
Extracts non-chunks from an Rmd file.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/extracton_section_text.R
\name{extract_section_text}
\alias{extract_section_text}
\title{Extract section text from Rmd}
\usage{
extract_section_text(rmd_name, hdr_pat, ignore_case = TRUE)
}
\arguments{
\item{rmd_name}{The filename of the Rmd script.}

\item{hdr_pat}{The regular expression pattern to pick up the section title.}

\item{ignore_case}{A boolean - whether or not to ignore case when matching for
the section title.}
}
\value{
Returns a character vector containing all the text written in the section
that begins with the specified pattern.

The pattern should pick up a unique section/sub-section/sub-sub-section.
Otherwise, it will stop and raise an error.
}
\description{
Extract section text from Rmd
}
\details{
The text that is picked up begins with the specified section, and ends with
the next string of pound symbols (#)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{replace_sp_chars_filename}
\alias{replace_sp_chars_filename}
\title{Replace Special Characters in File Name}
\usage{
replace_sp_chars_filename(dir_name, return_df = TRUE)
}
\arguments{
\item{dir_name}{A character string, referring to the directory of Rmd
files whose names should be replaced.}

\item{return_df}{A logical value, indicating if the old and new names should
be returned (in a tibble).}
}
\value{
A tibble containing the old and new names.
}
\description{
Replaces special characters in the name of an R or Rmd script.
}
\details{
If a filename contains one of the following special characters 
(ignore the quotes here): "\code{[ <>()|\\:&;#?*']}", the  
\link[knitr]{knit} function will replace 
them with underscores. Hence the filenames in the autoharp input directory 
and the output directory will not match, even allowing for the change in 
file extension. This will cause problems when we try to run 
\code{\link{render_one}} again on the same input directory.

This function renames the files in the input directory by replacing all 
special characters there.

The NUS LMS (LumiNUS) introduces parenthesized names or numbers in order 
to make filenames unique, so this function is necessary for NUS 
instructors.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/forestharp.R
\name{apply_can_improve_this}
\alias{apply_can_improve_this}
\title{Identifies if an apply function can improve the code}
\usage{
apply_can_improve_this(fname, window_size = 2)
}
\arguments{
\item{fname}{A filename: either Rmd/qmd or R.}

\item{window_size}{A window size to analyse.}
}
\value{
A numeric vector corresponding to windows for review is returned. If 
none are found, then NULL is returned.
}
\description{
Identifies if an apply function can improve the code
}
\details{
First, all library() calls and read() calls are removed. Next,
\code{only_actual_args_differ} is run on all pairwise expressions. The 
matrix is searched for blocks of 1's along the diagonal.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/th_getter-length.R
\name{get_child_ids}
\alias{get_child_ids}
\alias{get_child_ids,TreeHarp-method}
\alias{get_child_ids,list-method}
\title{Generic for Getting Child Node Ids}
\usage{
get_child_ids(x, node_num)

\S4method{get_child_ids}{TreeHarp}(x, node_num)

\S4method{get_child_ids}{list}(x, node_num)
}
\arguments{
\item{x}{An object of class TreeHarp.}

\item{node_num}{An integer, length 1. This the node whose children we are
after. If the specified node is a leaf, the NULL is returned.}
}
\value{
An integer vector, indicating the children node ids.
}
\description{
The generic method definition for getting child node ids.
}
\section{Methods (by class)}{
\itemize{
\item \code{get_child_ids(TreeHarp)}: Obtain child nodes.

Allows user to extract the child nodes from a specified node from TreeHarp
object.

\item \code{get_child_ids(list)}: Obtain child nodes.

Allows user to extract the child nodes from a specified node from an 
adjacency list.

}}
\seealso{
\code{\link{get_parent_id}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/to_BFS.R
\name{to_BFS}
\alias{to_BFS}
\title{Function to rearrage nodes in BFS}
\usage{
to_BFS(adj_list, node_info)
}
\arguments{
\item{adj_list}{The output of lang_2_tree.}

\item{node_info}{The output of lang_2_tree.}
}
\value{
An adjacency list and nodes info data frame in BFS order.
}
\description{
Function to rearrage nodes in BFS
}
\details{
This function is for an internal TreeHarp constructor use. It is
not exported.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_kernel.R
\name{tree_sim}
\alias{tree_sim}
\title{Compute tree similarity}
\usage{
tree_sim(t1, t2, norm = FALSE, ...)
}
\arguments{
\item{t1}{A TreeHarp object}

\item{t2}{Anothe TreeHarp object.}

\item{norm}{A logical value, indicating if the kernel function should be
normalised, to account for different tree lengths.}

\item{...}{Unused arguments, reserved for mcmapply}
}
\value{
A numerical value between 0 and 1 (if normed).
}
\description{
Computes similarity between two trees (non-recursively)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{get_libraries}
\alias{get_libraries}
\title{Extracts the Packages Used in An Rmd File.}
\usage{
get_libraries(fname)
}
\arguments{
\item{fname}{The Rmd filename or R script.}
}
\value{
A character vector containing the packages used within the Rmd
document.
}
\description{
The input filename could correspond to an R script or an Rmd file.
}
\details{
The file is assumed to be either an R script or an Rmd file.
If it is found to be an Rmd file using \link{extract_chunks}, it is purl-ed before
libraries are extracted. If it is found to be NOT an Rmd, it is assumed
to be an R script and nothing is done to process it.

The file is not parsed, so even text files will work with this function.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree_routines.R
\name{get_parent_call_id}
\alias{get_parent_call_id}
\title{Get Node Id of Parent Call}
\usage{
get_parent_call_id(x, node_id)
}
\arguments{
\item{x}{A TreeHarp object.}

\item{node_id}{The id of the node whose parent call is to be found. An 
integer value.}
}
\value{
An integer corresponding to the node id of the calling function.
}
\description{
Get the node id of the parent call for a given node.
}
\details{
When we need to go up the parse tree to obtain the function that 
called this node, we use this function. It is similar to \code{get_parent_id},
except that that function only returns the immediate parent.

It is not useful to call this function when the TreeHarp object is not 
constructed from a language object.

Perhaps this function is necessary only because of the way language objects 
are represented by the autoharp: formal arguments are included in the 
tree representation. When we wish to find the calling function, we have to 
walk up the branches till we reach a function call.
}
\examples{
ex3 <- quote(x <- f(y = g(3, 4), z=1L))
t1 <- TreeHarp(ex3, TRUE)

# get the function that calls g:
get_parent_call_id(t1, 6) 
#contrast with this:
get_parent_id(t1, 6)
}
\seealso{
\code{\link{get_parent_id}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{clean_dir}
\alias{clean_dir}
\title{Removes md Files when no HTML Present}
\usage{
clean_dir(dir_name, verbose = FALSE)
}
\arguments{
\item{dir_name}{The directory containing the files to be cleaned.}

\item{verbose}{If TRUE, then the files and directories being removed will be 
printed.}
}
\value{
No return value.
}
\description{
Cleans up the autoharp output directory.
}
\details{
When batch rendering Rmd files, it is inevitable that some files fail. These
files would have their knit.md and utf.md present, but they would not have a
corresponding html file generated.

This function is called for its' side-effect, to remove those lonely md files.

If this clean-up is not done, when we try to re-run the files (perhaps with
some of the errors fixed), these straggling md files will cause problems. 
The most crucial one is that the Rmd files will not be re-knitted, even 
though they have been changed.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_lints.R
\name{count_lints_one}
\alias{count_lints_one}
\title{File lint counter}
\usage{
count_lints_one(rmd_file, lint_list, lint_labels)
}
\arguments{
\item{rmd_file}{The path to the rmd file to check for lints.}

\item{lint_list}{List of lints to check for.}

\item{lint_labels}{List of labels to name the vector to return.}
}
\value{
Vector containing the lints.
}
\description{
Count number of lints in one file
}
\details{
The function will count the number of lints in a file. The lints to be 
checked can be passed as an argument. Else, the default will be used. 
The defaults are as follows: 
* T_and_F_symbol_linter
* line_length_linter
* assignment_linter
* absolute_path_linter
* pipe_continuation_linter
Note that labels would also need to be given if the non-default lints 
are chosen.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/log_summary.R
\name{log_summary}
\alias{log_summary}
\title{Generate a dataframe from the log file.}
\usage{
log_summary(log_file)
}
\arguments{
\item{log_file}{The name of the log file generated from 
\code{\link{render_one}}.}
}
\value{
The function returns a dataframe summarising the details in the
log file.
}
\description{
Generate a dataframe from the log file.
}
\details{
This provides a table view of the log file, which is updated 
in a more natural format by simply concatenating new updates. The output 
of this function makes it easier to group entries by filename, time, or
status, or even error message.

The output table does not contain correctness output. It only contains the 
columns name, timestamp, status (SUCCESS/FAIL), error message, 
number of libraries used and number of libraries installed.
}
\seealso{
\code{\link{render_one}}
}
